package drr.enrichment.common;

import com.google.common.collect.ImmutableList;
import com.rosetta.model.lib.RosettaModelObject;
import com.rosetta.model.lib.RosettaModelObjectBuilder;
import com.rosetta.model.lib.annotations.RosettaAttribute;
import com.rosetta.model.lib.annotations.RosettaDataType;
import com.rosetta.model.lib.meta.RosettaMetaData;
import com.rosetta.model.lib.path.RosettaPath;
import com.rosetta.model.lib.process.BuilderMerger;
import com.rosetta.model.lib.process.BuilderProcessor;
import com.rosetta.model.lib.process.Processor;
import com.rosetta.util.ListEquals;
import drr.enrichment.common.EnrichmentData;
import drr.enrichment.common.EnrichmentData.EnrichmentDataBuilder;
import drr.enrichment.common.EnrichmentData.EnrichmentDataBuilderImpl;
import drr.enrichment.common.EnrichmentData.EnrichmentDataImpl;
import drr.enrichment.common.meta.EnrichmentDataMeta;
import drr.enrichment.lei.LeiData;
import drr.standards.iso.MicData;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import static java.util.Optional.ofNullable;

/**
 * Optionally populated by the pre-enrichment process.
 * @version 4.2.1
 */
@RosettaDataType(value="EnrichmentData", builder=EnrichmentData.EnrichmentDataBuilderImpl.class, version="4.2.1")
public interface EnrichmentData extends RosettaModelObject {

	EnrichmentDataMeta metaData = new EnrichmentDataMeta();

	/*********************** Getter Methods  ***********************/
	List<? extends MicData> getMicData();
	List<? extends LeiData> getLeiData();

	/*********************** Build Methods  ***********************/
	EnrichmentData build();
	
	EnrichmentData.EnrichmentDataBuilder toBuilder();
	
	static EnrichmentData.EnrichmentDataBuilder builder() {
		return new EnrichmentData.EnrichmentDataBuilderImpl();
	}

	/*********************** Utility Methods  ***********************/
	@Override
	default RosettaMetaData<? extends EnrichmentData> metaData() {
		return metaData;
	}
	
	@Override
	default Class<? extends EnrichmentData> getType() {
		return EnrichmentData.class;
	}
	
	
	@Override
	default void process(RosettaPath path, Processor processor) {
		processRosetta(path.newSubPath("micData"), processor, MicData.class, getMicData());
		processRosetta(path.newSubPath("leiData"), processor, LeiData.class, getLeiData());
	}
	

	/*********************** Builder Interface  ***********************/
	interface EnrichmentDataBuilder extends EnrichmentData, RosettaModelObjectBuilder {
		MicData.MicDataBuilder getOrCreateMicData(int _index);
		List<? extends MicData.MicDataBuilder> getMicData();
		LeiData.LeiDataBuilder getOrCreateLeiData(int _index);
		List<? extends LeiData.LeiDataBuilder> getLeiData();
		EnrichmentData.EnrichmentDataBuilder addMicData(MicData micData0);
		EnrichmentData.EnrichmentDataBuilder addMicData(MicData micData1, int _idx);
		EnrichmentData.EnrichmentDataBuilder addMicData(List<? extends MicData> micData2);
		EnrichmentData.EnrichmentDataBuilder setMicData(List<? extends MicData> micData3);
		EnrichmentData.EnrichmentDataBuilder addLeiData(LeiData leiData0);
		EnrichmentData.EnrichmentDataBuilder addLeiData(LeiData leiData1, int _idx);
		EnrichmentData.EnrichmentDataBuilder addLeiData(List<? extends LeiData> leiData2);
		EnrichmentData.EnrichmentDataBuilder setLeiData(List<? extends LeiData> leiData3);

		@Override
		default void process(RosettaPath path, BuilderProcessor processor) {
			processRosetta(path.newSubPath("micData"), processor, MicData.MicDataBuilder.class, getMicData());
			processRosetta(path.newSubPath("leiData"), processor, LeiData.LeiDataBuilder.class, getLeiData());
		}
		

		EnrichmentData.EnrichmentDataBuilder prune();
	}

	/*********************** Immutable Implementation of EnrichmentData  ***********************/
	class EnrichmentDataImpl implements EnrichmentData {
		private final List<? extends MicData> micData;
		private final List<? extends LeiData> leiData;
		
		protected EnrichmentDataImpl(EnrichmentData.EnrichmentDataBuilder builder) {
			this.micData = ofNullable(builder.getMicData()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.leiData = ofNullable(builder.getLeiData()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
		}
		
		@Override
		@RosettaAttribute("micData")
		public List<? extends MicData> getMicData() {
			return micData;
		}
		
		@Override
		@RosettaAttribute("leiData")
		public List<? extends LeiData> getLeiData() {
			return leiData;
		}
		
		@Override
		public EnrichmentData build() {
			return this;
		}
		
		@Override
		public EnrichmentData.EnrichmentDataBuilder toBuilder() {
			EnrichmentData.EnrichmentDataBuilder builder = builder();
			setBuilderFields(builder);
			return builder;
		}
		
		protected void setBuilderFields(EnrichmentData.EnrichmentDataBuilder builder) {
			ofNullable(getMicData()).ifPresent(builder::setMicData);
			ofNullable(getLeiData()).ifPresent(builder::setLeiData);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			EnrichmentData _that = getType().cast(o);
		
			if (!ListEquals.listEquals(micData, _that.getMicData())) return false;
			if (!ListEquals.listEquals(leiData, _that.getLeiData())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (micData != null ? micData.hashCode() : 0);
			_result = 31 * _result + (leiData != null ? leiData.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "EnrichmentData {" +
				"micData=" + this.micData + ", " +
				"leiData=" + this.leiData +
			'}';
		}
	}

	/*********************** Builder Implementation of EnrichmentData  ***********************/
	class EnrichmentDataBuilderImpl implements EnrichmentData.EnrichmentDataBuilder {
	
		protected List<MicData.MicDataBuilder> micData = new ArrayList<>();
		protected List<LeiData.LeiDataBuilder> leiData = new ArrayList<>();
	
		public EnrichmentDataBuilderImpl() {
		}
	
		@Override
		@RosettaAttribute("micData")
		public List<? extends MicData.MicDataBuilder> getMicData() {
			return micData;
		}
		
		public MicData.MicDataBuilder getOrCreateMicData(int _index) {
		
			if (micData==null) {
				this.micData = new ArrayList<>();
			}
			MicData.MicDataBuilder result;
			return getIndex(micData, _index, () -> {
						MicData.MicDataBuilder newMicData = MicData.builder();
						return newMicData;
					});
		}
		
		@Override
		@RosettaAttribute("leiData")
		public List<? extends LeiData.LeiDataBuilder> getLeiData() {
			return leiData;
		}
		
		public LeiData.LeiDataBuilder getOrCreateLeiData(int _index) {
		
			if (leiData==null) {
				this.leiData = new ArrayList<>();
			}
			LeiData.LeiDataBuilder result;
			return getIndex(leiData, _index, () -> {
						LeiData.LeiDataBuilder newLeiData = LeiData.builder();
						return newLeiData;
					});
		}
		
	
		@Override
		public EnrichmentData.EnrichmentDataBuilder addMicData(MicData micData) {
			if (micData!=null) this.micData.add(micData.toBuilder());
			return this;
		}
		
		@Override
		public EnrichmentData.EnrichmentDataBuilder addMicData(MicData micData, int _idx) {
			getIndex(this.micData, _idx, () -> micData.toBuilder());
			return this;
		}
		@Override 
		public EnrichmentData.EnrichmentDataBuilder addMicData(List<? extends MicData> micDatas) {
			if (micDatas != null) {
				for (MicData toAdd : micDatas) {
					this.micData.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("micData")
		public EnrichmentData.EnrichmentDataBuilder setMicData(List<? extends MicData> micDatas) {
			if (micDatas == null)  {
				this.micData = new ArrayList<>();
			}
			else {
				this.micData = micDatas.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		public EnrichmentData.EnrichmentDataBuilder addLeiData(LeiData leiData) {
			if (leiData!=null) this.leiData.add(leiData.toBuilder());
			return this;
		}
		
		@Override
		public EnrichmentData.EnrichmentDataBuilder addLeiData(LeiData leiData, int _idx) {
			getIndex(this.leiData, _idx, () -> leiData.toBuilder());
			return this;
		}
		@Override 
		public EnrichmentData.EnrichmentDataBuilder addLeiData(List<? extends LeiData> leiDatas) {
			if (leiDatas != null) {
				for (LeiData toAdd : leiDatas) {
					this.leiData.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("leiData")
		public EnrichmentData.EnrichmentDataBuilder setLeiData(List<? extends LeiData> leiDatas) {
			if (leiDatas == null)  {
				this.leiData = new ArrayList<>();
			}
			else {
				this.leiData = leiDatas.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		
		@Override
		public EnrichmentData build() {
			return new EnrichmentData.EnrichmentDataImpl(this);
		}
		
		@Override
		public EnrichmentData.EnrichmentDataBuilder toBuilder() {
			return this;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public EnrichmentData.EnrichmentDataBuilder prune() {
			micData = micData.stream().filter(b->b!=null).<MicData.MicDataBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			leiData = leiData.stream().filter(b->b!=null).<LeiData.LeiDataBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			return this;
		}
		
		@Override
		public boolean hasData() {
			if (getMicData()!=null && getMicData().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getLeiData()!=null && getLeiData().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			return false;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public EnrichmentData.EnrichmentDataBuilder merge(RosettaModelObjectBuilder other, BuilderMerger merger) {
			EnrichmentData.EnrichmentDataBuilder o = (EnrichmentData.EnrichmentDataBuilder) other;
			
			merger.mergeRosetta(getMicData(), o.getMicData(), this::getOrCreateMicData);
			merger.mergeRosetta(getLeiData(), o.getLeiData(), this::getOrCreateLeiData);
			
			return this;
		}
	
		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			EnrichmentData _that = getType().cast(o);
		
			if (!ListEquals.listEquals(micData, _that.getMicData())) return false;
			if (!ListEquals.listEquals(leiData, _that.getLeiData())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (micData != null ? micData.hashCode() : 0);
			_result = 31 * _result + (leiData != null ? leiData.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "EnrichmentDataBuilder {" +
				"micData=" + this.micData + ", " +
				"leiData=" + this.leiData +
			'}';
		}
	}
}
