package drr.enrichment.lei;

import com.rosetta.model.lib.RosettaModelObject;
import com.rosetta.model.lib.RosettaModelObjectBuilder;
import com.rosetta.model.lib.annotations.RosettaAttribute;
import com.rosetta.model.lib.annotations.RosettaDataType;
import com.rosetta.model.lib.meta.RosettaMetaData;
import com.rosetta.model.lib.path.RosettaPath;
import com.rosetta.model.lib.process.BuilderMerger;
import com.rosetta.model.lib.process.BuilderProcessor;
import com.rosetta.model.lib.process.Processor;
import drr.enrichment.lei.LeiCategoryEnum;
import drr.enrichment.lei.LeiData;
import drr.enrichment.lei.LeiData.LeiDataBuilder;
import drr.enrichment.lei.LeiData.LeiDataBuilderImpl;
import drr.enrichment.lei.LeiData.LeiDataImpl;
import drr.enrichment.lei.LeiGeneralCategoryEnum;
import drr.enrichment.lei.LeiRegistrationStatusEnum;
import drr.enrichment.lei.LeiStatusEnum;
import drr.enrichment.lei.meta.LeiDataMeta;
import java.time.ZonedDateTime;
import java.util.Objects;

import static java.util.Optional.ofNullable;

/**
 * LEI data from GLEIF.
 * @version 4.2.1
 */
@RosettaDataType(value="LeiData", builder=LeiData.LeiDataBuilderImpl.class, version="4.2.1")
public interface LeiData extends RosettaModelObject {

	LeiDataMeta metaData = new LeiDataMeta();

	/*********************** Getter Methods  ***********************/
	/**
	 * LEI identifier.
	 */
	String getLei();
	/**
	 * LEI legal entity name.
	 */
	String getEntityName();
	/**
	 * LEI legal entity category.
	 */
	LeiCategoryEnum getEntityCategory();
	/**
	 * LEI legal entity status.
	 */
	LeiStatusEnum getEntityStatus();
	/**
	 * LEI branch entity status.
	 */
	LeiStatusEnum getBranchEntityStatus();
	LeiGeneralCategoryEnum getGeneralCategory();
	/**
	 * LEI registration status.
	 */
	LeiRegistrationStatusEnum getRegistrationStatus();
	/**
	 * LEI registration date.
	 */
	ZonedDateTime getRegistrationDate();
	/**
	 * The LEI is included in the LEI data as published by GLEIF.
	 */
	Boolean getPublished();

	/*********************** Build Methods  ***********************/
	LeiData build();
	
	LeiData.LeiDataBuilder toBuilder();
	
	static LeiData.LeiDataBuilder builder() {
		return new LeiData.LeiDataBuilderImpl();
	}

	/*********************** Utility Methods  ***********************/
	@Override
	default RosettaMetaData<? extends LeiData> metaData() {
		return metaData;
	}
	
	@Override
	default Class<? extends LeiData> getType() {
		return LeiData.class;
	}
	
	
	@Override
	default void process(RosettaPath path, Processor processor) {
		processor.processBasic(path.newSubPath("lei"), String.class, getLei(), this);
		processor.processBasic(path.newSubPath("entityName"), String.class, getEntityName(), this);
		processor.processBasic(path.newSubPath("entityCategory"), LeiCategoryEnum.class, getEntityCategory(), this);
		processor.processBasic(path.newSubPath("entityStatus"), LeiStatusEnum.class, getEntityStatus(), this);
		processor.processBasic(path.newSubPath("branchEntityStatus"), LeiStatusEnum.class, getBranchEntityStatus(), this);
		processor.processBasic(path.newSubPath("generalCategory"), LeiGeneralCategoryEnum.class, getGeneralCategory(), this);
		processor.processBasic(path.newSubPath("registrationStatus"), LeiRegistrationStatusEnum.class, getRegistrationStatus(), this);
		processor.processBasic(path.newSubPath("registrationDate"), ZonedDateTime.class, getRegistrationDate(), this);
		processor.processBasic(path.newSubPath("published"), Boolean.class, getPublished(), this);
	}
	

	/*********************** Builder Interface  ***********************/
	interface LeiDataBuilder extends LeiData, RosettaModelObjectBuilder {
		LeiData.LeiDataBuilder setLei(String lei);
		LeiData.LeiDataBuilder setEntityName(String entityName);
		LeiData.LeiDataBuilder setEntityCategory(LeiCategoryEnum entityCategory);
		LeiData.LeiDataBuilder setEntityStatus(LeiStatusEnum entityStatus);
		LeiData.LeiDataBuilder setBranchEntityStatus(LeiStatusEnum branchEntityStatus);
		LeiData.LeiDataBuilder setGeneralCategory(LeiGeneralCategoryEnum generalCategory);
		LeiData.LeiDataBuilder setRegistrationStatus(LeiRegistrationStatusEnum registrationStatus);
		LeiData.LeiDataBuilder setRegistrationDate(ZonedDateTime registrationDate);
		LeiData.LeiDataBuilder setPublished(Boolean published);

		@Override
		default void process(RosettaPath path, BuilderProcessor processor) {
			processor.processBasic(path.newSubPath("lei"), String.class, getLei(), this);
			processor.processBasic(path.newSubPath("entityName"), String.class, getEntityName(), this);
			processor.processBasic(path.newSubPath("entityCategory"), LeiCategoryEnum.class, getEntityCategory(), this);
			processor.processBasic(path.newSubPath("entityStatus"), LeiStatusEnum.class, getEntityStatus(), this);
			processor.processBasic(path.newSubPath("branchEntityStatus"), LeiStatusEnum.class, getBranchEntityStatus(), this);
			processor.processBasic(path.newSubPath("generalCategory"), LeiGeneralCategoryEnum.class, getGeneralCategory(), this);
			processor.processBasic(path.newSubPath("registrationStatus"), LeiRegistrationStatusEnum.class, getRegistrationStatus(), this);
			processor.processBasic(path.newSubPath("registrationDate"), ZonedDateTime.class, getRegistrationDate(), this);
			processor.processBasic(path.newSubPath("published"), Boolean.class, getPublished(), this);
		}
		

		LeiData.LeiDataBuilder prune();
	}

	/*********************** Immutable Implementation of LeiData  ***********************/
	class LeiDataImpl implements LeiData {
		private final String lei;
		private final String entityName;
		private final LeiCategoryEnum entityCategory;
		private final LeiStatusEnum entityStatus;
		private final LeiStatusEnum branchEntityStatus;
		private final LeiGeneralCategoryEnum generalCategory;
		private final LeiRegistrationStatusEnum registrationStatus;
		private final ZonedDateTime registrationDate;
		private final Boolean published;
		
		protected LeiDataImpl(LeiData.LeiDataBuilder builder) {
			this.lei = builder.getLei();
			this.entityName = builder.getEntityName();
			this.entityCategory = builder.getEntityCategory();
			this.entityStatus = builder.getEntityStatus();
			this.branchEntityStatus = builder.getBranchEntityStatus();
			this.generalCategory = builder.getGeneralCategory();
			this.registrationStatus = builder.getRegistrationStatus();
			this.registrationDate = builder.getRegistrationDate();
			this.published = builder.getPublished();
		}
		
		@Override
		@RosettaAttribute("lei")
		public String getLei() {
			return lei;
		}
		
		@Override
		@RosettaAttribute("entityName")
		public String getEntityName() {
			return entityName;
		}
		
		@Override
		@RosettaAttribute("entityCategory")
		public LeiCategoryEnum getEntityCategory() {
			return entityCategory;
		}
		
		@Override
		@RosettaAttribute("entityStatus")
		public LeiStatusEnum getEntityStatus() {
			return entityStatus;
		}
		
		@Override
		@RosettaAttribute("branchEntityStatus")
		public LeiStatusEnum getBranchEntityStatus() {
			return branchEntityStatus;
		}
		
		@Override
		@RosettaAttribute("generalCategory")
		public LeiGeneralCategoryEnum getGeneralCategory() {
			return generalCategory;
		}
		
		@Override
		@RosettaAttribute("registrationStatus")
		public LeiRegistrationStatusEnum getRegistrationStatus() {
			return registrationStatus;
		}
		
		@Override
		@RosettaAttribute("registrationDate")
		public ZonedDateTime getRegistrationDate() {
			return registrationDate;
		}
		
		@Override
		@RosettaAttribute("published")
		public Boolean getPublished() {
			return published;
		}
		
		@Override
		public LeiData build() {
			return this;
		}
		
		@Override
		public LeiData.LeiDataBuilder toBuilder() {
			LeiData.LeiDataBuilder builder = builder();
			setBuilderFields(builder);
			return builder;
		}
		
		protected void setBuilderFields(LeiData.LeiDataBuilder builder) {
			ofNullable(getLei()).ifPresent(builder::setLei);
			ofNullable(getEntityName()).ifPresent(builder::setEntityName);
			ofNullable(getEntityCategory()).ifPresent(builder::setEntityCategory);
			ofNullable(getEntityStatus()).ifPresent(builder::setEntityStatus);
			ofNullable(getBranchEntityStatus()).ifPresent(builder::setBranchEntityStatus);
			ofNullable(getGeneralCategory()).ifPresent(builder::setGeneralCategory);
			ofNullable(getRegistrationStatus()).ifPresent(builder::setRegistrationStatus);
			ofNullable(getRegistrationDate()).ifPresent(builder::setRegistrationDate);
			ofNullable(getPublished()).ifPresent(builder::setPublished);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			LeiData _that = getType().cast(o);
		
			if (!Objects.equals(lei, _that.getLei())) return false;
			if (!Objects.equals(entityName, _that.getEntityName())) return false;
			if (!Objects.equals(entityCategory, _that.getEntityCategory())) return false;
			if (!Objects.equals(entityStatus, _that.getEntityStatus())) return false;
			if (!Objects.equals(branchEntityStatus, _that.getBranchEntityStatus())) return false;
			if (!Objects.equals(generalCategory, _that.getGeneralCategory())) return false;
			if (!Objects.equals(registrationStatus, _that.getRegistrationStatus())) return false;
			if (!Objects.equals(registrationDate, _that.getRegistrationDate())) return false;
			if (!Objects.equals(published, _that.getPublished())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (lei != null ? lei.hashCode() : 0);
			_result = 31 * _result + (entityName != null ? entityName.hashCode() : 0);
			_result = 31 * _result + (entityCategory != null ? entityCategory.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (entityStatus != null ? entityStatus.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (branchEntityStatus != null ? branchEntityStatus.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (generalCategory != null ? generalCategory.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (registrationStatus != null ? registrationStatus.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (registrationDate != null ? registrationDate.hashCode() : 0);
			_result = 31 * _result + (published != null ? published.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "LeiData {" +
				"lei=" + this.lei + ", " +
				"entityName=" + this.entityName + ", " +
				"entityCategory=" + this.entityCategory + ", " +
				"entityStatus=" + this.entityStatus + ", " +
				"branchEntityStatus=" + this.branchEntityStatus + ", " +
				"generalCategory=" + this.generalCategory + ", " +
				"registrationStatus=" + this.registrationStatus + ", " +
				"registrationDate=" + this.registrationDate + ", " +
				"published=" + this.published +
			'}';
		}
	}

	/*********************** Builder Implementation of LeiData  ***********************/
	class LeiDataBuilderImpl implements LeiData.LeiDataBuilder {
	
		protected String lei;
		protected String entityName;
		protected LeiCategoryEnum entityCategory;
		protected LeiStatusEnum entityStatus;
		protected LeiStatusEnum branchEntityStatus;
		protected LeiGeneralCategoryEnum generalCategory;
		protected LeiRegistrationStatusEnum registrationStatus;
		protected ZonedDateTime registrationDate;
		protected Boolean published;
	
		public LeiDataBuilderImpl() {
		}
	
		@Override
		@RosettaAttribute("lei")
		public String getLei() {
			return lei;
		}
		
		@Override
		@RosettaAttribute("entityName")
		public String getEntityName() {
			return entityName;
		}
		
		@Override
		@RosettaAttribute("entityCategory")
		public LeiCategoryEnum getEntityCategory() {
			return entityCategory;
		}
		
		@Override
		@RosettaAttribute("entityStatus")
		public LeiStatusEnum getEntityStatus() {
			return entityStatus;
		}
		
		@Override
		@RosettaAttribute("branchEntityStatus")
		public LeiStatusEnum getBranchEntityStatus() {
			return branchEntityStatus;
		}
		
		@Override
		@RosettaAttribute("generalCategory")
		public LeiGeneralCategoryEnum getGeneralCategory() {
			return generalCategory;
		}
		
		@Override
		@RosettaAttribute("registrationStatus")
		public LeiRegistrationStatusEnum getRegistrationStatus() {
			return registrationStatus;
		}
		
		@Override
		@RosettaAttribute("registrationDate")
		public ZonedDateTime getRegistrationDate() {
			return registrationDate;
		}
		
		@Override
		@RosettaAttribute("published")
		public Boolean getPublished() {
			return published;
		}
		
	
		@Override
		@RosettaAttribute("lei")
		public LeiData.LeiDataBuilder setLei(String lei) {
			this.lei = lei==null?null:lei;
			return this;
		}
		@Override
		@RosettaAttribute("entityName")
		public LeiData.LeiDataBuilder setEntityName(String entityName) {
			this.entityName = entityName==null?null:entityName;
			return this;
		}
		@Override
		@RosettaAttribute("entityCategory")
		public LeiData.LeiDataBuilder setEntityCategory(LeiCategoryEnum entityCategory) {
			this.entityCategory = entityCategory==null?null:entityCategory;
			return this;
		}
		@Override
		@RosettaAttribute("entityStatus")
		public LeiData.LeiDataBuilder setEntityStatus(LeiStatusEnum entityStatus) {
			this.entityStatus = entityStatus==null?null:entityStatus;
			return this;
		}
		@Override
		@RosettaAttribute("branchEntityStatus")
		public LeiData.LeiDataBuilder setBranchEntityStatus(LeiStatusEnum branchEntityStatus) {
			this.branchEntityStatus = branchEntityStatus==null?null:branchEntityStatus;
			return this;
		}
		@Override
		@RosettaAttribute("generalCategory")
		public LeiData.LeiDataBuilder setGeneralCategory(LeiGeneralCategoryEnum generalCategory) {
			this.generalCategory = generalCategory==null?null:generalCategory;
			return this;
		}
		@Override
		@RosettaAttribute("registrationStatus")
		public LeiData.LeiDataBuilder setRegistrationStatus(LeiRegistrationStatusEnum registrationStatus) {
			this.registrationStatus = registrationStatus==null?null:registrationStatus;
			return this;
		}
		@Override
		@RosettaAttribute("registrationDate")
		public LeiData.LeiDataBuilder setRegistrationDate(ZonedDateTime registrationDate) {
			this.registrationDate = registrationDate==null?null:registrationDate;
			return this;
		}
		@Override
		@RosettaAttribute("published")
		public LeiData.LeiDataBuilder setPublished(Boolean published) {
			this.published = published==null?null:published;
			return this;
		}
		
		@Override
		public LeiData build() {
			return new LeiData.LeiDataImpl(this);
		}
		
		@Override
		public LeiData.LeiDataBuilder toBuilder() {
			return this;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public LeiData.LeiDataBuilder prune() {
			return this;
		}
		
		@Override
		public boolean hasData() {
			if (getLei()!=null) return true;
			if (getEntityName()!=null) return true;
			if (getEntityCategory()!=null) return true;
			if (getEntityStatus()!=null) return true;
			if (getBranchEntityStatus()!=null) return true;
			if (getGeneralCategory()!=null) return true;
			if (getRegistrationStatus()!=null) return true;
			if (getRegistrationDate()!=null) return true;
			if (getPublished()!=null) return true;
			return false;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public LeiData.LeiDataBuilder merge(RosettaModelObjectBuilder other, BuilderMerger merger) {
			LeiData.LeiDataBuilder o = (LeiData.LeiDataBuilder) other;
			
			
			merger.mergeBasic(getLei(), o.getLei(), this::setLei);
			merger.mergeBasic(getEntityName(), o.getEntityName(), this::setEntityName);
			merger.mergeBasic(getEntityCategory(), o.getEntityCategory(), this::setEntityCategory);
			merger.mergeBasic(getEntityStatus(), o.getEntityStatus(), this::setEntityStatus);
			merger.mergeBasic(getBranchEntityStatus(), o.getBranchEntityStatus(), this::setBranchEntityStatus);
			merger.mergeBasic(getGeneralCategory(), o.getGeneralCategory(), this::setGeneralCategory);
			merger.mergeBasic(getRegistrationStatus(), o.getRegistrationStatus(), this::setRegistrationStatus);
			merger.mergeBasic(getRegistrationDate(), o.getRegistrationDate(), this::setRegistrationDate);
			merger.mergeBasic(getPublished(), o.getPublished(), this::setPublished);
			return this;
		}
	
		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			LeiData _that = getType().cast(o);
		
			if (!Objects.equals(lei, _that.getLei())) return false;
			if (!Objects.equals(entityName, _that.getEntityName())) return false;
			if (!Objects.equals(entityCategory, _that.getEntityCategory())) return false;
			if (!Objects.equals(entityStatus, _that.getEntityStatus())) return false;
			if (!Objects.equals(branchEntityStatus, _that.getBranchEntityStatus())) return false;
			if (!Objects.equals(generalCategory, _that.getGeneralCategory())) return false;
			if (!Objects.equals(registrationStatus, _that.getRegistrationStatus())) return false;
			if (!Objects.equals(registrationDate, _that.getRegistrationDate())) return false;
			if (!Objects.equals(published, _that.getPublished())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (lei != null ? lei.hashCode() : 0);
			_result = 31 * _result + (entityName != null ? entityName.hashCode() : 0);
			_result = 31 * _result + (entityCategory != null ? entityCategory.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (entityStatus != null ? entityStatus.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (branchEntityStatus != null ? branchEntityStatus.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (generalCategory != null ? generalCategory.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (registrationStatus != null ? registrationStatus.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (registrationDate != null ? registrationDate.hashCode() : 0);
			_result = 31 * _result + (published != null ? published.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "LeiDataBuilder {" +
				"lei=" + this.lei + ", " +
				"entityName=" + this.entityName + ", " +
				"entityCategory=" + this.entityCategory + ", " +
				"entityStatus=" + this.entityStatus + ", " +
				"branchEntityStatus=" + this.branchEntityStatus + ", " +
				"generalCategory=" + this.generalCategory + ", " +
				"registrationStatus=" + this.registrationStatus + ", " +
				"registrationDate=" + this.registrationDate + ", " +
				"published=" + this.published +
			'}';
		}
	}
}
