package drr.regulation.common;

import cdm.base.staticdata.party.NaturalPersonRole;
import cdm.base.staticdata.party.Party;
import cdm.base.staticdata.party.RelatedParty;
import cdm.base.staticdata.party.metafields.ReferenceWithMetaParty;
import cdm.base.staticdata.party.metafields.ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder;
import com.google.common.collect.ImmutableList;
import com.rosetta.model.lib.RosettaModelObject;
import com.rosetta.model.lib.RosettaModelObjectBuilder;
import com.rosetta.model.lib.annotations.RosettaAttribute;
import com.rosetta.model.lib.annotations.RosettaDataType;
import com.rosetta.model.lib.meta.RosettaMetaData;
import com.rosetta.model.lib.path.RosettaPath;
import com.rosetta.model.lib.process.BuilderMerger;
import com.rosetta.model.lib.process.BuilderProcessor;
import com.rosetta.model.lib.process.Processor;
import com.rosetta.util.ListEquals;
import drr.regulation.common.PartyInformation;
import drr.regulation.common.PartyInformation.PartyInformationBuilder;
import drr.regulation.common.PartyInformation.PartyInformationBuilderImpl;
import drr.regulation.common.PartyInformation.PartyInformationImpl;
import drr.regulation.common.ReportingRegime;
import drr.regulation.common.meta.PartyInformationMeta;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import static java.util.Optional.ofNullable;

/**
 * Specifies party specific information required for reporting of the transaction
 * @version 4.2.1
 */
@RosettaDataType(value="PartyInformation", builder=PartyInformation.PartyInformationBuilderImpl.class, version="4.2.1")
public interface PartyInformation extends RosettaModelObject {

	PartyInformationMeta metaData = new PartyInformationMeta();

	/*********************** Getter Methods  ***********************/
	/**
	 * Specifies the party that is associated with the enriched information.
	 */
	ReferenceWithMetaParty getPartyReference();
	/**
	 * The role(s) that natural person(s) may have in relation to the transaction.
	 */
	List<? extends NaturalPersonRole> getRelatedPerson();
	/**
	 * Specifies one or more parties that perform a role within the transaction. The related party performs the role with respect to the party identified by the &#39;partyReference&#39;.
	 */
	List<? extends RelatedParty> getRelatedParty();
	/**
	 * Specifies the regulatory regime requiring reporting of the transaction. A list of regime information is provided when an event is reportable to more than one regime.
	 */
	List<? extends ReportingRegime> getRegimeInformation();

	/*********************** Build Methods  ***********************/
	PartyInformation build();
	
	PartyInformation.PartyInformationBuilder toBuilder();
	
	static PartyInformation.PartyInformationBuilder builder() {
		return new PartyInformation.PartyInformationBuilderImpl();
	}

	/*********************** Utility Methods  ***********************/
	@Override
	default RosettaMetaData<? extends PartyInformation> metaData() {
		return metaData;
	}
	
	@Override
	default Class<? extends PartyInformation> getType() {
		return PartyInformation.class;
	}
	
	
	@Override
	default void process(RosettaPath path, Processor processor) {
		processRosetta(path.newSubPath("partyReference"), processor, ReferenceWithMetaParty.class, getPartyReference());
		processRosetta(path.newSubPath("relatedPerson"), processor, NaturalPersonRole.class, getRelatedPerson());
		processRosetta(path.newSubPath("relatedParty"), processor, RelatedParty.class, getRelatedParty());
		processRosetta(path.newSubPath("regimeInformation"), processor, ReportingRegime.class, getRegimeInformation());
	}
	

	/*********************** Builder Interface  ***********************/
	interface PartyInformationBuilder extends PartyInformation, RosettaModelObjectBuilder {
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreatePartyReference();
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getPartyReference();
		NaturalPersonRole.NaturalPersonRoleBuilder getOrCreateRelatedPerson(int _index);
		List<? extends NaturalPersonRole.NaturalPersonRoleBuilder> getRelatedPerson();
		RelatedParty.RelatedPartyBuilder getOrCreateRelatedParty(int _index);
		List<? extends RelatedParty.RelatedPartyBuilder> getRelatedParty();
		ReportingRegime.ReportingRegimeBuilder getOrCreateRegimeInformation(int _index);
		List<? extends ReportingRegime.ReportingRegimeBuilder> getRegimeInformation();
		PartyInformation.PartyInformationBuilder setPartyReference(ReferenceWithMetaParty partyReference0);
		PartyInformation.PartyInformationBuilder setPartyReferenceValue(Party partyReference1);
		PartyInformation.PartyInformationBuilder addRelatedPerson(NaturalPersonRole relatedPerson0);
		PartyInformation.PartyInformationBuilder addRelatedPerson(NaturalPersonRole relatedPerson1, int _idx);
		PartyInformation.PartyInformationBuilder addRelatedPerson(List<? extends NaturalPersonRole> relatedPerson2);
		PartyInformation.PartyInformationBuilder setRelatedPerson(List<? extends NaturalPersonRole> relatedPerson3);
		PartyInformation.PartyInformationBuilder addRelatedParty(RelatedParty relatedParty0);
		PartyInformation.PartyInformationBuilder addRelatedParty(RelatedParty relatedParty1, int _idx);
		PartyInformation.PartyInformationBuilder addRelatedParty(List<? extends RelatedParty> relatedParty2);
		PartyInformation.PartyInformationBuilder setRelatedParty(List<? extends RelatedParty> relatedParty3);
		PartyInformation.PartyInformationBuilder addRegimeInformation(ReportingRegime regimeInformation0);
		PartyInformation.PartyInformationBuilder addRegimeInformation(ReportingRegime regimeInformation1, int _idx);
		PartyInformation.PartyInformationBuilder addRegimeInformation(List<? extends ReportingRegime> regimeInformation2);
		PartyInformation.PartyInformationBuilder setRegimeInformation(List<? extends ReportingRegime> regimeInformation3);

		@Override
		default void process(RosettaPath path, BuilderProcessor processor) {
			processRosetta(path.newSubPath("partyReference"), processor, ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder.class, getPartyReference());
			processRosetta(path.newSubPath("relatedPerson"), processor, NaturalPersonRole.NaturalPersonRoleBuilder.class, getRelatedPerson());
			processRosetta(path.newSubPath("relatedParty"), processor, RelatedParty.RelatedPartyBuilder.class, getRelatedParty());
			processRosetta(path.newSubPath("regimeInformation"), processor, ReportingRegime.ReportingRegimeBuilder.class, getRegimeInformation());
		}
		

		PartyInformation.PartyInformationBuilder prune();
	}

	/*********************** Immutable Implementation of PartyInformation  ***********************/
	class PartyInformationImpl implements PartyInformation {
		private final ReferenceWithMetaParty partyReference;
		private final List<? extends NaturalPersonRole> relatedPerson;
		private final List<? extends RelatedParty> relatedParty;
		private final List<? extends ReportingRegime> regimeInformation;
		
		protected PartyInformationImpl(PartyInformation.PartyInformationBuilder builder) {
			this.partyReference = ofNullable(builder.getPartyReference()).map(f->f.build()).orElse(null);
			this.relatedPerson = ofNullable(builder.getRelatedPerson()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.relatedParty = ofNullable(builder.getRelatedParty()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.regimeInformation = ofNullable(builder.getRegimeInformation()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
		}
		
		@Override
		@RosettaAttribute("partyReference")
		public ReferenceWithMetaParty getPartyReference() {
			return partyReference;
		}
		
		@Override
		@RosettaAttribute("relatedPerson")
		public List<? extends NaturalPersonRole> getRelatedPerson() {
			return relatedPerson;
		}
		
		@Override
		@RosettaAttribute("relatedParty")
		public List<? extends RelatedParty> getRelatedParty() {
			return relatedParty;
		}
		
		@Override
		@RosettaAttribute("regimeInformation")
		public List<? extends ReportingRegime> getRegimeInformation() {
			return regimeInformation;
		}
		
		@Override
		public PartyInformation build() {
			return this;
		}
		
		@Override
		public PartyInformation.PartyInformationBuilder toBuilder() {
			PartyInformation.PartyInformationBuilder builder = builder();
			setBuilderFields(builder);
			return builder;
		}
		
		protected void setBuilderFields(PartyInformation.PartyInformationBuilder builder) {
			ofNullable(getPartyReference()).ifPresent(builder::setPartyReference);
			ofNullable(getRelatedPerson()).ifPresent(builder::setRelatedPerson);
			ofNullable(getRelatedParty()).ifPresent(builder::setRelatedParty);
			ofNullable(getRegimeInformation()).ifPresent(builder::setRegimeInformation);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			PartyInformation _that = getType().cast(o);
		
			if (!Objects.equals(partyReference, _that.getPartyReference())) return false;
			if (!ListEquals.listEquals(relatedPerson, _that.getRelatedPerson())) return false;
			if (!ListEquals.listEquals(relatedParty, _that.getRelatedParty())) return false;
			if (!ListEquals.listEquals(regimeInformation, _that.getRegimeInformation())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (partyReference != null ? partyReference.hashCode() : 0);
			_result = 31 * _result + (relatedPerson != null ? relatedPerson.hashCode() : 0);
			_result = 31 * _result + (relatedParty != null ? relatedParty.hashCode() : 0);
			_result = 31 * _result + (regimeInformation != null ? regimeInformation.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "PartyInformation {" +
				"partyReference=" + this.partyReference + ", " +
				"relatedPerson=" + this.relatedPerson + ", " +
				"relatedParty=" + this.relatedParty + ", " +
				"regimeInformation=" + this.regimeInformation +
			'}';
		}
	}

	/*********************** Builder Implementation of PartyInformation  ***********************/
	class PartyInformationBuilderImpl implements PartyInformation.PartyInformationBuilder {
	
		protected ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder partyReference;
		protected List<NaturalPersonRole.NaturalPersonRoleBuilder> relatedPerson = new ArrayList<>();
		protected List<RelatedParty.RelatedPartyBuilder> relatedParty = new ArrayList<>();
		protected List<ReportingRegime.ReportingRegimeBuilder> regimeInformation = new ArrayList<>();
	
		public PartyInformationBuilderImpl() {
		}
	
		@Override
		@RosettaAttribute("partyReference")
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getPartyReference() {
			return partyReference;
		}
		
		@Override
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreatePartyReference() {
			ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder result;
			if (partyReference!=null) {
				result = partyReference;
			}
			else {
				result = partyReference = ReferenceWithMetaParty.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("relatedPerson")
		public List<? extends NaturalPersonRole.NaturalPersonRoleBuilder> getRelatedPerson() {
			return relatedPerson;
		}
		
		public NaturalPersonRole.NaturalPersonRoleBuilder getOrCreateRelatedPerson(int _index) {
		
			if (relatedPerson==null) {
				this.relatedPerson = new ArrayList<>();
			}
			NaturalPersonRole.NaturalPersonRoleBuilder result;
			return getIndex(relatedPerson, _index, () -> {
						NaturalPersonRole.NaturalPersonRoleBuilder newRelatedPerson = NaturalPersonRole.builder();
						return newRelatedPerson;
					});
		}
		
		@Override
		@RosettaAttribute("relatedParty")
		public List<? extends RelatedParty.RelatedPartyBuilder> getRelatedParty() {
			return relatedParty;
		}
		
		public RelatedParty.RelatedPartyBuilder getOrCreateRelatedParty(int _index) {
		
			if (relatedParty==null) {
				this.relatedParty = new ArrayList<>();
			}
			RelatedParty.RelatedPartyBuilder result;
			return getIndex(relatedParty, _index, () -> {
						RelatedParty.RelatedPartyBuilder newRelatedParty = RelatedParty.builder();
						return newRelatedParty;
					});
		}
		
		@Override
		@RosettaAttribute("regimeInformation")
		public List<? extends ReportingRegime.ReportingRegimeBuilder> getRegimeInformation() {
			return regimeInformation;
		}
		
		public ReportingRegime.ReportingRegimeBuilder getOrCreateRegimeInformation(int _index) {
		
			if (regimeInformation==null) {
				this.regimeInformation = new ArrayList<>();
			}
			ReportingRegime.ReportingRegimeBuilder result;
			return getIndex(regimeInformation, _index, () -> {
						ReportingRegime.ReportingRegimeBuilder newRegimeInformation = ReportingRegime.builder();
						return newRegimeInformation;
					});
		}
		
	
		@Override
		@RosettaAttribute("partyReference")
		public PartyInformation.PartyInformationBuilder setPartyReference(ReferenceWithMetaParty partyReference) {
			this.partyReference = partyReference==null?null:partyReference.toBuilder();
			return this;
		}
		@Override
		public PartyInformation.PartyInformationBuilder setPartyReferenceValue(Party partyReference) {
			this.getOrCreatePartyReference().setValue(partyReference);
			return this;
		}
		@Override
		public PartyInformation.PartyInformationBuilder addRelatedPerson(NaturalPersonRole relatedPerson) {
			if (relatedPerson!=null) this.relatedPerson.add(relatedPerson.toBuilder());
			return this;
		}
		
		@Override
		public PartyInformation.PartyInformationBuilder addRelatedPerson(NaturalPersonRole relatedPerson, int _idx) {
			getIndex(this.relatedPerson, _idx, () -> relatedPerson.toBuilder());
			return this;
		}
		@Override 
		public PartyInformation.PartyInformationBuilder addRelatedPerson(List<? extends NaturalPersonRole> relatedPersons) {
			if (relatedPersons != null) {
				for (NaturalPersonRole toAdd : relatedPersons) {
					this.relatedPerson.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("relatedPerson")
		public PartyInformation.PartyInformationBuilder setRelatedPerson(List<? extends NaturalPersonRole> relatedPersons) {
			if (relatedPersons == null)  {
				this.relatedPerson = new ArrayList<>();
			}
			else {
				this.relatedPerson = relatedPersons.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		public PartyInformation.PartyInformationBuilder addRelatedParty(RelatedParty relatedParty) {
			if (relatedParty!=null) this.relatedParty.add(relatedParty.toBuilder());
			return this;
		}
		
		@Override
		public PartyInformation.PartyInformationBuilder addRelatedParty(RelatedParty relatedParty, int _idx) {
			getIndex(this.relatedParty, _idx, () -> relatedParty.toBuilder());
			return this;
		}
		@Override 
		public PartyInformation.PartyInformationBuilder addRelatedParty(List<? extends RelatedParty> relatedPartys) {
			if (relatedPartys != null) {
				for (RelatedParty toAdd : relatedPartys) {
					this.relatedParty.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("relatedParty")
		public PartyInformation.PartyInformationBuilder setRelatedParty(List<? extends RelatedParty> relatedPartys) {
			if (relatedPartys == null)  {
				this.relatedParty = new ArrayList<>();
			}
			else {
				this.relatedParty = relatedPartys.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		public PartyInformation.PartyInformationBuilder addRegimeInformation(ReportingRegime regimeInformation) {
			if (regimeInformation!=null) this.regimeInformation.add(regimeInformation.toBuilder());
			return this;
		}
		
		@Override
		public PartyInformation.PartyInformationBuilder addRegimeInformation(ReportingRegime regimeInformation, int _idx) {
			getIndex(this.regimeInformation, _idx, () -> regimeInformation.toBuilder());
			return this;
		}
		@Override 
		public PartyInformation.PartyInformationBuilder addRegimeInformation(List<? extends ReportingRegime> regimeInformations) {
			if (regimeInformations != null) {
				for (ReportingRegime toAdd : regimeInformations) {
					this.regimeInformation.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("regimeInformation")
		public PartyInformation.PartyInformationBuilder setRegimeInformation(List<? extends ReportingRegime> regimeInformations) {
			if (regimeInformations == null)  {
				this.regimeInformation = new ArrayList<>();
			}
			else {
				this.regimeInformation = regimeInformations.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		
		@Override
		public PartyInformation build() {
			return new PartyInformation.PartyInformationImpl(this);
		}
		
		@Override
		public PartyInformation.PartyInformationBuilder toBuilder() {
			return this;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public PartyInformation.PartyInformationBuilder prune() {
			if (partyReference!=null && !partyReference.prune().hasData()) partyReference = null;
			relatedPerson = relatedPerson.stream().filter(b->b!=null).<NaturalPersonRole.NaturalPersonRoleBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			relatedParty = relatedParty.stream().filter(b->b!=null).<RelatedParty.RelatedPartyBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			regimeInformation = regimeInformation.stream().filter(b->b!=null).<ReportingRegime.ReportingRegimeBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			return this;
		}
		
		@Override
		public boolean hasData() {
			if (getPartyReference()!=null && getPartyReference().hasData()) return true;
			if (getRelatedPerson()!=null && getRelatedPerson().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getRelatedParty()!=null && getRelatedParty().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getRegimeInformation()!=null && getRegimeInformation().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			return false;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public PartyInformation.PartyInformationBuilder merge(RosettaModelObjectBuilder other, BuilderMerger merger) {
			PartyInformation.PartyInformationBuilder o = (PartyInformation.PartyInformationBuilder) other;
			
			merger.mergeRosetta(getPartyReference(), o.getPartyReference(), this::setPartyReference);
			merger.mergeRosetta(getRelatedPerson(), o.getRelatedPerson(), this::getOrCreateRelatedPerson);
			merger.mergeRosetta(getRelatedParty(), o.getRelatedParty(), this::getOrCreateRelatedParty);
			merger.mergeRosetta(getRegimeInformation(), o.getRegimeInformation(), this::getOrCreateRegimeInformation);
			
			return this;
		}
	
		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			PartyInformation _that = getType().cast(o);
		
			if (!Objects.equals(partyReference, _that.getPartyReference())) return false;
			if (!ListEquals.listEquals(relatedPerson, _that.getRelatedPerson())) return false;
			if (!ListEquals.listEquals(relatedParty, _that.getRelatedParty())) return false;
			if (!ListEquals.listEquals(regimeInformation, _that.getRegimeInformation())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (partyReference != null ? partyReference.hashCode() : 0);
			_result = 31 * _result + (relatedPerson != null ? relatedPerson.hashCode() : 0);
			_result = 31 * _result + (relatedParty != null ? relatedParty.hashCode() : 0);
			_result = 31 * _result + (regimeInformation != null ? regimeInformation.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "PartyInformationBuilder {" +
				"partyReference=" + this.partyReference + ", " +
				"relatedPerson=" + this.relatedPerson + ", " +
				"relatedParty=" + this.relatedParty + ", " +
				"regimeInformation=" + this.regimeInformation +
			'}';
		}
	}
}
