package drr.regulation.common;

import cdm.event.common.PositionIdentifier;
import cdm.observable.asset.PriceSchedule;
import com.google.common.collect.ImmutableList;
import com.rosetta.model.lib.RosettaModelObject;
import com.rosetta.model.lib.RosettaModelObjectBuilder;
import com.rosetta.model.lib.annotations.RosettaAttribute;
import com.rosetta.model.lib.annotations.RosettaDataType;
import com.rosetta.model.lib.meta.RosettaMetaData;
import com.rosetta.model.lib.path.RosettaPath;
import com.rosetta.model.lib.process.BuilderMerger;
import com.rosetta.model.lib.process.BuilderProcessor;
import com.rosetta.model.lib.process.Processor;
import com.rosetta.util.ListEquals;
import drr.enrichment.common.EnrichmentData;
import drr.regulation.common.ConfirmationMethodEnum;
import drr.regulation.common.ExecutionVenueTypeEnum;
import drr.regulation.common.PartyInformation;
import drr.regulation.common.ReportableActionEnum;
import drr.regulation.common.ReportableInformation;
import drr.regulation.common.ReportableInformation.ReportableInformationBuilder;
import drr.regulation.common.ReportableInformation.ReportableInformationBuilderImpl;
import drr.regulation.common.ReportableInformation.ReportableInformationImpl;
import drr.regulation.common.TransactionInformation;
import drr.regulation.common.meta.ReportableInformationMeta;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import static java.util.Optional.ofNullable;

/**
 * Specifies the relevant regulators or other supervisory bodies the transaction is relevant for, and regime specific information required relating to the transaction and the parties.
 * @version 4.2.1
 */
@RosettaDataType(value="ReportableInformation", builder=ReportableInformation.ReportableInformationBuilderImpl.class, version="4.2.1")
public interface ReportableInformation extends RosettaModelObject {

	ReportableInformationMeta metaData = new ReportableInformationMeta();

	/*********************** Getter Methods  ***********************/
	/**
	 * Specifies the method for confirmation of the transacion.
	 */
	ConfirmationMethodEnum getConfirmationMethod();
	/**
	 * Specifies the type of venue where a trade could be executed.
	 */
	ExecutionVenueTypeEnum getExecutionVenueType();
	/**
	 * Specifies whether the contract was entered into as an intragroup transaction, defined in Article 3 of Regulation (EU) No 648/2012.
	 */
	Boolean getIntragroup();
	/**
	 * Specifies whether the sender of this trade considers it to be a large notional trade or block trade for reporting purposes, and thus eligible for delayed public reporting. Normally this will only be applicable for off-facility trades.
	 */
	Boolean getLargeSizeTrade();
	/**
	 * Specifies party specific information required for reporting of the transaction
	 */
	List<? extends PartyInformation> getPartyInformation();
	/**
	 * Indicator of whether the swap was executed anonymously on a SEF or DCM. When set to true, counterparty details will be masked on reporting.
	 */
	Boolean getSefOrDcmAnonymousIndicator();
	/**
	 * Date and time a transaction was originally executed, resulting in the generation of a new UTI. This data element remains unchanged throughout the life of the UTI.
	 */
	ZonedDateTime getOriginalExecutionTimestamp();
	EnrichmentData getEnrichment();
	/**
	 * Indicator whether the derivative is based on crypto-assets.
	 */
	Boolean getCryptoBased();
	/**
	 * Last Available Spot Price of the underlying asset. This price is only needed for a regulatory purpose, that being the reporting of the notional amount for commodity basis products
	 */
	PriceSchedule getUnderlyingAssetLastAvailableSpotPrice();
	List<? extends TransactionInformation> getTransactionInformation();
	/**
	 * Specifies the regulatory lifecycle event action for the ReportableEvent
	 */
	ReportableActionEnum getReportableAction();
	/**
	 * The UTI of the position in which a derivative transaction is included.
	 */
	PositionIdentifier getSubsequentPositionUTI();
	/**
	 * The prior position UTI (eg in case of a corporate action of type StockIdentifierChange).
	 */
	PositionIdentifier getCounterpartyPositionPriorUTI();

	/*********************** Build Methods  ***********************/
	ReportableInformation build();
	
	ReportableInformation.ReportableInformationBuilder toBuilder();
	
	static ReportableInformation.ReportableInformationBuilder builder() {
		return new ReportableInformation.ReportableInformationBuilderImpl();
	}

	/*********************** Utility Methods  ***********************/
	@Override
	default RosettaMetaData<? extends ReportableInformation> metaData() {
		return metaData;
	}
	
	@Override
	default Class<? extends ReportableInformation> getType() {
		return ReportableInformation.class;
	}
	
	
	@Override
	default void process(RosettaPath path, Processor processor) {
		processor.processBasic(path.newSubPath("confirmationMethod"), ConfirmationMethodEnum.class, getConfirmationMethod(), this);
		processor.processBasic(path.newSubPath("executionVenueType"), ExecutionVenueTypeEnum.class, getExecutionVenueType(), this);
		processor.processBasic(path.newSubPath("intragroup"), Boolean.class, getIntragroup(), this);
		processor.processBasic(path.newSubPath("largeSizeTrade"), Boolean.class, getLargeSizeTrade(), this);
		processRosetta(path.newSubPath("partyInformation"), processor, PartyInformation.class, getPartyInformation());
		processor.processBasic(path.newSubPath("sefOrDcmAnonymousIndicator"), Boolean.class, getSefOrDcmAnonymousIndicator(), this);
		processor.processBasic(path.newSubPath("originalExecutionTimestamp"), ZonedDateTime.class, getOriginalExecutionTimestamp(), this);
		processRosetta(path.newSubPath("enrichment"), processor, EnrichmentData.class, getEnrichment());
		processor.processBasic(path.newSubPath("cryptoBased"), Boolean.class, getCryptoBased(), this);
		processRosetta(path.newSubPath("underlyingAssetLastAvailableSpotPrice"), processor, PriceSchedule.class, getUnderlyingAssetLastAvailableSpotPrice());
		processRosetta(path.newSubPath("transactionInformation"), processor, TransactionInformation.class, getTransactionInformation());
		processor.processBasic(path.newSubPath("reportableAction"), ReportableActionEnum.class, getReportableAction(), this);
		processRosetta(path.newSubPath("subsequentPositionUTI"), processor, PositionIdentifier.class, getSubsequentPositionUTI());
		processRosetta(path.newSubPath("counterpartyPositionPriorUTI"), processor, PositionIdentifier.class, getCounterpartyPositionPriorUTI());
	}
	

	/*********************** Builder Interface  ***********************/
	interface ReportableInformationBuilder extends ReportableInformation, RosettaModelObjectBuilder {
		PartyInformation.PartyInformationBuilder getOrCreatePartyInformation(int _index);
		List<? extends PartyInformation.PartyInformationBuilder> getPartyInformation();
		EnrichmentData.EnrichmentDataBuilder getOrCreateEnrichment();
		EnrichmentData.EnrichmentDataBuilder getEnrichment();
		PriceSchedule.PriceScheduleBuilder getOrCreateUnderlyingAssetLastAvailableSpotPrice();
		PriceSchedule.PriceScheduleBuilder getUnderlyingAssetLastAvailableSpotPrice();
		TransactionInformation.TransactionInformationBuilder getOrCreateTransactionInformation(int _index);
		List<? extends TransactionInformation.TransactionInformationBuilder> getTransactionInformation();
		PositionIdentifier.PositionIdentifierBuilder getOrCreateSubsequentPositionUTI();
		PositionIdentifier.PositionIdentifierBuilder getSubsequentPositionUTI();
		PositionIdentifier.PositionIdentifierBuilder getOrCreateCounterpartyPositionPriorUTI();
		PositionIdentifier.PositionIdentifierBuilder getCounterpartyPositionPriorUTI();
		ReportableInformation.ReportableInformationBuilder setConfirmationMethod(ConfirmationMethodEnum confirmationMethod);
		ReportableInformation.ReportableInformationBuilder setExecutionVenueType(ExecutionVenueTypeEnum executionVenueType);
		ReportableInformation.ReportableInformationBuilder setIntragroup(Boolean intragroup);
		ReportableInformation.ReportableInformationBuilder setLargeSizeTrade(Boolean largeSizeTrade);
		ReportableInformation.ReportableInformationBuilder addPartyInformation(PartyInformation partyInformation0);
		ReportableInformation.ReportableInformationBuilder addPartyInformation(PartyInformation partyInformation1, int _idx);
		ReportableInformation.ReportableInformationBuilder addPartyInformation(List<? extends PartyInformation> partyInformation2);
		ReportableInformation.ReportableInformationBuilder setPartyInformation(List<? extends PartyInformation> partyInformation3);
		ReportableInformation.ReportableInformationBuilder setSefOrDcmAnonymousIndicator(Boolean sefOrDcmAnonymousIndicator);
		ReportableInformation.ReportableInformationBuilder setOriginalExecutionTimestamp(ZonedDateTime originalExecutionTimestamp);
		ReportableInformation.ReportableInformationBuilder setEnrichment(EnrichmentData enrichment);
		ReportableInformation.ReportableInformationBuilder setCryptoBased(Boolean cryptoBased);
		ReportableInformation.ReportableInformationBuilder setUnderlyingAssetLastAvailableSpotPrice(PriceSchedule underlyingAssetLastAvailableSpotPrice);
		ReportableInformation.ReportableInformationBuilder addTransactionInformation(TransactionInformation transactionInformation0);
		ReportableInformation.ReportableInformationBuilder addTransactionInformation(TransactionInformation transactionInformation1, int _idx);
		ReportableInformation.ReportableInformationBuilder addTransactionInformation(List<? extends TransactionInformation> transactionInformation2);
		ReportableInformation.ReportableInformationBuilder setTransactionInformation(List<? extends TransactionInformation> transactionInformation3);
		ReportableInformation.ReportableInformationBuilder setReportableAction(ReportableActionEnum reportableAction);
		ReportableInformation.ReportableInformationBuilder setSubsequentPositionUTI(PositionIdentifier subsequentPositionUTI);
		ReportableInformation.ReportableInformationBuilder setCounterpartyPositionPriorUTI(PositionIdentifier counterpartyPositionPriorUTI);

		@Override
		default void process(RosettaPath path, BuilderProcessor processor) {
			processor.processBasic(path.newSubPath("confirmationMethod"), ConfirmationMethodEnum.class, getConfirmationMethod(), this);
			processor.processBasic(path.newSubPath("executionVenueType"), ExecutionVenueTypeEnum.class, getExecutionVenueType(), this);
			processor.processBasic(path.newSubPath("intragroup"), Boolean.class, getIntragroup(), this);
			processor.processBasic(path.newSubPath("largeSizeTrade"), Boolean.class, getLargeSizeTrade(), this);
			processRosetta(path.newSubPath("partyInformation"), processor, PartyInformation.PartyInformationBuilder.class, getPartyInformation());
			processor.processBasic(path.newSubPath("sefOrDcmAnonymousIndicator"), Boolean.class, getSefOrDcmAnonymousIndicator(), this);
			processor.processBasic(path.newSubPath("originalExecutionTimestamp"), ZonedDateTime.class, getOriginalExecutionTimestamp(), this);
			processRosetta(path.newSubPath("enrichment"), processor, EnrichmentData.EnrichmentDataBuilder.class, getEnrichment());
			processor.processBasic(path.newSubPath("cryptoBased"), Boolean.class, getCryptoBased(), this);
			processRosetta(path.newSubPath("underlyingAssetLastAvailableSpotPrice"), processor, PriceSchedule.PriceScheduleBuilder.class, getUnderlyingAssetLastAvailableSpotPrice());
			processRosetta(path.newSubPath("transactionInformation"), processor, TransactionInformation.TransactionInformationBuilder.class, getTransactionInformation());
			processor.processBasic(path.newSubPath("reportableAction"), ReportableActionEnum.class, getReportableAction(), this);
			processRosetta(path.newSubPath("subsequentPositionUTI"), processor, PositionIdentifier.PositionIdentifierBuilder.class, getSubsequentPositionUTI());
			processRosetta(path.newSubPath("counterpartyPositionPriorUTI"), processor, PositionIdentifier.PositionIdentifierBuilder.class, getCounterpartyPositionPriorUTI());
		}
		

		ReportableInformation.ReportableInformationBuilder prune();
	}

	/*********************** Immutable Implementation of ReportableInformation  ***********************/
	class ReportableInformationImpl implements ReportableInformation {
		private final ConfirmationMethodEnum confirmationMethod;
		private final ExecutionVenueTypeEnum executionVenueType;
		private final Boolean intragroup;
		private final Boolean largeSizeTrade;
		private final List<? extends PartyInformation> partyInformation;
		private final Boolean sefOrDcmAnonymousIndicator;
		private final ZonedDateTime originalExecutionTimestamp;
		private final EnrichmentData enrichment;
		private final Boolean cryptoBased;
		private final PriceSchedule underlyingAssetLastAvailableSpotPrice;
		private final List<? extends TransactionInformation> transactionInformation;
		private final ReportableActionEnum reportableAction;
		private final PositionIdentifier subsequentPositionUTI;
		private final PositionIdentifier counterpartyPositionPriorUTI;
		
		protected ReportableInformationImpl(ReportableInformation.ReportableInformationBuilder builder) {
			this.confirmationMethod = builder.getConfirmationMethod();
			this.executionVenueType = builder.getExecutionVenueType();
			this.intragroup = builder.getIntragroup();
			this.largeSizeTrade = builder.getLargeSizeTrade();
			this.partyInformation = ofNullable(builder.getPartyInformation()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.sefOrDcmAnonymousIndicator = builder.getSefOrDcmAnonymousIndicator();
			this.originalExecutionTimestamp = builder.getOriginalExecutionTimestamp();
			this.enrichment = ofNullable(builder.getEnrichment()).map(f->f.build()).orElse(null);
			this.cryptoBased = builder.getCryptoBased();
			this.underlyingAssetLastAvailableSpotPrice = ofNullable(builder.getUnderlyingAssetLastAvailableSpotPrice()).map(f->f.build()).orElse(null);
			this.transactionInformation = ofNullable(builder.getTransactionInformation()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.reportableAction = builder.getReportableAction();
			this.subsequentPositionUTI = ofNullable(builder.getSubsequentPositionUTI()).map(f->f.build()).orElse(null);
			this.counterpartyPositionPriorUTI = ofNullable(builder.getCounterpartyPositionPriorUTI()).map(f->f.build()).orElse(null);
		}
		
		@Override
		@RosettaAttribute("confirmationMethod")
		public ConfirmationMethodEnum getConfirmationMethod() {
			return confirmationMethod;
		}
		
		@Override
		@RosettaAttribute("executionVenueType")
		public ExecutionVenueTypeEnum getExecutionVenueType() {
			return executionVenueType;
		}
		
		@Override
		@RosettaAttribute("intragroup")
		public Boolean getIntragroup() {
			return intragroup;
		}
		
		@Override
		@RosettaAttribute("largeSizeTrade")
		public Boolean getLargeSizeTrade() {
			return largeSizeTrade;
		}
		
		@Override
		@RosettaAttribute("partyInformation")
		public List<? extends PartyInformation> getPartyInformation() {
			return partyInformation;
		}
		
		@Override
		@RosettaAttribute("sefOrDcmAnonymousIndicator")
		public Boolean getSefOrDcmAnonymousIndicator() {
			return sefOrDcmAnonymousIndicator;
		}
		
		@Override
		@RosettaAttribute("originalExecutionTimestamp")
		public ZonedDateTime getOriginalExecutionTimestamp() {
			return originalExecutionTimestamp;
		}
		
		@Override
		@RosettaAttribute("enrichment")
		public EnrichmentData getEnrichment() {
			return enrichment;
		}
		
		@Override
		@RosettaAttribute("cryptoBased")
		public Boolean getCryptoBased() {
			return cryptoBased;
		}
		
		@Override
		@RosettaAttribute("underlyingAssetLastAvailableSpotPrice")
		public PriceSchedule getUnderlyingAssetLastAvailableSpotPrice() {
			return underlyingAssetLastAvailableSpotPrice;
		}
		
		@Override
		@RosettaAttribute("transactionInformation")
		public List<? extends TransactionInformation> getTransactionInformation() {
			return transactionInformation;
		}
		
		@Override
		@RosettaAttribute("reportableAction")
		public ReportableActionEnum getReportableAction() {
			return reportableAction;
		}
		
		@Override
		@RosettaAttribute("subsequentPositionUTI")
		public PositionIdentifier getSubsequentPositionUTI() {
			return subsequentPositionUTI;
		}
		
		@Override
		@RosettaAttribute("counterpartyPositionPriorUTI")
		public PositionIdentifier getCounterpartyPositionPriorUTI() {
			return counterpartyPositionPriorUTI;
		}
		
		@Override
		public ReportableInformation build() {
			return this;
		}
		
		@Override
		public ReportableInformation.ReportableInformationBuilder toBuilder() {
			ReportableInformation.ReportableInformationBuilder builder = builder();
			setBuilderFields(builder);
			return builder;
		}
		
		protected void setBuilderFields(ReportableInformation.ReportableInformationBuilder builder) {
			ofNullable(getConfirmationMethod()).ifPresent(builder::setConfirmationMethod);
			ofNullable(getExecutionVenueType()).ifPresent(builder::setExecutionVenueType);
			ofNullable(getIntragroup()).ifPresent(builder::setIntragroup);
			ofNullable(getLargeSizeTrade()).ifPresent(builder::setLargeSizeTrade);
			ofNullable(getPartyInformation()).ifPresent(builder::setPartyInformation);
			ofNullable(getSefOrDcmAnonymousIndicator()).ifPresent(builder::setSefOrDcmAnonymousIndicator);
			ofNullable(getOriginalExecutionTimestamp()).ifPresent(builder::setOriginalExecutionTimestamp);
			ofNullable(getEnrichment()).ifPresent(builder::setEnrichment);
			ofNullable(getCryptoBased()).ifPresent(builder::setCryptoBased);
			ofNullable(getUnderlyingAssetLastAvailableSpotPrice()).ifPresent(builder::setUnderlyingAssetLastAvailableSpotPrice);
			ofNullable(getTransactionInformation()).ifPresent(builder::setTransactionInformation);
			ofNullable(getReportableAction()).ifPresent(builder::setReportableAction);
			ofNullable(getSubsequentPositionUTI()).ifPresent(builder::setSubsequentPositionUTI);
			ofNullable(getCounterpartyPositionPriorUTI()).ifPresent(builder::setCounterpartyPositionPriorUTI);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			ReportableInformation _that = getType().cast(o);
		
			if (!Objects.equals(confirmationMethod, _that.getConfirmationMethod())) return false;
			if (!Objects.equals(executionVenueType, _that.getExecutionVenueType())) return false;
			if (!Objects.equals(intragroup, _that.getIntragroup())) return false;
			if (!Objects.equals(largeSizeTrade, _that.getLargeSizeTrade())) return false;
			if (!ListEquals.listEquals(partyInformation, _that.getPartyInformation())) return false;
			if (!Objects.equals(sefOrDcmAnonymousIndicator, _that.getSefOrDcmAnonymousIndicator())) return false;
			if (!Objects.equals(originalExecutionTimestamp, _that.getOriginalExecutionTimestamp())) return false;
			if (!Objects.equals(enrichment, _that.getEnrichment())) return false;
			if (!Objects.equals(cryptoBased, _that.getCryptoBased())) return false;
			if (!Objects.equals(underlyingAssetLastAvailableSpotPrice, _that.getUnderlyingAssetLastAvailableSpotPrice())) return false;
			if (!ListEquals.listEquals(transactionInformation, _that.getTransactionInformation())) return false;
			if (!Objects.equals(reportableAction, _that.getReportableAction())) return false;
			if (!Objects.equals(subsequentPositionUTI, _that.getSubsequentPositionUTI())) return false;
			if (!Objects.equals(counterpartyPositionPriorUTI, _that.getCounterpartyPositionPriorUTI())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (confirmationMethod != null ? confirmationMethod.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (executionVenueType != null ? executionVenueType.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (intragroup != null ? intragroup.hashCode() : 0);
			_result = 31 * _result + (largeSizeTrade != null ? largeSizeTrade.hashCode() : 0);
			_result = 31 * _result + (partyInformation != null ? partyInformation.hashCode() : 0);
			_result = 31 * _result + (sefOrDcmAnonymousIndicator != null ? sefOrDcmAnonymousIndicator.hashCode() : 0);
			_result = 31 * _result + (originalExecutionTimestamp != null ? originalExecutionTimestamp.hashCode() : 0);
			_result = 31 * _result + (enrichment != null ? enrichment.hashCode() : 0);
			_result = 31 * _result + (cryptoBased != null ? cryptoBased.hashCode() : 0);
			_result = 31 * _result + (underlyingAssetLastAvailableSpotPrice != null ? underlyingAssetLastAvailableSpotPrice.hashCode() : 0);
			_result = 31 * _result + (transactionInformation != null ? transactionInformation.hashCode() : 0);
			_result = 31 * _result + (reportableAction != null ? reportableAction.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (subsequentPositionUTI != null ? subsequentPositionUTI.hashCode() : 0);
			_result = 31 * _result + (counterpartyPositionPriorUTI != null ? counterpartyPositionPriorUTI.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "ReportableInformation {" +
				"confirmationMethod=" + this.confirmationMethod + ", " +
				"executionVenueType=" + this.executionVenueType + ", " +
				"intragroup=" + this.intragroup + ", " +
				"largeSizeTrade=" + this.largeSizeTrade + ", " +
				"partyInformation=" + this.partyInformation + ", " +
				"sefOrDcmAnonymousIndicator=" + this.sefOrDcmAnonymousIndicator + ", " +
				"originalExecutionTimestamp=" + this.originalExecutionTimestamp + ", " +
				"enrichment=" + this.enrichment + ", " +
				"cryptoBased=" + this.cryptoBased + ", " +
				"underlyingAssetLastAvailableSpotPrice=" + this.underlyingAssetLastAvailableSpotPrice + ", " +
				"transactionInformation=" + this.transactionInformation + ", " +
				"reportableAction=" + this.reportableAction + ", " +
				"subsequentPositionUTI=" + this.subsequentPositionUTI + ", " +
				"counterpartyPositionPriorUTI=" + this.counterpartyPositionPriorUTI +
			'}';
		}
	}

	/*********************** Builder Implementation of ReportableInformation  ***********************/
	class ReportableInformationBuilderImpl implements ReportableInformation.ReportableInformationBuilder {
	
		protected ConfirmationMethodEnum confirmationMethod;
		protected ExecutionVenueTypeEnum executionVenueType;
		protected Boolean intragroup;
		protected Boolean largeSizeTrade;
		protected List<PartyInformation.PartyInformationBuilder> partyInformation = new ArrayList<>();
		protected Boolean sefOrDcmAnonymousIndicator;
		protected ZonedDateTime originalExecutionTimestamp;
		protected EnrichmentData.EnrichmentDataBuilder enrichment;
		protected Boolean cryptoBased;
		protected PriceSchedule.PriceScheduleBuilder underlyingAssetLastAvailableSpotPrice;
		protected List<TransactionInformation.TransactionInformationBuilder> transactionInformation = new ArrayList<>();
		protected ReportableActionEnum reportableAction;
		protected PositionIdentifier.PositionIdentifierBuilder subsequentPositionUTI;
		protected PositionIdentifier.PositionIdentifierBuilder counterpartyPositionPriorUTI;
	
		public ReportableInformationBuilderImpl() {
		}
	
		@Override
		@RosettaAttribute("confirmationMethod")
		public ConfirmationMethodEnum getConfirmationMethod() {
			return confirmationMethod;
		}
		
		@Override
		@RosettaAttribute("executionVenueType")
		public ExecutionVenueTypeEnum getExecutionVenueType() {
			return executionVenueType;
		}
		
		@Override
		@RosettaAttribute("intragroup")
		public Boolean getIntragroup() {
			return intragroup;
		}
		
		@Override
		@RosettaAttribute("largeSizeTrade")
		public Boolean getLargeSizeTrade() {
			return largeSizeTrade;
		}
		
		@Override
		@RosettaAttribute("partyInformation")
		public List<? extends PartyInformation.PartyInformationBuilder> getPartyInformation() {
			return partyInformation;
		}
		
		public PartyInformation.PartyInformationBuilder getOrCreatePartyInformation(int _index) {
		
			if (partyInformation==null) {
				this.partyInformation = new ArrayList<>();
			}
			PartyInformation.PartyInformationBuilder result;
			return getIndex(partyInformation, _index, () -> {
						PartyInformation.PartyInformationBuilder newPartyInformation = PartyInformation.builder();
						return newPartyInformation;
					});
		}
		
		@Override
		@RosettaAttribute("sefOrDcmAnonymousIndicator")
		public Boolean getSefOrDcmAnonymousIndicator() {
			return sefOrDcmAnonymousIndicator;
		}
		
		@Override
		@RosettaAttribute("originalExecutionTimestamp")
		public ZonedDateTime getOriginalExecutionTimestamp() {
			return originalExecutionTimestamp;
		}
		
		@Override
		@RosettaAttribute("enrichment")
		public EnrichmentData.EnrichmentDataBuilder getEnrichment() {
			return enrichment;
		}
		
		@Override
		public EnrichmentData.EnrichmentDataBuilder getOrCreateEnrichment() {
			EnrichmentData.EnrichmentDataBuilder result;
			if (enrichment!=null) {
				result = enrichment;
			}
			else {
				result = enrichment = EnrichmentData.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("cryptoBased")
		public Boolean getCryptoBased() {
			return cryptoBased;
		}
		
		@Override
		@RosettaAttribute("underlyingAssetLastAvailableSpotPrice")
		public PriceSchedule.PriceScheduleBuilder getUnderlyingAssetLastAvailableSpotPrice() {
			return underlyingAssetLastAvailableSpotPrice;
		}
		
		@Override
		public PriceSchedule.PriceScheduleBuilder getOrCreateUnderlyingAssetLastAvailableSpotPrice() {
			PriceSchedule.PriceScheduleBuilder result;
			if (underlyingAssetLastAvailableSpotPrice!=null) {
				result = underlyingAssetLastAvailableSpotPrice;
			}
			else {
				result = underlyingAssetLastAvailableSpotPrice = PriceSchedule.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("transactionInformation")
		public List<? extends TransactionInformation.TransactionInformationBuilder> getTransactionInformation() {
			return transactionInformation;
		}
		
		public TransactionInformation.TransactionInformationBuilder getOrCreateTransactionInformation(int _index) {
		
			if (transactionInformation==null) {
				this.transactionInformation = new ArrayList<>();
			}
			TransactionInformation.TransactionInformationBuilder result;
			return getIndex(transactionInformation, _index, () -> {
						TransactionInformation.TransactionInformationBuilder newTransactionInformation = TransactionInformation.builder();
						return newTransactionInformation;
					});
		}
		
		@Override
		@RosettaAttribute("reportableAction")
		public ReportableActionEnum getReportableAction() {
			return reportableAction;
		}
		
		@Override
		@RosettaAttribute("subsequentPositionUTI")
		public PositionIdentifier.PositionIdentifierBuilder getSubsequentPositionUTI() {
			return subsequentPositionUTI;
		}
		
		@Override
		public PositionIdentifier.PositionIdentifierBuilder getOrCreateSubsequentPositionUTI() {
			PositionIdentifier.PositionIdentifierBuilder result;
			if (subsequentPositionUTI!=null) {
				result = subsequentPositionUTI;
			}
			else {
				result = subsequentPositionUTI = PositionIdentifier.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("counterpartyPositionPriorUTI")
		public PositionIdentifier.PositionIdentifierBuilder getCounterpartyPositionPriorUTI() {
			return counterpartyPositionPriorUTI;
		}
		
		@Override
		public PositionIdentifier.PositionIdentifierBuilder getOrCreateCounterpartyPositionPriorUTI() {
			PositionIdentifier.PositionIdentifierBuilder result;
			if (counterpartyPositionPriorUTI!=null) {
				result = counterpartyPositionPriorUTI;
			}
			else {
				result = counterpartyPositionPriorUTI = PositionIdentifier.builder();
			}
			
			return result;
		}
	
		@Override
		@RosettaAttribute("confirmationMethod")
		public ReportableInformation.ReportableInformationBuilder setConfirmationMethod(ConfirmationMethodEnum confirmationMethod) {
			this.confirmationMethod = confirmationMethod==null?null:confirmationMethod;
			return this;
		}
		@Override
		@RosettaAttribute("executionVenueType")
		public ReportableInformation.ReportableInformationBuilder setExecutionVenueType(ExecutionVenueTypeEnum executionVenueType) {
			this.executionVenueType = executionVenueType==null?null:executionVenueType;
			return this;
		}
		@Override
		@RosettaAttribute("intragroup")
		public ReportableInformation.ReportableInformationBuilder setIntragroup(Boolean intragroup) {
			this.intragroup = intragroup==null?null:intragroup;
			return this;
		}
		@Override
		@RosettaAttribute("largeSizeTrade")
		public ReportableInformation.ReportableInformationBuilder setLargeSizeTrade(Boolean largeSizeTrade) {
			this.largeSizeTrade = largeSizeTrade==null?null:largeSizeTrade;
			return this;
		}
		@Override
		public ReportableInformation.ReportableInformationBuilder addPartyInformation(PartyInformation partyInformation) {
			if (partyInformation!=null) this.partyInformation.add(partyInformation.toBuilder());
			return this;
		}
		
		@Override
		public ReportableInformation.ReportableInformationBuilder addPartyInformation(PartyInformation partyInformation, int _idx) {
			getIndex(this.partyInformation, _idx, () -> partyInformation.toBuilder());
			return this;
		}
		@Override 
		public ReportableInformation.ReportableInformationBuilder addPartyInformation(List<? extends PartyInformation> partyInformations) {
			if (partyInformations != null) {
				for (PartyInformation toAdd : partyInformations) {
					this.partyInformation.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("partyInformation")
		public ReportableInformation.ReportableInformationBuilder setPartyInformation(List<? extends PartyInformation> partyInformations) {
			if (partyInformations == null)  {
				this.partyInformation = new ArrayList<>();
			}
			else {
				this.partyInformation = partyInformations.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		@RosettaAttribute("sefOrDcmAnonymousIndicator")
		public ReportableInformation.ReportableInformationBuilder setSefOrDcmAnonymousIndicator(Boolean sefOrDcmAnonymousIndicator) {
			this.sefOrDcmAnonymousIndicator = sefOrDcmAnonymousIndicator==null?null:sefOrDcmAnonymousIndicator;
			return this;
		}
		@Override
		@RosettaAttribute("originalExecutionTimestamp")
		public ReportableInformation.ReportableInformationBuilder setOriginalExecutionTimestamp(ZonedDateTime originalExecutionTimestamp) {
			this.originalExecutionTimestamp = originalExecutionTimestamp==null?null:originalExecutionTimestamp;
			return this;
		}
		@Override
		@RosettaAttribute("enrichment")
		public ReportableInformation.ReportableInformationBuilder setEnrichment(EnrichmentData enrichment) {
			this.enrichment = enrichment==null?null:enrichment.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("cryptoBased")
		public ReportableInformation.ReportableInformationBuilder setCryptoBased(Boolean cryptoBased) {
			this.cryptoBased = cryptoBased==null?null:cryptoBased;
			return this;
		}
		@Override
		@RosettaAttribute("underlyingAssetLastAvailableSpotPrice")
		public ReportableInformation.ReportableInformationBuilder setUnderlyingAssetLastAvailableSpotPrice(PriceSchedule underlyingAssetLastAvailableSpotPrice) {
			this.underlyingAssetLastAvailableSpotPrice = underlyingAssetLastAvailableSpotPrice==null?null:underlyingAssetLastAvailableSpotPrice.toBuilder();
			return this;
		}
		@Override
		public ReportableInformation.ReportableInformationBuilder addTransactionInformation(TransactionInformation transactionInformation) {
			if (transactionInformation!=null) this.transactionInformation.add(transactionInformation.toBuilder());
			return this;
		}
		
		@Override
		public ReportableInformation.ReportableInformationBuilder addTransactionInformation(TransactionInformation transactionInformation, int _idx) {
			getIndex(this.transactionInformation, _idx, () -> transactionInformation.toBuilder());
			return this;
		}
		@Override 
		public ReportableInformation.ReportableInformationBuilder addTransactionInformation(List<? extends TransactionInformation> transactionInformations) {
			if (transactionInformations != null) {
				for (TransactionInformation toAdd : transactionInformations) {
					this.transactionInformation.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("transactionInformation")
		public ReportableInformation.ReportableInformationBuilder setTransactionInformation(List<? extends TransactionInformation> transactionInformations) {
			if (transactionInformations == null)  {
				this.transactionInformation = new ArrayList<>();
			}
			else {
				this.transactionInformation = transactionInformations.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		@RosettaAttribute("reportableAction")
		public ReportableInformation.ReportableInformationBuilder setReportableAction(ReportableActionEnum reportableAction) {
			this.reportableAction = reportableAction==null?null:reportableAction;
			return this;
		}
		@Override
		@RosettaAttribute("subsequentPositionUTI")
		public ReportableInformation.ReportableInformationBuilder setSubsequentPositionUTI(PositionIdentifier subsequentPositionUTI) {
			this.subsequentPositionUTI = subsequentPositionUTI==null?null:subsequentPositionUTI.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("counterpartyPositionPriorUTI")
		public ReportableInformation.ReportableInformationBuilder setCounterpartyPositionPriorUTI(PositionIdentifier counterpartyPositionPriorUTI) {
			this.counterpartyPositionPriorUTI = counterpartyPositionPriorUTI==null?null:counterpartyPositionPriorUTI.toBuilder();
			return this;
		}
		
		@Override
		public ReportableInformation build() {
			return new ReportableInformation.ReportableInformationImpl(this);
		}
		
		@Override
		public ReportableInformation.ReportableInformationBuilder toBuilder() {
			return this;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public ReportableInformation.ReportableInformationBuilder prune() {
			partyInformation = partyInformation.stream().filter(b->b!=null).<PartyInformation.PartyInformationBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			if (enrichment!=null && !enrichment.prune().hasData()) enrichment = null;
			if (underlyingAssetLastAvailableSpotPrice!=null && !underlyingAssetLastAvailableSpotPrice.prune().hasData()) underlyingAssetLastAvailableSpotPrice = null;
			transactionInformation = transactionInformation.stream().filter(b->b!=null).<TransactionInformation.TransactionInformationBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			if (subsequentPositionUTI!=null && !subsequentPositionUTI.prune().hasData()) subsequentPositionUTI = null;
			if (counterpartyPositionPriorUTI!=null && !counterpartyPositionPriorUTI.prune().hasData()) counterpartyPositionPriorUTI = null;
			return this;
		}
		
		@Override
		public boolean hasData() {
			if (getConfirmationMethod()!=null) return true;
			if (getExecutionVenueType()!=null) return true;
			if (getIntragroup()!=null) return true;
			if (getLargeSizeTrade()!=null) return true;
			if (getPartyInformation()!=null && getPartyInformation().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getSefOrDcmAnonymousIndicator()!=null) return true;
			if (getOriginalExecutionTimestamp()!=null) return true;
			if (getEnrichment()!=null && getEnrichment().hasData()) return true;
			if (getCryptoBased()!=null) return true;
			if (getUnderlyingAssetLastAvailableSpotPrice()!=null && getUnderlyingAssetLastAvailableSpotPrice().hasData()) return true;
			if (getTransactionInformation()!=null && getTransactionInformation().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getReportableAction()!=null) return true;
			if (getSubsequentPositionUTI()!=null && getSubsequentPositionUTI().hasData()) return true;
			if (getCounterpartyPositionPriorUTI()!=null && getCounterpartyPositionPriorUTI().hasData()) return true;
			return false;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public ReportableInformation.ReportableInformationBuilder merge(RosettaModelObjectBuilder other, BuilderMerger merger) {
			ReportableInformation.ReportableInformationBuilder o = (ReportableInformation.ReportableInformationBuilder) other;
			
			merger.mergeRosetta(getPartyInformation(), o.getPartyInformation(), this::getOrCreatePartyInformation);
			merger.mergeRosetta(getEnrichment(), o.getEnrichment(), this::setEnrichment);
			merger.mergeRosetta(getUnderlyingAssetLastAvailableSpotPrice(), o.getUnderlyingAssetLastAvailableSpotPrice(), this::setUnderlyingAssetLastAvailableSpotPrice);
			merger.mergeRosetta(getTransactionInformation(), o.getTransactionInformation(), this::getOrCreateTransactionInformation);
			merger.mergeRosetta(getSubsequentPositionUTI(), o.getSubsequentPositionUTI(), this::setSubsequentPositionUTI);
			merger.mergeRosetta(getCounterpartyPositionPriorUTI(), o.getCounterpartyPositionPriorUTI(), this::setCounterpartyPositionPriorUTI);
			
			merger.mergeBasic(getConfirmationMethod(), o.getConfirmationMethod(), this::setConfirmationMethod);
			merger.mergeBasic(getExecutionVenueType(), o.getExecutionVenueType(), this::setExecutionVenueType);
			merger.mergeBasic(getIntragroup(), o.getIntragroup(), this::setIntragroup);
			merger.mergeBasic(getLargeSizeTrade(), o.getLargeSizeTrade(), this::setLargeSizeTrade);
			merger.mergeBasic(getSefOrDcmAnonymousIndicator(), o.getSefOrDcmAnonymousIndicator(), this::setSefOrDcmAnonymousIndicator);
			merger.mergeBasic(getOriginalExecutionTimestamp(), o.getOriginalExecutionTimestamp(), this::setOriginalExecutionTimestamp);
			merger.mergeBasic(getCryptoBased(), o.getCryptoBased(), this::setCryptoBased);
			merger.mergeBasic(getReportableAction(), o.getReportableAction(), this::setReportableAction);
			return this;
		}
	
		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			ReportableInformation _that = getType().cast(o);
		
			if (!Objects.equals(confirmationMethod, _that.getConfirmationMethod())) return false;
			if (!Objects.equals(executionVenueType, _that.getExecutionVenueType())) return false;
			if (!Objects.equals(intragroup, _that.getIntragroup())) return false;
			if (!Objects.equals(largeSizeTrade, _that.getLargeSizeTrade())) return false;
			if (!ListEquals.listEquals(partyInformation, _that.getPartyInformation())) return false;
			if (!Objects.equals(sefOrDcmAnonymousIndicator, _that.getSefOrDcmAnonymousIndicator())) return false;
			if (!Objects.equals(originalExecutionTimestamp, _that.getOriginalExecutionTimestamp())) return false;
			if (!Objects.equals(enrichment, _that.getEnrichment())) return false;
			if (!Objects.equals(cryptoBased, _that.getCryptoBased())) return false;
			if (!Objects.equals(underlyingAssetLastAvailableSpotPrice, _that.getUnderlyingAssetLastAvailableSpotPrice())) return false;
			if (!ListEquals.listEquals(transactionInformation, _that.getTransactionInformation())) return false;
			if (!Objects.equals(reportableAction, _that.getReportableAction())) return false;
			if (!Objects.equals(subsequentPositionUTI, _that.getSubsequentPositionUTI())) return false;
			if (!Objects.equals(counterpartyPositionPriorUTI, _that.getCounterpartyPositionPriorUTI())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (confirmationMethod != null ? confirmationMethod.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (executionVenueType != null ? executionVenueType.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (intragroup != null ? intragroup.hashCode() : 0);
			_result = 31 * _result + (largeSizeTrade != null ? largeSizeTrade.hashCode() : 0);
			_result = 31 * _result + (partyInformation != null ? partyInformation.hashCode() : 0);
			_result = 31 * _result + (sefOrDcmAnonymousIndicator != null ? sefOrDcmAnonymousIndicator.hashCode() : 0);
			_result = 31 * _result + (originalExecutionTimestamp != null ? originalExecutionTimestamp.hashCode() : 0);
			_result = 31 * _result + (enrichment != null ? enrichment.hashCode() : 0);
			_result = 31 * _result + (cryptoBased != null ? cryptoBased.hashCode() : 0);
			_result = 31 * _result + (underlyingAssetLastAvailableSpotPrice != null ? underlyingAssetLastAvailableSpotPrice.hashCode() : 0);
			_result = 31 * _result + (transactionInformation != null ? transactionInformation.hashCode() : 0);
			_result = 31 * _result + (reportableAction != null ? reportableAction.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (subsequentPositionUTI != null ? subsequentPositionUTI.hashCode() : 0);
			_result = 31 * _result + (counterpartyPositionPriorUTI != null ? counterpartyPositionPriorUTI.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "ReportableInformationBuilder {" +
				"confirmationMethod=" + this.confirmationMethod + ", " +
				"executionVenueType=" + this.executionVenueType + ", " +
				"intragroup=" + this.intragroup + ", " +
				"largeSizeTrade=" + this.largeSizeTrade + ", " +
				"partyInformation=" + this.partyInformation + ", " +
				"sefOrDcmAnonymousIndicator=" + this.sefOrDcmAnonymousIndicator + ", " +
				"originalExecutionTimestamp=" + this.originalExecutionTimestamp + ", " +
				"enrichment=" + this.enrichment + ", " +
				"cryptoBased=" + this.cryptoBased + ", " +
				"underlyingAssetLastAvailableSpotPrice=" + this.underlyingAssetLastAvailableSpotPrice + ", " +
				"transactionInformation=" + this.transactionInformation + ", " +
				"reportableAction=" + this.reportableAction + ", " +
				"subsequentPositionUTI=" + this.subsequentPositionUTI + ", " +
				"counterpartyPositionPriorUTI=" + this.counterpartyPositionPriorUTI +
			'}';
		}
	}
}
