package drr.regulation.common;

import com.rosetta.model.lib.RosettaModelObject;
import com.rosetta.model.lib.RosettaModelObjectBuilder;
import com.rosetta.model.lib.annotations.RosettaAttribute;
import com.rosetta.model.lib.annotations.RosettaDataType;
import com.rosetta.model.lib.meta.RosettaMetaData;
import com.rosetta.model.lib.path.RosettaPath;
import com.rosetta.model.lib.process.BuilderMerger;
import com.rosetta.model.lib.process.BuilderProcessor;
import com.rosetta.model.lib.process.Processor;
import drr.regulation.common.CFTCPartyInformation;
import drr.regulation.common.ClearingException;
import drr.regulation.common.ESMAPartyInformation;
import drr.regulation.common.MandatorilyClearableEnum;
import drr.regulation.common.RegimeNameEnum;
import drr.regulation.common.ReportingRegime;
import drr.regulation.common.ReportingRegime.ReportingRegimeBuilder;
import drr.regulation.common.ReportingRegime.ReportingRegimeBuilderImpl;
import drr.regulation.common.ReportingRegime.ReportingRegimeImpl;
import drr.regulation.common.ReportingRoleEnum;
import drr.regulation.common.SECPartyInformation;
import drr.regulation.common.SupervisoryBodyEnum;
import drr.regulation.common.meta.ReportingRegimeMeta;
import drr.regulation.common.metafields.FieldWithMetaRegimeNameEnum;
import drr.regulation.common.metafields.FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder;
import drr.regulation.common.metafields.FieldWithMetaSupervisoryBodyEnum;
import drr.regulation.common.metafields.FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder;
import java.util.Objects;

import static java.util.Optional.ofNullable;

/**
 * Specifies the regulatory regime requiring reporting of the transaction.
 * @version 4.2.1
 */
@RosettaDataType(value="ReportingRegime", builder=ReportingRegime.ReportingRegimeBuilderImpl.class, version="4.2.1")
public interface ReportingRegime extends RosettaModelObject {

	ReportingRegimeMeta metaData = new ReportingRegimeMeta();

	/*********************** Getter Methods  ***********************/
	/**
	 * Identifies the reporting regime under which this data is reported. For example, Dodd-Frank, MiFID, HongKongOTCDRepository, ODRF
	 */
	FieldWithMetaRegimeNameEnum getRegimeName();
	/**
	 * Identifies the specific regulator or other supervisory body for which this data is produced. For example, CFTC, SEC, UKFSA, ODRF, SFC, ESMA.
	 */
	FieldWithMetaSupervisoryBodyEnum getSupervisoryBody();
	/**
	 * Indicator of whether the transaction is mandatorily clearable.
	 */
	MandatorilyClearableEnum getMandatorilyClearable();
	/**
	 * Identifies the type of clearing exception or exemption that the counterparty has elected.
	 */
	ClearingException getClearingException();
	/**
	 * Indication of whether the transaction satisfies the definition of &#39;post-priced swap&#39;.
	 */
	Boolean getPostPricedIndicator();
	/**
	 * Party information required for reporting a transaction to the CFTC.
	 */
	CFTCPartyInformation getCftcPartyInformation();
	/**
	 * Party information required for reporting a transaction under ESMA.
	 */
	ESMAPartyInformation getEsmaPartyInformation();
	/**
	 * Party information required for reporting a transaction under SEC.
	 */
	SECPartyInformation getSecPartyInformation();
	/**
	 * Specifies the reporting role of the party with respect to the regime.
	 */
	ReportingRoleEnum getReportingRole();

	/*********************** Build Methods  ***********************/
	ReportingRegime build();
	
	ReportingRegime.ReportingRegimeBuilder toBuilder();
	
	static ReportingRegime.ReportingRegimeBuilder builder() {
		return new ReportingRegime.ReportingRegimeBuilderImpl();
	}

	/*********************** Utility Methods  ***********************/
	@Override
	default RosettaMetaData<? extends ReportingRegime> metaData() {
		return metaData;
	}
	
	@Override
	default Class<? extends ReportingRegime> getType() {
		return ReportingRegime.class;
	}
	
	
	@Override
	default void process(RosettaPath path, Processor processor) {
		processRosetta(path.newSubPath("regimeName"), processor, FieldWithMetaRegimeNameEnum.class, getRegimeName());
		processRosetta(path.newSubPath("supervisoryBody"), processor, FieldWithMetaSupervisoryBodyEnum.class, getSupervisoryBody());
		processor.processBasic(path.newSubPath("mandatorilyClearable"), MandatorilyClearableEnum.class, getMandatorilyClearable(), this);
		processRosetta(path.newSubPath("clearingException"), processor, ClearingException.class, getClearingException());
		processor.processBasic(path.newSubPath("postPricedIndicator"), Boolean.class, getPostPricedIndicator(), this);
		processRosetta(path.newSubPath("cftcPartyInformation"), processor, CFTCPartyInformation.class, getCftcPartyInformation());
		processRosetta(path.newSubPath("esmaPartyInformation"), processor, ESMAPartyInformation.class, getEsmaPartyInformation());
		processRosetta(path.newSubPath("secPartyInformation"), processor, SECPartyInformation.class, getSecPartyInformation());
		processor.processBasic(path.newSubPath("reportingRole"), ReportingRoleEnum.class, getReportingRole(), this);
	}
	

	/*********************** Builder Interface  ***********************/
	interface ReportingRegimeBuilder extends ReportingRegime, RosettaModelObjectBuilder {
		FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder getOrCreateRegimeName();
		FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder getRegimeName();
		FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder getOrCreateSupervisoryBody();
		FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder getSupervisoryBody();
		ClearingException.ClearingExceptionBuilder getOrCreateClearingException();
		ClearingException.ClearingExceptionBuilder getClearingException();
		CFTCPartyInformation.CFTCPartyInformationBuilder getOrCreateCftcPartyInformation();
		CFTCPartyInformation.CFTCPartyInformationBuilder getCftcPartyInformation();
		ESMAPartyInformation.ESMAPartyInformationBuilder getOrCreateEsmaPartyInformation();
		ESMAPartyInformation.ESMAPartyInformationBuilder getEsmaPartyInformation();
		SECPartyInformation.SECPartyInformationBuilder getOrCreateSecPartyInformation();
		SECPartyInformation.SECPartyInformationBuilder getSecPartyInformation();
		ReportingRegime.ReportingRegimeBuilder setRegimeName(FieldWithMetaRegimeNameEnum regimeName0);
		ReportingRegime.ReportingRegimeBuilder setRegimeNameValue(RegimeNameEnum regimeName1);
		ReportingRegime.ReportingRegimeBuilder setSupervisoryBody(FieldWithMetaSupervisoryBodyEnum supervisoryBody0);
		ReportingRegime.ReportingRegimeBuilder setSupervisoryBodyValue(SupervisoryBodyEnum supervisoryBody1);
		ReportingRegime.ReportingRegimeBuilder setMandatorilyClearable(MandatorilyClearableEnum mandatorilyClearable);
		ReportingRegime.ReportingRegimeBuilder setClearingException(ClearingException clearingException);
		ReportingRegime.ReportingRegimeBuilder setPostPricedIndicator(Boolean postPricedIndicator);
		ReportingRegime.ReportingRegimeBuilder setCftcPartyInformation(CFTCPartyInformation cftcPartyInformation);
		ReportingRegime.ReportingRegimeBuilder setEsmaPartyInformation(ESMAPartyInformation esmaPartyInformation);
		ReportingRegime.ReportingRegimeBuilder setSecPartyInformation(SECPartyInformation secPartyInformation);
		ReportingRegime.ReportingRegimeBuilder setReportingRole(ReportingRoleEnum reportingRole);

		@Override
		default void process(RosettaPath path, BuilderProcessor processor) {
			processRosetta(path.newSubPath("regimeName"), processor, FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder.class, getRegimeName());
			processRosetta(path.newSubPath("supervisoryBody"), processor, FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder.class, getSupervisoryBody());
			processor.processBasic(path.newSubPath("mandatorilyClearable"), MandatorilyClearableEnum.class, getMandatorilyClearable(), this);
			processRosetta(path.newSubPath("clearingException"), processor, ClearingException.ClearingExceptionBuilder.class, getClearingException());
			processor.processBasic(path.newSubPath("postPricedIndicator"), Boolean.class, getPostPricedIndicator(), this);
			processRosetta(path.newSubPath("cftcPartyInformation"), processor, CFTCPartyInformation.CFTCPartyInformationBuilder.class, getCftcPartyInformation());
			processRosetta(path.newSubPath("esmaPartyInformation"), processor, ESMAPartyInformation.ESMAPartyInformationBuilder.class, getEsmaPartyInformation());
			processRosetta(path.newSubPath("secPartyInformation"), processor, SECPartyInformation.SECPartyInformationBuilder.class, getSecPartyInformation());
			processor.processBasic(path.newSubPath("reportingRole"), ReportingRoleEnum.class, getReportingRole(), this);
		}
		

		ReportingRegime.ReportingRegimeBuilder prune();
	}

	/*********************** Immutable Implementation of ReportingRegime  ***********************/
	class ReportingRegimeImpl implements ReportingRegime {
		private final FieldWithMetaRegimeNameEnum regimeName;
		private final FieldWithMetaSupervisoryBodyEnum supervisoryBody;
		private final MandatorilyClearableEnum mandatorilyClearable;
		private final ClearingException clearingException;
		private final Boolean postPricedIndicator;
		private final CFTCPartyInformation cftcPartyInformation;
		private final ESMAPartyInformation esmaPartyInformation;
		private final SECPartyInformation secPartyInformation;
		private final ReportingRoleEnum reportingRole;
		
		protected ReportingRegimeImpl(ReportingRegime.ReportingRegimeBuilder builder) {
			this.regimeName = ofNullable(builder.getRegimeName()).map(f->f.build()).orElse(null);
			this.supervisoryBody = ofNullable(builder.getSupervisoryBody()).map(f->f.build()).orElse(null);
			this.mandatorilyClearable = builder.getMandatorilyClearable();
			this.clearingException = ofNullable(builder.getClearingException()).map(f->f.build()).orElse(null);
			this.postPricedIndicator = builder.getPostPricedIndicator();
			this.cftcPartyInformation = ofNullable(builder.getCftcPartyInformation()).map(f->f.build()).orElse(null);
			this.esmaPartyInformation = ofNullable(builder.getEsmaPartyInformation()).map(f->f.build()).orElse(null);
			this.secPartyInformation = ofNullable(builder.getSecPartyInformation()).map(f->f.build()).orElse(null);
			this.reportingRole = builder.getReportingRole();
		}
		
		@Override
		@RosettaAttribute("regimeName")
		public FieldWithMetaRegimeNameEnum getRegimeName() {
			return regimeName;
		}
		
		@Override
		@RosettaAttribute("supervisoryBody")
		public FieldWithMetaSupervisoryBodyEnum getSupervisoryBody() {
			return supervisoryBody;
		}
		
		@Override
		@RosettaAttribute("mandatorilyClearable")
		public MandatorilyClearableEnum getMandatorilyClearable() {
			return mandatorilyClearable;
		}
		
		@Override
		@RosettaAttribute("clearingException")
		public ClearingException getClearingException() {
			return clearingException;
		}
		
		@Override
		@RosettaAttribute("postPricedIndicator")
		public Boolean getPostPricedIndicator() {
			return postPricedIndicator;
		}
		
		@Override
		@RosettaAttribute("cftcPartyInformation")
		public CFTCPartyInformation getCftcPartyInformation() {
			return cftcPartyInformation;
		}
		
		@Override
		@RosettaAttribute("esmaPartyInformation")
		public ESMAPartyInformation getEsmaPartyInformation() {
			return esmaPartyInformation;
		}
		
		@Override
		@RosettaAttribute("secPartyInformation")
		public SECPartyInformation getSecPartyInformation() {
			return secPartyInformation;
		}
		
		@Override
		@RosettaAttribute("reportingRole")
		public ReportingRoleEnum getReportingRole() {
			return reportingRole;
		}
		
		@Override
		public ReportingRegime build() {
			return this;
		}
		
		@Override
		public ReportingRegime.ReportingRegimeBuilder toBuilder() {
			ReportingRegime.ReportingRegimeBuilder builder = builder();
			setBuilderFields(builder);
			return builder;
		}
		
		protected void setBuilderFields(ReportingRegime.ReportingRegimeBuilder builder) {
			ofNullable(getRegimeName()).ifPresent(builder::setRegimeName);
			ofNullable(getSupervisoryBody()).ifPresent(builder::setSupervisoryBody);
			ofNullable(getMandatorilyClearable()).ifPresent(builder::setMandatorilyClearable);
			ofNullable(getClearingException()).ifPresent(builder::setClearingException);
			ofNullable(getPostPricedIndicator()).ifPresent(builder::setPostPricedIndicator);
			ofNullable(getCftcPartyInformation()).ifPresent(builder::setCftcPartyInformation);
			ofNullable(getEsmaPartyInformation()).ifPresent(builder::setEsmaPartyInformation);
			ofNullable(getSecPartyInformation()).ifPresent(builder::setSecPartyInformation);
			ofNullable(getReportingRole()).ifPresent(builder::setReportingRole);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			ReportingRegime _that = getType().cast(o);
		
			if (!Objects.equals(regimeName, _that.getRegimeName())) return false;
			if (!Objects.equals(supervisoryBody, _that.getSupervisoryBody())) return false;
			if (!Objects.equals(mandatorilyClearable, _that.getMandatorilyClearable())) return false;
			if (!Objects.equals(clearingException, _that.getClearingException())) return false;
			if (!Objects.equals(postPricedIndicator, _that.getPostPricedIndicator())) return false;
			if (!Objects.equals(cftcPartyInformation, _that.getCftcPartyInformation())) return false;
			if (!Objects.equals(esmaPartyInformation, _that.getEsmaPartyInformation())) return false;
			if (!Objects.equals(secPartyInformation, _that.getSecPartyInformation())) return false;
			if (!Objects.equals(reportingRole, _that.getReportingRole())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (regimeName != null ? regimeName.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (supervisoryBody != null ? supervisoryBody.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (mandatorilyClearable != null ? mandatorilyClearable.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (clearingException != null ? clearingException.hashCode() : 0);
			_result = 31 * _result + (postPricedIndicator != null ? postPricedIndicator.hashCode() : 0);
			_result = 31 * _result + (cftcPartyInformation != null ? cftcPartyInformation.hashCode() : 0);
			_result = 31 * _result + (esmaPartyInformation != null ? esmaPartyInformation.hashCode() : 0);
			_result = 31 * _result + (secPartyInformation != null ? secPartyInformation.hashCode() : 0);
			_result = 31 * _result + (reportingRole != null ? reportingRole.getClass().getName().hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "ReportingRegime {" +
				"regimeName=" + this.regimeName + ", " +
				"supervisoryBody=" + this.supervisoryBody + ", " +
				"mandatorilyClearable=" + this.mandatorilyClearable + ", " +
				"clearingException=" + this.clearingException + ", " +
				"postPricedIndicator=" + this.postPricedIndicator + ", " +
				"cftcPartyInformation=" + this.cftcPartyInformation + ", " +
				"esmaPartyInformation=" + this.esmaPartyInformation + ", " +
				"secPartyInformation=" + this.secPartyInformation + ", " +
				"reportingRole=" + this.reportingRole +
			'}';
		}
	}

	/*********************** Builder Implementation of ReportingRegime  ***********************/
	class ReportingRegimeBuilderImpl implements ReportingRegime.ReportingRegimeBuilder {
	
		protected FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder regimeName;
		protected FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder supervisoryBody;
		protected MandatorilyClearableEnum mandatorilyClearable;
		protected ClearingException.ClearingExceptionBuilder clearingException;
		protected Boolean postPricedIndicator;
		protected CFTCPartyInformation.CFTCPartyInformationBuilder cftcPartyInformation;
		protected ESMAPartyInformation.ESMAPartyInformationBuilder esmaPartyInformation;
		protected SECPartyInformation.SECPartyInformationBuilder secPartyInformation;
		protected ReportingRoleEnum reportingRole;
	
		public ReportingRegimeBuilderImpl() {
		}
	
		@Override
		@RosettaAttribute("regimeName")
		public FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder getRegimeName() {
			return regimeName;
		}
		
		@Override
		public FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder getOrCreateRegimeName() {
			FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder result;
			if (regimeName!=null) {
				result = regimeName;
			}
			else {
				result = regimeName = FieldWithMetaRegimeNameEnum.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("supervisoryBody")
		public FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder getSupervisoryBody() {
			return supervisoryBody;
		}
		
		@Override
		public FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder getOrCreateSupervisoryBody() {
			FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder result;
			if (supervisoryBody!=null) {
				result = supervisoryBody;
			}
			else {
				result = supervisoryBody = FieldWithMetaSupervisoryBodyEnum.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("mandatorilyClearable")
		public MandatorilyClearableEnum getMandatorilyClearable() {
			return mandatorilyClearable;
		}
		
		@Override
		@RosettaAttribute("clearingException")
		public ClearingException.ClearingExceptionBuilder getClearingException() {
			return clearingException;
		}
		
		@Override
		public ClearingException.ClearingExceptionBuilder getOrCreateClearingException() {
			ClearingException.ClearingExceptionBuilder result;
			if (clearingException!=null) {
				result = clearingException;
			}
			else {
				result = clearingException = ClearingException.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("postPricedIndicator")
		public Boolean getPostPricedIndicator() {
			return postPricedIndicator;
		}
		
		@Override
		@RosettaAttribute("cftcPartyInformation")
		public CFTCPartyInformation.CFTCPartyInformationBuilder getCftcPartyInformation() {
			return cftcPartyInformation;
		}
		
		@Override
		public CFTCPartyInformation.CFTCPartyInformationBuilder getOrCreateCftcPartyInformation() {
			CFTCPartyInformation.CFTCPartyInformationBuilder result;
			if (cftcPartyInformation!=null) {
				result = cftcPartyInformation;
			}
			else {
				result = cftcPartyInformation = CFTCPartyInformation.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("esmaPartyInformation")
		public ESMAPartyInformation.ESMAPartyInformationBuilder getEsmaPartyInformation() {
			return esmaPartyInformation;
		}
		
		@Override
		public ESMAPartyInformation.ESMAPartyInformationBuilder getOrCreateEsmaPartyInformation() {
			ESMAPartyInformation.ESMAPartyInformationBuilder result;
			if (esmaPartyInformation!=null) {
				result = esmaPartyInformation;
			}
			else {
				result = esmaPartyInformation = ESMAPartyInformation.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("secPartyInformation")
		public SECPartyInformation.SECPartyInformationBuilder getSecPartyInformation() {
			return secPartyInformation;
		}
		
		@Override
		public SECPartyInformation.SECPartyInformationBuilder getOrCreateSecPartyInformation() {
			SECPartyInformation.SECPartyInformationBuilder result;
			if (secPartyInformation!=null) {
				result = secPartyInformation;
			}
			else {
				result = secPartyInformation = SECPartyInformation.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("reportingRole")
		public ReportingRoleEnum getReportingRole() {
			return reportingRole;
		}
		
	
		@Override
		@RosettaAttribute("regimeName")
		public ReportingRegime.ReportingRegimeBuilder setRegimeName(FieldWithMetaRegimeNameEnum regimeName) {
			this.regimeName = regimeName==null?null:regimeName.toBuilder();
			return this;
		}
		@Override
		public ReportingRegime.ReportingRegimeBuilder setRegimeNameValue(RegimeNameEnum regimeName) {
			this.getOrCreateRegimeName().setValue(regimeName);
			return this;
		}
		@Override
		@RosettaAttribute("supervisoryBody")
		public ReportingRegime.ReportingRegimeBuilder setSupervisoryBody(FieldWithMetaSupervisoryBodyEnum supervisoryBody) {
			this.supervisoryBody = supervisoryBody==null?null:supervisoryBody.toBuilder();
			return this;
		}
		@Override
		public ReportingRegime.ReportingRegimeBuilder setSupervisoryBodyValue(SupervisoryBodyEnum supervisoryBody) {
			this.getOrCreateSupervisoryBody().setValue(supervisoryBody);
			return this;
		}
		@Override
		@RosettaAttribute("mandatorilyClearable")
		public ReportingRegime.ReportingRegimeBuilder setMandatorilyClearable(MandatorilyClearableEnum mandatorilyClearable) {
			this.mandatorilyClearable = mandatorilyClearable==null?null:mandatorilyClearable;
			return this;
		}
		@Override
		@RosettaAttribute("clearingException")
		public ReportingRegime.ReportingRegimeBuilder setClearingException(ClearingException clearingException) {
			this.clearingException = clearingException==null?null:clearingException.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("postPricedIndicator")
		public ReportingRegime.ReportingRegimeBuilder setPostPricedIndicator(Boolean postPricedIndicator) {
			this.postPricedIndicator = postPricedIndicator==null?null:postPricedIndicator;
			return this;
		}
		@Override
		@RosettaAttribute("cftcPartyInformation")
		public ReportingRegime.ReportingRegimeBuilder setCftcPartyInformation(CFTCPartyInformation cftcPartyInformation) {
			this.cftcPartyInformation = cftcPartyInformation==null?null:cftcPartyInformation.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("esmaPartyInformation")
		public ReportingRegime.ReportingRegimeBuilder setEsmaPartyInformation(ESMAPartyInformation esmaPartyInformation) {
			this.esmaPartyInformation = esmaPartyInformation==null?null:esmaPartyInformation.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("secPartyInformation")
		public ReportingRegime.ReportingRegimeBuilder setSecPartyInformation(SECPartyInformation secPartyInformation) {
			this.secPartyInformation = secPartyInformation==null?null:secPartyInformation.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("reportingRole")
		public ReportingRegime.ReportingRegimeBuilder setReportingRole(ReportingRoleEnum reportingRole) {
			this.reportingRole = reportingRole==null?null:reportingRole;
			return this;
		}
		
		@Override
		public ReportingRegime build() {
			return new ReportingRegime.ReportingRegimeImpl(this);
		}
		
		@Override
		public ReportingRegime.ReportingRegimeBuilder toBuilder() {
			return this;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public ReportingRegime.ReportingRegimeBuilder prune() {
			if (regimeName!=null && !regimeName.prune().hasData()) regimeName = null;
			if (supervisoryBody!=null && !supervisoryBody.prune().hasData()) supervisoryBody = null;
			if (clearingException!=null && !clearingException.prune().hasData()) clearingException = null;
			if (cftcPartyInformation!=null && !cftcPartyInformation.prune().hasData()) cftcPartyInformation = null;
			if (esmaPartyInformation!=null && !esmaPartyInformation.prune().hasData()) esmaPartyInformation = null;
			if (secPartyInformation!=null && !secPartyInformation.prune().hasData()) secPartyInformation = null;
			return this;
		}
		
		@Override
		public boolean hasData() {
			if (getRegimeName()!=null) return true;
			if (getSupervisoryBody()!=null) return true;
			if (getMandatorilyClearable()!=null) return true;
			if (getClearingException()!=null && getClearingException().hasData()) return true;
			if (getPostPricedIndicator()!=null) return true;
			if (getCftcPartyInformation()!=null && getCftcPartyInformation().hasData()) return true;
			if (getEsmaPartyInformation()!=null && getEsmaPartyInformation().hasData()) return true;
			if (getSecPartyInformation()!=null && getSecPartyInformation().hasData()) return true;
			if (getReportingRole()!=null) return true;
			return false;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public ReportingRegime.ReportingRegimeBuilder merge(RosettaModelObjectBuilder other, BuilderMerger merger) {
			ReportingRegime.ReportingRegimeBuilder o = (ReportingRegime.ReportingRegimeBuilder) other;
			
			merger.mergeRosetta(getRegimeName(), o.getRegimeName(), this::setRegimeName);
			merger.mergeRosetta(getSupervisoryBody(), o.getSupervisoryBody(), this::setSupervisoryBody);
			merger.mergeRosetta(getClearingException(), o.getClearingException(), this::setClearingException);
			merger.mergeRosetta(getCftcPartyInformation(), o.getCftcPartyInformation(), this::setCftcPartyInformation);
			merger.mergeRosetta(getEsmaPartyInformation(), o.getEsmaPartyInformation(), this::setEsmaPartyInformation);
			merger.mergeRosetta(getSecPartyInformation(), o.getSecPartyInformation(), this::setSecPartyInformation);
			
			merger.mergeBasic(getMandatorilyClearable(), o.getMandatorilyClearable(), this::setMandatorilyClearable);
			merger.mergeBasic(getPostPricedIndicator(), o.getPostPricedIndicator(), this::setPostPricedIndicator);
			merger.mergeBasic(getReportingRole(), o.getReportingRole(), this::setReportingRole);
			return this;
		}
	
		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			ReportingRegime _that = getType().cast(o);
		
			if (!Objects.equals(regimeName, _that.getRegimeName())) return false;
			if (!Objects.equals(supervisoryBody, _that.getSupervisoryBody())) return false;
			if (!Objects.equals(mandatorilyClearable, _that.getMandatorilyClearable())) return false;
			if (!Objects.equals(clearingException, _that.getClearingException())) return false;
			if (!Objects.equals(postPricedIndicator, _that.getPostPricedIndicator())) return false;
			if (!Objects.equals(cftcPartyInformation, _that.getCftcPartyInformation())) return false;
			if (!Objects.equals(esmaPartyInformation, _that.getEsmaPartyInformation())) return false;
			if (!Objects.equals(secPartyInformation, _that.getSecPartyInformation())) return false;
			if (!Objects.equals(reportingRole, _that.getReportingRole())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (regimeName != null ? regimeName.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (supervisoryBody != null ? supervisoryBody.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (mandatorilyClearable != null ? mandatorilyClearable.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (clearingException != null ? clearingException.hashCode() : 0);
			_result = 31 * _result + (postPricedIndicator != null ? postPricedIndicator.hashCode() : 0);
			_result = 31 * _result + (cftcPartyInformation != null ? cftcPartyInformation.hashCode() : 0);
			_result = 31 * _result + (esmaPartyInformation != null ? esmaPartyInformation.hashCode() : 0);
			_result = 31 * _result + (secPartyInformation != null ? secPartyInformation.hashCode() : 0);
			_result = 31 * _result + (reportingRole != null ? reportingRole.getClass().getName().hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "ReportingRegimeBuilder {" +
				"regimeName=" + this.regimeName + ", " +
				"supervisoryBody=" + this.supervisoryBody + ", " +
				"mandatorilyClearable=" + this.mandatorilyClearable + ", " +
				"clearingException=" + this.clearingException + ", " +
				"postPricedIndicator=" + this.postPricedIndicator + ", " +
				"cftcPartyInformation=" + this.cftcPartyInformation + ", " +
				"esmaPartyInformation=" + this.esmaPartyInformation + ", " +
				"secPartyInformation=" + this.secPartyInformation + ", " +
				"reportingRole=" + this.reportingRole +
			'}';
		}
	}
}
