package drr.regulation.common;

import cdm.base.staticdata.party.Party;
import cdm.base.staticdata.party.metafields.ReferenceWithMetaParty;
import cdm.base.staticdata.party.metafields.ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder;
import com.rosetta.model.lib.RosettaModelObject;
import com.rosetta.model.lib.RosettaModelObjectBuilder;
import com.rosetta.model.lib.annotations.RosettaAttribute;
import com.rosetta.model.lib.annotations.RosettaDataType;
import com.rosetta.model.lib.meta.RosettaMetaData;
import com.rosetta.model.lib.path.RosettaPath;
import com.rosetta.model.lib.process.BuilderMerger;
import com.rosetta.model.lib.process.BuilderProcessor;
import com.rosetta.model.lib.process.Processor;
import drr.regulation.common.ReportingSide;
import drr.regulation.common.ReportingSide.ReportingSideBuilder;
import drr.regulation.common.ReportingSide.ReportingSideBuilderImpl;
import drr.regulation.common.ReportingSide.ReportingSideImpl;
import drr.regulation.common.meta.ReportingSideMeta;
import java.util.Objects;

import static java.util.Optional.ofNullable;

/**
 * Specifies the reporting roles for a ReportableEvent or ReportableCollateral.
 * @version 4.2.1
 */
@RosettaDataType(value="ReportingSide", builder=ReportingSide.ReportingSideBuilderImpl.class, version="4.2.1")
public interface ReportingSide extends RosettaModelObject {

	ReportingSideMeta metaData = new ReportingSideMeta();

	/*********************** Getter Methods  ***********************/
	/**
	 * The party responsible for reporting this transaction.
	 */
	ReferenceWithMetaParty getReportingParty();
	ReferenceWithMetaParty getReportingCounterparty();
	/**
	 * The party responsibility for submitting the transaction report to the authority.
	 */
	ReferenceWithMetaParty getPartyResponsibleForReporting();
	/**
	 * The party submitting the transaction report to the authority.
	 */
	ReferenceWithMetaParty getReportSubmittingParty();

	/*********************** Build Methods  ***********************/
	ReportingSide build();
	
	ReportingSide.ReportingSideBuilder toBuilder();
	
	static ReportingSide.ReportingSideBuilder builder() {
		return new ReportingSide.ReportingSideBuilderImpl();
	}

	/*********************** Utility Methods  ***********************/
	@Override
	default RosettaMetaData<? extends ReportingSide> metaData() {
		return metaData;
	}
	
	@Override
	default Class<? extends ReportingSide> getType() {
		return ReportingSide.class;
	}
	
	
	@Override
	default void process(RosettaPath path, Processor processor) {
		processRosetta(path.newSubPath("reportingParty"), processor, ReferenceWithMetaParty.class, getReportingParty());
		processRosetta(path.newSubPath("reportingCounterparty"), processor, ReferenceWithMetaParty.class, getReportingCounterparty());
		processRosetta(path.newSubPath("partyResponsibleForReporting"), processor, ReferenceWithMetaParty.class, getPartyResponsibleForReporting());
		processRosetta(path.newSubPath("reportSubmittingParty"), processor, ReferenceWithMetaParty.class, getReportSubmittingParty());
	}
	

	/*********************** Builder Interface  ***********************/
	interface ReportingSideBuilder extends ReportingSide, RosettaModelObjectBuilder {
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreateReportingParty();
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getReportingParty();
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreateReportingCounterparty();
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getReportingCounterparty();
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreatePartyResponsibleForReporting();
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getPartyResponsibleForReporting();
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreateReportSubmittingParty();
		ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getReportSubmittingParty();
		ReportingSide.ReportingSideBuilder setReportingParty(ReferenceWithMetaParty reportingParty0);
		ReportingSide.ReportingSideBuilder setReportingPartyValue(Party reportingParty1);
		ReportingSide.ReportingSideBuilder setReportingCounterparty(ReferenceWithMetaParty reportingCounterparty0);
		ReportingSide.ReportingSideBuilder setReportingCounterpartyValue(Party reportingCounterparty1);
		ReportingSide.ReportingSideBuilder setPartyResponsibleForReporting(ReferenceWithMetaParty partyResponsibleForReporting0);
		ReportingSide.ReportingSideBuilder setPartyResponsibleForReportingValue(Party partyResponsibleForReporting1);
		ReportingSide.ReportingSideBuilder setReportSubmittingParty(ReferenceWithMetaParty reportSubmittingParty0);
		ReportingSide.ReportingSideBuilder setReportSubmittingPartyValue(Party reportSubmittingParty1);

		@Override
		default void process(RosettaPath path, BuilderProcessor processor) {
			processRosetta(path.newSubPath("reportingParty"), processor, ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder.class, getReportingParty());
			processRosetta(path.newSubPath("reportingCounterparty"), processor, ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder.class, getReportingCounterparty());
			processRosetta(path.newSubPath("partyResponsibleForReporting"), processor, ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder.class, getPartyResponsibleForReporting());
			processRosetta(path.newSubPath("reportSubmittingParty"), processor, ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder.class, getReportSubmittingParty());
		}
		

		ReportingSide.ReportingSideBuilder prune();
	}

	/*********************** Immutable Implementation of ReportingSide  ***********************/
	class ReportingSideImpl implements ReportingSide {
		private final ReferenceWithMetaParty reportingParty;
		private final ReferenceWithMetaParty reportingCounterparty;
		private final ReferenceWithMetaParty partyResponsibleForReporting;
		private final ReferenceWithMetaParty reportSubmittingParty;
		
		protected ReportingSideImpl(ReportingSide.ReportingSideBuilder builder) {
			this.reportingParty = ofNullable(builder.getReportingParty()).map(f->f.build()).orElse(null);
			this.reportingCounterparty = ofNullable(builder.getReportingCounterparty()).map(f->f.build()).orElse(null);
			this.partyResponsibleForReporting = ofNullable(builder.getPartyResponsibleForReporting()).map(f->f.build()).orElse(null);
			this.reportSubmittingParty = ofNullable(builder.getReportSubmittingParty()).map(f->f.build()).orElse(null);
		}
		
		@Override
		@RosettaAttribute("reportingParty")
		public ReferenceWithMetaParty getReportingParty() {
			return reportingParty;
		}
		
		@Override
		@RosettaAttribute("reportingCounterparty")
		public ReferenceWithMetaParty getReportingCounterparty() {
			return reportingCounterparty;
		}
		
		@Override
		@RosettaAttribute("partyResponsibleForReporting")
		public ReferenceWithMetaParty getPartyResponsibleForReporting() {
			return partyResponsibleForReporting;
		}
		
		@Override
		@RosettaAttribute("reportSubmittingParty")
		public ReferenceWithMetaParty getReportSubmittingParty() {
			return reportSubmittingParty;
		}
		
		@Override
		public ReportingSide build() {
			return this;
		}
		
		@Override
		public ReportingSide.ReportingSideBuilder toBuilder() {
			ReportingSide.ReportingSideBuilder builder = builder();
			setBuilderFields(builder);
			return builder;
		}
		
		protected void setBuilderFields(ReportingSide.ReportingSideBuilder builder) {
			ofNullable(getReportingParty()).ifPresent(builder::setReportingParty);
			ofNullable(getReportingCounterparty()).ifPresent(builder::setReportingCounterparty);
			ofNullable(getPartyResponsibleForReporting()).ifPresent(builder::setPartyResponsibleForReporting);
			ofNullable(getReportSubmittingParty()).ifPresent(builder::setReportSubmittingParty);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			ReportingSide _that = getType().cast(o);
		
			if (!Objects.equals(reportingParty, _that.getReportingParty())) return false;
			if (!Objects.equals(reportingCounterparty, _that.getReportingCounterparty())) return false;
			if (!Objects.equals(partyResponsibleForReporting, _that.getPartyResponsibleForReporting())) return false;
			if (!Objects.equals(reportSubmittingParty, _that.getReportSubmittingParty())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (reportingParty != null ? reportingParty.hashCode() : 0);
			_result = 31 * _result + (reportingCounterparty != null ? reportingCounterparty.hashCode() : 0);
			_result = 31 * _result + (partyResponsibleForReporting != null ? partyResponsibleForReporting.hashCode() : 0);
			_result = 31 * _result + (reportSubmittingParty != null ? reportSubmittingParty.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "ReportingSide {" +
				"reportingParty=" + this.reportingParty + ", " +
				"reportingCounterparty=" + this.reportingCounterparty + ", " +
				"partyResponsibleForReporting=" + this.partyResponsibleForReporting + ", " +
				"reportSubmittingParty=" + this.reportSubmittingParty +
			'}';
		}
	}

	/*********************** Builder Implementation of ReportingSide  ***********************/
	class ReportingSideBuilderImpl implements ReportingSide.ReportingSideBuilder {
	
		protected ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder reportingParty;
		protected ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder reportingCounterparty;
		protected ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder partyResponsibleForReporting;
		protected ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder reportSubmittingParty;
	
		public ReportingSideBuilderImpl() {
		}
	
		@Override
		@RosettaAttribute("reportingParty")
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getReportingParty() {
			return reportingParty;
		}
		
		@Override
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreateReportingParty() {
			ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder result;
			if (reportingParty!=null) {
				result = reportingParty;
			}
			else {
				result = reportingParty = ReferenceWithMetaParty.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("reportingCounterparty")
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getReportingCounterparty() {
			return reportingCounterparty;
		}
		
		@Override
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreateReportingCounterparty() {
			ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder result;
			if (reportingCounterparty!=null) {
				result = reportingCounterparty;
			}
			else {
				result = reportingCounterparty = ReferenceWithMetaParty.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("partyResponsibleForReporting")
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getPartyResponsibleForReporting() {
			return partyResponsibleForReporting;
		}
		
		@Override
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreatePartyResponsibleForReporting() {
			ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder result;
			if (partyResponsibleForReporting!=null) {
				result = partyResponsibleForReporting;
			}
			else {
				result = partyResponsibleForReporting = ReferenceWithMetaParty.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("reportSubmittingParty")
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getReportSubmittingParty() {
			return reportSubmittingParty;
		}
		
		@Override
		public ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder getOrCreateReportSubmittingParty() {
			ReferenceWithMetaParty.ReferenceWithMetaPartyBuilder result;
			if (reportSubmittingParty!=null) {
				result = reportSubmittingParty;
			}
			else {
				result = reportSubmittingParty = ReferenceWithMetaParty.builder();
			}
			
			return result;
		}
	
		@Override
		@RosettaAttribute("reportingParty")
		public ReportingSide.ReportingSideBuilder setReportingParty(ReferenceWithMetaParty reportingParty) {
			this.reportingParty = reportingParty==null?null:reportingParty.toBuilder();
			return this;
		}
		@Override
		public ReportingSide.ReportingSideBuilder setReportingPartyValue(Party reportingParty) {
			this.getOrCreateReportingParty().setValue(reportingParty);
			return this;
		}
		@Override
		@RosettaAttribute("reportingCounterparty")
		public ReportingSide.ReportingSideBuilder setReportingCounterparty(ReferenceWithMetaParty reportingCounterparty) {
			this.reportingCounterparty = reportingCounterparty==null?null:reportingCounterparty.toBuilder();
			return this;
		}
		@Override
		public ReportingSide.ReportingSideBuilder setReportingCounterpartyValue(Party reportingCounterparty) {
			this.getOrCreateReportingCounterparty().setValue(reportingCounterparty);
			return this;
		}
		@Override
		@RosettaAttribute("partyResponsibleForReporting")
		public ReportingSide.ReportingSideBuilder setPartyResponsibleForReporting(ReferenceWithMetaParty partyResponsibleForReporting) {
			this.partyResponsibleForReporting = partyResponsibleForReporting==null?null:partyResponsibleForReporting.toBuilder();
			return this;
		}
		@Override
		public ReportingSide.ReportingSideBuilder setPartyResponsibleForReportingValue(Party partyResponsibleForReporting) {
			this.getOrCreatePartyResponsibleForReporting().setValue(partyResponsibleForReporting);
			return this;
		}
		@Override
		@RosettaAttribute("reportSubmittingParty")
		public ReportingSide.ReportingSideBuilder setReportSubmittingParty(ReferenceWithMetaParty reportSubmittingParty) {
			this.reportSubmittingParty = reportSubmittingParty==null?null:reportSubmittingParty.toBuilder();
			return this;
		}
		@Override
		public ReportingSide.ReportingSideBuilder setReportSubmittingPartyValue(Party reportSubmittingParty) {
			this.getOrCreateReportSubmittingParty().setValue(reportSubmittingParty);
			return this;
		}
		
		@Override
		public ReportingSide build() {
			return new ReportingSide.ReportingSideImpl(this);
		}
		
		@Override
		public ReportingSide.ReportingSideBuilder toBuilder() {
			return this;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public ReportingSide.ReportingSideBuilder prune() {
			if (reportingParty!=null && !reportingParty.prune().hasData()) reportingParty = null;
			if (reportingCounterparty!=null && !reportingCounterparty.prune().hasData()) reportingCounterparty = null;
			if (partyResponsibleForReporting!=null && !partyResponsibleForReporting.prune().hasData()) partyResponsibleForReporting = null;
			if (reportSubmittingParty!=null && !reportSubmittingParty.prune().hasData()) reportSubmittingParty = null;
			return this;
		}
		
		@Override
		public boolean hasData() {
			if (getReportingParty()!=null && getReportingParty().hasData()) return true;
			if (getReportingCounterparty()!=null && getReportingCounterparty().hasData()) return true;
			if (getPartyResponsibleForReporting()!=null && getPartyResponsibleForReporting().hasData()) return true;
			if (getReportSubmittingParty()!=null && getReportSubmittingParty().hasData()) return true;
			return false;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public ReportingSide.ReportingSideBuilder merge(RosettaModelObjectBuilder other, BuilderMerger merger) {
			ReportingSide.ReportingSideBuilder o = (ReportingSide.ReportingSideBuilder) other;
			
			merger.mergeRosetta(getReportingParty(), o.getReportingParty(), this::setReportingParty);
			merger.mergeRosetta(getReportingCounterparty(), o.getReportingCounterparty(), this::setReportingCounterparty);
			merger.mergeRosetta(getPartyResponsibleForReporting(), o.getPartyResponsibleForReporting(), this::setPartyResponsibleForReporting);
			merger.mergeRosetta(getReportSubmittingParty(), o.getReportSubmittingParty(), this::setReportSubmittingParty);
			
			return this;
		}
	
		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			ReportingSide _that = getType().cast(o);
		
			if (!Objects.equals(reportingParty, _that.getReportingParty())) return false;
			if (!Objects.equals(reportingCounterparty, _that.getReportingCounterparty())) return false;
			if (!Objects.equals(partyResponsibleForReporting, _that.getPartyResponsibleForReporting())) return false;
			if (!Objects.equals(reportSubmittingParty, _that.getReportSubmittingParty())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (reportingParty != null ? reportingParty.hashCode() : 0);
			_result = 31 * _result + (reportingCounterparty != null ? reportingCounterparty.hashCode() : 0);
			_result = 31 * _result + (partyResponsibleForReporting != null ? partyResponsibleForReporting.hashCode() : 0);
			_result = 31 * _result + (reportSubmittingParty != null ? reportSubmittingParty.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "ReportingSideBuilder {" +
				"reportingParty=" + this.reportingParty + ", " +
				"reportingCounterparty=" + this.reportingCounterparty + ", " +
				"partyResponsibleForReporting=" + this.partyResponsibleForReporting + ", " +
				"reportSubmittingParty=" + this.reportSubmittingParty +
			'}';
		}
	}
}
