package drr.regulation.common;

import cdm.event.common.TradeIdentifier;
import com.rosetta.model.lib.RosettaModelObject;
import com.rosetta.model.lib.RosettaModelObjectBuilder;
import com.rosetta.model.lib.annotations.RosettaAttribute;
import com.rosetta.model.lib.annotations.RosettaDataType;
import com.rosetta.model.lib.meta.RosettaMetaData;
import com.rosetta.model.lib.path.RosettaPath;
import com.rosetta.model.lib.process.BuilderMerger;
import com.rosetta.model.lib.process.BuilderProcessor;
import com.rosetta.model.lib.process.Processor;
import drr.regulation.common.ESMATransactionInformation;
import drr.regulation.common.RegimeNameEnum;
import drr.regulation.common.SupervisoryBodyEnum;
import drr.regulation.common.TransactionInformation;
import drr.regulation.common.TransactionInformation.TransactionInformationBuilder;
import drr.regulation.common.TransactionInformation.TransactionInformationBuilderImpl;
import drr.regulation.common.TransactionInformation.TransactionInformationImpl;
import drr.regulation.common.meta.TransactionInformationMeta;
import drr.regulation.common.metafields.FieldWithMetaRegimeNameEnum;
import drr.regulation.common.metafields.FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder;
import drr.regulation.common.metafields.FieldWithMetaSupervisoryBodyEnum;
import drr.regulation.common.metafields.FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder;
import java.util.Objects;

import static java.util.Optional.ofNullable;

/**
 * @version 4.2.1
 */
@RosettaDataType(value="TransactionInformation", builder=TransactionInformation.TransactionInformationBuilderImpl.class, version="4.2.1")
public interface TransactionInformation extends RosettaModelObject {

	TransactionInformationMeta metaData = new TransactionInformationMeta();

	/*********************** Getter Methods  ***********************/
	/**
	 * Identifies the reporting regime under which this data is reported. For example, Dodd-Frank, MiFID, HongKongOTCDRepository, ODRF
	 */
	FieldWithMetaRegimeNameEnum getRegimeName();
	/**
	 * Identifies the specific regulator or other supervisory body for which this data is produced. For example, CFTC, SEC, UKFSA, ODRF, SFC, ESMA.
	 */
	FieldWithMetaSupervisoryBodyEnum getSupervisoryBody();
	/**
	 * Identifier to use specifically in the context of reporting to a specific regime
	 */
	TradeIdentifier getTransactionIdentifier();
	/**
	 * Transaction information required for reporting a transaction under ESMA.
	 */
	ESMATransactionInformation getEsmaTransactionInformation();

	/*********************** Build Methods  ***********************/
	TransactionInformation build();
	
	TransactionInformation.TransactionInformationBuilder toBuilder();
	
	static TransactionInformation.TransactionInformationBuilder builder() {
		return new TransactionInformation.TransactionInformationBuilderImpl();
	}

	/*********************** Utility Methods  ***********************/
	@Override
	default RosettaMetaData<? extends TransactionInformation> metaData() {
		return metaData;
	}
	
	@Override
	default Class<? extends TransactionInformation> getType() {
		return TransactionInformation.class;
	}
	
	
	@Override
	default void process(RosettaPath path, Processor processor) {
		processRosetta(path.newSubPath("regimeName"), processor, FieldWithMetaRegimeNameEnum.class, getRegimeName());
		processRosetta(path.newSubPath("supervisoryBody"), processor, FieldWithMetaSupervisoryBodyEnum.class, getSupervisoryBody());
		processRosetta(path.newSubPath("transactionIdentifier"), processor, TradeIdentifier.class, getTransactionIdentifier());
		processRosetta(path.newSubPath("esmaTransactionInformation"), processor, ESMATransactionInformation.class, getEsmaTransactionInformation());
	}
	

	/*********************** Builder Interface  ***********************/
	interface TransactionInformationBuilder extends TransactionInformation, RosettaModelObjectBuilder {
		FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder getOrCreateRegimeName();
		FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder getRegimeName();
		FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder getOrCreateSupervisoryBody();
		FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder getSupervisoryBody();
		TradeIdentifier.TradeIdentifierBuilder getOrCreateTransactionIdentifier();
		TradeIdentifier.TradeIdentifierBuilder getTransactionIdentifier();
		ESMATransactionInformation.ESMATransactionInformationBuilder getOrCreateEsmaTransactionInformation();
		ESMATransactionInformation.ESMATransactionInformationBuilder getEsmaTransactionInformation();
		TransactionInformation.TransactionInformationBuilder setRegimeName(FieldWithMetaRegimeNameEnum regimeName0);
		TransactionInformation.TransactionInformationBuilder setRegimeNameValue(RegimeNameEnum regimeName1);
		TransactionInformation.TransactionInformationBuilder setSupervisoryBody(FieldWithMetaSupervisoryBodyEnum supervisoryBody0);
		TransactionInformation.TransactionInformationBuilder setSupervisoryBodyValue(SupervisoryBodyEnum supervisoryBody1);
		TransactionInformation.TransactionInformationBuilder setTransactionIdentifier(TradeIdentifier transactionIdentifier);
		TransactionInformation.TransactionInformationBuilder setEsmaTransactionInformation(ESMATransactionInformation esmaTransactionInformation);

		@Override
		default void process(RosettaPath path, BuilderProcessor processor) {
			processRosetta(path.newSubPath("regimeName"), processor, FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder.class, getRegimeName());
			processRosetta(path.newSubPath("supervisoryBody"), processor, FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder.class, getSupervisoryBody());
			processRosetta(path.newSubPath("transactionIdentifier"), processor, TradeIdentifier.TradeIdentifierBuilder.class, getTransactionIdentifier());
			processRosetta(path.newSubPath("esmaTransactionInformation"), processor, ESMATransactionInformation.ESMATransactionInformationBuilder.class, getEsmaTransactionInformation());
		}
		

		TransactionInformation.TransactionInformationBuilder prune();
	}

	/*********************** Immutable Implementation of TransactionInformation  ***********************/
	class TransactionInformationImpl implements TransactionInformation {
		private final FieldWithMetaRegimeNameEnum regimeName;
		private final FieldWithMetaSupervisoryBodyEnum supervisoryBody;
		private final TradeIdentifier transactionIdentifier;
		private final ESMATransactionInformation esmaTransactionInformation;
		
		protected TransactionInformationImpl(TransactionInformation.TransactionInformationBuilder builder) {
			this.regimeName = ofNullable(builder.getRegimeName()).map(f->f.build()).orElse(null);
			this.supervisoryBody = ofNullable(builder.getSupervisoryBody()).map(f->f.build()).orElse(null);
			this.transactionIdentifier = ofNullable(builder.getTransactionIdentifier()).map(f->f.build()).orElse(null);
			this.esmaTransactionInformation = ofNullable(builder.getEsmaTransactionInformation()).map(f->f.build()).orElse(null);
		}
		
		@Override
		@RosettaAttribute("regimeName")
		public FieldWithMetaRegimeNameEnum getRegimeName() {
			return regimeName;
		}
		
		@Override
		@RosettaAttribute("supervisoryBody")
		public FieldWithMetaSupervisoryBodyEnum getSupervisoryBody() {
			return supervisoryBody;
		}
		
		@Override
		@RosettaAttribute("transactionIdentifier")
		public TradeIdentifier getTransactionIdentifier() {
			return transactionIdentifier;
		}
		
		@Override
		@RosettaAttribute("esmaTransactionInformation")
		public ESMATransactionInformation getEsmaTransactionInformation() {
			return esmaTransactionInformation;
		}
		
		@Override
		public TransactionInformation build() {
			return this;
		}
		
		@Override
		public TransactionInformation.TransactionInformationBuilder toBuilder() {
			TransactionInformation.TransactionInformationBuilder builder = builder();
			setBuilderFields(builder);
			return builder;
		}
		
		protected void setBuilderFields(TransactionInformation.TransactionInformationBuilder builder) {
			ofNullable(getRegimeName()).ifPresent(builder::setRegimeName);
			ofNullable(getSupervisoryBody()).ifPresent(builder::setSupervisoryBody);
			ofNullable(getTransactionIdentifier()).ifPresent(builder::setTransactionIdentifier);
			ofNullable(getEsmaTransactionInformation()).ifPresent(builder::setEsmaTransactionInformation);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			TransactionInformation _that = getType().cast(o);
		
			if (!Objects.equals(regimeName, _that.getRegimeName())) return false;
			if (!Objects.equals(supervisoryBody, _that.getSupervisoryBody())) return false;
			if (!Objects.equals(transactionIdentifier, _that.getTransactionIdentifier())) return false;
			if (!Objects.equals(esmaTransactionInformation, _that.getEsmaTransactionInformation())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (regimeName != null ? regimeName.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (supervisoryBody != null ? supervisoryBody.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (transactionIdentifier != null ? transactionIdentifier.hashCode() : 0);
			_result = 31 * _result + (esmaTransactionInformation != null ? esmaTransactionInformation.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "TransactionInformation {" +
				"regimeName=" + this.regimeName + ", " +
				"supervisoryBody=" + this.supervisoryBody + ", " +
				"transactionIdentifier=" + this.transactionIdentifier + ", " +
				"esmaTransactionInformation=" + this.esmaTransactionInformation +
			'}';
		}
	}

	/*********************** Builder Implementation of TransactionInformation  ***********************/
	class TransactionInformationBuilderImpl implements TransactionInformation.TransactionInformationBuilder {
	
		protected FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder regimeName;
		protected FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder supervisoryBody;
		protected TradeIdentifier.TradeIdentifierBuilder transactionIdentifier;
		protected ESMATransactionInformation.ESMATransactionInformationBuilder esmaTransactionInformation;
	
		public TransactionInformationBuilderImpl() {
		}
	
		@Override
		@RosettaAttribute("regimeName")
		public FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder getRegimeName() {
			return regimeName;
		}
		
		@Override
		public FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder getOrCreateRegimeName() {
			FieldWithMetaRegimeNameEnum.FieldWithMetaRegimeNameEnumBuilder result;
			if (regimeName!=null) {
				result = regimeName;
			}
			else {
				result = regimeName = FieldWithMetaRegimeNameEnum.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("supervisoryBody")
		public FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder getSupervisoryBody() {
			return supervisoryBody;
		}
		
		@Override
		public FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder getOrCreateSupervisoryBody() {
			FieldWithMetaSupervisoryBodyEnum.FieldWithMetaSupervisoryBodyEnumBuilder result;
			if (supervisoryBody!=null) {
				result = supervisoryBody;
			}
			else {
				result = supervisoryBody = FieldWithMetaSupervisoryBodyEnum.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("transactionIdentifier")
		public TradeIdentifier.TradeIdentifierBuilder getTransactionIdentifier() {
			return transactionIdentifier;
		}
		
		@Override
		public TradeIdentifier.TradeIdentifierBuilder getOrCreateTransactionIdentifier() {
			TradeIdentifier.TradeIdentifierBuilder result;
			if (transactionIdentifier!=null) {
				result = transactionIdentifier;
			}
			else {
				result = transactionIdentifier = TradeIdentifier.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("esmaTransactionInformation")
		public ESMATransactionInformation.ESMATransactionInformationBuilder getEsmaTransactionInformation() {
			return esmaTransactionInformation;
		}
		
		@Override
		public ESMATransactionInformation.ESMATransactionInformationBuilder getOrCreateEsmaTransactionInformation() {
			ESMATransactionInformation.ESMATransactionInformationBuilder result;
			if (esmaTransactionInformation!=null) {
				result = esmaTransactionInformation;
			}
			else {
				result = esmaTransactionInformation = ESMATransactionInformation.builder();
			}
			
			return result;
		}
	
		@Override
		@RosettaAttribute("regimeName")
		public TransactionInformation.TransactionInformationBuilder setRegimeName(FieldWithMetaRegimeNameEnum regimeName) {
			this.regimeName = regimeName==null?null:regimeName.toBuilder();
			return this;
		}
		@Override
		public TransactionInformation.TransactionInformationBuilder setRegimeNameValue(RegimeNameEnum regimeName) {
			this.getOrCreateRegimeName().setValue(regimeName);
			return this;
		}
		@Override
		@RosettaAttribute("supervisoryBody")
		public TransactionInformation.TransactionInformationBuilder setSupervisoryBody(FieldWithMetaSupervisoryBodyEnum supervisoryBody) {
			this.supervisoryBody = supervisoryBody==null?null:supervisoryBody.toBuilder();
			return this;
		}
		@Override
		public TransactionInformation.TransactionInformationBuilder setSupervisoryBodyValue(SupervisoryBodyEnum supervisoryBody) {
			this.getOrCreateSupervisoryBody().setValue(supervisoryBody);
			return this;
		}
		@Override
		@RosettaAttribute("transactionIdentifier")
		public TransactionInformation.TransactionInformationBuilder setTransactionIdentifier(TradeIdentifier transactionIdentifier) {
			this.transactionIdentifier = transactionIdentifier==null?null:transactionIdentifier.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("esmaTransactionInformation")
		public TransactionInformation.TransactionInformationBuilder setEsmaTransactionInformation(ESMATransactionInformation esmaTransactionInformation) {
			this.esmaTransactionInformation = esmaTransactionInformation==null?null:esmaTransactionInformation.toBuilder();
			return this;
		}
		
		@Override
		public TransactionInformation build() {
			return new TransactionInformation.TransactionInformationImpl(this);
		}
		
		@Override
		public TransactionInformation.TransactionInformationBuilder toBuilder() {
			return this;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public TransactionInformation.TransactionInformationBuilder prune() {
			if (regimeName!=null && !regimeName.prune().hasData()) regimeName = null;
			if (supervisoryBody!=null && !supervisoryBody.prune().hasData()) supervisoryBody = null;
			if (transactionIdentifier!=null && !transactionIdentifier.prune().hasData()) transactionIdentifier = null;
			if (esmaTransactionInformation!=null && !esmaTransactionInformation.prune().hasData()) esmaTransactionInformation = null;
			return this;
		}
		
		@Override
		public boolean hasData() {
			if (getRegimeName()!=null) return true;
			if (getSupervisoryBody()!=null) return true;
			if (getTransactionIdentifier()!=null && getTransactionIdentifier().hasData()) return true;
			if (getEsmaTransactionInformation()!=null && getEsmaTransactionInformation().hasData()) return true;
			return false;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public TransactionInformation.TransactionInformationBuilder merge(RosettaModelObjectBuilder other, BuilderMerger merger) {
			TransactionInformation.TransactionInformationBuilder o = (TransactionInformation.TransactionInformationBuilder) other;
			
			merger.mergeRosetta(getRegimeName(), o.getRegimeName(), this::setRegimeName);
			merger.mergeRosetta(getSupervisoryBody(), o.getSupervisoryBody(), this::setSupervisoryBody);
			merger.mergeRosetta(getTransactionIdentifier(), o.getTransactionIdentifier(), this::setTransactionIdentifier);
			merger.mergeRosetta(getEsmaTransactionInformation(), o.getEsmaTransactionInformation(), this::setEsmaTransactionInformation);
			
			return this;
		}
	
		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			TransactionInformation _that = getType().cast(o);
		
			if (!Objects.equals(regimeName, _that.getRegimeName())) return false;
			if (!Objects.equals(supervisoryBody, _that.getSupervisoryBody())) return false;
			if (!Objects.equals(transactionIdentifier, _that.getTransactionIdentifier())) return false;
			if (!Objects.equals(esmaTransactionInformation, _that.getEsmaTransactionInformation())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (regimeName != null ? regimeName.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (supervisoryBody != null ? supervisoryBody.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (transactionIdentifier != null ? transactionIdentifier.hashCode() : 0);
			_result = 31 * _result + (esmaTransactionInformation != null ? esmaTransactionInformation.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "TransactionInformationBuilder {" +
				"regimeName=" + this.regimeName + ", " +
				"supervisoryBody=" + this.supervisoryBody + ", " +
				"transactionIdentifier=" + this.transactionIdentifier + ", " +
				"esmaTransactionInformation=" + this.esmaTransactionInformation +
			'}';
		}
	}
}
