package drr.regulation.common.functions;

import cdm.base.datetime.DayOfWeekEnum;
import cdm.base.math.Quantity;
import cdm.base.math.UnitType;
import cdm.observable.asset.Price;
import cdm.product.asset.AssetDeliveryInformation;
import cdm.product.asset.AssetDeliveryPeriods;
import cdm.product.asset.AssetDeliveryProfile;
import cdm.product.asset.AssetDeliveryProfileBlock;
import cdm.product.asset.BankHolidayTreatmentEnum;
import cdm.product.asset.CalculationScheduleDeliveryPeriods;
import cdm.product.asset.CommodityPayout;
import cdm.product.qualification.functions.Qualify_Commodity_Forward;
import cdm.product.qualification.functions.Qualify_Commodity_Option;
import cdm.product.qualification.functions.Qualify_Commodity_Swap_Basis;
import cdm.product.qualification.functions.Qualify_Commodity_Swap_FixedFloat;
import cdm.product.qualification.functions.Qualify_Commodity_Swaption;
import cdm.product.template.CalculationSchedule;
import cdm.product.template.EconomicTerms;
import cdm.product.template.ForwardPayout;
import cdm.product.template.OptionPayout;
import cdm.product.template.Payout;
import cdm.product.template.Product;
import cdm.product.template.SchedulePeriod;
import com.google.inject.ImplementedBy;
import com.rosetta.model.lib.expression.ComparisonResult;
import com.rosetta.model.lib.functions.ModelObjectValidator;
import com.rosetta.model.lib.functions.RosettaFunction;
import com.rosetta.model.lib.mapper.MapperListOfLists;
import com.rosetta.model.lib.mapper.MapperS;
import com.rosetta.model.lib.records.Date;
import com.rosetta.model.metafields.FieldWithMetaString;
import drr.regulation.common.CommodityTimeUnitEnum;
import drr.regulation.common.ESMATransactionInformation;
import drr.regulation.common.ReportableDelivery;
import drr.regulation.common.ReportableDelivery.ReportableDeliveryBuilder;
import drr.regulation.common.ReportableEvent;
import drr.regulation.common.ReportableInformation;
import drr.regulation.common.TransactionInformation;
import java.math.BigDecimal;
import java.time.LocalTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.inject.Inject;

import static com.rosetta.model.lib.expression.ExpressionOperators.*;

@ImplementedBy(GetReportableDelivery.GetReportableDeliveryDefault.class)
public abstract class GetReportableDelivery implements RosettaFunction {
	
	@Inject protected ModelObjectValidator objectValidator;
	
	// RosettaFunction dependencies
	//
	@Inject protected Create_DeliveryBlock create_DeliveryBlock;
	@Inject protected EconomicTermsForProduct economicTermsForProduct;
	@Inject protected IsSingleOnlyCommodityPayout isSingleOnlyCommodityPayout;
	@Inject protected ProductForEvent productForEvent;
	@Inject protected Qualify_Commodity_Forward qualify_Commodity_Forward;
	@Inject protected Qualify_Commodity_Option qualify_Commodity_Option;
	@Inject protected Qualify_Commodity_Swap_Basis qualify_Commodity_Swap_Basis;
	@Inject protected Qualify_Commodity_Swap_FixedFloat qualify_Commodity_Swap_FixedFloat;
	@Inject protected Qualify_Commodity_Swaption qualify_Commodity_Swaption;

	/**
	* @param reportableEvent 
	* @return reportableDelivery 
	*/
	public List<? extends ReportableDelivery> evaluate(ReportableEvent reportableEvent) {
		List<ReportableDelivery.ReportableDeliveryBuilder> reportableDeliveryBuilder = doEvaluate(reportableEvent);
		
		final List<? extends ReportableDelivery> reportableDelivery;
		if (reportableDeliveryBuilder == null) {
			reportableDelivery = null;
		} else {
			reportableDelivery = reportableDeliveryBuilder.stream().map(ReportableDelivery::build).collect(Collectors.toList());
			objectValidator.validate(ReportableDelivery.class, reportableDelivery);
		}
		
		return reportableDelivery;
	}

	protected abstract List<ReportableDelivery.ReportableDeliveryBuilder> doEvaluate(ReportableEvent reportableEvent);

	protected abstract MapperS<? extends Product> product(ReportableEvent reportableEvent);

	protected abstract MapperS<? extends EconomicTerms> economicTerms(ReportableEvent reportableEvent);

	protected abstract MapperS<? extends AssetDeliveryInformation> delivery(ReportableEvent reportableEvent);

	protected abstract MapperS<? extends CalculationSchedule> customizedSchedule(ReportableEvent reportableEvent);

	protected abstract MapperS<CommodityTimeUnitEnum> duration(ReportableEvent reportableEvent);

	public static class GetReportableDeliveryDefault extends GetReportableDelivery {
		@Override
		protected List<ReportableDelivery.ReportableDeliveryBuilder> doEvaluate(ReportableEvent reportableEvent) {
			List<ReportableDelivery.ReportableDeliveryBuilder> reportableDelivery = new ArrayList<>();
			return assignOutput(reportableDelivery, reportableEvent);
		}
		
		protected List<ReportableDelivery.ReportableDeliveryBuilder> assignOutput(List<ReportableDelivery.ReportableDeliveryBuilder> reportableDelivery, ReportableEvent reportableEvent) {
			if (exists(customizedSchedule(reportableEvent)).getOrDefault(false)) {
				final MapperListOfLists<ReportableDelivery> thenResult = customizedSchedule(reportableEvent).<SchedulePeriod>mapC("getSchedulePeriod", calculationSchedule -> calculationSchedule.getSchedulePeriod())
					.mapItemToList(schedulePeriod -> {
						final MapperListOfLists<ReportableDelivery> _thenResult = schedulePeriod.<CalculationScheduleDeliveryPeriods>map("getDeliveryPeriod", _schedulePeriod -> _schedulePeriod.getDeliveryPeriod()).<AssetDeliveryProfile>mapC("getProfile", assetDeliveryPeriods -> assetDeliveryPeriods.getProfile())
							.mapItemToList(deliveryProfile -> deliveryProfile.<AssetDeliveryProfileBlock>mapC("getBlock", assetDeliveryProfile -> assetDeliveryProfile.getBlock())
								.mapItem(deliveryBlock -> {
									final Quantity ifThenElseResult0;
									if (exists(schedulePeriod.<CalculationScheduleDeliveryPeriods>map("getDeliveryPeriod", _schedulePeriod -> _schedulePeriod.getDeliveryPeriod()).<Quantity>map("getDeliveryCapacity", calculationScheduleDeliveryPeriods -> calculationScheduleDeliveryPeriods.getDeliveryCapacity())).getOrDefault(false)) {
										ifThenElseResult0 = schedulePeriod.<CalculationScheduleDeliveryPeriods>map("getDeliveryPeriod", _schedulePeriod -> _schedulePeriod.getDeliveryPeriod()).<Quantity>map("getDeliveryCapacity", calculationScheduleDeliveryPeriods -> calculationScheduleDeliveryPeriods.getDeliveryCapacity()).get();
									} else if (exists(deliveryBlock.<Quantity>map("getDeliveryCapacity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getDeliveryCapacity())).getOrDefault(false)) {
										ifThenElseResult0 = deliveryBlock.<Quantity>map("getDeliveryCapacity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getDeliveryCapacity()).get();
									} else {
										ifThenElseResult0 = null;
									}
									final BigDecimal ifThenElseResult1;
									if (exists(schedulePeriod.<CalculationScheduleDeliveryPeriods>map("getDeliveryPeriod", _schedulePeriod -> _schedulePeriod.getDeliveryPeriod()).<Price>map("getPriceTimeIntervalQuantity", calculationScheduleDeliveryPeriods -> calculationScheduleDeliveryPeriods.getPriceTimeIntervalQuantity())).getOrDefault(false)) {
										ifThenElseResult1 = schedulePeriod.<CalculationScheduleDeliveryPeriods>map("getDeliveryPeriod", _schedulePeriod -> _schedulePeriod.getDeliveryPeriod()).<Price>map("getPriceTimeIntervalQuantity", calculationScheduleDeliveryPeriods -> calculationScheduleDeliveryPeriods.getPriceTimeIntervalQuantity()).<BigDecimal>map("getValue", measureBase -> measureBase.getValue()).get();
									} else if (exists(deliveryBlock.<Price>map("getPriceTimeIntervalQuantity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getPriceTimeIntervalQuantity())).getOrDefault(false)) {
										ifThenElseResult1 = deliveryBlock.<Price>map("getPriceTimeIntervalQuantity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getPriceTimeIntervalQuantity()).<BigDecimal>map("getValue", measureBase -> measureBase.getValue()).get();
									} else {
										ifThenElseResult1 = null;
									}
									final String ifThenElseResult2;
									if (exists(schedulePeriod.<CalculationScheduleDeliveryPeriods>map("getDeliveryPeriod", _schedulePeriod -> _schedulePeriod.getDeliveryPeriod()).<Price>map("getPriceTimeIntervalQuantity", calculationScheduleDeliveryPeriods -> calculationScheduleDeliveryPeriods.getPriceTimeIntervalQuantity())).getOrDefault(false)) {
										ifThenElseResult2 = schedulePeriod.<CalculationScheduleDeliveryPeriods>map("getDeliveryPeriod", _schedulePeriod -> _schedulePeriod.getDeliveryPeriod()).<Price>map("getPriceTimeIntervalQuantity", calculationScheduleDeliveryPeriods -> calculationScheduleDeliveryPeriods.getPriceTimeIntervalQuantity()).<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FieldWithMetaString>map("getCurrency", unitType -> unitType.getCurrency()).<String>map("getValue", _f->_f.getValue()).get();
									} else if (exists(deliveryBlock.<Price>map("getPriceTimeIntervalQuantity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getPriceTimeIntervalQuantity())).getOrDefault(false)) {
										ifThenElseResult2 = deliveryBlock.<Price>map("getPriceTimeIntervalQuantity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getPriceTimeIntervalQuantity()).<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FieldWithMetaString>map("getCurrency", unitType -> unitType.getCurrency()).<String>map("getValue", _f->_f.getValue()).get();
									} else {
										ifThenElseResult2 = null;
									}
									return MapperS.of(create_DeliveryBlock.evaluate(deliveryBlock.<LocalTime>map("getStartTime", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getStartTime()).get(), deliveryBlock.<LocalTime>map("getEndTime", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getEndTime()).get(), schedulePeriod.<CalculationScheduleDeliveryPeriods>map("getDeliveryPeriod", _schedulePeriod -> _schedulePeriod.getDeliveryPeriod()).<Date>map("getStartDate", assetDeliveryPeriods -> assetDeliveryPeriods.getStartDate()).get(), schedulePeriod.<CalculationScheduleDeliveryPeriods>map("getDeliveryPeriod", _schedulePeriod -> _schedulePeriod.getDeliveryPeriod()).<Date>map("getEndDate", assetDeliveryPeriods -> assetDeliveryPeriods.getEndDate()).get(), deliveryProfile.<BankHolidayTreatmentEnum>map("getBankHolidaysTreatment", assetDeliveryProfile -> assetDeliveryProfile.getBankHolidaysTreatment()).get(), duration(reportableEvent).get(), deliveryBlock.<DayOfWeekEnum>mapC("getDayOfWeek", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getDayOfWeek()).getMulti(), ifThenElseResult0, ifThenElseResult1, ifThenElseResult2));
								}));
						return _thenResult
							.flattenList();
					});
				reportableDelivery.addAll(toBuilder(thenResult
					.flattenList().getMulti()));
			} else {
				reportableDelivery.addAll(toBuilder(delivery(reportableEvent).<AssetDeliveryPeriods>map("getPeriods", assetDeliveryInformation -> assetDeliveryInformation.getPeriods())
					.mapSingleToList(periods -> {
						final MapperListOfLists<ReportableDelivery> _thenResult = periods.<AssetDeliveryProfile>mapC("getProfile", assetDeliveryPeriods -> assetDeliveryPeriods.getProfile())
							.mapItemToList(deliveryProfile -> deliveryProfile.<AssetDeliveryProfileBlock>mapC("getBlock", assetDeliveryProfile -> assetDeliveryProfile.getBlock())
								.mapItem(deliveryBlock -> {
									final Quantity ifThenElseResult;
									if (exists(delivery(reportableEvent).<Quantity>map("getDeliveryCapacity", assetDeliveryInformation -> assetDeliveryInformation.getDeliveryCapacity())).getOrDefault(false)) {
										ifThenElseResult = delivery(reportableEvent).<Quantity>map("getDeliveryCapacity", assetDeliveryInformation -> assetDeliveryInformation.getDeliveryCapacity()).get();
									} else if (exists(deliveryBlock.<Quantity>map("getDeliveryCapacity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getDeliveryCapacity())).getOrDefault(false)) {
										ifThenElseResult = deliveryBlock.<Quantity>map("getDeliveryCapacity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getDeliveryCapacity()).get();
									} else {
										ifThenElseResult = null;
									}
									return MapperS.of(create_DeliveryBlock.evaluate(deliveryBlock.<LocalTime>map("getStartTime", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getStartTime()).get(), deliveryBlock.<LocalTime>map("getEndTime", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getEndTime()).get(), periods.<Date>map("getStartDate", assetDeliveryPeriods -> assetDeliveryPeriods.getStartDate()).get(), periods.<Date>map("getEndDate", assetDeliveryPeriods -> assetDeliveryPeriods.getEndDate()).get(), deliveryProfile.<BankHolidayTreatmentEnum>map("getBankHolidaysTreatment", assetDeliveryProfile -> assetDeliveryProfile.getBankHolidaysTreatment()).get(), duration(reportableEvent).get(), deliveryBlock.<DayOfWeekEnum>mapC("getDayOfWeek", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getDayOfWeek()).getMulti(), ifThenElseResult, deliveryBlock.<Price>map("getPriceTimeIntervalQuantity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getPriceTimeIntervalQuantity()).<BigDecimal>map("getValue", measureBase -> measureBase.getValue()).get(), deliveryBlock.<Price>map("getPriceTimeIntervalQuantity", assetDeliveryProfileBlock -> assetDeliveryProfileBlock.getPriceTimeIntervalQuantity()).<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FieldWithMetaString>map("getCurrency", unitType -> unitType.getCurrency()).<String>map("getValue", _f->_f.getValue()).get()));
								}));
						return _thenResult
							.flattenList();
					}).getMulti()));
			}
			
			return Optional.ofNullable(reportableDelivery)
				.map(o -> o.stream().map(i -> i.prune()).collect(Collectors.toList()))
				.orElse(null);
		}
		
		@Override
		protected MapperS<? extends Product> product(ReportableEvent reportableEvent) {
			return MapperS.of(productForEvent.evaluate(reportableEvent));
		}
		
		@Override
		protected MapperS<? extends EconomicTerms> economicTerms(ReportableEvent reportableEvent) {
			return MapperS.of(economicTermsForProduct.evaluate(product(reportableEvent).get()));
		}
		
		@Override
		protected MapperS<? extends AssetDeliveryInformation> delivery(ReportableEvent reportableEvent) {
			if (ComparisonResult.of(MapperS.of(qualify_Commodity_Swap_FixedFloat.evaluate(economicTerms(reportableEvent).get()))).or(ComparisonResult.of(MapperS.of(qualify_Commodity_Swap_Basis.evaluate(economicTerms(reportableEvent).get())))).getOrDefault(false)) {
				return economicTerms(reportableEvent).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<CommodityPayout>mapC("getCommodityPayout", payout -> payout.getCommodityPayout()).<AssetDeliveryInformation>map("getDelivery", commodityPayout -> commodityPayout.getDelivery())
					.first();
			}
			final Boolean boolean0 = qualify_Commodity_Option.evaluate(economicTerms(reportableEvent).get());
			if ((boolean0 == null ? false : boolean0)) {
				return MapperS.of(economicTerms(reportableEvent).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).<AssetDeliveryInformation>map("getDelivery", optionPayout -> optionPayout.getDelivery()).get());
			}
			final Boolean boolean1 = qualify_Commodity_Swaption.evaluate(economicTerms(reportableEvent).get());
			if ((boolean1 == null ? false : boolean1)) {
				return MapperS.of(economicTermsForProduct.evaluate(economicTerms(reportableEvent).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).<Product>map("getUnderlier", optionPayout -> optionPayout.getUnderlier()).get())).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<CommodityPayout>mapC("getCommodityPayout", payout -> payout.getCommodityPayout()).<AssetDeliveryInformation>map("getDelivery", commodityPayout -> commodityPayout.getDelivery())
					.first();
			}
			final Boolean boolean2 = isSingleOnlyCommodityPayout.evaluate(product(reportableEvent).get());
			if ((boolean2 == null ? false : boolean2)) {
				return MapperS.of(economicTerms(reportableEvent).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<CommodityPayout>mapC("getCommodityPayout", payout -> payout.getCommodityPayout()).get()).<AssetDeliveryInformation>map("getDelivery", commodityPayout -> commodityPayout.getDelivery());
			}
			final Boolean boolean3 = qualify_Commodity_Forward.evaluate(economicTerms(reportableEvent).get());
			if ((boolean3 == null ? false : boolean3)) {
				return MapperS.of(economicTerms(reportableEvent).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<ForwardPayout>mapC("getForwardPayout", payout -> payout.getForwardPayout()).get()).<AssetDeliveryInformation>map("getDelivery", forwardPayout -> forwardPayout.getDelivery());
			}
			return MapperS.<AssetDeliveryInformation>ofNull();
		}
		
		@Override
		protected MapperS<? extends CalculationSchedule> customizedSchedule(ReportableEvent reportableEvent) {
			if (ComparisonResult.of(MapperS.of(qualify_Commodity_Swap_FixedFloat.evaluate(economicTerms(reportableEvent).get()))).or(ComparisonResult.of(MapperS.of(qualify_Commodity_Swap_Basis.evaluate(economicTerms(reportableEvent).get())))).getOrDefault(false)) {
				return economicTerms(reportableEvent).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<CommodityPayout>mapC("getCommodityPayout", payout -> payout.getCommodityPayout()).<CalculationSchedule>map("getSchedule", commodityPayout -> commodityPayout.getSchedule())
					.first();
			}
			if (ComparisonResult.of(MapperS.of(qualify_Commodity_Option.evaluate(economicTerms(reportableEvent).get()))).or(ComparisonResult.of(MapperS.of(qualify_Commodity_Swaption.evaluate(economicTerms(reportableEvent).get())))).getOrDefault(false)) {
				return MapperS.of(economicTerms(reportableEvent).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).get()).<CalculationSchedule>map("getSchedule", optionPayout -> optionPayout.getSchedule());
			}
			final Boolean boolean0 = isSingleOnlyCommodityPayout.evaluate(product(reportableEvent).get());
			if ((boolean0 == null ? false : boolean0)) {
				return MapperS.of(economicTerms(reportableEvent).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<CommodityPayout>mapC("getCommodityPayout", payout -> payout.getCommodityPayout()).get()).<CalculationSchedule>map("getSchedule", commodityPayout -> commodityPayout.getSchedule());
			}
			final Boolean boolean1 = qualify_Commodity_Forward.evaluate(economicTerms(reportableEvent).get());
			if ((boolean1 == null ? false : boolean1)) {
				return MapperS.of(economicTerms(reportableEvent).<Payout>map("getPayout", _economicTerms -> _economicTerms.getPayout()).<ForwardPayout>mapC("getForwardPayout", payout -> payout.getForwardPayout()).get()).<CalculationSchedule>map("getSchedule", forwardPayout -> forwardPayout.getSchedule());
			}
			return MapperS.<CalculationSchedule>ofNull();
		}
		
		@Override
		protected MapperS<CommodityTimeUnitEnum> duration(ReportableEvent reportableEvent) {
			return MapperS.of(reportableEvent).<ReportableInformation>map("getReportableInformation", _reportableEvent -> _reportableEvent.getReportableInformation()).<TransactionInformation>mapC("getTransactionInformation", reportableInformation -> reportableInformation.getTransactionInformation()).<ESMATransactionInformation>map("getEsmaTransactionInformation", transactionInformation -> transactionInformation.getEsmaTransactionInformation()).<CommodityTimeUnitEnum>map("getDuration", eSMATransactionInformation -> eSMATransactionInformation.getDuration())
				.first();
		}
	}
}
