package drr.regulation.common.functions;

import cdm.base.datetime.DateRange;
import cdm.base.math.DatedValue;
import cdm.observable.asset.PriceSchedule;
import cdm.product.qualification.functions.Qualify_Commodity_Swaption;
import cdm.product.template.CalculationSchedule;
import cdm.product.template.OptionPayout;
import cdm.product.template.Payout;
import cdm.product.template.SchedulePeriod;
import com.google.inject.ImplementedBy;
import com.rosetta.model.lib.expression.CardinalityOperator;
import com.rosetta.model.lib.expression.ComparisonResult;
import com.rosetta.model.lib.functions.ModelObjectValidator;
import com.rosetta.model.lib.functions.RosettaFunction;
import com.rosetta.model.lib.mapper.MapperC;
import com.rosetta.model.lib.mapper.MapperS;
import com.rosetta.model.lib.records.Date;
import drr.regulation.common.PriceReportablePeriod;
import drr.regulation.common.PriceReportablePeriod.PriceReportablePeriodBuilder;
import drr.regulation.common.ReportableEvent;
import drr.standards.iso.PriceNotationEnum;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.inject.Inject;

import static com.rosetta.model.lib.expression.ExpressionOperators.*;

@ImplementedBy(GetReportableStrikePricePeriod.GetReportableStrikePricePeriodDefault.class)
public abstract class GetReportableStrikePricePeriod implements RosettaFunction {
	
	@Inject protected ModelObjectValidator objectValidator;
	
	// RosettaFunction dependencies
	//
	@Inject protected Contract_StrikePrice contract_StrikePrice;
	@Inject protected Create_PriceReportablePeriod create_PriceReportablePeriod;
	@Inject protected EconomicTermsForProduct economicTermsForProduct;
	@Inject protected GetPriceNotation getPriceNotation;
	@Inject protected IsCommodityOption isCommodityOption;
	@Inject protected ProductForTrade productForTrade;
	@Inject protected Qualify_Commodity_Swaption qualify_Commodity_Swaption;
	@Inject protected TradeForEvent tradeForEvent;

	/**
	* @param reportableEvent 
	* @return reportablePeriod 
	*/
	public List<? extends PriceReportablePeriod> evaluate(ReportableEvent reportableEvent) {
		List<PriceReportablePeriod.PriceReportablePeriodBuilder> reportablePeriodBuilder = doEvaluate(reportableEvent);
		
		final List<? extends PriceReportablePeriod> reportablePeriod;
		if (reportablePeriodBuilder == null) {
			reportablePeriod = null;
		} else {
			reportablePeriod = reportablePeriodBuilder.stream().map(PriceReportablePeriod::build).collect(Collectors.toList());
			objectValidator.validate(PriceReportablePeriod.class, reportablePeriod);
		}
		
		return reportablePeriod;
	}

	protected abstract List<PriceReportablePeriod.PriceReportablePeriodBuilder> doEvaluate(ReportableEvent reportableEvent);

	protected abstract MapperS<? extends CalculationSchedule> customizedSchedule(ReportableEvent reportableEvent);

	protected abstract MapperS<? extends PriceSchedule> strikePriceSchedule(ReportableEvent reportableEvent);

	protected abstract MapperS<PriceNotationEnum> priceNotation(ReportableEvent reportableEvent);

	public static class GetReportableStrikePricePeriodDefault extends GetReportableStrikePricePeriod {
		@Override
		protected List<PriceReportablePeriod.PriceReportablePeriodBuilder> doEvaluate(ReportableEvent reportableEvent) {
			List<PriceReportablePeriod.PriceReportablePeriodBuilder> reportablePeriod = new ArrayList<>();
			return assignOutput(reportablePeriod, reportableEvent);
		}
		
		protected List<PriceReportablePeriod.PriceReportablePeriodBuilder> assignOutput(List<PriceReportablePeriod.PriceReportablePeriodBuilder> reportablePeriod, ReportableEvent reportableEvent) {
			if (exists(customizedSchedule(reportableEvent)).getOrDefault(false)) {
				reportablePeriod.addAll(toBuilder(strikePriceSchedule(reportableEvent).<DatedValue>mapC("getDatedValue", measureSchedule -> measureSchedule.getDatedValue())
					.mapItem(pricePeriod -> {
						final MapperC<SchedulePeriod> thenResult0 = customizedSchedule(reportableEvent).<SchedulePeriod>mapC("getSchedulePeriod", calculationSchedule -> calculationSchedule.getSchedulePeriod())
							.filterItemNullSafe(customPeriod -> areEqual(pricePeriod.<Date>map("getDate", datedValue -> datedValue.getDate()), customPeriod.<DateRange>map("getCalculationPeriod", schedulePeriod -> schedulePeriod.getCalculationPeriod()).<Date>map("getStartDate", dateRange -> dateRange.getStartDate()), CardinalityOperator.All).get());
						final MapperS<SchedulePeriod> thenResult1 = MapperS.of(thenResult0.get());
						return thenResult1
							.mapSingleToItem(matchingCustomPeriod -> MapperS.of(create_PriceReportablePeriod.evaluate(pricePeriod.<BigDecimal>map("getValue", datedValue -> datedValue.getValue()).get(), matchingCustomPeriod.<DateRange>map("getCalculationPeriod", schedulePeriod -> schedulePeriod.getCalculationPeriod()).<Date>map("getStartDate", dateRange -> dateRange.getStartDate()).get(), matchingCustomPeriod.<DateRange>map("getCalculationPeriod", schedulePeriod -> schedulePeriod.getCalculationPeriod()).<Date>map("getEndDate", dateRange -> dateRange.getEndDate()).get(), priceNotation(reportableEvent).get())));
					}).getMulti()));
			} else {
				reportablePeriod.addAll(toBuilder(strikePriceSchedule(reportableEvent).<DatedValue>mapC("getDatedValue", measureSchedule -> measureSchedule.getDatedValue())
					.mapItem(pricePeriod -> MapperS.of(create_PriceReportablePeriod.evaluate(pricePeriod.<BigDecimal>map("getValue", datedValue -> datedValue.getValue()).get(), pricePeriod.<Date>map("getDate", datedValue -> datedValue.getDate()).get(), null, priceNotation(reportableEvent).get()))).getMulti()));
			}
			
			return Optional.ofNullable(reportablePeriod)
				.map(o -> o.stream().map(i -> i.prune()).collect(Collectors.toList()))
				.orElse(null);
		}
		
		@Override
		protected MapperS<? extends CalculationSchedule> customizedSchedule(ReportableEvent reportableEvent) {
			if (ComparisonResult.of(MapperS.of(isCommodityOption.evaluate(productForTrade.evaluate(tradeForEvent.evaluate(reportableEvent))))).or(ComparisonResult.of(MapperS.of(qualify_Commodity_Swaption.evaluate(economicTermsForProduct.evaluate(productForTrade.evaluate(tradeForEvent.evaluate(reportableEvent))))))).getOrDefault(false)) {
				return MapperS.of(MapperS.of(economicTermsForProduct.evaluate(productForTrade.evaluate(tradeForEvent.evaluate(reportableEvent)))).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).<CalculationSchedule>map("getSchedule", optionPayout -> optionPayout.getSchedule()).get());
			}
			return MapperS.<CalculationSchedule>ofNull();
		}
		
		@Override
		protected MapperS<? extends PriceSchedule> strikePriceSchedule(ReportableEvent reportableEvent) {
			return MapperS.of(MapperC.of(contract_StrikePrice.evaluate(reportableEvent)).get());
		}
		
		@Override
		protected MapperS<PriceNotationEnum> priceNotation(ReportableEvent reportableEvent) {
			return MapperS.of(getPriceNotation.evaluate(strikePriceSchedule(reportableEvent).get()));
		}
	}
}
