package drr.regulation.common.functions;

import cdm.base.staticdata.asset.common.DebtEconomics;
import cdm.base.staticdata.asset.common.DebtSeniorityEnum;
import cdm.base.staticdata.asset.common.DebtType;
import cdm.base.staticdata.asset.common.Security;
import cdm.product.asset.CreditDefaultPayout;
import cdm.product.asset.CreditIndexReferenceInformation;
import cdm.product.asset.CreditSeniorityEnum;
import cdm.product.asset.GeneralTerms;
import cdm.product.asset.ReferenceInformation;
import cdm.product.asset.ReferenceObligation;
import cdm.product.template.Payout;
import cdm.product.template.Product;
import com.google.inject.ImplementedBy;
import com.rosetta.model.lib.expression.CardinalityOperator;
import com.rosetta.model.lib.functions.RosettaFunction;
import com.rosetta.model.lib.mapper.MapperC;
import com.rosetta.model.lib.mapper.MapperS;
import drr.regulation.common.ReportableEvent;
import javax.inject.Inject;

import static com.rosetta.model.lib.expression.ExpressionOperators.*;

@ImplementedBy(GetSeniority.GetSeniorityDefault.class)
public abstract class GetSeniority implements RosettaFunction {
	
	// RosettaFunction dependencies
	//
	@Inject protected EconomicTermsForProduct economicTermsForProduct;
	@Inject protected IsCreditDefaultSwap isCreditDefaultSwap;
	@Inject protected IsCreditSwaption isCreditSwaption;
	@Inject protected ProductForEvent productForEvent;
	@Inject protected UnderlierForProduct underlierForProduct;

	/**
	* @param reportableEvent 
	* @return result 
	*/
	public String evaluate(ReportableEvent reportableEvent) {
		String result = doEvaluate(reportableEvent);
		
		return result;
	}

	protected abstract String doEvaluate(ReportableEvent reportableEvent);

	protected abstract MapperS<? extends Product> product(ReportableEvent reportableEvent);

	protected abstract MapperS<? extends CreditIndexReferenceInformation> indexReferenceInformation(ReportableEvent reportableEvent);

	protected abstract MapperS<? extends ReferenceInformation> referenceInformation(ReportableEvent reportableEvent);

	protected abstract MapperS<CreditSeniorityEnum> creditSeniority(ReportableEvent reportableEvent);

	protected abstract MapperC<DebtSeniorityEnum> debtSeniority(ReportableEvent reportableEvent);

	public static class GetSeniorityDefault extends GetSeniority {
		@Override
		protected String doEvaluate(ReportableEvent reportableEvent) {
			String result = null;
			return assignOutput(result, reportableEvent);
		}
		
		protected String assignOutput(String result, ReportableEvent reportableEvent) {
			if (areEqual(creditSeniority(reportableEvent), MapperS.of(CreditSeniorityEnum.SENIOR_UN_SEC), CardinalityOperator.All).or(areEqual(creditSeniority(reportableEvent), MapperS.of(CreditSeniorityEnum.SENIOR_SEC), CardinalityOperator.All)).or(areEqual(creditSeniority(reportableEvent), MapperS.of(CreditSeniorityEnum.SENIOR_LOSS_ABSORBING_CAPACITY), CardinalityOperator.All)).or(areEqual(debtSeniority(reportableEvent), MapperS.of(DebtSeniorityEnum.SENIOR), CardinalityOperator.Any)).getOrDefault(false)) {
				result = "SNDB";
			} else if (areEqual(creditSeniority(reportableEvent), MapperS.of(CreditSeniorityEnum.SUB_TIER_3), CardinalityOperator.All).or(areEqual(creditSeniority(reportableEvent), MapperS.of(CreditSeniorityEnum.SUB_UPPER_TIER_2), CardinalityOperator.All)).or(areEqual(creditSeniority(reportableEvent), MapperS.of(CreditSeniorityEnum.SUB_LOWER_TIER_2), CardinalityOperator.All)).or(areEqual(creditSeniority(reportableEvent), MapperS.of(CreditSeniorityEnum.SUB_TIER_1), CardinalityOperator.All)).or(areEqual(debtSeniority(reportableEvent), MapperS.of(DebtSeniorityEnum.SUBORDINATED), CardinalityOperator.Any)).getOrDefault(false)) {
				result = "SBOD";
			} else if (areEqual(creditSeniority(reportableEvent), MapperS.of(CreditSeniorityEnum.OTHER), CardinalityOperator.All).or(areEqual(debtSeniority(reportableEvent), MapperS.of(DebtSeniorityEnum.SECURED), CardinalityOperator.Any)).getOrDefault(false)) {
				result = "OTHR";
			} else {
				result = null;
			}
			
			return result;
		}
		
		@Override
		protected MapperS<? extends Product> product(ReportableEvent reportableEvent) {
			return MapperS.of(productForEvent.evaluate(reportableEvent));
		}
		
		@Override
		protected MapperS<? extends CreditIndexReferenceInformation> indexReferenceInformation(ReportableEvent reportableEvent) {
			final Boolean boolean0 = isCreditDefaultSwap.evaluate(product(reportableEvent).get());
			if ((boolean0 == null ? false : boolean0)) {
				return MapperS.of(economicTermsForProduct.evaluate(product(reportableEvent).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<CreditDefaultPayout>map("getCreditDefaultPayout", payout -> payout.getCreditDefaultPayout()).<GeneralTerms>map("getGeneralTerms", creditDefaultPayout -> creditDefaultPayout.getGeneralTerms()).<CreditIndexReferenceInformation>map("getIndexReferenceInformation", generalTerms -> generalTerms.getIndexReferenceInformation());
			}
			final Boolean boolean1 = isCreditSwaption.evaluate(product(reportableEvent).get());
			if ((boolean1 == null ? false : boolean1)) {
				return MapperS.of(economicTermsForProduct.evaluate(underlierForProduct.evaluate(product(reportableEvent).get()))).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<CreditDefaultPayout>map("getCreditDefaultPayout", payout -> payout.getCreditDefaultPayout()).<GeneralTerms>map("getGeneralTerms", creditDefaultPayout -> creditDefaultPayout.getGeneralTerms()).<CreditIndexReferenceInformation>map("getIndexReferenceInformation", generalTerms -> generalTerms.getIndexReferenceInformation());
			}
			return MapperS.<CreditIndexReferenceInformation>ofNull();
		}
		
		@Override
		protected MapperS<? extends ReferenceInformation> referenceInformation(ReportableEvent reportableEvent) {
			final Boolean boolean0 = isCreditDefaultSwap.evaluate(product(reportableEvent).get());
			if ((boolean0 == null ? false : boolean0)) {
				return MapperS.of(economicTermsForProduct.evaluate(product(reportableEvent).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<CreditDefaultPayout>map("getCreditDefaultPayout", payout -> payout.getCreditDefaultPayout()).<GeneralTerms>map("getGeneralTerms", creditDefaultPayout -> creditDefaultPayout.getGeneralTerms()).<ReferenceInformation>map("getReferenceInformation", generalTerms -> generalTerms.getReferenceInformation());
			}
			final Boolean boolean1 = isCreditSwaption.evaluate(product(reportableEvent).get());
			if ((boolean1 == null ? false : boolean1)) {
				return MapperS.of(economicTermsForProduct.evaluate(underlierForProduct.evaluate(product(reportableEvent).get()))).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<CreditDefaultPayout>map("getCreditDefaultPayout", payout -> payout.getCreditDefaultPayout()).<GeneralTerms>map("getGeneralTerms", creditDefaultPayout -> creditDefaultPayout.getGeneralTerms()).<ReferenceInformation>map("getReferenceInformation", generalTerms -> generalTerms.getReferenceInformation());
			}
			return MapperS.<ReferenceInformation>ofNull();
		}
		
		@Override
		protected MapperS<CreditSeniorityEnum> creditSeniority(ReportableEvent reportableEvent) {
			return indexReferenceInformation(reportableEvent).<CreditSeniorityEnum>map("getSeniority", creditIndexReferenceInformation -> creditIndexReferenceInformation.getSeniority());
		}
		
		@Override
		protected MapperC<DebtSeniorityEnum> debtSeniority(ReportableEvent reportableEvent) {
			return referenceInformation(reportableEvent).<ReferenceObligation>mapC("getReferenceObligation", _referenceInformation -> _referenceInformation.getReferenceObligation()).<Security>map("getSecurity", referenceObligation -> referenceObligation.getSecurity()).<DebtType>map("getDebtType", security -> security.getDebtType()).<DebtEconomics>mapC("getDebtEconomics", debtType -> debtType.getDebtEconomics()).<DebtSeniorityEnum>map("getDebtSeniority", debtEconomics -> debtEconomics.getDebtSeniority());
		}
	}
}
