package drr.regulation.common.processor;

import cdm.base.staticdata.party.metafields.ReferenceWithMetaParty;
import com.regnosys.rosetta.common.translation.Mapping;
import com.regnosys.rosetta.common.translation.MappingContext;
import com.regnosys.rosetta.common.translation.MappingProcessor;
import com.regnosys.rosetta.common.translation.Path;
import com.rosetta.model.lib.RosettaModelObjectBuilder;
import com.rosetta.model.lib.path.RosettaPath;
import drr.regulation.common.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

import static com.regnosys.rosetta.common.translation.MappingProcessorUtils.getNonNullMappedValue;
import static com.regnosys.rosetta.common.translation.MappingProcessorUtils.setValueAndOptionallyUpdateMappings;
import static drr.regulation.common.PartyInformation.PartyInformationBuilder;
import static drr.regulation.common.ReportingRegime.ReportingRegimeBuilder;

/**
 * FpML mapping processor.
 * Specs for the mapper:
 * - Access to partyTradeInformation list and extract the partyReference
 * - Relate the partyReference "href" with the party "id" bloc and extract the classification within it
 * - Map it int CDM in ReportableEvent->reportableInformation->partyInformation->regimeInformation->esmaPartyInformation->corporateSector->financialSector/nonFinancialSector->nonFinancialSectorIndicator
 * - Take into account that partyInformation in CDM is a list (1..*), financialSector is type FinancialSectorEnum and nonFinancialSector is a list 0..* of type NonFinancialSector and nonFinancialSectorIndicator is type NonFinancialSectorEnum
 * - For populating into the enums, use the new scheme Jan provided (fpml-schemeDefinitions) in order to decide how to populate the elements.
 */
@SuppressWarnings("unused")
public class CorporateSectorMappingProcessor extends MappingProcessor {
    private static final Logger LOGGER = LoggerFactory.getLogger(CorporateSectorMappingProcessor.class);
    public static final String ESMA_CORPORATE_SECTOR_SCHEME = "http://www.fpml.org/coding-scheme/esma-emir-refit-regulatory-corporate-sector";

    public CorporateSectorMappingProcessor(RosettaPath modelPath, List<Path> synonymPaths, MappingContext context) {
        super(modelPath, synonymPaths, context);
    }

    @Override
    public void map(Path synonymPath, List<? extends RosettaModelObjectBuilder> builders, RosettaModelObjectBuilder parent) {
        PartyInformationBuilder partyInformationBuilder = (PartyInformationBuilder) parent;
        List<ReportingRegimeBuilder> reportingRegimeBuilders = (List<ReportingRegimeBuilder>) builders;

        Optional.ofNullable(partyInformationBuilder.getPartyReference())
                .map(ReferenceWithMetaParty::getExternalReference)
                .map(partyReference -> getClassificationPaths(partyReference))
                .ifPresent(classificationPaths -> {
                    CorporateSector.CorporateSectorBuilder corporateSectorBuilder = CorporateSector.builder();
                    for (Path classificationPath : classificationPaths) {
                        setValueAndOptionallyUpdateMappings(classificationPath,
                                classificationValue -> {
                                    reportingRegimeBuilders.stream()
                                            .filter(reportingRegimeBuilder -> reportingRegimeBuilder.getSupervisoryBody() != null)
                                            .forEach(reportingRegimeBuilder -> {
                                                SupervisoryBodyEnum supervisoryBody = reportingRegimeBuilder.getSupervisoryBody().getValue();

                                                if (supervisoryBody == SupervisoryBodyEnum.ESMA) {
                                                    // Include the reportingRegimeBuilder in the report
                                                    // Perform the desired actions for ESMA case
                                                    getFinancialSectorEnum(classificationValue)
                                                            .ifPresent(financialSector ->
                                                                    corporateSectorBuilder.addFinancialSector(financialSector));

                                                    getNonFinancialSector(classificationValue)
                                                            .ifPresent(nonFinancialSector ->
                                                                    corporateSectorBuilder.addNonFinancialSector(nonFinancialSector));
                                                }
                                            });
                                    // Return true if the mappings should be updated successfully
                                    return corporateSectorBuilder.hasData();
                                },
                                getMappings(),
                                getModelPath());
                    }
                    if (corporateSectorBuilder.hasData()) {
                        classificationPaths.forEach(classificationPath -> {
                            Path classificationSchemePath = classificationPath.addElement("industryClassificationScheme");
                            String industryClassificationScheme = getNonNullMappedValue(classificationSchemePath, getMappings()).orElse(null);
                            if (ESMA_CORPORATE_SECTOR_SCHEME.equals(industryClassificationScheme)) {
                                reportingRegimeBuilders.stream()
                                        .filter(reportingRegimeBuilder ->
                                                reportingRegimeBuilder.getSupervisoryBody() != null &&
                                                        reportingRegimeBuilder.getSupervisoryBody().getValue() == SupervisoryBodyEnum.ESMA)
                                        .forEach(reportingRegimeBuilder ->
                                                reportingRegimeBuilder.getOrCreateEsmaPartyInformation().setCorporateSector(corporateSectorBuilder));
                            }
                        });
                    }
                });
    }

    private List<Path> getClassificationPaths(String partyReference) {
        List<Path> classificationPaths = new ArrayList<>();

        Optional<Path> partyPath = getPartyPath(partyReference, "party", "id");
        partyPath.ifPresent(path -> {
            String partyPathStr = path.toString();
            for (Mapping mapping : getMappings()) {
                Path xmlPath = mapping.getXmlPath();
                if (mapping.getXmlValue() != null
                        && xmlPath != null
                        && xmlPath.toString().startsWith(partyPathStr)
                        && xmlPath.endsWith("classification")) {
                    classificationPaths.add(xmlPath);
                }
            }
        });

        return classificationPaths;
    }

    private Optional<Path> getPartyPath(String xmlValue, String... xmlPathEndsWith) {
        return getMappings().stream()
                .filter(m -> m.getXmlPath().endsWith(xmlPathEndsWith))
                .filter(m -> {
                    Object xmlValueObj = m.getXmlValue();
                    String xmlValueStr = (xmlValueObj != null) ? xmlValueObj.toString() : null;
                    return Objects.equals(xmlValueStr, xmlValue);
                })
                .map(m -> m.getXmlPath()) // e.g. nonpublicExecutionReport.party(0).id
                .map(p -> p.getParent()) // e.g. nonpublicExecutionReport.party(0)
                .findFirst();
    }

    private Optional<FinancialSectorEnum> getFinancialSectorEnum(String value) {
        try {
            return Optional.of(FinancialSectorEnum.valueOf(value));
        } catch (IllegalArgumentException e) {
            return Optional.empty();
        }
    }

    private Optional<NonFinancialSector> getNonFinancialSector(String value) {
        try {
            return Optional.of(NonFinancialSectorEnum.valueOf(value))
                    .map(e ->
                            NonFinancialSector.builder()
                                    .setNonFinancialSectorIndicator(e)
                                    .setOrdinal(e.ordinal() + 1));
        } catch (IllegalArgumentException e) {
            return Optional.empty();
        }
    }
}
