package drr.regulation.esma.emir.refit.trade.reports;

import com.google.inject.ImplementedBy;
import com.rosetta.model.lib.functions.ModelObjectValidator;
import com.rosetta.model.lib.mapper.MapperC;
import com.rosetta.model.lib.mapper.MapperS;
import com.rosetta.model.lib.reports.ReportFunction;
import drr.regulation.common.ReportableDelivery;
import drr.regulation.common.TransactionReportInstruction;
import drr.regulation.common.functions.GetReportableDelivery;
import drr.regulation.esma.emir.refit.trade.DeliveryReport;
import drr.regulation.esma.emir.refit.trade.DeliveryReport.DeliveryReportBuilder;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.inject.Inject;


@ImplementedBy(DeliveryRule.DeliveryRuleDefault.class)
public abstract class DeliveryRule implements ReportFunction<TransactionReportInstruction, List<? extends DeliveryReport>> {
	
	@Inject protected ModelObjectValidator objectValidator;
	
	// RosettaFunction dependencies
	//
	@Inject protected CurrencyOfThePriceTimeIntervalQuantityRule currencyOfThePriceTimeIntervalQuantity;
	@Inject protected DaysOfTheWeekRule daysOfTheWeek;
	@Inject protected DeliveryCapacityRule deliveryCapacity;
	@Inject protected DeliveryEndDateRule deliveryEndDate;
	@Inject protected DeliveryIntervalEndTimeRule deliveryIntervalEndTime;
	@Inject protected DeliveryIntervalStartTimeRule deliveryIntervalStartTime;
	@Inject protected DeliveryStartDateRule deliveryStartDate;
	@Inject protected DurationRule duration;
	@Inject protected GetReportableDelivery getReportableDelivery;
	@Inject protected PriceTimeIntervalQuantityRule priceTimeIntervalQuantity;
	@Inject protected QuantityUnitRule quantityUnit;

	/**
	* @param input 
	* @return output 
	*/
	@Override
	public List<? extends DeliveryReport> evaluate(TransactionReportInstruction input) {
		List<DeliveryReport.DeliveryReportBuilder> outputBuilder = doEvaluate(input);
		
		final List<? extends DeliveryReport> output;
		if (outputBuilder == null) {
			output = null;
		} else {
			output = outputBuilder.stream().map(DeliveryReport::build).collect(Collectors.toList());
			objectValidator.validate(DeliveryReport.class, output);
		}
		
		return output;
	}

	protected abstract List<DeliveryReport.DeliveryReportBuilder> doEvaluate(TransactionReportInstruction input);

	public static class DeliveryRuleDefault extends DeliveryRule {
		@Override
		protected List<DeliveryReport.DeliveryReportBuilder> doEvaluate(TransactionReportInstruction input) {
			List<DeliveryReport.DeliveryReportBuilder> output = new ArrayList<>();
			return assignOutput(output, input);
		}
		
		protected List<DeliveryReport.DeliveryReportBuilder> assignOutput(List<DeliveryReport.DeliveryReportBuilder> output, TransactionReportInstruction input) {
			final MapperC<ReportableDelivery> thenResult = MapperS.of(input)
				.mapSingleToList(item -> MapperC.<ReportableDelivery>of(getReportableDelivery.evaluate(item.get())));
			output = toBuilder(thenResult
				.mapItem(item -> MapperS.of(DeliveryReport.builder()
					.setDeliveryStartTime(deliveryIntervalStartTime.evaluate(item.get()))
					.setDeliveryEndTime(deliveryIntervalEndTime.evaluate(item.get()))
					.setDeliveryStartDate(deliveryStartDate.evaluate(item.get()))
					.setDeliveryEndDate(deliveryEndDate.evaluate(item.get()))
					.setDuration(duration.evaluate(item.get()))
					.setDaysOfTheWeek(daysOfTheWeek.evaluate(item.get()))
					.setDeliveryCapacity(deliveryCapacity.evaluate(item.get()))
					.setQuantityUnit(quantityUnit.evaluate(item.get()))
					.setPriceTimeIntervalQuantity(priceTimeIntervalQuantity.evaluate(item.get()))
					.setCurrencyOfThePriceTimeIntervalQuantity(currencyOfThePriceTimeIntervalQuantity.evaluate(item.get()))
					.build()
				)).getMulti());
			
			return Optional.ofNullable(output)
				.map(o -> o.stream().map(i -> i.prune()).collect(Collectors.toList()))
				.orElse(null);
		}
	}
}
