package drr.regulation.esma.emir.refit.trade.reports;

import cdm.base.math.FinancialUnitEnum;
import cdm.base.math.NonNegativeQuantitySchedule;
import cdm.base.math.RoundingDirectionEnum;
import cdm.base.math.UnitType;
import cdm.base.math.functions.RoundToPrecision;
import cdm.base.math.metafields.FieldWithMetaNonNegativeQuantitySchedule;
import cdm.base.math.metafields.ReferenceWithMetaNonNegativeQuantitySchedule;
import cdm.observable.asset.Price;
import cdm.observable.asset.PriceSchedule;
import cdm.observable.asset.metafields.ReferenceWithMetaPriceSchedule;
import cdm.product.asset.CommodityPayout;
import cdm.product.asset.CreditDefaultPayout;
import cdm.product.common.settlement.FixedPrice;
import cdm.product.common.settlement.PriceQuantity;
import cdm.product.common.settlement.ResolvablePriceQuantity;
import cdm.product.qualification.functions.Qualify_AssetClass_InterestRate;
import cdm.product.qualification.functions.Qualify_Commodity_Swap_Basis;
import cdm.product.qualification.functions.Qualify_Commodity_Swap_FixedFloat;
import cdm.product.qualification.functions.Qualify_Commodity_Swaption;
import cdm.product.qualification.functions.Qualify_InterestRate_Forward_Debt;
import cdm.product.qualification.functions.Qualify_InterestRate_Option_Swaption;
import cdm.product.template.AveragingStrikeFeature;
import cdm.product.template.FixedPricePayout;
import cdm.product.template.ForwardPayout;
import cdm.product.template.OptionExercise;
import cdm.product.template.OptionPayout;
import cdm.product.template.OptionStrike;
import cdm.product.template.Payout;
import cdm.product.template.PerformancePayout;
import cdm.product.template.Product;
import cdm.product.template.TradableProduct;
import cdm.product.template.TradeLot;
import com.google.inject.ImplementedBy;
import com.rosetta.model.lib.expression.ComparisonResult;
import com.rosetta.model.lib.expression.MapperMaths;
import com.rosetta.model.lib.mapper.MapperC;
import com.rosetta.model.lib.mapper.MapperS;
import com.rosetta.model.lib.reports.ReportFunction;
import com.rosetta.model.metafields.FieldWithMetaString;
import drr.regulation.common.TransactionReportInstruction;
import drr.regulation.common.functions.BeforeTradeForEvent;
import drr.regulation.common.functions.CommodityBasisLegWithNoSpread;
import drr.regulation.common.functions.CommodityLeg1;
import drr.regulation.common.functions.CommodityObservablePriceQuantity;
import drr.regulation.common.functions.EconomicTermsForProduct;
import drr.regulation.common.functions.FXLeg1;
import drr.regulation.common.functions.FixedPriceLeg1;
import drr.regulation.common.functions.InterestRateLeg1;
import drr.regulation.common.functions.InterestRateLeg2FixedFloatSingleCurrency;
import drr.regulation.common.functions.IsActionTypeTERM;
import drr.regulation.common.functions.IsAllowableAction;
import drr.regulation.common.functions.IsCommodityFixedPriceForward;
import drr.regulation.common.functions.IsCommodityFloatingPriceForward;
import drr.regulation.common.functions.IsCommodityOption;
import drr.regulation.common.functions.IsCreditDefaultSwap;
import drr.regulation.common.functions.IsCreditSwaption;
import drr.regulation.common.functions.IsEquity;
import drr.regulation.common.functions.IsEquityForward;
import drr.regulation.common.functions.IsEquityOption;
import drr.regulation.common.functions.IsFXForward;
import drr.regulation.common.functions.IsFXOption;
import drr.regulation.common.functions.IsFixedFloatZeroCouponSwapKnownAmount;
import drr.regulation.common.functions.IsInterestRateDebtOption;
import drr.regulation.common.functions.IsSingleOnlyCommodityPayout;
import drr.regulation.common.functions.LastAvailableSpotPrice;
import drr.regulation.common.functions.TradableProductForEvent;
import drr.regulation.common.functions.UnderlierForProduct;
import drr.standards.iosco.cde.functions.CDECommodityNotional;
import drr.standards.iosco.cde.functions.CDEEquityForwardNotional;
import drr.standards.iosco.cde.functions.CDEEquityOptionNotional;
import drr.standards.iosco.cde.functions.CDEPerformanceNotional;
import drr.standards.iosco.cde.reports.CDECommodityOptionNotionalRule;
import drr.standards.iosco.cde.reports.CDECreditNotionalRule;
import drr.standards.iosco.cde.reports.CDEFXNotionalRule;
import drr.standards.iosco.cde.reports.CDEForwardNotionalRule;
import drr.standards.iosco.cde.reports.CDEInterestRateNotionalRule;
import java.math.BigDecimal;
import javax.inject.Inject;

import static com.rosetta.model.lib.expression.ExpressionOperators.*;

@ImplementedBy(NotionalAmountOfLeg1Rule.NotionalAmountOfLeg1RuleDefault.class)
public abstract class NotionalAmountOfLeg1Rule implements ReportFunction<TransactionReportInstruction, BigDecimal> {
	
	// RosettaFunction dependencies
	//
	@Inject protected BeforeTradeForEvent beforeTradeForEvent;
	@Inject protected CDECommodityNotional cDECommodityNotional;
	@Inject protected CDECommodityOptionNotionalRule cDECommodityOptionNotional;
	@Inject protected CDECreditNotionalRule cDECreditNotional;
	@Inject protected CDEEquityForwardNotional cDEEquityForwardNotional;
	@Inject protected CDEEquityOptionNotional cDEEquityOptionNotional;
	@Inject protected CDEFXNotionalRule cDEFXNotional;
	@Inject protected CDEForwardNotionalRule cDEForwardNotional;
	@Inject protected CDEInterestRateNotionalRule cDEInterestRateNotional;
	@Inject protected CDEPerformanceNotional cDEPerformanceNotional;
	@Inject protected CommodityBasisLegWithNoSpread commodityBasisLegWithNoSpread;
	@Inject protected CommodityLeg1 commodityLeg1;
	@Inject protected CommodityObservablePriceQuantity commodityObservablePriceQuantity;
	@Inject protected EconomicTermsForProduct economicTermsForProduct;
	@Inject protected FXLeg1 fXLeg1;
	@Inject protected FixedPriceLeg1 fixedPriceLeg1;
	@Inject protected InterestRateLeg1 interestRateLeg1;
	@Inject protected InterestRateLeg2FixedFloatSingleCurrency interestRateLeg2FixedFloatSingleCurrency;
	@Inject protected IsActionTypeTERM isActionTypeTERM;
	@Inject protected IsAllowableAction isAllowableAction;
	@Inject protected IsCommodityFixedPriceForward isCommodityFixedPriceForward;
	@Inject protected IsCommodityFloatingPriceForward isCommodityFloatingPriceForward;
	@Inject protected IsCommodityOption isCommodityOption;
	@Inject protected IsCreditDefaultSwap isCreditDefaultSwap;
	@Inject protected IsCreditSwaption isCreditSwaption;
	@Inject protected IsEquity isEquity;
	@Inject protected IsEquityForward isEquityForward;
	@Inject protected IsEquityOption isEquityOption;
	@Inject protected IsFXForward isFXForward;
	@Inject protected IsFXOption isFXOption;
	@Inject protected IsFixedFloatZeroCouponSwapKnownAmount isFixedFloatZeroCouponSwapKnownAmount;
	@Inject protected IsInterestRateDebtOption isInterestRateDebtOption;
	@Inject protected IsSingleOnlyCommodityPayout isSingleOnlyCommodityPayout;
	@Inject protected LastAvailableSpotPrice lastAvailableSpotPrice;
	@Inject protected Qualify_AssetClass_InterestRate qualify_AssetClass_InterestRate;
	@Inject protected Qualify_Commodity_Swap_Basis qualify_Commodity_Swap_Basis;
	@Inject protected Qualify_Commodity_Swap_FixedFloat qualify_Commodity_Swap_FixedFloat;
	@Inject protected Qualify_Commodity_Swaption qualify_Commodity_Swaption;
	@Inject protected Qualify_InterestRate_Forward_Debt qualify_InterestRate_Forward_Debt;
	@Inject protected Qualify_InterestRate_Option_Swaption qualify_InterestRate_Option_Swaption;
	@Inject protected RoundToPrecision roundToPrecision;
	@Inject protected TradableProductForEvent tradableProductForEvent;
	@Inject protected UnderlierForProduct underlierForProduct;

	/**
	* @param input 
	* @return output 
	*/
	@Override
	public BigDecimal evaluate(TransactionReportInstruction input) {
		BigDecimal output = doEvaluate(input);
		
		return output;
	}

	protected abstract BigDecimal doEvaluate(TransactionReportInstruction input);

	public static class NotionalAmountOfLeg1RuleDefault extends NotionalAmountOfLeg1Rule {
		@Override
		protected BigDecimal doEvaluate(TransactionReportInstruction input) {
			BigDecimal output = null;
			return assignOutput(output, input);
		}
		
		protected BigDecimal assignOutput(BigDecimal output, TransactionReportInstruction input) {
			final MapperS<TransactionReportInstruction> thenResult0 = MapperS.of(input)
				.filterSingleNullSafe(item -> isAllowableAction.evaluate(item.get()));
			final MapperS<BigDecimal> thenResult1 = thenResult0
				.mapSingleToItem(reportableEvent -> {
					final Boolean _boolean = isActionTypeTERM.evaluate(thenResult0.get());
					final MapperS<TradableProduct> thenResult;
					if ((_boolean == null ? false : _boolean)) {
						thenResult = MapperS.of(beforeTradeForEvent.evaluate(thenResult0.get())).<TradableProduct>map("getTradableProduct", trade -> trade.getTradableProduct());
					} else {
						thenResult = MapperS.of(tradableProductForEvent.evaluate(thenResult0.get()));
					}
					return thenResult
						.mapSingleToItem(item -> {
							final Boolean boolean0 = isCommodityOption.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean0 == null ? false : boolean0)) {
								if (exists(MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).get()).<OptionExercise>map("getExerciseTerms", optionPayout -> optionPayout.getExerciseTerms()).<OptionStrike>map("getStrike", optionExercise -> optionExercise.getStrike()).<Price>map("getStrikePrice", optionStrike -> optionStrike.getStrikePrice())).getOrDefault(false)) {
									return MapperS.of(cDECommodityOptionNotional.evaluate(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).get()));
								}
								if (exists(MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).get()).<OptionExercise>map("getExerciseTerms", optionPayout -> optionPayout.getExerciseTerms()).<OptionStrike>map("getStrike", optionExercise -> optionExercise.getStrike()).<AveragingStrikeFeature>map("getAveragingStrikeFeature", optionStrike -> optionStrike.getAveragingStrikeFeature())).getOrDefault(false)) {
									if (exists(MapperS.of(lastAvailableSpotPrice.evaluate(reportableEvent.get()))).getOrDefault(false)) {
										return MapperS.of(cDECommodityNotional.evaluate(MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).get()).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).get(), lastAvailableSpotPrice.evaluate(reportableEvent.get())));
									}
									final MapperC<NonNegativeQuantitySchedule> _thenResult0 = MapperS.of(item.<TradeLot>mapC("getTradeLot", tradableProduct -> tradableProduct.getTradeLot()).<PriceQuantity>mapC("getPriceQuantity", tradeLot -> tradeLot.getPriceQuantity()).get()).<FieldWithMetaNonNegativeQuantitySchedule>mapC("getQuantity", priceQuantity -> priceQuantity.getQuantity()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue())
										.filterItemNullSafe(_item -> exists(_item.<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FieldWithMetaString>map("getCurrency", unitType -> unitType.getCurrency()).<String>map("getValue", _f->_f.getValue())).get());
									final MapperC<BigDecimal> _thenResult1 = _thenResult0
										.mapItem(_item -> _item.<BigDecimal>map("getValue", measureBase -> measureBase.getValue()));
									return MapperS.of(_thenResult1.get());
								}
								return MapperS.<BigDecimal>ofNull();
							}
							final Boolean boolean1 = isCreditSwaption.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean1 == null ? false : boolean1)) {
								return MapperS.of(cDECreditNotional.evaluate(MapperS.of(economicTermsForProduct.evaluate(underlierForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()))).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<CreditDefaultPayout>map("getCreditDefaultPayout", payout -> payout.getCreditDefaultPayout()).get()));
							}
							if (ComparisonResult.of(MapperS.of(isFXForward.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()))).or(ComparisonResult.of(MapperS.of(isFXOption.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())))).getOrDefault(false)) {
								return MapperS.of(cDEFXNotional.evaluate(fXLeg1.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())));
							}
							final Boolean boolean2 = isEquityOption.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean2 == null ? false : boolean2)) {
								return MapperS.of(cDEEquityOptionNotional.evaluate(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).getMulti()));
							}
							final Boolean boolean3 = isEquityForward.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean3 == null ? false : boolean3)) {
								return MapperS.of(cDEEquityForwardNotional.evaluate(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<ForwardPayout>mapC("getForwardPayout", payout -> payout.getForwardPayout()).getMulti()));
							}
							final Boolean boolean4 = qualify_InterestRate_Option_Swaption.evaluate(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()));
							if ((boolean4 == null ? false : boolean4)) {
								return MapperS.of(cDEInterestRateNotional.evaluate(interestRateLeg1.evaluate(underlierForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()))));
							}
							if (ComparisonResult.of(MapperS.of(isInterestRateDebtOption.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()))).and(exists(MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).get()).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FieldWithMetaString>map("getCurrency", unitType -> unitType.getCurrency()).<String>map("getValue", _f->_f.getValue()))).getOrDefault(false)) {
								return MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).get()).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).<BigDecimal>map("getValue", measureBase -> measureBase.getValue());
							}
							if (ComparisonResult.of(MapperS.of(isInterestRateDebtOption.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()))).and(exists(MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).get()).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FinancialUnitEnum>map("getFinancialUnit", unitType -> unitType.getFinancialUnit()))).getOrDefault(false)) {
								return MapperMaths.<BigDecimal, BigDecimal, BigDecimal>multiply(MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).get()).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).<BigDecimal>map("getValue", measureBase -> measureBase.getValue()), MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<OptionPayout>mapC("getOptionPayout", payout -> payout.getOptionPayout()).<OptionExercise>map("getExerciseTerms", optionPayout -> optionPayout.getExerciseTerms()).<OptionStrike>map("getStrike", optionExercise -> optionExercise.getStrike()).<Price>map("getStrikePrice", optionStrike -> optionStrike.getStrikePrice()).<BigDecimal>map("getValue", measureBase -> measureBase.getValue()));
							}
							final Boolean boolean5 = isEquity.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean5 == null ? false : boolean5)) {
								return MapperS.of(cDEPerformanceNotional.evaluate(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<PerformancePayout>mapC("getPerformancePayout", payout -> payout.getPerformancePayout()).getMulti()));
							}
							final Boolean boolean6 = isCreditDefaultSwap.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean6 == null ? false : boolean6)) {
								return MapperS.of(cDECreditNotional.evaluate(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<CreditDefaultPayout>map("getCreditDefaultPayout", payout -> payout.getCreditDefaultPayout()).get()));
							}
							final Boolean boolean7 = qualify_Commodity_Swap_FixedFloat.evaluate(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()));
							if ((boolean7 == null ? false : boolean7)) {
								return MapperS.of(cDECommodityNotional.evaluate(MapperS.of(fixedPriceLeg1.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).get(), MapperS.of(fixedPriceLeg1.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<FixedPrice>map("getFixedPrice", fixedPricePayout -> fixedPricePayout.getFixedPrice()).<ReferenceWithMetaPriceSchedule>map("getPrice", fixedPrice -> fixedPrice.getPrice()).<PriceSchedule>map("getValue", _f->_f.getValue()).get()));
							}
							if (ComparisonResult.of(MapperS.of(qualify_Commodity_Swaption.evaluate(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())))).and(ComparisonResult.of(MapperS.of(qualify_Commodity_Swap_FixedFloat.evaluate(economicTermsForProduct.evaluate(underlierForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())))))).getOrDefault(false)) {
								return MapperS.of(cDECommodityNotional.evaluate(MapperS.of(fixedPriceLeg1.evaluate(underlierForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()))).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).get(), MapperS.of(fixedPriceLeg1.evaluate(underlierForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()))).<FixedPrice>map("getFixedPrice", fixedPricePayout -> fixedPricePayout.getFixedPrice()).<ReferenceWithMetaPriceSchedule>map("getPrice", fixedPrice -> fixedPrice.getPrice()).<PriceSchedule>map("getValue", _f->_f.getValue()).get()));
							}
							if (ComparisonResult.of(MapperS.of(qualify_Commodity_Swaption.evaluate(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())))).and(ComparisonResult.of(MapperS.of(qualify_Commodity_Swap_Basis.evaluate(economicTermsForProduct.evaluate(underlierForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())))))).getOrDefault(false)) {
								if (exists(MapperS.of(lastAvailableSpotPrice.evaluate(reportableEvent.get()))).getOrDefault(false)) {
									return MapperS.of(cDECommodityNotional.evaluate(MapperS.of(commodityBasisLegWithNoSpread.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).get(), lastAvailableSpotPrice.evaluate(reportableEvent.get())));
								}
								final MapperC<NonNegativeQuantitySchedule> thenResult2 = MapperS.of(commodityObservablePriceQuantity.evaluate(item.<TradeLot>mapC("getTradeLot", tradableProduct -> tradableProduct.getTradeLot()).get(), commodityLeg1.evaluate(underlierForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())))).<FieldWithMetaNonNegativeQuantitySchedule>mapC("getQuantity", priceQuantity -> priceQuantity.getQuantity()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue())
									.filterItemNullSafe(_item -> exists(_item.<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FieldWithMetaString>map("getCurrency", unitType -> unitType.getCurrency()).<String>map("getValue", _f->_f.getValue())).get());
								final MapperC<BigDecimal> thenResult3 = thenResult2
									.mapItem(_item -> _item.<BigDecimal>map("getValue", measureBase -> measureBase.getValue()));
								return MapperS.of(thenResult3.get());
							}
							final Boolean boolean8 = qualify_Commodity_Swap_Basis.evaluate(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()));
							if ((boolean8 == null ? false : boolean8)) {
								if (exists(MapperS.of(lastAvailableSpotPrice.evaluate(reportableEvent.get()))).getOrDefault(false)) {
									return MapperS.of(cDECommodityNotional.evaluate(MapperS.of(commodityBasisLegWithNoSpread.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).get(), lastAvailableSpotPrice.evaluate(reportableEvent.get())));
								}
								final MapperC<NonNegativeQuantitySchedule> thenResult4 = MapperS.of(commodityObservablePriceQuantity.evaluate(item.<TradeLot>mapC("getTradeLot", tradableProduct -> tradableProduct.getTradeLot()).get(), commodityLeg1.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()))).<FieldWithMetaNonNegativeQuantitySchedule>mapC("getQuantity", priceQuantity -> priceQuantity.getQuantity()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue())
									.filterItemNullSafe(_item -> exists(_item.<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FieldWithMetaString>map("getCurrency", unitType -> unitType.getCurrency()).<String>map("getValue", _f->_f.getValue())).get());
								final MapperC<BigDecimal> thenResult5 = thenResult4
									.mapItem(_item -> _item.<BigDecimal>map("getValue", measureBase -> measureBase.getValue()));
								return MapperS.of(thenResult5.get());
							}
							final Boolean boolean9 = isSingleOnlyCommodityPayout.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean9 == null ? false : boolean9)) {
								if (exists(MapperS.of(lastAvailableSpotPrice.evaluate(reportableEvent.get()))).getOrDefault(false)) {
									return MapperS.of(cDECommodityNotional.evaluate(MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<CommodityPayout>mapC("getCommodityPayout", payout -> payout.getCommodityPayout()).get()).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).get(), lastAvailableSpotPrice.evaluate(reportableEvent.get())));
								}
								final MapperC<NonNegativeQuantitySchedule> thenResult6 = MapperS.of(item.<TradeLot>mapC("getTradeLot", tradableProduct -> tradableProduct.getTradeLot()).<PriceQuantity>mapC("getPriceQuantity", tradeLot -> tradeLot.getPriceQuantity()).get()).<FieldWithMetaNonNegativeQuantitySchedule>mapC("getQuantity", priceQuantity -> priceQuantity.getQuantity()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue())
									.filterItemNullSafe(_item -> exists(_item.<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FieldWithMetaString>map("getCurrency", unitType -> unitType.getCurrency()).<String>map("getValue", _f->_f.getValue())).get());
								final MapperC<BigDecimal> thenResult7 = thenResult6
									.mapItem(_item -> _item.<BigDecimal>map("getValue", measureBase -> measureBase.getValue()));
								return MapperS.of(thenResult7.get());
							}
							final Boolean boolean10 = isCommodityFixedPriceForward.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean10 == null ? false : boolean10)) {
								return MapperS.of(cDECommodityNotional.evaluate(MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<FixedPricePayout>mapC("getFixedPricePayout", payout -> payout.getFixedPricePayout()).get()).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).get(), MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<FixedPricePayout>mapC("getFixedPricePayout", payout -> payout.getFixedPricePayout()).get()).<FixedPrice>map("getFixedPrice", fixedPricePayout -> fixedPricePayout.getFixedPrice()).<ReferenceWithMetaPriceSchedule>map("getPrice", fixedPrice -> fixedPrice.getPrice()).<PriceSchedule>map("getValue", _f->_f.getValue()).get()));
							}
							final Boolean boolean11 = isCommodityFloatingPriceForward.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean11 == null ? false : boolean11)) {
								if (exists(MapperS.of(lastAvailableSpotPrice.evaluate(reportableEvent.get()))).getOrDefault(false)) {
									return MapperS.of(cDECommodityNotional.evaluate(MapperS.of(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<CommodityPayout>mapC("getCommodityPayout", payout -> payout.getCommodityPayout()).get()).<ResolvablePriceQuantity>map("getPriceQuantity", payoutBase -> payoutBase.getPriceQuantity()).<ReferenceWithMetaNonNegativeQuantitySchedule>map("getQuantitySchedule", resolvablePriceQuantity -> resolvablePriceQuantity.getQuantitySchedule()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue()).get(), lastAvailableSpotPrice.evaluate(reportableEvent.get())));
								}
								final MapperC<NonNegativeQuantitySchedule> thenResult8 = MapperS.of(commodityObservablePriceQuantity.evaluate(item.<TradeLot>mapC("getTradeLot", tradableProduct -> tradableProduct.getTradeLot()).get(), MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<CommodityPayout>mapC("getCommodityPayout", payout -> payout.getCommodityPayout()).get())).<FieldWithMetaNonNegativeQuantitySchedule>mapC("getQuantity", priceQuantity -> priceQuantity.getQuantity()).<NonNegativeQuantitySchedule>map("getValue", _f->_f.getValue())
									.filterItemNullSafe(_item -> exists(_item.<UnitType>map("getUnit", measureBase -> measureBase.getUnit()).<FieldWithMetaString>map("getCurrency", unitType -> unitType.getCurrency()).<String>map("getValue", _f->_f.getValue())).get());
								final MapperC<BigDecimal> thenResult9 = thenResult8
									.mapItem(_item -> _item.<BigDecimal>map("getValue", measureBase -> measureBase.getValue()));
								return MapperS.of(thenResult9.get());
							}
							final Boolean boolean12 = isFixedFloatZeroCouponSwapKnownAmount.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get());
							if ((boolean12 == null ? false : boolean12)) {
								return MapperS.of(cDEInterestRateNotional.evaluate(interestRateLeg2FixedFloatSingleCurrency.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())));
							}
							final Boolean boolean13 = qualify_InterestRate_Forward_Debt.evaluate(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()));
							if ((boolean13 == null ? false : boolean13)) {
								return MapperS.of(cDEForwardNotional.evaluate(MapperS.of(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())).<Payout>map("getPayout", economicTerms -> economicTerms.getPayout()).<ForwardPayout>mapC("getForwardPayout", payout -> payout.getForwardPayout()).get()));
							}
							final Boolean boolean14 = qualify_AssetClass_InterestRate.evaluate(economicTermsForProduct.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get()));
							if ((boolean14 == null ? false : boolean14)) {
								return MapperS.of(cDEInterestRateNotional.evaluate(interestRateLeg1.evaluate(item.<Product>map("getProduct", tradableProduct -> tradableProduct.getProduct()).get())));
							}
							return MapperS.<BigDecimal>ofNull();
						});
				});
			output = thenResult1
				.mapSingleToItem(item -> MapperS.of(roundToPrecision.evaluate(item.get(), 5, RoundingDirectionEnum.NEAREST))).get();
			
			return output;
		}
	}
}
