package drr.regulation.jfsa.rewrite.trade.reports;

import com.google.inject.ImplementedBy;
import com.rosetta.model.lib.functions.ModelObjectValidator;
import com.rosetta.model.lib.mapper.MapperC;
import com.rosetta.model.lib.mapper.MapperS;
import com.rosetta.model.lib.records.Date;
import com.rosetta.model.lib.reports.ReportFunction;
import drr.regulation.common.PriceReportablePeriod;
import drr.regulation.common.TransactionReportInstruction;
import drr.regulation.common.functions.GetReportablePricePeriod;
import drr.regulation.jfsa.rewrite.trade.PriceScheduleReport;
import drr.regulation.jfsa.rewrite.trade.PriceScheduleReport.PriceScheduleReportBuilder;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.inject.Inject;


@ImplementedBy(PriceScheduleRule.PriceScheduleRuleDefault.class)
public abstract class PriceScheduleRule implements ReportFunction<TransactionReportInstruction, List<? extends PriceScheduleReport>> {
	
	@Inject protected ModelObjectValidator objectValidator;
	
	// RosettaFunction dependencies
	//
	@Inject protected ContractPriceScheduleEffectiveDateRule contractPriceScheduleEffectiveDate;
	@Inject protected ContractPriceScheduleEndDateRule contractPriceScheduleEndDate;
	@Inject protected GetReportablePricePeriod getReportablePricePeriod;
	@Inject protected PriceScheduleAmountFormatMonetaryRule priceScheduleAmountFormatMonetary;
	@Inject protected PriceScheduleAmountFormatPercentageRule priceScheduleAmountFormatPercentage;

	/**
	* @param input 
	* @return output 
	*/
	@Override
	public List<? extends PriceScheduleReport> evaluate(TransactionReportInstruction input) {
		List<PriceScheduleReport.PriceScheduleReportBuilder> outputBuilder = doEvaluate(input);
		
		final List<? extends PriceScheduleReport> output;
		if (outputBuilder == null) {
			output = null;
		} else {
			output = outputBuilder.stream().map(PriceScheduleReport::build).collect(Collectors.toList());
			objectValidator.validate(PriceScheduleReport.class, output);
		}
		
		return output;
	}

	protected abstract List<PriceScheduleReport.PriceScheduleReportBuilder> doEvaluate(TransactionReportInstruction input);

	public static class PriceScheduleRuleDefault extends PriceScheduleRule {
		@Override
		protected List<PriceScheduleReport.PriceScheduleReportBuilder> doEvaluate(TransactionReportInstruction input) {
			List<PriceScheduleReport.PriceScheduleReportBuilder> output = new ArrayList<>();
			return assignOutput(output, input);
		}
		
		protected List<PriceScheduleReport.PriceScheduleReportBuilder> assignOutput(List<PriceScheduleReport.PriceScheduleReportBuilder> output, TransactionReportInstruction input) {
			final MapperC<PriceReportablePeriod> thenResult0 = MapperS.of(input)
				.mapSingleToList(item -> MapperC.<PriceReportablePeriod>of(getReportablePricePeriod.evaluate(item.get())));
			final MapperC<PriceScheduleReport> thenResult1 = thenResult0
				.mapItem(item -> MapperS.of(PriceScheduleReport.builder()
					.setEffectiveDate(contractPriceScheduleEffectiveDate.evaluate(item.get()))
					.setEndDate(contractPriceScheduleEndDate.evaluate(item.get()))
					.setPriceMonetary(priceScheduleAmountFormatMonetary.evaluate(item.get()))
					.setPricePercentage(priceScheduleAmountFormatPercentage.evaluate(item.get()))
					.build()
				));
			output = toBuilder(thenResult1
				.sort(item -> item.<Date>map("getEffectiveDate", priceScheduleReport -> priceScheduleReport.getEffectiveDate())).getMulti());
			
			return Optional.ofNullable(output)
				.map(o -> o.stream().map(i -> i.prune()).collect(Collectors.toList()))
				.orElse(null);
		}
	}
}
