package drr.regulation.jfsa.rewrite.trade;

import cdm.base.staticdata.asset.common.ISOCurrencyCodeEnum;
import cdm.base.staticdata.asset.common.ProductIdTypeEnum;
import com.google.common.collect.ImmutableList;
import com.rosetta.model.lib.RosettaModelObject;
import com.rosetta.model.lib.RosettaModelObjectBuilder;
import com.rosetta.model.lib.annotations.RosettaAttribute;
import com.rosetta.model.lib.annotations.RosettaDataType;
import com.rosetta.model.lib.meta.RosettaMetaData;
import com.rosetta.model.lib.path.RosettaPath;
import com.rosetta.model.lib.process.BuilderMerger;
import com.rosetta.model.lib.process.BuilderProcessor;
import com.rosetta.model.lib.process.Processor;
import com.rosetta.model.lib.records.Date;
import com.rosetta.util.ListEquals;
import drr.regulation.jfsa.rewrite.trade.JFSANonReportable;
import drr.regulation.jfsa.rewrite.trade.JFSATransactionReport;
import drr.regulation.jfsa.rewrite.trade.JFSATransactionReport.JFSATransactionReportBuilder;
import drr.regulation.jfsa.rewrite.trade.JFSATransactionReport.JFSATransactionReportBuilderImpl;
import drr.regulation.jfsa.rewrite.trade.JFSATransactionReport.JFSATransactionReportImpl;
import drr.regulation.jfsa.rewrite.trade.NotionalAmountScheduleLeg1Report;
import drr.regulation.jfsa.rewrite.trade.NotionalAmountScheduleLeg2Report;
import drr.regulation.jfsa.rewrite.trade.NotionalQuantityScheduleLeg1Report;
import drr.regulation.jfsa.rewrite.trade.NotionalQuantityScheduleLeg2Report;
import drr.regulation.jfsa.rewrite.trade.OtherPaymentReport;
import drr.regulation.jfsa.rewrite.trade.PackageTransactionPrice;
import drr.regulation.jfsa.rewrite.trade.PackageTransactionSpread;
import drr.regulation.jfsa.rewrite.trade.PriceAndNotation;
import drr.regulation.jfsa.rewrite.trade.PriceScheduleReport;
import drr.regulation.jfsa.rewrite.trade.SpreadOfLeg1;
import drr.regulation.jfsa.rewrite.trade.SpreadOfLeg2;
import drr.regulation.jfsa.rewrite.trade.StrikePriceAndNotation;
import drr.regulation.jfsa.rewrite.trade.StrikePriceScheduleReport;
import drr.regulation.jfsa.rewrite.trade.meta.JFSATransactionReportMeta;
import iso20022.auth030.esma.InterestComputationMethod4Code;
import iso20022.auth030.esma.ValuationType1Code;
import java.math.BigDecimal;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import static java.util.Optional.ofNullable;

/**
 * @version 4.2.1
 */
@RosettaDataType(value="JFSATransactionReport", builder=JFSATransactionReport.JFSATransactionReportBuilderImpl.class, version="4.2.1")
public interface JFSATransactionReport extends RosettaModelObject {

	JFSATransactionReportMeta metaData = new JFSATransactionReportMeta();

	/*********************** Getter Methods  ***********************/
	Date getEffectiveDate();
	Date getExpirationDate();
	Date getEarlyTerminationDate();
	ZonedDateTime getReportingTimestamp();
	ZonedDateTime getExecutionTimestamp();
	String getEntityResponsibleForReporting();
	String getCounterparty1();
	String getCounterparty2();
	Boolean getCounterparty2IdentifierType();
	String getDirection1();
	String getDirection2Leg1();
	String getDirection2Leg2();
	String getSubmitterIdentifier();
	String getNewSDRIdentifier();
	String getOriginalSwapSDRIdentifier();
	String getCleared();
	String getCentralCounterparty();
	String getClearingMember();
	String getPlatformIdentifier();
	String getConfirmed();
	Date getFinalContractualSettlementDate();
	ZonedDateTime getConfirmationTimestamp();
	ISOCurrencyCodeEnum getSettlementCurrencyLeg1();
	ISOCurrencyCodeEnum getSettlementCurrencyLeg2();
	String getClearingAccountOrigin();
	String getOriginalSwapUTI();
	ZonedDateTime getClearingReceiptTimestamp();
	String getUti();
	String getUtiProprietary();
	String getPriorUti();
	String getPriorUtiProprietary();
	InterestComputationMethod4Code getFixedRateDayCountConventionLeg1();
	InterestComputationMethod4Code getFloatingRateDayCountConventionOfLeg1();
	InterestComputationMethod4Code getFixedRateDayCountConventionLeg2();
	InterestComputationMethod4Code getFloatingRateDayCountConventionOfLeg2();
	String getFixedRatePaymentFrequencyPeriodLeg1();
	String getFixedRatePaymentFrequencyPeriodLeg2();
	String getFloatingRatePaymentFrequencyPeriodLeg1();
	String getFloatingRatePaymentFrequencyPeriodLeg2();
	Integer getFixedRatePaymentFrequencyPeriodMultiplierLeg1();
	Integer getFixedRatePaymentFrequencyPeriodMultiplierLeg2();
	Integer getFloatingRatePaymentFrequencyPeriodMultiplierLeg1();
	Integer getFloatingRatePaymentFrequencyPeriodMultiplierLeg2();
	ZonedDateTime getFixingDateLeg1();
	ZonedDateTime getFixingDateLeg2();
	String getFloatingRateResetFrequencyPeriodLeg1();
	String getFloatingRateResetFrequencyPeriodOfLeg2();
	Integer getFloatingRateResetFrequencyPeriodMultiplierLeg1();
	Integer getFloatingRateResetFrequencyPeriodMultiplierLeg2();
	List<? extends OtherPaymentReport> getOtherPayment();
	BigDecimal getValuationAmount();
	ISOCurrencyCodeEnum getValuationCurrency();
	ValuationType1Code getValuationMethod();
	ZonedDateTime getValuationTimestamp();
	Boolean getCollateralPortfolioIndicator();
	String getInitialMarginCollateralPortfolioCode();
	String getVariationMarginCollateralPortfolioCode();
	PriceAndNotation getPrice();
	String getPriceNotation();
	String getPriceUnitOfMeasure();
	List<? extends PriceScheduleReport> getPriceSchedule();
	BigDecimal getFixedRateLeg1();
	BigDecimal getFixedRateLeg2();
	SpreadOfLeg1 getSpreadOfLeg1();
	String getSpreadNotationOfLeg1();
	SpreadOfLeg2 getSpreadOfLeg2();
	String getSpreadNotationOfLeg2();
	StrikePriceAndNotation getStrikePrice();
	String getStrikePriceNotation();
	List<? extends StrikePriceScheduleReport> getStrikePriceSchedule();
	BigDecimal getOptionPremiumAmount();
	ISOCurrencyCodeEnum getOptionPremiumCurrency();
	Date getOptionPremiumPaymentDate();
	Date getFirstExerciseDate();
	BigDecimal getExchangeRate();
	String getExchangeRateBasis();
	BigDecimal getNotionalLeg1();
	BigDecimal getNotionalLeg2();
	BigDecimal getCallAmount();
	BigDecimal getPutAmount();
	ISOCurrencyCodeEnum getNotionalCurrencyOfLeg1();
	ISOCurrencyCodeEnum getNotionalCurrencyOfLeg2();
	ISOCurrencyCodeEnum getCallCurrency();
	ISOCurrencyCodeEnum getPutCurrency();
	String getQuantityUnitOfMeasureLeg1();
	String getQuantityUnitOfMeasureLeg2();
	List<? extends NotionalAmountScheduleLeg1Report> getNotionalAmountScheduleLeg1();
	List<? extends NotionalAmountScheduleLeg2Report> getNotionalAmountScheduleLeg2();
	BigDecimal getTotalNotionalQuantityLeg1();
	BigDecimal getTotalNotionalQuantityLeg2();
	List<? extends NotionalQuantityScheduleLeg1Report> getNotionalQuantityScheduleLeg1();
	List<? extends NotionalQuantityScheduleLeg2Report> getNotionalQuantityScheduleLeg2();
	String getActionType();
	String getEventType();
	ZonedDateTime getEventTimestamp();
	String getEventIdentifier();
	BigDecimal getIndexFactor();
	String getEmbeddedOptionType();
	String getUpi();
	String getDeliveryType();
	String getAssetClass();
	String getUnderlyingIdentificationType();
	String getUnderlyingIdentification();
	String getUnderlyingIndexIndicator();
	String getNameOfTheUnderlyingIndex();
	String getReferenceEntity();
	String getIndicatorOfTheFloatingRateOfLeg1();
	String getIndicatorOfTheFloatingRateOfLeg2();
	String getNameOfTheFloatingRateOfLeg1();
	String getNameOfTheFloatingRateOfLeg2();
	String getFloatingRateReferencePeriodOfLeg1TimePeriod();
	String getFloatingRateReferencePeriodOfLeg2TimePeriod();
	Integer getFloatingRateReferencePeriodOfLeg1Multiplier();
	Integer getFloatingRateReferencePeriodOfLeg2Multiplier();
	Boolean getIsCrypto();
	Date getMaturityDateOfTheUnderlying();
	String getSeniority();
	Integer getSeries();
	Integer getVersion();
	BigDecimal getCdSIndexAttachmentPoint();
	BigDecimal getCdSIndexDetachmentPoint();
	String getCustomBasketCode();
	List<String> getIdentifierOfBasketConstituents();
	List<ProductIdTypeEnum> getBasketConstituentIdentifierSource();
	String getContractType();
	String getOptionStyle();
	String getOptionType();
	String getPackageIdentifier();
	PackageTransactionPrice getPackageTransactionPrice();
	ISOCurrencyCodeEnum getPackageTransactionPriceCurrency();
	String getPackageTransactionPriceNotation();
	PackageTransactionSpread getPackageTransactionSpread();
	ISOCurrencyCodeEnum getPackageTransactionSpreadCurrency();
	String getPackageTransactionSpreadNotation();
	String getTechnicalRecordId();
	JFSANonReportable getNonReportable();

	/*********************** Build Methods  ***********************/
	JFSATransactionReport build();
	
	JFSATransactionReport.JFSATransactionReportBuilder toBuilder();
	
	static JFSATransactionReport.JFSATransactionReportBuilder builder() {
		return new JFSATransactionReport.JFSATransactionReportBuilderImpl();
	}

	/*********************** Utility Methods  ***********************/
	@Override
	default RosettaMetaData<? extends JFSATransactionReport> metaData() {
		return metaData;
	}
	
	@Override
	default Class<? extends JFSATransactionReport> getType() {
		return JFSATransactionReport.class;
	}
	
	
	@Override
	default void process(RosettaPath path, Processor processor) {
		processor.processBasic(path.newSubPath("effectiveDate"), Date.class, getEffectiveDate(), this);
		processor.processBasic(path.newSubPath("expirationDate"), Date.class, getExpirationDate(), this);
		processor.processBasic(path.newSubPath("earlyTerminationDate"), Date.class, getEarlyTerminationDate(), this);
		processor.processBasic(path.newSubPath("reportingTimestamp"), ZonedDateTime.class, getReportingTimestamp(), this);
		processor.processBasic(path.newSubPath("executionTimestamp"), ZonedDateTime.class, getExecutionTimestamp(), this);
		processor.processBasic(path.newSubPath("entityResponsibleForReporting"), String.class, getEntityResponsibleForReporting(), this);
		processor.processBasic(path.newSubPath("counterparty1"), String.class, getCounterparty1(), this);
		processor.processBasic(path.newSubPath("counterparty2"), String.class, getCounterparty2(), this);
		processor.processBasic(path.newSubPath("counterparty2IdentifierType"), Boolean.class, getCounterparty2IdentifierType(), this);
		processor.processBasic(path.newSubPath("direction1"), String.class, getDirection1(), this);
		processor.processBasic(path.newSubPath("direction2Leg1"), String.class, getDirection2Leg1(), this);
		processor.processBasic(path.newSubPath("direction2Leg2"), String.class, getDirection2Leg2(), this);
		processor.processBasic(path.newSubPath("submitterIdentifier"), String.class, getSubmitterIdentifier(), this);
		processor.processBasic(path.newSubPath("newSDRIdentifier"), String.class, getNewSDRIdentifier(), this);
		processor.processBasic(path.newSubPath("originalSwapSDRIdentifier"), String.class, getOriginalSwapSDRIdentifier(), this);
		processor.processBasic(path.newSubPath("cleared"), String.class, getCleared(), this);
		processor.processBasic(path.newSubPath("centralCounterparty"), String.class, getCentralCounterparty(), this);
		processor.processBasic(path.newSubPath("clearingMember"), String.class, getClearingMember(), this);
		processor.processBasic(path.newSubPath("platformIdentifier"), String.class, getPlatformIdentifier(), this);
		processor.processBasic(path.newSubPath("confirmed"), String.class, getConfirmed(), this);
		processor.processBasic(path.newSubPath("finalContractualSettlementDate"), Date.class, getFinalContractualSettlementDate(), this);
		processor.processBasic(path.newSubPath("confirmationTimestamp"), ZonedDateTime.class, getConfirmationTimestamp(), this);
		processor.processBasic(path.newSubPath("settlementCurrencyLeg1"), ISOCurrencyCodeEnum.class, getSettlementCurrencyLeg1(), this);
		processor.processBasic(path.newSubPath("settlementCurrencyLeg2"), ISOCurrencyCodeEnum.class, getSettlementCurrencyLeg2(), this);
		processor.processBasic(path.newSubPath("clearingAccountOrigin"), String.class, getClearingAccountOrigin(), this);
		processor.processBasic(path.newSubPath("originalSwapUTI"), String.class, getOriginalSwapUTI(), this);
		processor.processBasic(path.newSubPath("clearingReceiptTimestamp"), ZonedDateTime.class, getClearingReceiptTimestamp(), this);
		processor.processBasic(path.newSubPath("uti"), String.class, getUti(), this);
		processor.processBasic(path.newSubPath("utiProprietary"), String.class, getUtiProprietary(), this);
		processor.processBasic(path.newSubPath("priorUti"), String.class, getPriorUti(), this);
		processor.processBasic(path.newSubPath("priorUtiProprietary"), String.class, getPriorUtiProprietary(), this);
		processor.processBasic(path.newSubPath("fixedRateDayCountConventionLeg1"), InterestComputationMethod4Code.class, getFixedRateDayCountConventionLeg1(), this);
		processor.processBasic(path.newSubPath("floatingRateDayCountConventionOfLeg1"), InterestComputationMethod4Code.class, getFloatingRateDayCountConventionOfLeg1(), this);
		processor.processBasic(path.newSubPath("fixedRateDayCountConventionLeg2"), InterestComputationMethod4Code.class, getFixedRateDayCountConventionLeg2(), this);
		processor.processBasic(path.newSubPath("floatingRateDayCountConventionOfLeg2"), InterestComputationMethod4Code.class, getFloatingRateDayCountConventionOfLeg2(), this);
		processor.processBasic(path.newSubPath("fixedRatePaymentFrequencyPeriodLeg1"), String.class, getFixedRatePaymentFrequencyPeriodLeg1(), this);
		processor.processBasic(path.newSubPath("fixedRatePaymentFrequencyPeriodLeg2"), String.class, getFixedRatePaymentFrequencyPeriodLeg2(), this);
		processor.processBasic(path.newSubPath("floatingRatePaymentFrequencyPeriodLeg1"), String.class, getFloatingRatePaymentFrequencyPeriodLeg1(), this);
		processor.processBasic(path.newSubPath("floatingRatePaymentFrequencyPeriodLeg2"), String.class, getFloatingRatePaymentFrequencyPeriodLeg2(), this);
		processor.processBasic(path.newSubPath("fixedRatePaymentFrequencyPeriodMultiplierLeg1"), Integer.class, getFixedRatePaymentFrequencyPeriodMultiplierLeg1(), this);
		processor.processBasic(path.newSubPath("fixedRatePaymentFrequencyPeriodMultiplierLeg2"), Integer.class, getFixedRatePaymentFrequencyPeriodMultiplierLeg2(), this);
		processor.processBasic(path.newSubPath("floatingRatePaymentFrequencyPeriodMultiplierLeg1"), Integer.class, getFloatingRatePaymentFrequencyPeriodMultiplierLeg1(), this);
		processor.processBasic(path.newSubPath("floatingRatePaymentFrequencyPeriodMultiplierLeg2"), Integer.class, getFloatingRatePaymentFrequencyPeriodMultiplierLeg2(), this);
		processor.processBasic(path.newSubPath("fixingDateLeg1"), ZonedDateTime.class, getFixingDateLeg1(), this);
		processor.processBasic(path.newSubPath("fixingDateLeg2"), ZonedDateTime.class, getFixingDateLeg2(), this);
		processor.processBasic(path.newSubPath("floatingRateResetFrequencyPeriodLeg1"), String.class, getFloatingRateResetFrequencyPeriodLeg1(), this);
		processor.processBasic(path.newSubPath("floatingRateResetFrequencyPeriodOfLeg2"), String.class, getFloatingRateResetFrequencyPeriodOfLeg2(), this);
		processor.processBasic(path.newSubPath("floatingRateResetFrequencyPeriodMultiplierLeg1"), Integer.class, getFloatingRateResetFrequencyPeriodMultiplierLeg1(), this);
		processor.processBasic(path.newSubPath("floatingRateResetFrequencyPeriodMultiplierLeg2"), Integer.class, getFloatingRateResetFrequencyPeriodMultiplierLeg2(), this);
		processRosetta(path.newSubPath("otherPayment"), processor, OtherPaymentReport.class, getOtherPayment());
		processor.processBasic(path.newSubPath("valuationAmount"), BigDecimal.class, getValuationAmount(), this);
		processor.processBasic(path.newSubPath("valuationCurrency"), ISOCurrencyCodeEnum.class, getValuationCurrency(), this);
		processor.processBasic(path.newSubPath("valuationMethod"), ValuationType1Code.class, getValuationMethod(), this);
		processor.processBasic(path.newSubPath("valuationTimestamp"), ZonedDateTime.class, getValuationTimestamp(), this);
		processor.processBasic(path.newSubPath("collateralPortfolioIndicator"), Boolean.class, getCollateralPortfolioIndicator(), this);
		processor.processBasic(path.newSubPath("initialMarginCollateralPortfolioCode"), String.class, getInitialMarginCollateralPortfolioCode(), this);
		processor.processBasic(path.newSubPath("variationMarginCollateralPortfolioCode"), String.class, getVariationMarginCollateralPortfolioCode(), this);
		processRosetta(path.newSubPath("price"), processor, PriceAndNotation.class, getPrice());
		processor.processBasic(path.newSubPath("priceNotation"), String.class, getPriceNotation(), this);
		processor.processBasic(path.newSubPath("priceUnitOfMeasure"), String.class, getPriceUnitOfMeasure(), this);
		processRosetta(path.newSubPath("priceSchedule"), processor, PriceScheduleReport.class, getPriceSchedule());
		processor.processBasic(path.newSubPath("fixedRateLeg1"), BigDecimal.class, getFixedRateLeg1(), this);
		processor.processBasic(path.newSubPath("fixedRateLeg2"), BigDecimal.class, getFixedRateLeg2(), this);
		processRosetta(path.newSubPath("spreadOfLeg1"), processor, SpreadOfLeg1.class, getSpreadOfLeg1());
		processor.processBasic(path.newSubPath("spreadNotationOfLeg1"), String.class, getSpreadNotationOfLeg1(), this);
		processRosetta(path.newSubPath("spreadOfLeg2"), processor, SpreadOfLeg2.class, getSpreadOfLeg2());
		processor.processBasic(path.newSubPath("spreadNotationOfLeg2"), String.class, getSpreadNotationOfLeg2(), this);
		processRosetta(path.newSubPath("strikePrice"), processor, StrikePriceAndNotation.class, getStrikePrice());
		processor.processBasic(path.newSubPath("strikePriceNotation"), String.class, getStrikePriceNotation(), this);
		processRosetta(path.newSubPath("strikePriceSchedule"), processor, StrikePriceScheduleReport.class, getStrikePriceSchedule());
		processor.processBasic(path.newSubPath("optionPremiumAmount"), BigDecimal.class, getOptionPremiumAmount(), this);
		processor.processBasic(path.newSubPath("optionPremiumCurrency"), ISOCurrencyCodeEnum.class, getOptionPremiumCurrency(), this);
		processor.processBasic(path.newSubPath("optionPremiumPaymentDate"), Date.class, getOptionPremiumPaymentDate(), this);
		processor.processBasic(path.newSubPath("firstExerciseDate"), Date.class, getFirstExerciseDate(), this);
		processor.processBasic(path.newSubPath("exchangeRate"), BigDecimal.class, getExchangeRate(), this);
		processor.processBasic(path.newSubPath("exchangeRateBasis"), String.class, getExchangeRateBasis(), this);
		processor.processBasic(path.newSubPath("notionalLeg1"), BigDecimal.class, getNotionalLeg1(), this);
		processor.processBasic(path.newSubPath("notionalLeg2"), BigDecimal.class, getNotionalLeg2(), this);
		processor.processBasic(path.newSubPath("callAmount"), BigDecimal.class, getCallAmount(), this);
		processor.processBasic(path.newSubPath("putAmount"), BigDecimal.class, getPutAmount(), this);
		processor.processBasic(path.newSubPath("notionalCurrencyOfLeg1"), ISOCurrencyCodeEnum.class, getNotionalCurrencyOfLeg1(), this);
		processor.processBasic(path.newSubPath("notionalCurrencyOfLeg2"), ISOCurrencyCodeEnum.class, getNotionalCurrencyOfLeg2(), this);
		processor.processBasic(path.newSubPath("callCurrency"), ISOCurrencyCodeEnum.class, getCallCurrency(), this);
		processor.processBasic(path.newSubPath("putCurrency"), ISOCurrencyCodeEnum.class, getPutCurrency(), this);
		processor.processBasic(path.newSubPath("quantityUnitOfMeasureLeg1"), String.class, getQuantityUnitOfMeasureLeg1(), this);
		processor.processBasic(path.newSubPath("quantityUnitOfMeasureLeg2"), String.class, getQuantityUnitOfMeasureLeg2(), this);
		processRosetta(path.newSubPath("notionalAmountScheduleLeg1"), processor, NotionalAmountScheduleLeg1Report.class, getNotionalAmountScheduleLeg1());
		processRosetta(path.newSubPath("notionalAmountScheduleLeg2"), processor, NotionalAmountScheduleLeg2Report.class, getNotionalAmountScheduleLeg2());
		processor.processBasic(path.newSubPath("totalNotionalQuantityLeg1"), BigDecimal.class, getTotalNotionalQuantityLeg1(), this);
		processor.processBasic(path.newSubPath("totalNotionalQuantityLeg2"), BigDecimal.class, getTotalNotionalQuantityLeg2(), this);
		processRosetta(path.newSubPath("notionalQuantityScheduleLeg1"), processor, NotionalQuantityScheduleLeg1Report.class, getNotionalQuantityScheduleLeg1());
		processRosetta(path.newSubPath("notionalQuantityScheduleLeg2"), processor, NotionalQuantityScheduleLeg2Report.class, getNotionalQuantityScheduleLeg2());
		processor.processBasic(path.newSubPath("actionType"), String.class, getActionType(), this);
		processor.processBasic(path.newSubPath("eventType"), String.class, getEventType(), this);
		processor.processBasic(path.newSubPath("eventTimestamp"), ZonedDateTime.class, getEventTimestamp(), this);
		processor.processBasic(path.newSubPath("eventIdentifier"), String.class, getEventIdentifier(), this);
		processor.processBasic(path.newSubPath("indexFactor"), BigDecimal.class, getIndexFactor(), this);
		processor.processBasic(path.newSubPath("embeddedOptionType"), String.class, getEmbeddedOptionType(), this);
		processor.processBasic(path.newSubPath("upi"), String.class, getUpi(), this);
		processor.processBasic(path.newSubPath("deliveryType"), String.class, getDeliveryType(), this);
		processor.processBasic(path.newSubPath("assetClass"), String.class, getAssetClass(), this);
		processor.processBasic(path.newSubPath("underlyingIdentificationType"), String.class, getUnderlyingIdentificationType(), this);
		processor.processBasic(path.newSubPath("underlyingIdentification"), String.class, getUnderlyingIdentification(), this);
		processor.processBasic(path.newSubPath("underlyingIndexIndicator"), String.class, getUnderlyingIndexIndicator(), this);
		processor.processBasic(path.newSubPath("nameOfTheUnderlyingIndex"), String.class, getNameOfTheUnderlyingIndex(), this);
		processor.processBasic(path.newSubPath("referenceEntity"), String.class, getReferenceEntity(), this);
		processor.processBasic(path.newSubPath("indicatorOfTheFloatingRateOfLeg1"), String.class, getIndicatorOfTheFloatingRateOfLeg1(), this);
		processor.processBasic(path.newSubPath("indicatorOfTheFloatingRateOfLeg2"), String.class, getIndicatorOfTheFloatingRateOfLeg2(), this);
		processor.processBasic(path.newSubPath("nameOfTheFloatingRateOfLeg1"), String.class, getNameOfTheFloatingRateOfLeg1(), this);
		processor.processBasic(path.newSubPath("nameOfTheFloatingRateOfLeg2"), String.class, getNameOfTheFloatingRateOfLeg2(), this);
		processor.processBasic(path.newSubPath("floatingRateReferencePeriodOfLeg1TimePeriod"), String.class, getFloatingRateReferencePeriodOfLeg1TimePeriod(), this);
		processor.processBasic(path.newSubPath("floatingRateReferencePeriodOfLeg2TimePeriod"), String.class, getFloatingRateReferencePeriodOfLeg2TimePeriod(), this);
		processor.processBasic(path.newSubPath("floatingRateReferencePeriodOfLeg1Multiplier"), Integer.class, getFloatingRateReferencePeriodOfLeg1Multiplier(), this);
		processor.processBasic(path.newSubPath("floatingRateReferencePeriodOfLeg2Multiplier"), Integer.class, getFloatingRateReferencePeriodOfLeg2Multiplier(), this);
		processor.processBasic(path.newSubPath("isCrypto"), Boolean.class, getIsCrypto(), this);
		processor.processBasic(path.newSubPath("maturityDateOfTheUnderlying"), Date.class, getMaturityDateOfTheUnderlying(), this);
		processor.processBasic(path.newSubPath("seniority"), String.class, getSeniority(), this);
		processor.processBasic(path.newSubPath("series"), Integer.class, getSeries(), this);
		processor.processBasic(path.newSubPath("version"), Integer.class, getVersion(), this);
		processor.processBasic(path.newSubPath("cdSIndexAttachmentPoint"), BigDecimal.class, getCdSIndexAttachmentPoint(), this);
		processor.processBasic(path.newSubPath("cdSIndexDetachmentPoint"), BigDecimal.class, getCdSIndexDetachmentPoint(), this);
		processor.processBasic(path.newSubPath("customBasketCode"), String.class, getCustomBasketCode(), this);
		processor.processBasic(path.newSubPath("identifierOfBasketConstituents"), String.class, getIdentifierOfBasketConstituents(), this);
		processor.processBasic(path.newSubPath("basketConstituentIdentifierSource"), ProductIdTypeEnum.class, getBasketConstituentIdentifierSource(), this);
		processor.processBasic(path.newSubPath("contractType"), String.class, getContractType(), this);
		processor.processBasic(path.newSubPath("optionStyle"), String.class, getOptionStyle(), this);
		processor.processBasic(path.newSubPath("optionType"), String.class, getOptionType(), this);
		processor.processBasic(path.newSubPath("packageIdentifier"), String.class, getPackageIdentifier(), this);
		processRosetta(path.newSubPath("packageTransactionPrice"), processor, PackageTransactionPrice.class, getPackageTransactionPrice());
		processor.processBasic(path.newSubPath("packageTransactionPriceCurrency"), ISOCurrencyCodeEnum.class, getPackageTransactionPriceCurrency(), this);
		processor.processBasic(path.newSubPath("packageTransactionPriceNotation"), String.class, getPackageTransactionPriceNotation(), this);
		processRosetta(path.newSubPath("packageTransactionSpread"), processor, PackageTransactionSpread.class, getPackageTransactionSpread());
		processor.processBasic(path.newSubPath("packageTransactionSpreadCurrency"), ISOCurrencyCodeEnum.class, getPackageTransactionSpreadCurrency(), this);
		processor.processBasic(path.newSubPath("packageTransactionSpreadNotation"), String.class, getPackageTransactionSpreadNotation(), this);
		processor.processBasic(path.newSubPath("technicalRecordId"), String.class, getTechnicalRecordId(), this);
		processRosetta(path.newSubPath("nonReportable"), processor, JFSANonReportable.class, getNonReportable());
	}
	

	/*********************** Builder Interface  ***********************/
	interface JFSATransactionReportBuilder extends JFSATransactionReport, RosettaModelObjectBuilder {
		OtherPaymentReport.OtherPaymentReportBuilder getOrCreateOtherPayment(int _index);
		List<? extends OtherPaymentReport.OtherPaymentReportBuilder> getOtherPayment();
		PriceAndNotation.PriceAndNotationBuilder getOrCreatePrice();
		PriceAndNotation.PriceAndNotationBuilder getPrice();
		PriceScheduleReport.PriceScheduleReportBuilder getOrCreatePriceSchedule(int _index);
		List<? extends PriceScheduleReport.PriceScheduleReportBuilder> getPriceSchedule();
		SpreadOfLeg1.SpreadOfLeg1Builder getOrCreateSpreadOfLeg1();
		SpreadOfLeg1.SpreadOfLeg1Builder getSpreadOfLeg1();
		SpreadOfLeg2.SpreadOfLeg2Builder getOrCreateSpreadOfLeg2();
		SpreadOfLeg2.SpreadOfLeg2Builder getSpreadOfLeg2();
		StrikePriceAndNotation.StrikePriceAndNotationBuilder getOrCreateStrikePrice();
		StrikePriceAndNotation.StrikePriceAndNotationBuilder getStrikePrice();
		StrikePriceScheduleReport.StrikePriceScheduleReportBuilder getOrCreateStrikePriceSchedule(int _index);
		List<? extends StrikePriceScheduleReport.StrikePriceScheduleReportBuilder> getStrikePriceSchedule();
		NotionalAmountScheduleLeg1Report.NotionalAmountScheduleLeg1ReportBuilder getOrCreateNotionalAmountScheduleLeg1(int _index);
		List<? extends NotionalAmountScheduleLeg1Report.NotionalAmountScheduleLeg1ReportBuilder> getNotionalAmountScheduleLeg1();
		NotionalAmountScheduleLeg2Report.NotionalAmountScheduleLeg2ReportBuilder getOrCreateNotionalAmountScheduleLeg2(int _index);
		List<? extends NotionalAmountScheduleLeg2Report.NotionalAmountScheduleLeg2ReportBuilder> getNotionalAmountScheduleLeg2();
		NotionalQuantityScheduleLeg1Report.NotionalQuantityScheduleLeg1ReportBuilder getOrCreateNotionalQuantityScheduleLeg1(int _index);
		List<? extends NotionalQuantityScheduleLeg1Report.NotionalQuantityScheduleLeg1ReportBuilder> getNotionalQuantityScheduleLeg1();
		NotionalQuantityScheduleLeg2Report.NotionalQuantityScheduleLeg2ReportBuilder getOrCreateNotionalQuantityScheduleLeg2(int _index);
		List<? extends NotionalQuantityScheduleLeg2Report.NotionalQuantityScheduleLeg2ReportBuilder> getNotionalQuantityScheduleLeg2();
		PackageTransactionPrice.PackageTransactionPriceBuilder getOrCreatePackageTransactionPrice();
		PackageTransactionPrice.PackageTransactionPriceBuilder getPackageTransactionPrice();
		PackageTransactionSpread.PackageTransactionSpreadBuilder getOrCreatePackageTransactionSpread();
		PackageTransactionSpread.PackageTransactionSpreadBuilder getPackageTransactionSpread();
		JFSANonReportable.JFSANonReportableBuilder getOrCreateNonReportable();
		JFSANonReportable.JFSANonReportableBuilder getNonReportable();
		JFSATransactionReport.JFSATransactionReportBuilder setEffectiveDate(Date effectiveDate);
		JFSATransactionReport.JFSATransactionReportBuilder setExpirationDate(Date expirationDate);
		JFSATransactionReport.JFSATransactionReportBuilder setEarlyTerminationDate(Date earlyTerminationDate);
		JFSATransactionReport.JFSATransactionReportBuilder setReportingTimestamp(ZonedDateTime reportingTimestamp);
		JFSATransactionReport.JFSATransactionReportBuilder setExecutionTimestamp(ZonedDateTime executionTimestamp);
		JFSATransactionReport.JFSATransactionReportBuilder setEntityResponsibleForReporting(String entityResponsibleForReporting);
		JFSATransactionReport.JFSATransactionReportBuilder setCounterparty1(String counterparty1);
		JFSATransactionReport.JFSATransactionReportBuilder setCounterparty2(String counterparty2);
		JFSATransactionReport.JFSATransactionReportBuilder setCounterparty2IdentifierType(Boolean counterparty2IdentifierType);
		JFSATransactionReport.JFSATransactionReportBuilder setDirection1(String direction1);
		JFSATransactionReport.JFSATransactionReportBuilder setDirection2Leg1(String direction2Leg1);
		JFSATransactionReport.JFSATransactionReportBuilder setDirection2Leg2(String direction2Leg2);
		JFSATransactionReport.JFSATransactionReportBuilder setSubmitterIdentifier(String submitterIdentifier);
		JFSATransactionReport.JFSATransactionReportBuilder setNewSDRIdentifier(String newSDRIdentifier);
		JFSATransactionReport.JFSATransactionReportBuilder setOriginalSwapSDRIdentifier(String originalSwapSDRIdentifier);
		JFSATransactionReport.JFSATransactionReportBuilder setCleared(String cleared);
		JFSATransactionReport.JFSATransactionReportBuilder setCentralCounterparty(String centralCounterparty);
		JFSATransactionReport.JFSATransactionReportBuilder setClearingMember(String clearingMember);
		JFSATransactionReport.JFSATransactionReportBuilder setPlatformIdentifier(String platformIdentifier);
		JFSATransactionReport.JFSATransactionReportBuilder setConfirmed(String confirmed);
		JFSATransactionReport.JFSATransactionReportBuilder setFinalContractualSettlementDate(Date finalContractualSettlementDate);
		JFSATransactionReport.JFSATransactionReportBuilder setConfirmationTimestamp(ZonedDateTime confirmationTimestamp);
		JFSATransactionReport.JFSATransactionReportBuilder setSettlementCurrencyLeg1(ISOCurrencyCodeEnum settlementCurrencyLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setSettlementCurrencyLeg2(ISOCurrencyCodeEnum settlementCurrencyLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setClearingAccountOrigin(String clearingAccountOrigin);
		JFSATransactionReport.JFSATransactionReportBuilder setOriginalSwapUTI(String originalSwapUTI);
		JFSATransactionReport.JFSATransactionReportBuilder setClearingReceiptTimestamp(ZonedDateTime clearingReceiptTimestamp);
		JFSATransactionReport.JFSATransactionReportBuilder setUti(String uti);
		JFSATransactionReport.JFSATransactionReportBuilder setUtiProprietary(String utiProprietary);
		JFSATransactionReport.JFSATransactionReportBuilder setPriorUti(String priorUti);
		JFSATransactionReport.JFSATransactionReportBuilder setPriorUtiProprietary(String priorUtiProprietary);
		JFSATransactionReport.JFSATransactionReportBuilder setFixedRateDayCountConventionLeg1(InterestComputationMethod4Code fixedRateDayCountConventionLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateDayCountConventionOfLeg1(InterestComputationMethod4Code floatingRateDayCountConventionOfLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFixedRateDayCountConventionLeg2(InterestComputationMethod4Code fixedRateDayCountConventionLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateDayCountConventionOfLeg2(InterestComputationMethod4Code floatingRateDayCountConventionOfLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setFixedRatePaymentFrequencyPeriodLeg1(String fixedRatePaymentFrequencyPeriodLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFixedRatePaymentFrequencyPeriodLeg2(String fixedRatePaymentFrequencyPeriodLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRatePaymentFrequencyPeriodLeg1(String floatingRatePaymentFrequencyPeriodLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRatePaymentFrequencyPeriodLeg2(String floatingRatePaymentFrequencyPeriodLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setFixedRatePaymentFrequencyPeriodMultiplierLeg1(Integer fixedRatePaymentFrequencyPeriodMultiplierLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFixedRatePaymentFrequencyPeriodMultiplierLeg2(Integer fixedRatePaymentFrequencyPeriodMultiplierLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRatePaymentFrequencyPeriodMultiplierLeg1(Integer floatingRatePaymentFrequencyPeriodMultiplierLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRatePaymentFrequencyPeriodMultiplierLeg2(Integer floatingRatePaymentFrequencyPeriodMultiplierLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setFixingDateLeg1(ZonedDateTime fixingDateLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFixingDateLeg2(ZonedDateTime fixingDateLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateResetFrequencyPeriodLeg1(String floatingRateResetFrequencyPeriodLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateResetFrequencyPeriodOfLeg2(String floatingRateResetFrequencyPeriodOfLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateResetFrequencyPeriodMultiplierLeg1(Integer floatingRateResetFrequencyPeriodMultiplierLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateResetFrequencyPeriodMultiplierLeg2(Integer floatingRateResetFrequencyPeriodMultiplierLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder addOtherPayment(OtherPaymentReport otherPayment0);
		JFSATransactionReport.JFSATransactionReportBuilder addOtherPayment(OtherPaymentReport otherPayment1, int _idx);
		JFSATransactionReport.JFSATransactionReportBuilder addOtherPayment(List<? extends OtherPaymentReport> otherPayment2);
		JFSATransactionReport.JFSATransactionReportBuilder setOtherPayment(List<? extends OtherPaymentReport> otherPayment3);
		JFSATransactionReport.JFSATransactionReportBuilder setValuationAmount(BigDecimal valuationAmount);
		JFSATransactionReport.JFSATransactionReportBuilder setValuationCurrency(ISOCurrencyCodeEnum valuationCurrency);
		JFSATransactionReport.JFSATransactionReportBuilder setValuationMethod(ValuationType1Code valuationMethod);
		JFSATransactionReport.JFSATransactionReportBuilder setValuationTimestamp(ZonedDateTime valuationTimestamp);
		JFSATransactionReport.JFSATransactionReportBuilder setCollateralPortfolioIndicator(Boolean collateralPortfolioIndicator);
		JFSATransactionReport.JFSATransactionReportBuilder setInitialMarginCollateralPortfolioCode(String initialMarginCollateralPortfolioCode);
		JFSATransactionReport.JFSATransactionReportBuilder setVariationMarginCollateralPortfolioCode(String variationMarginCollateralPortfolioCode);
		JFSATransactionReport.JFSATransactionReportBuilder setPrice(PriceAndNotation price);
		JFSATransactionReport.JFSATransactionReportBuilder setPriceNotation(String priceNotation);
		JFSATransactionReport.JFSATransactionReportBuilder setPriceUnitOfMeasure(String priceUnitOfMeasure);
		JFSATransactionReport.JFSATransactionReportBuilder addPriceSchedule(PriceScheduleReport priceSchedule0);
		JFSATransactionReport.JFSATransactionReportBuilder addPriceSchedule(PriceScheduleReport priceSchedule1, int _idx);
		JFSATransactionReport.JFSATransactionReportBuilder addPriceSchedule(List<? extends PriceScheduleReport> priceSchedule2);
		JFSATransactionReport.JFSATransactionReportBuilder setPriceSchedule(List<? extends PriceScheduleReport> priceSchedule3);
		JFSATransactionReport.JFSATransactionReportBuilder setFixedRateLeg1(BigDecimal fixedRateLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setFixedRateLeg2(BigDecimal fixedRateLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setSpreadOfLeg1(SpreadOfLeg1 spreadOfLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setSpreadNotationOfLeg1(String spreadNotationOfLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setSpreadOfLeg2(SpreadOfLeg2 spreadOfLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setSpreadNotationOfLeg2(String spreadNotationOfLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setStrikePrice(StrikePriceAndNotation strikePrice);
		JFSATransactionReport.JFSATransactionReportBuilder setStrikePriceNotation(String strikePriceNotation);
		JFSATransactionReport.JFSATransactionReportBuilder addStrikePriceSchedule(StrikePriceScheduleReport strikePriceSchedule0);
		JFSATransactionReport.JFSATransactionReportBuilder addStrikePriceSchedule(StrikePriceScheduleReport strikePriceSchedule1, int _idx);
		JFSATransactionReport.JFSATransactionReportBuilder addStrikePriceSchedule(List<? extends StrikePriceScheduleReport> strikePriceSchedule2);
		JFSATransactionReport.JFSATransactionReportBuilder setStrikePriceSchedule(List<? extends StrikePriceScheduleReport> strikePriceSchedule3);
		JFSATransactionReport.JFSATransactionReportBuilder setOptionPremiumAmount(BigDecimal optionPremiumAmount);
		JFSATransactionReport.JFSATransactionReportBuilder setOptionPremiumCurrency(ISOCurrencyCodeEnum optionPremiumCurrency);
		JFSATransactionReport.JFSATransactionReportBuilder setOptionPremiumPaymentDate(Date optionPremiumPaymentDate);
		JFSATransactionReport.JFSATransactionReportBuilder setFirstExerciseDate(Date firstExerciseDate);
		JFSATransactionReport.JFSATransactionReportBuilder setExchangeRate(BigDecimal exchangeRate);
		JFSATransactionReport.JFSATransactionReportBuilder setExchangeRateBasis(String exchangeRateBasis);
		JFSATransactionReport.JFSATransactionReportBuilder setNotionalLeg1(BigDecimal notionalLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setNotionalLeg2(BigDecimal notionalLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setCallAmount(BigDecimal callAmount);
		JFSATransactionReport.JFSATransactionReportBuilder setPutAmount(BigDecimal putAmount);
		JFSATransactionReport.JFSATransactionReportBuilder setNotionalCurrencyOfLeg1(ISOCurrencyCodeEnum notionalCurrencyOfLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setNotionalCurrencyOfLeg2(ISOCurrencyCodeEnum notionalCurrencyOfLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setCallCurrency(ISOCurrencyCodeEnum callCurrency);
		JFSATransactionReport.JFSATransactionReportBuilder setPutCurrency(ISOCurrencyCodeEnum putCurrency);
		JFSATransactionReport.JFSATransactionReportBuilder setQuantityUnitOfMeasureLeg1(String quantityUnitOfMeasureLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setQuantityUnitOfMeasureLeg2(String quantityUnitOfMeasureLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg1(NotionalAmountScheduleLeg1Report notionalAmountScheduleLeg10);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg1(NotionalAmountScheduleLeg1Report notionalAmountScheduleLeg11, int _idx);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg1(List<? extends NotionalAmountScheduleLeg1Report> notionalAmountScheduleLeg12);
		JFSATransactionReport.JFSATransactionReportBuilder setNotionalAmountScheduleLeg1(List<? extends NotionalAmountScheduleLeg1Report> notionalAmountScheduleLeg13);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg2(NotionalAmountScheduleLeg2Report notionalAmountScheduleLeg20);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg2(NotionalAmountScheduleLeg2Report notionalAmountScheduleLeg21, int _idx);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg2(List<? extends NotionalAmountScheduleLeg2Report> notionalAmountScheduleLeg22);
		JFSATransactionReport.JFSATransactionReportBuilder setNotionalAmountScheduleLeg2(List<? extends NotionalAmountScheduleLeg2Report> notionalAmountScheduleLeg23);
		JFSATransactionReport.JFSATransactionReportBuilder setTotalNotionalQuantityLeg1(BigDecimal totalNotionalQuantityLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setTotalNotionalQuantityLeg2(BigDecimal totalNotionalQuantityLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg1(NotionalQuantityScheduleLeg1Report notionalQuantityScheduleLeg10);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg1(NotionalQuantityScheduleLeg1Report notionalQuantityScheduleLeg11, int _idx);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg1(List<? extends NotionalQuantityScheduleLeg1Report> notionalQuantityScheduleLeg12);
		JFSATransactionReport.JFSATransactionReportBuilder setNotionalQuantityScheduleLeg1(List<? extends NotionalQuantityScheduleLeg1Report> notionalQuantityScheduleLeg13);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg2(NotionalQuantityScheduleLeg2Report notionalQuantityScheduleLeg20);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg2(NotionalQuantityScheduleLeg2Report notionalQuantityScheduleLeg21, int _idx);
		JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg2(List<? extends NotionalQuantityScheduleLeg2Report> notionalQuantityScheduleLeg22);
		JFSATransactionReport.JFSATransactionReportBuilder setNotionalQuantityScheduleLeg2(List<? extends NotionalQuantityScheduleLeg2Report> notionalQuantityScheduleLeg23);
		JFSATransactionReport.JFSATransactionReportBuilder setActionType(String actionType);
		JFSATransactionReport.JFSATransactionReportBuilder setEventType(String eventType);
		JFSATransactionReport.JFSATransactionReportBuilder setEventTimestamp(ZonedDateTime eventTimestamp);
		JFSATransactionReport.JFSATransactionReportBuilder setEventIdentifier(String eventIdentifier);
		JFSATransactionReport.JFSATransactionReportBuilder setIndexFactor(BigDecimal indexFactor);
		JFSATransactionReport.JFSATransactionReportBuilder setEmbeddedOptionType(String embeddedOptionType);
		JFSATransactionReport.JFSATransactionReportBuilder setUpi(String upi);
		JFSATransactionReport.JFSATransactionReportBuilder setDeliveryType(String deliveryType);
		JFSATransactionReport.JFSATransactionReportBuilder setAssetClass(String assetClass);
		JFSATransactionReport.JFSATransactionReportBuilder setUnderlyingIdentificationType(String underlyingIdentificationType);
		JFSATransactionReport.JFSATransactionReportBuilder setUnderlyingIdentification(String underlyingIdentification);
		JFSATransactionReport.JFSATransactionReportBuilder setUnderlyingIndexIndicator(String underlyingIndexIndicator);
		JFSATransactionReport.JFSATransactionReportBuilder setNameOfTheUnderlyingIndex(String nameOfTheUnderlyingIndex);
		JFSATransactionReport.JFSATransactionReportBuilder setReferenceEntity(String referenceEntity);
		JFSATransactionReport.JFSATransactionReportBuilder setIndicatorOfTheFloatingRateOfLeg1(String indicatorOfTheFloatingRateOfLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setIndicatorOfTheFloatingRateOfLeg2(String indicatorOfTheFloatingRateOfLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setNameOfTheFloatingRateOfLeg1(String nameOfTheFloatingRateOfLeg1);
		JFSATransactionReport.JFSATransactionReportBuilder setNameOfTheFloatingRateOfLeg2(String nameOfTheFloatingRateOfLeg2);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateReferencePeriodOfLeg1TimePeriod(String floatingRateReferencePeriodOfLeg1TimePeriod);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateReferencePeriodOfLeg2TimePeriod(String floatingRateReferencePeriodOfLeg2TimePeriod);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateReferencePeriodOfLeg1Multiplier(Integer floatingRateReferencePeriodOfLeg1Multiplier);
		JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateReferencePeriodOfLeg2Multiplier(Integer floatingRateReferencePeriodOfLeg2Multiplier);
		JFSATransactionReport.JFSATransactionReportBuilder setIsCrypto(Boolean isCrypto);
		JFSATransactionReport.JFSATransactionReportBuilder setMaturityDateOfTheUnderlying(Date maturityDateOfTheUnderlying);
		JFSATransactionReport.JFSATransactionReportBuilder setSeniority(String seniority);
		JFSATransactionReport.JFSATransactionReportBuilder setSeries(Integer series);
		JFSATransactionReport.JFSATransactionReportBuilder setVersion(Integer version);
		JFSATransactionReport.JFSATransactionReportBuilder setCdSIndexAttachmentPoint(BigDecimal cdSIndexAttachmentPoint);
		JFSATransactionReport.JFSATransactionReportBuilder setCdSIndexDetachmentPoint(BigDecimal cdSIndexDetachmentPoint);
		JFSATransactionReport.JFSATransactionReportBuilder setCustomBasketCode(String customBasketCode);
		JFSATransactionReport.JFSATransactionReportBuilder addIdentifierOfBasketConstituents(String identifierOfBasketConstituents0);
		JFSATransactionReport.JFSATransactionReportBuilder addIdentifierOfBasketConstituents(String identifierOfBasketConstituents1, int _idx);
		JFSATransactionReport.JFSATransactionReportBuilder addIdentifierOfBasketConstituents(List<? extends String> identifierOfBasketConstituents2);
		JFSATransactionReport.JFSATransactionReportBuilder setIdentifierOfBasketConstituents(List<? extends String> identifierOfBasketConstituents3);
		JFSATransactionReport.JFSATransactionReportBuilder addBasketConstituentIdentifierSource(ProductIdTypeEnum basketConstituentIdentifierSource0);
		JFSATransactionReport.JFSATransactionReportBuilder addBasketConstituentIdentifierSource(ProductIdTypeEnum basketConstituentIdentifierSource1, int _idx);
		JFSATransactionReport.JFSATransactionReportBuilder addBasketConstituentIdentifierSource(List<? extends ProductIdTypeEnum> basketConstituentIdentifierSource2);
		JFSATransactionReport.JFSATransactionReportBuilder setBasketConstituentIdentifierSource(List<? extends ProductIdTypeEnum> basketConstituentIdentifierSource3);
		JFSATransactionReport.JFSATransactionReportBuilder setContractType(String contractType);
		JFSATransactionReport.JFSATransactionReportBuilder setOptionStyle(String optionStyle);
		JFSATransactionReport.JFSATransactionReportBuilder setOptionType(String optionType);
		JFSATransactionReport.JFSATransactionReportBuilder setPackageIdentifier(String packageIdentifier);
		JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionPrice(PackageTransactionPrice packageTransactionPrice);
		JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionPriceCurrency(ISOCurrencyCodeEnum packageTransactionPriceCurrency);
		JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionPriceNotation(String packageTransactionPriceNotation);
		JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionSpread(PackageTransactionSpread packageTransactionSpread);
		JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionSpreadCurrency(ISOCurrencyCodeEnum packageTransactionSpreadCurrency);
		JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionSpreadNotation(String packageTransactionSpreadNotation);
		JFSATransactionReport.JFSATransactionReportBuilder setTechnicalRecordId(String technicalRecordId);
		JFSATransactionReport.JFSATransactionReportBuilder setNonReportable(JFSANonReportable nonReportable);

		@Override
		default void process(RosettaPath path, BuilderProcessor processor) {
			processor.processBasic(path.newSubPath("effectiveDate"), Date.class, getEffectiveDate(), this);
			processor.processBasic(path.newSubPath("expirationDate"), Date.class, getExpirationDate(), this);
			processor.processBasic(path.newSubPath("earlyTerminationDate"), Date.class, getEarlyTerminationDate(), this);
			processor.processBasic(path.newSubPath("reportingTimestamp"), ZonedDateTime.class, getReportingTimestamp(), this);
			processor.processBasic(path.newSubPath("executionTimestamp"), ZonedDateTime.class, getExecutionTimestamp(), this);
			processor.processBasic(path.newSubPath("entityResponsibleForReporting"), String.class, getEntityResponsibleForReporting(), this);
			processor.processBasic(path.newSubPath("counterparty1"), String.class, getCounterparty1(), this);
			processor.processBasic(path.newSubPath("counterparty2"), String.class, getCounterparty2(), this);
			processor.processBasic(path.newSubPath("counterparty2IdentifierType"), Boolean.class, getCounterparty2IdentifierType(), this);
			processor.processBasic(path.newSubPath("direction1"), String.class, getDirection1(), this);
			processor.processBasic(path.newSubPath("direction2Leg1"), String.class, getDirection2Leg1(), this);
			processor.processBasic(path.newSubPath("direction2Leg2"), String.class, getDirection2Leg2(), this);
			processor.processBasic(path.newSubPath("submitterIdentifier"), String.class, getSubmitterIdentifier(), this);
			processor.processBasic(path.newSubPath("newSDRIdentifier"), String.class, getNewSDRIdentifier(), this);
			processor.processBasic(path.newSubPath("originalSwapSDRIdentifier"), String.class, getOriginalSwapSDRIdentifier(), this);
			processor.processBasic(path.newSubPath("cleared"), String.class, getCleared(), this);
			processor.processBasic(path.newSubPath("centralCounterparty"), String.class, getCentralCounterparty(), this);
			processor.processBasic(path.newSubPath("clearingMember"), String.class, getClearingMember(), this);
			processor.processBasic(path.newSubPath("platformIdentifier"), String.class, getPlatformIdentifier(), this);
			processor.processBasic(path.newSubPath("confirmed"), String.class, getConfirmed(), this);
			processor.processBasic(path.newSubPath("finalContractualSettlementDate"), Date.class, getFinalContractualSettlementDate(), this);
			processor.processBasic(path.newSubPath("confirmationTimestamp"), ZonedDateTime.class, getConfirmationTimestamp(), this);
			processor.processBasic(path.newSubPath("settlementCurrencyLeg1"), ISOCurrencyCodeEnum.class, getSettlementCurrencyLeg1(), this);
			processor.processBasic(path.newSubPath("settlementCurrencyLeg2"), ISOCurrencyCodeEnum.class, getSettlementCurrencyLeg2(), this);
			processor.processBasic(path.newSubPath("clearingAccountOrigin"), String.class, getClearingAccountOrigin(), this);
			processor.processBasic(path.newSubPath("originalSwapUTI"), String.class, getOriginalSwapUTI(), this);
			processor.processBasic(path.newSubPath("clearingReceiptTimestamp"), ZonedDateTime.class, getClearingReceiptTimestamp(), this);
			processor.processBasic(path.newSubPath("uti"), String.class, getUti(), this);
			processor.processBasic(path.newSubPath("utiProprietary"), String.class, getUtiProprietary(), this);
			processor.processBasic(path.newSubPath("priorUti"), String.class, getPriorUti(), this);
			processor.processBasic(path.newSubPath("priorUtiProprietary"), String.class, getPriorUtiProprietary(), this);
			processor.processBasic(path.newSubPath("fixedRateDayCountConventionLeg1"), InterestComputationMethod4Code.class, getFixedRateDayCountConventionLeg1(), this);
			processor.processBasic(path.newSubPath("floatingRateDayCountConventionOfLeg1"), InterestComputationMethod4Code.class, getFloatingRateDayCountConventionOfLeg1(), this);
			processor.processBasic(path.newSubPath("fixedRateDayCountConventionLeg2"), InterestComputationMethod4Code.class, getFixedRateDayCountConventionLeg2(), this);
			processor.processBasic(path.newSubPath("floatingRateDayCountConventionOfLeg2"), InterestComputationMethod4Code.class, getFloatingRateDayCountConventionOfLeg2(), this);
			processor.processBasic(path.newSubPath("fixedRatePaymentFrequencyPeriodLeg1"), String.class, getFixedRatePaymentFrequencyPeriodLeg1(), this);
			processor.processBasic(path.newSubPath("fixedRatePaymentFrequencyPeriodLeg2"), String.class, getFixedRatePaymentFrequencyPeriodLeg2(), this);
			processor.processBasic(path.newSubPath("floatingRatePaymentFrequencyPeriodLeg1"), String.class, getFloatingRatePaymentFrequencyPeriodLeg1(), this);
			processor.processBasic(path.newSubPath("floatingRatePaymentFrequencyPeriodLeg2"), String.class, getFloatingRatePaymentFrequencyPeriodLeg2(), this);
			processor.processBasic(path.newSubPath("fixedRatePaymentFrequencyPeriodMultiplierLeg1"), Integer.class, getFixedRatePaymentFrequencyPeriodMultiplierLeg1(), this);
			processor.processBasic(path.newSubPath("fixedRatePaymentFrequencyPeriodMultiplierLeg2"), Integer.class, getFixedRatePaymentFrequencyPeriodMultiplierLeg2(), this);
			processor.processBasic(path.newSubPath("floatingRatePaymentFrequencyPeriodMultiplierLeg1"), Integer.class, getFloatingRatePaymentFrequencyPeriodMultiplierLeg1(), this);
			processor.processBasic(path.newSubPath("floatingRatePaymentFrequencyPeriodMultiplierLeg2"), Integer.class, getFloatingRatePaymentFrequencyPeriodMultiplierLeg2(), this);
			processor.processBasic(path.newSubPath("fixingDateLeg1"), ZonedDateTime.class, getFixingDateLeg1(), this);
			processor.processBasic(path.newSubPath("fixingDateLeg2"), ZonedDateTime.class, getFixingDateLeg2(), this);
			processor.processBasic(path.newSubPath("floatingRateResetFrequencyPeriodLeg1"), String.class, getFloatingRateResetFrequencyPeriodLeg1(), this);
			processor.processBasic(path.newSubPath("floatingRateResetFrequencyPeriodOfLeg2"), String.class, getFloatingRateResetFrequencyPeriodOfLeg2(), this);
			processor.processBasic(path.newSubPath("floatingRateResetFrequencyPeriodMultiplierLeg1"), Integer.class, getFloatingRateResetFrequencyPeriodMultiplierLeg1(), this);
			processor.processBasic(path.newSubPath("floatingRateResetFrequencyPeriodMultiplierLeg2"), Integer.class, getFloatingRateResetFrequencyPeriodMultiplierLeg2(), this);
			processRosetta(path.newSubPath("otherPayment"), processor, OtherPaymentReport.OtherPaymentReportBuilder.class, getOtherPayment());
			processor.processBasic(path.newSubPath("valuationAmount"), BigDecimal.class, getValuationAmount(), this);
			processor.processBasic(path.newSubPath("valuationCurrency"), ISOCurrencyCodeEnum.class, getValuationCurrency(), this);
			processor.processBasic(path.newSubPath("valuationMethod"), ValuationType1Code.class, getValuationMethod(), this);
			processor.processBasic(path.newSubPath("valuationTimestamp"), ZonedDateTime.class, getValuationTimestamp(), this);
			processor.processBasic(path.newSubPath("collateralPortfolioIndicator"), Boolean.class, getCollateralPortfolioIndicator(), this);
			processor.processBasic(path.newSubPath("initialMarginCollateralPortfolioCode"), String.class, getInitialMarginCollateralPortfolioCode(), this);
			processor.processBasic(path.newSubPath("variationMarginCollateralPortfolioCode"), String.class, getVariationMarginCollateralPortfolioCode(), this);
			processRosetta(path.newSubPath("price"), processor, PriceAndNotation.PriceAndNotationBuilder.class, getPrice());
			processor.processBasic(path.newSubPath("priceNotation"), String.class, getPriceNotation(), this);
			processor.processBasic(path.newSubPath("priceUnitOfMeasure"), String.class, getPriceUnitOfMeasure(), this);
			processRosetta(path.newSubPath("priceSchedule"), processor, PriceScheduleReport.PriceScheduleReportBuilder.class, getPriceSchedule());
			processor.processBasic(path.newSubPath("fixedRateLeg1"), BigDecimal.class, getFixedRateLeg1(), this);
			processor.processBasic(path.newSubPath("fixedRateLeg2"), BigDecimal.class, getFixedRateLeg2(), this);
			processRosetta(path.newSubPath("spreadOfLeg1"), processor, SpreadOfLeg1.SpreadOfLeg1Builder.class, getSpreadOfLeg1());
			processor.processBasic(path.newSubPath("spreadNotationOfLeg1"), String.class, getSpreadNotationOfLeg1(), this);
			processRosetta(path.newSubPath("spreadOfLeg2"), processor, SpreadOfLeg2.SpreadOfLeg2Builder.class, getSpreadOfLeg2());
			processor.processBasic(path.newSubPath("spreadNotationOfLeg2"), String.class, getSpreadNotationOfLeg2(), this);
			processRosetta(path.newSubPath("strikePrice"), processor, StrikePriceAndNotation.StrikePriceAndNotationBuilder.class, getStrikePrice());
			processor.processBasic(path.newSubPath("strikePriceNotation"), String.class, getStrikePriceNotation(), this);
			processRosetta(path.newSubPath("strikePriceSchedule"), processor, StrikePriceScheduleReport.StrikePriceScheduleReportBuilder.class, getStrikePriceSchedule());
			processor.processBasic(path.newSubPath("optionPremiumAmount"), BigDecimal.class, getOptionPremiumAmount(), this);
			processor.processBasic(path.newSubPath("optionPremiumCurrency"), ISOCurrencyCodeEnum.class, getOptionPremiumCurrency(), this);
			processor.processBasic(path.newSubPath("optionPremiumPaymentDate"), Date.class, getOptionPremiumPaymentDate(), this);
			processor.processBasic(path.newSubPath("firstExerciseDate"), Date.class, getFirstExerciseDate(), this);
			processor.processBasic(path.newSubPath("exchangeRate"), BigDecimal.class, getExchangeRate(), this);
			processor.processBasic(path.newSubPath("exchangeRateBasis"), String.class, getExchangeRateBasis(), this);
			processor.processBasic(path.newSubPath("notionalLeg1"), BigDecimal.class, getNotionalLeg1(), this);
			processor.processBasic(path.newSubPath("notionalLeg2"), BigDecimal.class, getNotionalLeg2(), this);
			processor.processBasic(path.newSubPath("callAmount"), BigDecimal.class, getCallAmount(), this);
			processor.processBasic(path.newSubPath("putAmount"), BigDecimal.class, getPutAmount(), this);
			processor.processBasic(path.newSubPath("notionalCurrencyOfLeg1"), ISOCurrencyCodeEnum.class, getNotionalCurrencyOfLeg1(), this);
			processor.processBasic(path.newSubPath("notionalCurrencyOfLeg2"), ISOCurrencyCodeEnum.class, getNotionalCurrencyOfLeg2(), this);
			processor.processBasic(path.newSubPath("callCurrency"), ISOCurrencyCodeEnum.class, getCallCurrency(), this);
			processor.processBasic(path.newSubPath("putCurrency"), ISOCurrencyCodeEnum.class, getPutCurrency(), this);
			processor.processBasic(path.newSubPath("quantityUnitOfMeasureLeg1"), String.class, getQuantityUnitOfMeasureLeg1(), this);
			processor.processBasic(path.newSubPath("quantityUnitOfMeasureLeg2"), String.class, getQuantityUnitOfMeasureLeg2(), this);
			processRosetta(path.newSubPath("notionalAmountScheduleLeg1"), processor, NotionalAmountScheduleLeg1Report.NotionalAmountScheduleLeg1ReportBuilder.class, getNotionalAmountScheduleLeg1());
			processRosetta(path.newSubPath("notionalAmountScheduleLeg2"), processor, NotionalAmountScheduleLeg2Report.NotionalAmountScheduleLeg2ReportBuilder.class, getNotionalAmountScheduleLeg2());
			processor.processBasic(path.newSubPath("totalNotionalQuantityLeg1"), BigDecimal.class, getTotalNotionalQuantityLeg1(), this);
			processor.processBasic(path.newSubPath("totalNotionalQuantityLeg2"), BigDecimal.class, getTotalNotionalQuantityLeg2(), this);
			processRosetta(path.newSubPath("notionalQuantityScheduleLeg1"), processor, NotionalQuantityScheduleLeg1Report.NotionalQuantityScheduleLeg1ReportBuilder.class, getNotionalQuantityScheduleLeg1());
			processRosetta(path.newSubPath("notionalQuantityScheduleLeg2"), processor, NotionalQuantityScheduleLeg2Report.NotionalQuantityScheduleLeg2ReportBuilder.class, getNotionalQuantityScheduleLeg2());
			processor.processBasic(path.newSubPath("actionType"), String.class, getActionType(), this);
			processor.processBasic(path.newSubPath("eventType"), String.class, getEventType(), this);
			processor.processBasic(path.newSubPath("eventTimestamp"), ZonedDateTime.class, getEventTimestamp(), this);
			processor.processBasic(path.newSubPath("eventIdentifier"), String.class, getEventIdentifier(), this);
			processor.processBasic(path.newSubPath("indexFactor"), BigDecimal.class, getIndexFactor(), this);
			processor.processBasic(path.newSubPath("embeddedOptionType"), String.class, getEmbeddedOptionType(), this);
			processor.processBasic(path.newSubPath("upi"), String.class, getUpi(), this);
			processor.processBasic(path.newSubPath("deliveryType"), String.class, getDeliveryType(), this);
			processor.processBasic(path.newSubPath("assetClass"), String.class, getAssetClass(), this);
			processor.processBasic(path.newSubPath("underlyingIdentificationType"), String.class, getUnderlyingIdentificationType(), this);
			processor.processBasic(path.newSubPath("underlyingIdentification"), String.class, getUnderlyingIdentification(), this);
			processor.processBasic(path.newSubPath("underlyingIndexIndicator"), String.class, getUnderlyingIndexIndicator(), this);
			processor.processBasic(path.newSubPath("nameOfTheUnderlyingIndex"), String.class, getNameOfTheUnderlyingIndex(), this);
			processor.processBasic(path.newSubPath("referenceEntity"), String.class, getReferenceEntity(), this);
			processor.processBasic(path.newSubPath("indicatorOfTheFloatingRateOfLeg1"), String.class, getIndicatorOfTheFloatingRateOfLeg1(), this);
			processor.processBasic(path.newSubPath("indicatorOfTheFloatingRateOfLeg2"), String.class, getIndicatorOfTheFloatingRateOfLeg2(), this);
			processor.processBasic(path.newSubPath("nameOfTheFloatingRateOfLeg1"), String.class, getNameOfTheFloatingRateOfLeg1(), this);
			processor.processBasic(path.newSubPath("nameOfTheFloatingRateOfLeg2"), String.class, getNameOfTheFloatingRateOfLeg2(), this);
			processor.processBasic(path.newSubPath("floatingRateReferencePeriodOfLeg1TimePeriod"), String.class, getFloatingRateReferencePeriodOfLeg1TimePeriod(), this);
			processor.processBasic(path.newSubPath("floatingRateReferencePeriodOfLeg2TimePeriod"), String.class, getFloatingRateReferencePeriodOfLeg2TimePeriod(), this);
			processor.processBasic(path.newSubPath("floatingRateReferencePeriodOfLeg1Multiplier"), Integer.class, getFloatingRateReferencePeriodOfLeg1Multiplier(), this);
			processor.processBasic(path.newSubPath("floatingRateReferencePeriodOfLeg2Multiplier"), Integer.class, getFloatingRateReferencePeriodOfLeg2Multiplier(), this);
			processor.processBasic(path.newSubPath("isCrypto"), Boolean.class, getIsCrypto(), this);
			processor.processBasic(path.newSubPath("maturityDateOfTheUnderlying"), Date.class, getMaturityDateOfTheUnderlying(), this);
			processor.processBasic(path.newSubPath("seniority"), String.class, getSeniority(), this);
			processor.processBasic(path.newSubPath("series"), Integer.class, getSeries(), this);
			processor.processBasic(path.newSubPath("version"), Integer.class, getVersion(), this);
			processor.processBasic(path.newSubPath("cdSIndexAttachmentPoint"), BigDecimal.class, getCdSIndexAttachmentPoint(), this);
			processor.processBasic(path.newSubPath("cdSIndexDetachmentPoint"), BigDecimal.class, getCdSIndexDetachmentPoint(), this);
			processor.processBasic(path.newSubPath("customBasketCode"), String.class, getCustomBasketCode(), this);
			processor.processBasic(path.newSubPath("identifierOfBasketConstituents"), String.class, getIdentifierOfBasketConstituents(), this);
			processor.processBasic(path.newSubPath("basketConstituentIdentifierSource"), ProductIdTypeEnum.class, getBasketConstituentIdentifierSource(), this);
			processor.processBasic(path.newSubPath("contractType"), String.class, getContractType(), this);
			processor.processBasic(path.newSubPath("optionStyle"), String.class, getOptionStyle(), this);
			processor.processBasic(path.newSubPath("optionType"), String.class, getOptionType(), this);
			processor.processBasic(path.newSubPath("packageIdentifier"), String.class, getPackageIdentifier(), this);
			processRosetta(path.newSubPath("packageTransactionPrice"), processor, PackageTransactionPrice.PackageTransactionPriceBuilder.class, getPackageTransactionPrice());
			processor.processBasic(path.newSubPath("packageTransactionPriceCurrency"), ISOCurrencyCodeEnum.class, getPackageTransactionPriceCurrency(), this);
			processor.processBasic(path.newSubPath("packageTransactionPriceNotation"), String.class, getPackageTransactionPriceNotation(), this);
			processRosetta(path.newSubPath("packageTransactionSpread"), processor, PackageTransactionSpread.PackageTransactionSpreadBuilder.class, getPackageTransactionSpread());
			processor.processBasic(path.newSubPath("packageTransactionSpreadCurrency"), ISOCurrencyCodeEnum.class, getPackageTransactionSpreadCurrency(), this);
			processor.processBasic(path.newSubPath("packageTransactionSpreadNotation"), String.class, getPackageTransactionSpreadNotation(), this);
			processor.processBasic(path.newSubPath("technicalRecordId"), String.class, getTechnicalRecordId(), this);
			processRosetta(path.newSubPath("nonReportable"), processor, JFSANonReportable.JFSANonReportableBuilder.class, getNonReportable());
		}
		

		JFSATransactionReport.JFSATransactionReportBuilder prune();
	}

	/*********************** Immutable Implementation of JFSATransactionReport  ***********************/
	class JFSATransactionReportImpl implements JFSATransactionReport {
		private final Date effectiveDate;
		private final Date expirationDate;
		private final Date earlyTerminationDate;
		private final ZonedDateTime reportingTimestamp;
		private final ZonedDateTime executionTimestamp;
		private final String entityResponsibleForReporting;
		private final String counterparty1;
		private final String counterparty2;
		private final Boolean counterparty2IdentifierType;
		private final String direction1;
		private final String direction2Leg1;
		private final String direction2Leg2;
		private final String submitterIdentifier;
		private final String newSDRIdentifier;
		private final String originalSwapSDRIdentifier;
		private final String cleared;
		private final String centralCounterparty;
		private final String clearingMember;
		private final String platformIdentifier;
		private final String confirmed;
		private final Date finalContractualSettlementDate;
		private final ZonedDateTime confirmationTimestamp;
		private final ISOCurrencyCodeEnum settlementCurrencyLeg1;
		private final ISOCurrencyCodeEnum settlementCurrencyLeg2;
		private final String clearingAccountOrigin;
		private final String originalSwapUTI;
		private final ZonedDateTime clearingReceiptTimestamp;
		private final String uti;
		private final String utiProprietary;
		private final String priorUti;
		private final String priorUtiProprietary;
		private final InterestComputationMethod4Code fixedRateDayCountConventionLeg1;
		private final InterestComputationMethod4Code floatingRateDayCountConventionOfLeg1;
		private final InterestComputationMethod4Code fixedRateDayCountConventionLeg2;
		private final InterestComputationMethod4Code floatingRateDayCountConventionOfLeg2;
		private final String fixedRatePaymentFrequencyPeriodLeg1;
		private final String fixedRatePaymentFrequencyPeriodLeg2;
		private final String floatingRatePaymentFrequencyPeriodLeg1;
		private final String floatingRatePaymentFrequencyPeriodLeg2;
		private final Integer fixedRatePaymentFrequencyPeriodMultiplierLeg1;
		private final Integer fixedRatePaymentFrequencyPeriodMultiplierLeg2;
		private final Integer floatingRatePaymentFrequencyPeriodMultiplierLeg1;
		private final Integer floatingRatePaymentFrequencyPeriodMultiplierLeg2;
		private final ZonedDateTime fixingDateLeg1;
		private final ZonedDateTime fixingDateLeg2;
		private final String floatingRateResetFrequencyPeriodLeg1;
		private final String floatingRateResetFrequencyPeriodOfLeg2;
		private final Integer floatingRateResetFrequencyPeriodMultiplierLeg1;
		private final Integer floatingRateResetFrequencyPeriodMultiplierLeg2;
		private final List<? extends OtherPaymentReport> otherPayment;
		private final BigDecimal valuationAmount;
		private final ISOCurrencyCodeEnum valuationCurrency;
		private final ValuationType1Code valuationMethod;
		private final ZonedDateTime valuationTimestamp;
		private final Boolean collateralPortfolioIndicator;
		private final String initialMarginCollateralPortfolioCode;
		private final String variationMarginCollateralPortfolioCode;
		private final PriceAndNotation price;
		private final String priceNotation;
		private final String priceUnitOfMeasure;
		private final List<? extends PriceScheduleReport> priceSchedule;
		private final BigDecimal fixedRateLeg1;
		private final BigDecimal fixedRateLeg2;
		private final SpreadOfLeg1 spreadOfLeg1;
		private final String spreadNotationOfLeg1;
		private final SpreadOfLeg2 spreadOfLeg2;
		private final String spreadNotationOfLeg2;
		private final StrikePriceAndNotation strikePrice;
		private final String strikePriceNotation;
		private final List<? extends StrikePriceScheduleReport> strikePriceSchedule;
		private final BigDecimal optionPremiumAmount;
		private final ISOCurrencyCodeEnum optionPremiumCurrency;
		private final Date optionPremiumPaymentDate;
		private final Date firstExerciseDate;
		private final BigDecimal exchangeRate;
		private final String exchangeRateBasis;
		private final BigDecimal notionalLeg1;
		private final BigDecimal notionalLeg2;
		private final BigDecimal callAmount;
		private final BigDecimal putAmount;
		private final ISOCurrencyCodeEnum notionalCurrencyOfLeg1;
		private final ISOCurrencyCodeEnum notionalCurrencyOfLeg2;
		private final ISOCurrencyCodeEnum callCurrency;
		private final ISOCurrencyCodeEnum putCurrency;
		private final String quantityUnitOfMeasureLeg1;
		private final String quantityUnitOfMeasureLeg2;
		private final List<? extends NotionalAmountScheduleLeg1Report> notionalAmountScheduleLeg1;
		private final List<? extends NotionalAmountScheduleLeg2Report> notionalAmountScheduleLeg2;
		private final BigDecimal totalNotionalQuantityLeg1;
		private final BigDecimal totalNotionalQuantityLeg2;
		private final List<? extends NotionalQuantityScheduleLeg1Report> notionalQuantityScheduleLeg1;
		private final List<? extends NotionalQuantityScheduleLeg2Report> notionalQuantityScheduleLeg2;
		private final String actionType;
		private final String eventType;
		private final ZonedDateTime eventTimestamp;
		private final String eventIdentifier;
		private final BigDecimal indexFactor;
		private final String embeddedOptionType;
		private final String upi;
		private final String deliveryType;
		private final String assetClass;
		private final String underlyingIdentificationType;
		private final String underlyingIdentification;
		private final String underlyingIndexIndicator;
		private final String nameOfTheUnderlyingIndex;
		private final String referenceEntity;
		private final String indicatorOfTheFloatingRateOfLeg1;
		private final String indicatorOfTheFloatingRateOfLeg2;
		private final String nameOfTheFloatingRateOfLeg1;
		private final String nameOfTheFloatingRateOfLeg2;
		private final String floatingRateReferencePeriodOfLeg1TimePeriod;
		private final String floatingRateReferencePeriodOfLeg2TimePeriod;
		private final Integer floatingRateReferencePeriodOfLeg1Multiplier;
		private final Integer floatingRateReferencePeriodOfLeg2Multiplier;
		private final Boolean isCrypto;
		private final Date maturityDateOfTheUnderlying;
		private final String seniority;
		private final Integer series;
		private final Integer version;
		private final BigDecimal cdSIndexAttachmentPoint;
		private final BigDecimal cdSIndexDetachmentPoint;
		private final String customBasketCode;
		private final List<String> identifierOfBasketConstituents;
		private final List<ProductIdTypeEnum> basketConstituentIdentifierSource;
		private final String contractType;
		private final String optionStyle;
		private final String optionType;
		private final String packageIdentifier;
		private final PackageTransactionPrice packageTransactionPrice;
		private final ISOCurrencyCodeEnum packageTransactionPriceCurrency;
		private final String packageTransactionPriceNotation;
		private final PackageTransactionSpread packageTransactionSpread;
		private final ISOCurrencyCodeEnum packageTransactionSpreadCurrency;
		private final String packageTransactionSpreadNotation;
		private final String technicalRecordId;
		private final JFSANonReportable nonReportable;
		
		protected JFSATransactionReportImpl(JFSATransactionReport.JFSATransactionReportBuilder builder) {
			this.effectiveDate = builder.getEffectiveDate();
			this.expirationDate = builder.getExpirationDate();
			this.earlyTerminationDate = builder.getEarlyTerminationDate();
			this.reportingTimestamp = builder.getReportingTimestamp();
			this.executionTimestamp = builder.getExecutionTimestamp();
			this.entityResponsibleForReporting = builder.getEntityResponsibleForReporting();
			this.counterparty1 = builder.getCounterparty1();
			this.counterparty2 = builder.getCounterparty2();
			this.counterparty2IdentifierType = builder.getCounterparty2IdentifierType();
			this.direction1 = builder.getDirection1();
			this.direction2Leg1 = builder.getDirection2Leg1();
			this.direction2Leg2 = builder.getDirection2Leg2();
			this.submitterIdentifier = builder.getSubmitterIdentifier();
			this.newSDRIdentifier = builder.getNewSDRIdentifier();
			this.originalSwapSDRIdentifier = builder.getOriginalSwapSDRIdentifier();
			this.cleared = builder.getCleared();
			this.centralCounterparty = builder.getCentralCounterparty();
			this.clearingMember = builder.getClearingMember();
			this.platformIdentifier = builder.getPlatformIdentifier();
			this.confirmed = builder.getConfirmed();
			this.finalContractualSettlementDate = builder.getFinalContractualSettlementDate();
			this.confirmationTimestamp = builder.getConfirmationTimestamp();
			this.settlementCurrencyLeg1 = builder.getSettlementCurrencyLeg1();
			this.settlementCurrencyLeg2 = builder.getSettlementCurrencyLeg2();
			this.clearingAccountOrigin = builder.getClearingAccountOrigin();
			this.originalSwapUTI = builder.getOriginalSwapUTI();
			this.clearingReceiptTimestamp = builder.getClearingReceiptTimestamp();
			this.uti = builder.getUti();
			this.utiProprietary = builder.getUtiProprietary();
			this.priorUti = builder.getPriorUti();
			this.priorUtiProprietary = builder.getPriorUtiProprietary();
			this.fixedRateDayCountConventionLeg1 = builder.getFixedRateDayCountConventionLeg1();
			this.floatingRateDayCountConventionOfLeg1 = builder.getFloatingRateDayCountConventionOfLeg1();
			this.fixedRateDayCountConventionLeg2 = builder.getFixedRateDayCountConventionLeg2();
			this.floatingRateDayCountConventionOfLeg2 = builder.getFloatingRateDayCountConventionOfLeg2();
			this.fixedRatePaymentFrequencyPeriodLeg1 = builder.getFixedRatePaymentFrequencyPeriodLeg1();
			this.fixedRatePaymentFrequencyPeriodLeg2 = builder.getFixedRatePaymentFrequencyPeriodLeg2();
			this.floatingRatePaymentFrequencyPeriodLeg1 = builder.getFloatingRatePaymentFrequencyPeriodLeg1();
			this.floatingRatePaymentFrequencyPeriodLeg2 = builder.getFloatingRatePaymentFrequencyPeriodLeg2();
			this.fixedRatePaymentFrequencyPeriodMultiplierLeg1 = builder.getFixedRatePaymentFrequencyPeriodMultiplierLeg1();
			this.fixedRatePaymentFrequencyPeriodMultiplierLeg2 = builder.getFixedRatePaymentFrequencyPeriodMultiplierLeg2();
			this.floatingRatePaymentFrequencyPeriodMultiplierLeg1 = builder.getFloatingRatePaymentFrequencyPeriodMultiplierLeg1();
			this.floatingRatePaymentFrequencyPeriodMultiplierLeg2 = builder.getFloatingRatePaymentFrequencyPeriodMultiplierLeg2();
			this.fixingDateLeg1 = builder.getFixingDateLeg1();
			this.fixingDateLeg2 = builder.getFixingDateLeg2();
			this.floatingRateResetFrequencyPeriodLeg1 = builder.getFloatingRateResetFrequencyPeriodLeg1();
			this.floatingRateResetFrequencyPeriodOfLeg2 = builder.getFloatingRateResetFrequencyPeriodOfLeg2();
			this.floatingRateResetFrequencyPeriodMultiplierLeg1 = builder.getFloatingRateResetFrequencyPeriodMultiplierLeg1();
			this.floatingRateResetFrequencyPeriodMultiplierLeg2 = builder.getFloatingRateResetFrequencyPeriodMultiplierLeg2();
			this.otherPayment = ofNullable(builder.getOtherPayment()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.valuationAmount = builder.getValuationAmount();
			this.valuationCurrency = builder.getValuationCurrency();
			this.valuationMethod = builder.getValuationMethod();
			this.valuationTimestamp = builder.getValuationTimestamp();
			this.collateralPortfolioIndicator = builder.getCollateralPortfolioIndicator();
			this.initialMarginCollateralPortfolioCode = builder.getInitialMarginCollateralPortfolioCode();
			this.variationMarginCollateralPortfolioCode = builder.getVariationMarginCollateralPortfolioCode();
			this.price = ofNullable(builder.getPrice()).map(f->f.build()).orElse(null);
			this.priceNotation = builder.getPriceNotation();
			this.priceUnitOfMeasure = builder.getPriceUnitOfMeasure();
			this.priceSchedule = ofNullable(builder.getPriceSchedule()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.fixedRateLeg1 = builder.getFixedRateLeg1();
			this.fixedRateLeg2 = builder.getFixedRateLeg2();
			this.spreadOfLeg1 = ofNullable(builder.getSpreadOfLeg1()).map(f->f.build()).orElse(null);
			this.spreadNotationOfLeg1 = builder.getSpreadNotationOfLeg1();
			this.spreadOfLeg2 = ofNullable(builder.getSpreadOfLeg2()).map(f->f.build()).orElse(null);
			this.spreadNotationOfLeg2 = builder.getSpreadNotationOfLeg2();
			this.strikePrice = ofNullable(builder.getStrikePrice()).map(f->f.build()).orElse(null);
			this.strikePriceNotation = builder.getStrikePriceNotation();
			this.strikePriceSchedule = ofNullable(builder.getStrikePriceSchedule()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.optionPremiumAmount = builder.getOptionPremiumAmount();
			this.optionPremiumCurrency = builder.getOptionPremiumCurrency();
			this.optionPremiumPaymentDate = builder.getOptionPremiumPaymentDate();
			this.firstExerciseDate = builder.getFirstExerciseDate();
			this.exchangeRate = builder.getExchangeRate();
			this.exchangeRateBasis = builder.getExchangeRateBasis();
			this.notionalLeg1 = builder.getNotionalLeg1();
			this.notionalLeg2 = builder.getNotionalLeg2();
			this.callAmount = builder.getCallAmount();
			this.putAmount = builder.getPutAmount();
			this.notionalCurrencyOfLeg1 = builder.getNotionalCurrencyOfLeg1();
			this.notionalCurrencyOfLeg2 = builder.getNotionalCurrencyOfLeg2();
			this.callCurrency = builder.getCallCurrency();
			this.putCurrency = builder.getPutCurrency();
			this.quantityUnitOfMeasureLeg1 = builder.getQuantityUnitOfMeasureLeg1();
			this.quantityUnitOfMeasureLeg2 = builder.getQuantityUnitOfMeasureLeg2();
			this.notionalAmountScheduleLeg1 = ofNullable(builder.getNotionalAmountScheduleLeg1()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.notionalAmountScheduleLeg2 = ofNullable(builder.getNotionalAmountScheduleLeg2()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.totalNotionalQuantityLeg1 = builder.getTotalNotionalQuantityLeg1();
			this.totalNotionalQuantityLeg2 = builder.getTotalNotionalQuantityLeg2();
			this.notionalQuantityScheduleLeg1 = ofNullable(builder.getNotionalQuantityScheduleLeg1()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.notionalQuantityScheduleLeg2 = ofNullable(builder.getNotionalQuantityScheduleLeg2()).filter(_l->!_l.isEmpty()).map(list -> list.stream().filter(Objects::nonNull).map(f->f.build()).filter(Objects::nonNull).collect(ImmutableList.toImmutableList())).orElse(null);
			this.actionType = builder.getActionType();
			this.eventType = builder.getEventType();
			this.eventTimestamp = builder.getEventTimestamp();
			this.eventIdentifier = builder.getEventIdentifier();
			this.indexFactor = builder.getIndexFactor();
			this.embeddedOptionType = builder.getEmbeddedOptionType();
			this.upi = builder.getUpi();
			this.deliveryType = builder.getDeliveryType();
			this.assetClass = builder.getAssetClass();
			this.underlyingIdentificationType = builder.getUnderlyingIdentificationType();
			this.underlyingIdentification = builder.getUnderlyingIdentification();
			this.underlyingIndexIndicator = builder.getUnderlyingIndexIndicator();
			this.nameOfTheUnderlyingIndex = builder.getNameOfTheUnderlyingIndex();
			this.referenceEntity = builder.getReferenceEntity();
			this.indicatorOfTheFloatingRateOfLeg1 = builder.getIndicatorOfTheFloatingRateOfLeg1();
			this.indicatorOfTheFloatingRateOfLeg2 = builder.getIndicatorOfTheFloatingRateOfLeg2();
			this.nameOfTheFloatingRateOfLeg1 = builder.getNameOfTheFloatingRateOfLeg1();
			this.nameOfTheFloatingRateOfLeg2 = builder.getNameOfTheFloatingRateOfLeg2();
			this.floatingRateReferencePeriodOfLeg1TimePeriod = builder.getFloatingRateReferencePeriodOfLeg1TimePeriod();
			this.floatingRateReferencePeriodOfLeg2TimePeriod = builder.getFloatingRateReferencePeriodOfLeg2TimePeriod();
			this.floatingRateReferencePeriodOfLeg1Multiplier = builder.getFloatingRateReferencePeriodOfLeg1Multiplier();
			this.floatingRateReferencePeriodOfLeg2Multiplier = builder.getFloatingRateReferencePeriodOfLeg2Multiplier();
			this.isCrypto = builder.getIsCrypto();
			this.maturityDateOfTheUnderlying = builder.getMaturityDateOfTheUnderlying();
			this.seniority = builder.getSeniority();
			this.series = builder.getSeries();
			this.version = builder.getVersion();
			this.cdSIndexAttachmentPoint = builder.getCdSIndexAttachmentPoint();
			this.cdSIndexDetachmentPoint = builder.getCdSIndexDetachmentPoint();
			this.customBasketCode = builder.getCustomBasketCode();
			this.identifierOfBasketConstituents = ofNullable(builder.getIdentifierOfBasketConstituents()).filter(_l->!_l.isEmpty()).map(ImmutableList::copyOf).orElse(null);
			this.basketConstituentIdentifierSource = ofNullable(builder.getBasketConstituentIdentifierSource()).filter(_l->!_l.isEmpty()).map(ImmutableList::copyOf).orElse(null);
			this.contractType = builder.getContractType();
			this.optionStyle = builder.getOptionStyle();
			this.optionType = builder.getOptionType();
			this.packageIdentifier = builder.getPackageIdentifier();
			this.packageTransactionPrice = ofNullable(builder.getPackageTransactionPrice()).map(f->f.build()).orElse(null);
			this.packageTransactionPriceCurrency = builder.getPackageTransactionPriceCurrency();
			this.packageTransactionPriceNotation = builder.getPackageTransactionPriceNotation();
			this.packageTransactionSpread = ofNullable(builder.getPackageTransactionSpread()).map(f->f.build()).orElse(null);
			this.packageTransactionSpreadCurrency = builder.getPackageTransactionSpreadCurrency();
			this.packageTransactionSpreadNotation = builder.getPackageTransactionSpreadNotation();
			this.technicalRecordId = builder.getTechnicalRecordId();
			this.nonReportable = ofNullable(builder.getNonReportable()).map(f->f.build()).orElse(null);
		}
		
		@Override
		@RosettaAttribute("effectiveDate")
		public Date getEffectiveDate() {
			return effectiveDate;
		}
		
		@Override
		@RosettaAttribute("expirationDate")
		public Date getExpirationDate() {
			return expirationDate;
		}
		
		@Override
		@RosettaAttribute("earlyTerminationDate")
		public Date getEarlyTerminationDate() {
			return earlyTerminationDate;
		}
		
		@Override
		@RosettaAttribute("reportingTimestamp")
		public ZonedDateTime getReportingTimestamp() {
			return reportingTimestamp;
		}
		
		@Override
		@RosettaAttribute("executionTimestamp")
		public ZonedDateTime getExecutionTimestamp() {
			return executionTimestamp;
		}
		
		@Override
		@RosettaAttribute("entityResponsibleForReporting")
		public String getEntityResponsibleForReporting() {
			return entityResponsibleForReporting;
		}
		
		@Override
		@RosettaAttribute("counterparty1")
		public String getCounterparty1() {
			return counterparty1;
		}
		
		@Override
		@RosettaAttribute("counterparty2")
		public String getCounterparty2() {
			return counterparty2;
		}
		
		@Override
		@RosettaAttribute("counterparty2IdentifierType")
		public Boolean getCounterparty2IdentifierType() {
			return counterparty2IdentifierType;
		}
		
		@Override
		@RosettaAttribute("direction1")
		public String getDirection1() {
			return direction1;
		}
		
		@Override
		@RosettaAttribute("direction2Leg1")
		public String getDirection2Leg1() {
			return direction2Leg1;
		}
		
		@Override
		@RosettaAttribute("direction2Leg2")
		public String getDirection2Leg2() {
			return direction2Leg2;
		}
		
		@Override
		@RosettaAttribute("submitterIdentifier")
		public String getSubmitterIdentifier() {
			return submitterIdentifier;
		}
		
		@Override
		@RosettaAttribute("newSDRIdentifier")
		public String getNewSDRIdentifier() {
			return newSDRIdentifier;
		}
		
		@Override
		@RosettaAttribute("originalSwapSDRIdentifier")
		public String getOriginalSwapSDRIdentifier() {
			return originalSwapSDRIdentifier;
		}
		
		@Override
		@RosettaAttribute("cleared")
		public String getCleared() {
			return cleared;
		}
		
		@Override
		@RosettaAttribute("centralCounterparty")
		public String getCentralCounterparty() {
			return centralCounterparty;
		}
		
		@Override
		@RosettaAttribute("clearingMember")
		public String getClearingMember() {
			return clearingMember;
		}
		
		@Override
		@RosettaAttribute("platformIdentifier")
		public String getPlatformIdentifier() {
			return platformIdentifier;
		}
		
		@Override
		@RosettaAttribute("confirmed")
		public String getConfirmed() {
			return confirmed;
		}
		
		@Override
		@RosettaAttribute("finalContractualSettlementDate")
		public Date getFinalContractualSettlementDate() {
			return finalContractualSettlementDate;
		}
		
		@Override
		@RosettaAttribute("confirmationTimestamp")
		public ZonedDateTime getConfirmationTimestamp() {
			return confirmationTimestamp;
		}
		
		@Override
		@RosettaAttribute("settlementCurrencyLeg1")
		public ISOCurrencyCodeEnum getSettlementCurrencyLeg1() {
			return settlementCurrencyLeg1;
		}
		
		@Override
		@RosettaAttribute("settlementCurrencyLeg2")
		public ISOCurrencyCodeEnum getSettlementCurrencyLeg2() {
			return settlementCurrencyLeg2;
		}
		
		@Override
		@RosettaAttribute("clearingAccountOrigin")
		public String getClearingAccountOrigin() {
			return clearingAccountOrigin;
		}
		
		@Override
		@RosettaAttribute("originalSwapUTI")
		public String getOriginalSwapUTI() {
			return originalSwapUTI;
		}
		
		@Override
		@RosettaAttribute("clearingReceiptTimestamp")
		public ZonedDateTime getClearingReceiptTimestamp() {
			return clearingReceiptTimestamp;
		}
		
		@Override
		@RosettaAttribute("uti")
		public String getUti() {
			return uti;
		}
		
		@Override
		@RosettaAttribute("utiProprietary")
		public String getUtiProprietary() {
			return utiProprietary;
		}
		
		@Override
		@RosettaAttribute("priorUti")
		public String getPriorUti() {
			return priorUti;
		}
		
		@Override
		@RosettaAttribute("priorUtiProprietary")
		public String getPriorUtiProprietary() {
			return priorUtiProprietary;
		}
		
		@Override
		@RosettaAttribute("fixedRateDayCountConventionLeg1")
		public InterestComputationMethod4Code getFixedRateDayCountConventionLeg1() {
			return fixedRateDayCountConventionLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRateDayCountConventionOfLeg1")
		public InterestComputationMethod4Code getFloatingRateDayCountConventionOfLeg1() {
			return floatingRateDayCountConventionOfLeg1;
		}
		
		@Override
		@RosettaAttribute("fixedRateDayCountConventionLeg2")
		public InterestComputationMethod4Code getFixedRateDayCountConventionLeg2() {
			return fixedRateDayCountConventionLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRateDayCountConventionOfLeg2")
		public InterestComputationMethod4Code getFloatingRateDayCountConventionOfLeg2() {
			return floatingRateDayCountConventionOfLeg2;
		}
		
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodLeg1")
		public String getFixedRatePaymentFrequencyPeriodLeg1() {
			return fixedRatePaymentFrequencyPeriodLeg1;
		}
		
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodLeg2")
		public String getFixedRatePaymentFrequencyPeriodLeg2() {
			return fixedRatePaymentFrequencyPeriodLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodLeg1")
		public String getFloatingRatePaymentFrequencyPeriodLeg1() {
			return floatingRatePaymentFrequencyPeriodLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodLeg2")
		public String getFloatingRatePaymentFrequencyPeriodLeg2() {
			return floatingRatePaymentFrequencyPeriodLeg2;
		}
		
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodMultiplierLeg1")
		public Integer getFixedRatePaymentFrequencyPeriodMultiplierLeg1() {
			return fixedRatePaymentFrequencyPeriodMultiplierLeg1;
		}
		
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodMultiplierLeg2")
		public Integer getFixedRatePaymentFrequencyPeriodMultiplierLeg2() {
			return fixedRatePaymentFrequencyPeriodMultiplierLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodMultiplierLeg1")
		public Integer getFloatingRatePaymentFrequencyPeriodMultiplierLeg1() {
			return floatingRatePaymentFrequencyPeriodMultiplierLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodMultiplierLeg2")
		public Integer getFloatingRatePaymentFrequencyPeriodMultiplierLeg2() {
			return floatingRatePaymentFrequencyPeriodMultiplierLeg2;
		}
		
		@Override
		@RosettaAttribute("fixingDateLeg1")
		public ZonedDateTime getFixingDateLeg1() {
			return fixingDateLeg1;
		}
		
		@Override
		@RosettaAttribute("fixingDateLeg2")
		public ZonedDateTime getFixingDateLeg2() {
			return fixingDateLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodLeg1")
		public String getFloatingRateResetFrequencyPeriodLeg1() {
			return floatingRateResetFrequencyPeriodLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodOfLeg2")
		public String getFloatingRateResetFrequencyPeriodOfLeg2() {
			return floatingRateResetFrequencyPeriodOfLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodMultiplierLeg1")
		public Integer getFloatingRateResetFrequencyPeriodMultiplierLeg1() {
			return floatingRateResetFrequencyPeriodMultiplierLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodMultiplierLeg2")
		public Integer getFloatingRateResetFrequencyPeriodMultiplierLeg2() {
			return floatingRateResetFrequencyPeriodMultiplierLeg2;
		}
		
		@Override
		@RosettaAttribute("otherPayment")
		public List<? extends OtherPaymentReport> getOtherPayment() {
			return otherPayment;
		}
		
		@Override
		@RosettaAttribute("valuationAmount")
		public BigDecimal getValuationAmount() {
			return valuationAmount;
		}
		
		@Override
		@RosettaAttribute("valuationCurrency")
		public ISOCurrencyCodeEnum getValuationCurrency() {
			return valuationCurrency;
		}
		
		@Override
		@RosettaAttribute("valuationMethod")
		public ValuationType1Code getValuationMethod() {
			return valuationMethod;
		}
		
		@Override
		@RosettaAttribute("valuationTimestamp")
		public ZonedDateTime getValuationTimestamp() {
			return valuationTimestamp;
		}
		
		@Override
		@RosettaAttribute("collateralPortfolioIndicator")
		public Boolean getCollateralPortfolioIndicator() {
			return collateralPortfolioIndicator;
		}
		
		@Override
		@RosettaAttribute("initialMarginCollateralPortfolioCode")
		public String getInitialMarginCollateralPortfolioCode() {
			return initialMarginCollateralPortfolioCode;
		}
		
		@Override
		@RosettaAttribute("variationMarginCollateralPortfolioCode")
		public String getVariationMarginCollateralPortfolioCode() {
			return variationMarginCollateralPortfolioCode;
		}
		
		@Override
		@RosettaAttribute("price")
		public PriceAndNotation getPrice() {
			return price;
		}
		
		@Override
		@RosettaAttribute("priceNotation")
		public String getPriceNotation() {
			return priceNotation;
		}
		
		@Override
		@RosettaAttribute("priceUnitOfMeasure")
		public String getPriceUnitOfMeasure() {
			return priceUnitOfMeasure;
		}
		
		@Override
		@RosettaAttribute("priceSchedule")
		public List<? extends PriceScheduleReport> getPriceSchedule() {
			return priceSchedule;
		}
		
		@Override
		@RosettaAttribute("fixedRateLeg1")
		public BigDecimal getFixedRateLeg1() {
			return fixedRateLeg1;
		}
		
		@Override
		@RosettaAttribute("fixedRateLeg2")
		public BigDecimal getFixedRateLeg2() {
			return fixedRateLeg2;
		}
		
		@Override
		@RosettaAttribute("spreadOfLeg1")
		public SpreadOfLeg1 getSpreadOfLeg1() {
			return spreadOfLeg1;
		}
		
		@Override
		@RosettaAttribute("spreadNotationOfLeg1")
		public String getSpreadNotationOfLeg1() {
			return spreadNotationOfLeg1;
		}
		
		@Override
		@RosettaAttribute("spreadOfLeg2")
		public SpreadOfLeg2 getSpreadOfLeg2() {
			return spreadOfLeg2;
		}
		
		@Override
		@RosettaAttribute("spreadNotationOfLeg2")
		public String getSpreadNotationOfLeg2() {
			return spreadNotationOfLeg2;
		}
		
		@Override
		@RosettaAttribute("strikePrice")
		public StrikePriceAndNotation getStrikePrice() {
			return strikePrice;
		}
		
		@Override
		@RosettaAttribute("strikePriceNotation")
		public String getStrikePriceNotation() {
			return strikePriceNotation;
		}
		
		@Override
		@RosettaAttribute("strikePriceSchedule")
		public List<? extends StrikePriceScheduleReport> getStrikePriceSchedule() {
			return strikePriceSchedule;
		}
		
		@Override
		@RosettaAttribute("optionPremiumAmount")
		public BigDecimal getOptionPremiumAmount() {
			return optionPremiumAmount;
		}
		
		@Override
		@RosettaAttribute("optionPremiumCurrency")
		public ISOCurrencyCodeEnum getOptionPremiumCurrency() {
			return optionPremiumCurrency;
		}
		
		@Override
		@RosettaAttribute("optionPremiumPaymentDate")
		public Date getOptionPremiumPaymentDate() {
			return optionPremiumPaymentDate;
		}
		
		@Override
		@RosettaAttribute("firstExerciseDate")
		public Date getFirstExerciseDate() {
			return firstExerciseDate;
		}
		
		@Override
		@RosettaAttribute("exchangeRate")
		public BigDecimal getExchangeRate() {
			return exchangeRate;
		}
		
		@Override
		@RosettaAttribute("exchangeRateBasis")
		public String getExchangeRateBasis() {
			return exchangeRateBasis;
		}
		
		@Override
		@RosettaAttribute("notionalLeg1")
		public BigDecimal getNotionalLeg1() {
			return notionalLeg1;
		}
		
		@Override
		@RosettaAttribute("notionalLeg2")
		public BigDecimal getNotionalLeg2() {
			return notionalLeg2;
		}
		
		@Override
		@RosettaAttribute("callAmount")
		public BigDecimal getCallAmount() {
			return callAmount;
		}
		
		@Override
		@RosettaAttribute("putAmount")
		public BigDecimal getPutAmount() {
			return putAmount;
		}
		
		@Override
		@RosettaAttribute("notionalCurrencyOfLeg1")
		public ISOCurrencyCodeEnum getNotionalCurrencyOfLeg1() {
			return notionalCurrencyOfLeg1;
		}
		
		@Override
		@RosettaAttribute("notionalCurrencyOfLeg2")
		public ISOCurrencyCodeEnum getNotionalCurrencyOfLeg2() {
			return notionalCurrencyOfLeg2;
		}
		
		@Override
		@RosettaAttribute("callCurrency")
		public ISOCurrencyCodeEnum getCallCurrency() {
			return callCurrency;
		}
		
		@Override
		@RosettaAttribute("putCurrency")
		public ISOCurrencyCodeEnum getPutCurrency() {
			return putCurrency;
		}
		
		@Override
		@RosettaAttribute("quantityUnitOfMeasureLeg1")
		public String getQuantityUnitOfMeasureLeg1() {
			return quantityUnitOfMeasureLeg1;
		}
		
		@Override
		@RosettaAttribute("quantityUnitOfMeasureLeg2")
		public String getQuantityUnitOfMeasureLeg2() {
			return quantityUnitOfMeasureLeg2;
		}
		
		@Override
		@RosettaAttribute("notionalAmountScheduleLeg1")
		public List<? extends NotionalAmountScheduleLeg1Report> getNotionalAmountScheduleLeg1() {
			return notionalAmountScheduleLeg1;
		}
		
		@Override
		@RosettaAttribute("notionalAmountScheduleLeg2")
		public List<? extends NotionalAmountScheduleLeg2Report> getNotionalAmountScheduleLeg2() {
			return notionalAmountScheduleLeg2;
		}
		
		@Override
		@RosettaAttribute("totalNotionalQuantityLeg1")
		public BigDecimal getTotalNotionalQuantityLeg1() {
			return totalNotionalQuantityLeg1;
		}
		
		@Override
		@RosettaAttribute("totalNotionalQuantityLeg2")
		public BigDecimal getTotalNotionalQuantityLeg2() {
			return totalNotionalQuantityLeg2;
		}
		
		@Override
		@RosettaAttribute("notionalQuantityScheduleLeg1")
		public List<? extends NotionalQuantityScheduleLeg1Report> getNotionalQuantityScheduleLeg1() {
			return notionalQuantityScheduleLeg1;
		}
		
		@Override
		@RosettaAttribute("notionalQuantityScheduleLeg2")
		public List<? extends NotionalQuantityScheduleLeg2Report> getNotionalQuantityScheduleLeg2() {
			return notionalQuantityScheduleLeg2;
		}
		
		@Override
		@RosettaAttribute("actionType")
		public String getActionType() {
			return actionType;
		}
		
		@Override
		@RosettaAttribute("eventType")
		public String getEventType() {
			return eventType;
		}
		
		@Override
		@RosettaAttribute("eventTimestamp")
		public ZonedDateTime getEventTimestamp() {
			return eventTimestamp;
		}
		
		@Override
		@RosettaAttribute("eventIdentifier")
		public String getEventIdentifier() {
			return eventIdentifier;
		}
		
		@Override
		@RosettaAttribute("indexFactor")
		public BigDecimal getIndexFactor() {
			return indexFactor;
		}
		
		@Override
		@RosettaAttribute("embeddedOptionType")
		public String getEmbeddedOptionType() {
			return embeddedOptionType;
		}
		
		@Override
		@RosettaAttribute("upi")
		public String getUpi() {
			return upi;
		}
		
		@Override
		@RosettaAttribute("deliveryType")
		public String getDeliveryType() {
			return deliveryType;
		}
		
		@Override
		@RosettaAttribute("assetClass")
		public String getAssetClass() {
			return assetClass;
		}
		
		@Override
		@RosettaAttribute("underlyingIdentificationType")
		public String getUnderlyingIdentificationType() {
			return underlyingIdentificationType;
		}
		
		@Override
		@RosettaAttribute("underlyingIdentification")
		public String getUnderlyingIdentification() {
			return underlyingIdentification;
		}
		
		@Override
		@RosettaAttribute("underlyingIndexIndicator")
		public String getUnderlyingIndexIndicator() {
			return underlyingIndexIndicator;
		}
		
		@Override
		@RosettaAttribute("nameOfTheUnderlyingIndex")
		public String getNameOfTheUnderlyingIndex() {
			return nameOfTheUnderlyingIndex;
		}
		
		@Override
		@RosettaAttribute("referenceEntity")
		public String getReferenceEntity() {
			return referenceEntity;
		}
		
		@Override
		@RosettaAttribute("indicatorOfTheFloatingRateOfLeg1")
		public String getIndicatorOfTheFloatingRateOfLeg1() {
			return indicatorOfTheFloatingRateOfLeg1;
		}
		
		@Override
		@RosettaAttribute("indicatorOfTheFloatingRateOfLeg2")
		public String getIndicatorOfTheFloatingRateOfLeg2() {
			return indicatorOfTheFloatingRateOfLeg2;
		}
		
		@Override
		@RosettaAttribute("nameOfTheFloatingRateOfLeg1")
		public String getNameOfTheFloatingRateOfLeg1() {
			return nameOfTheFloatingRateOfLeg1;
		}
		
		@Override
		@RosettaAttribute("nameOfTheFloatingRateOfLeg2")
		public String getNameOfTheFloatingRateOfLeg2() {
			return nameOfTheFloatingRateOfLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg1TimePeriod")
		public String getFloatingRateReferencePeriodOfLeg1TimePeriod() {
			return floatingRateReferencePeriodOfLeg1TimePeriod;
		}
		
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg2TimePeriod")
		public String getFloatingRateReferencePeriodOfLeg2TimePeriod() {
			return floatingRateReferencePeriodOfLeg2TimePeriod;
		}
		
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg1Multiplier")
		public Integer getFloatingRateReferencePeriodOfLeg1Multiplier() {
			return floatingRateReferencePeriodOfLeg1Multiplier;
		}
		
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg2Multiplier")
		public Integer getFloatingRateReferencePeriodOfLeg2Multiplier() {
			return floatingRateReferencePeriodOfLeg2Multiplier;
		}
		
		@Override
		@RosettaAttribute("isCrypto")
		public Boolean getIsCrypto() {
			return isCrypto;
		}
		
		@Override
		@RosettaAttribute("maturityDateOfTheUnderlying")
		public Date getMaturityDateOfTheUnderlying() {
			return maturityDateOfTheUnderlying;
		}
		
		@Override
		@RosettaAttribute("seniority")
		public String getSeniority() {
			return seniority;
		}
		
		@Override
		@RosettaAttribute("series")
		public Integer getSeries() {
			return series;
		}
		
		@Override
		@RosettaAttribute("version")
		public Integer getVersion() {
			return version;
		}
		
		@Override
		@RosettaAttribute("cdSIndexAttachmentPoint")
		public BigDecimal getCdSIndexAttachmentPoint() {
			return cdSIndexAttachmentPoint;
		}
		
		@Override
		@RosettaAttribute("cdSIndexDetachmentPoint")
		public BigDecimal getCdSIndexDetachmentPoint() {
			return cdSIndexDetachmentPoint;
		}
		
		@Override
		@RosettaAttribute("customBasketCode")
		public String getCustomBasketCode() {
			return customBasketCode;
		}
		
		@Override
		@RosettaAttribute("identifierOfBasketConstituents")
		public List<String> getIdentifierOfBasketConstituents() {
			return identifierOfBasketConstituents;
		}
		
		@Override
		@RosettaAttribute("basketConstituentIdentifierSource")
		public List<ProductIdTypeEnum> getBasketConstituentIdentifierSource() {
			return basketConstituentIdentifierSource;
		}
		
		@Override
		@RosettaAttribute("contractType")
		public String getContractType() {
			return contractType;
		}
		
		@Override
		@RosettaAttribute("optionStyle")
		public String getOptionStyle() {
			return optionStyle;
		}
		
		@Override
		@RosettaAttribute("optionType")
		public String getOptionType() {
			return optionType;
		}
		
		@Override
		@RosettaAttribute("packageIdentifier")
		public String getPackageIdentifier() {
			return packageIdentifier;
		}
		
		@Override
		@RosettaAttribute("packageTransactionPrice")
		public PackageTransactionPrice getPackageTransactionPrice() {
			return packageTransactionPrice;
		}
		
		@Override
		@RosettaAttribute("packageTransactionPriceCurrency")
		public ISOCurrencyCodeEnum getPackageTransactionPriceCurrency() {
			return packageTransactionPriceCurrency;
		}
		
		@Override
		@RosettaAttribute("packageTransactionPriceNotation")
		public String getPackageTransactionPriceNotation() {
			return packageTransactionPriceNotation;
		}
		
		@Override
		@RosettaAttribute("packageTransactionSpread")
		public PackageTransactionSpread getPackageTransactionSpread() {
			return packageTransactionSpread;
		}
		
		@Override
		@RosettaAttribute("packageTransactionSpreadCurrency")
		public ISOCurrencyCodeEnum getPackageTransactionSpreadCurrency() {
			return packageTransactionSpreadCurrency;
		}
		
		@Override
		@RosettaAttribute("packageTransactionSpreadNotation")
		public String getPackageTransactionSpreadNotation() {
			return packageTransactionSpreadNotation;
		}
		
		@Override
		@RosettaAttribute("technicalRecordId")
		public String getTechnicalRecordId() {
			return technicalRecordId;
		}
		
		@Override
		@RosettaAttribute("nonReportable")
		public JFSANonReportable getNonReportable() {
			return nonReportable;
		}
		
		@Override
		public JFSATransactionReport build() {
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder toBuilder() {
			JFSATransactionReport.JFSATransactionReportBuilder builder = builder();
			setBuilderFields(builder);
			return builder;
		}
		
		protected void setBuilderFields(JFSATransactionReport.JFSATransactionReportBuilder builder) {
			ofNullable(getEffectiveDate()).ifPresent(builder::setEffectiveDate);
			ofNullable(getExpirationDate()).ifPresent(builder::setExpirationDate);
			ofNullable(getEarlyTerminationDate()).ifPresent(builder::setEarlyTerminationDate);
			ofNullable(getReportingTimestamp()).ifPresent(builder::setReportingTimestamp);
			ofNullable(getExecutionTimestamp()).ifPresent(builder::setExecutionTimestamp);
			ofNullable(getEntityResponsibleForReporting()).ifPresent(builder::setEntityResponsibleForReporting);
			ofNullable(getCounterparty1()).ifPresent(builder::setCounterparty1);
			ofNullable(getCounterparty2()).ifPresent(builder::setCounterparty2);
			ofNullable(getCounterparty2IdentifierType()).ifPresent(builder::setCounterparty2IdentifierType);
			ofNullable(getDirection1()).ifPresent(builder::setDirection1);
			ofNullable(getDirection2Leg1()).ifPresent(builder::setDirection2Leg1);
			ofNullable(getDirection2Leg2()).ifPresent(builder::setDirection2Leg2);
			ofNullable(getSubmitterIdentifier()).ifPresent(builder::setSubmitterIdentifier);
			ofNullable(getNewSDRIdentifier()).ifPresent(builder::setNewSDRIdentifier);
			ofNullable(getOriginalSwapSDRIdentifier()).ifPresent(builder::setOriginalSwapSDRIdentifier);
			ofNullable(getCleared()).ifPresent(builder::setCleared);
			ofNullable(getCentralCounterparty()).ifPresent(builder::setCentralCounterparty);
			ofNullable(getClearingMember()).ifPresent(builder::setClearingMember);
			ofNullable(getPlatformIdentifier()).ifPresent(builder::setPlatformIdentifier);
			ofNullable(getConfirmed()).ifPresent(builder::setConfirmed);
			ofNullable(getFinalContractualSettlementDate()).ifPresent(builder::setFinalContractualSettlementDate);
			ofNullable(getConfirmationTimestamp()).ifPresent(builder::setConfirmationTimestamp);
			ofNullable(getSettlementCurrencyLeg1()).ifPresent(builder::setSettlementCurrencyLeg1);
			ofNullable(getSettlementCurrencyLeg2()).ifPresent(builder::setSettlementCurrencyLeg2);
			ofNullable(getClearingAccountOrigin()).ifPresent(builder::setClearingAccountOrigin);
			ofNullable(getOriginalSwapUTI()).ifPresent(builder::setOriginalSwapUTI);
			ofNullable(getClearingReceiptTimestamp()).ifPresent(builder::setClearingReceiptTimestamp);
			ofNullable(getUti()).ifPresent(builder::setUti);
			ofNullable(getUtiProprietary()).ifPresent(builder::setUtiProprietary);
			ofNullable(getPriorUti()).ifPresent(builder::setPriorUti);
			ofNullable(getPriorUtiProprietary()).ifPresent(builder::setPriorUtiProprietary);
			ofNullable(getFixedRateDayCountConventionLeg1()).ifPresent(builder::setFixedRateDayCountConventionLeg1);
			ofNullable(getFloatingRateDayCountConventionOfLeg1()).ifPresent(builder::setFloatingRateDayCountConventionOfLeg1);
			ofNullable(getFixedRateDayCountConventionLeg2()).ifPresent(builder::setFixedRateDayCountConventionLeg2);
			ofNullable(getFloatingRateDayCountConventionOfLeg2()).ifPresent(builder::setFloatingRateDayCountConventionOfLeg2);
			ofNullable(getFixedRatePaymentFrequencyPeriodLeg1()).ifPresent(builder::setFixedRatePaymentFrequencyPeriodLeg1);
			ofNullable(getFixedRatePaymentFrequencyPeriodLeg2()).ifPresent(builder::setFixedRatePaymentFrequencyPeriodLeg2);
			ofNullable(getFloatingRatePaymentFrequencyPeriodLeg1()).ifPresent(builder::setFloatingRatePaymentFrequencyPeriodLeg1);
			ofNullable(getFloatingRatePaymentFrequencyPeriodLeg2()).ifPresent(builder::setFloatingRatePaymentFrequencyPeriodLeg2);
			ofNullable(getFixedRatePaymentFrequencyPeriodMultiplierLeg1()).ifPresent(builder::setFixedRatePaymentFrequencyPeriodMultiplierLeg1);
			ofNullable(getFixedRatePaymentFrequencyPeriodMultiplierLeg2()).ifPresent(builder::setFixedRatePaymentFrequencyPeriodMultiplierLeg2);
			ofNullable(getFloatingRatePaymentFrequencyPeriodMultiplierLeg1()).ifPresent(builder::setFloatingRatePaymentFrequencyPeriodMultiplierLeg1);
			ofNullable(getFloatingRatePaymentFrequencyPeriodMultiplierLeg2()).ifPresent(builder::setFloatingRatePaymentFrequencyPeriodMultiplierLeg2);
			ofNullable(getFixingDateLeg1()).ifPresent(builder::setFixingDateLeg1);
			ofNullable(getFixingDateLeg2()).ifPresent(builder::setFixingDateLeg2);
			ofNullable(getFloatingRateResetFrequencyPeriodLeg1()).ifPresent(builder::setFloatingRateResetFrequencyPeriodLeg1);
			ofNullable(getFloatingRateResetFrequencyPeriodOfLeg2()).ifPresent(builder::setFloatingRateResetFrequencyPeriodOfLeg2);
			ofNullable(getFloatingRateResetFrequencyPeriodMultiplierLeg1()).ifPresent(builder::setFloatingRateResetFrequencyPeriodMultiplierLeg1);
			ofNullable(getFloatingRateResetFrequencyPeriodMultiplierLeg2()).ifPresent(builder::setFloatingRateResetFrequencyPeriodMultiplierLeg2);
			ofNullable(getOtherPayment()).ifPresent(builder::setOtherPayment);
			ofNullable(getValuationAmount()).ifPresent(builder::setValuationAmount);
			ofNullable(getValuationCurrency()).ifPresent(builder::setValuationCurrency);
			ofNullable(getValuationMethod()).ifPresent(builder::setValuationMethod);
			ofNullable(getValuationTimestamp()).ifPresent(builder::setValuationTimestamp);
			ofNullable(getCollateralPortfolioIndicator()).ifPresent(builder::setCollateralPortfolioIndicator);
			ofNullable(getInitialMarginCollateralPortfolioCode()).ifPresent(builder::setInitialMarginCollateralPortfolioCode);
			ofNullable(getVariationMarginCollateralPortfolioCode()).ifPresent(builder::setVariationMarginCollateralPortfolioCode);
			ofNullable(getPrice()).ifPresent(builder::setPrice);
			ofNullable(getPriceNotation()).ifPresent(builder::setPriceNotation);
			ofNullable(getPriceUnitOfMeasure()).ifPresent(builder::setPriceUnitOfMeasure);
			ofNullable(getPriceSchedule()).ifPresent(builder::setPriceSchedule);
			ofNullable(getFixedRateLeg1()).ifPresent(builder::setFixedRateLeg1);
			ofNullable(getFixedRateLeg2()).ifPresent(builder::setFixedRateLeg2);
			ofNullable(getSpreadOfLeg1()).ifPresent(builder::setSpreadOfLeg1);
			ofNullable(getSpreadNotationOfLeg1()).ifPresent(builder::setSpreadNotationOfLeg1);
			ofNullable(getSpreadOfLeg2()).ifPresent(builder::setSpreadOfLeg2);
			ofNullable(getSpreadNotationOfLeg2()).ifPresent(builder::setSpreadNotationOfLeg2);
			ofNullable(getStrikePrice()).ifPresent(builder::setStrikePrice);
			ofNullable(getStrikePriceNotation()).ifPresent(builder::setStrikePriceNotation);
			ofNullable(getStrikePriceSchedule()).ifPresent(builder::setStrikePriceSchedule);
			ofNullable(getOptionPremiumAmount()).ifPresent(builder::setOptionPremiumAmount);
			ofNullable(getOptionPremiumCurrency()).ifPresent(builder::setOptionPremiumCurrency);
			ofNullable(getOptionPremiumPaymentDate()).ifPresent(builder::setOptionPremiumPaymentDate);
			ofNullable(getFirstExerciseDate()).ifPresent(builder::setFirstExerciseDate);
			ofNullable(getExchangeRate()).ifPresent(builder::setExchangeRate);
			ofNullable(getExchangeRateBasis()).ifPresent(builder::setExchangeRateBasis);
			ofNullable(getNotionalLeg1()).ifPresent(builder::setNotionalLeg1);
			ofNullable(getNotionalLeg2()).ifPresent(builder::setNotionalLeg2);
			ofNullable(getCallAmount()).ifPresent(builder::setCallAmount);
			ofNullable(getPutAmount()).ifPresent(builder::setPutAmount);
			ofNullable(getNotionalCurrencyOfLeg1()).ifPresent(builder::setNotionalCurrencyOfLeg1);
			ofNullable(getNotionalCurrencyOfLeg2()).ifPresent(builder::setNotionalCurrencyOfLeg2);
			ofNullable(getCallCurrency()).ifPresent(builder::setCallCurrency);
			ofNullable(getPutCurrency()).ifPresent(builder::setPutCurrency);
			ofNullable(getQuantityUnitOfMeasureLeg1()).ifPresent(builder::setQuantityUnitOfMeasureLeg1);
			ofNullable(getQuantityUnitOfMeasureLeg2()).ifPresent(builder::setQuantityUnitOfMeasureLeg2);
			ofNullable(getNotionalAmountScheduleLeg1()).ifPresent(builder::setNotionalAmountScheduleLeg1);
			ofNullable(getNotionalAmountScheduleLeg2()).ifPresent(builder::setNotionalAmountScheduleLeg2);
			ofNullable(getTotalNotionalQuantityLeg1()).ifPresent(builder::setTotalNotionalQuantityLeg1);
			ofNullable(getTotalNotionalQuantityLeg2()).ifPresent(builder::setTotalNotionalQuantityLeg2);
			ofNullable(getNotionalQuantityScheduleLeg1()).ifPresent(builder::setNotionalQuantityScheduleLeg1);
			ofNullable(getNotionalQuantityScheduleLeg2()).ifPresent(builder::setNotionalQuantityScheduleLeg2);
			ofNullable(getActionType()).ifPresent(builder::setActionType);
			ofNullable(getEventType()).ifPresent(builder::setEventType);
			ofNullable(getEventTimestamp()).ifPresent(builder::setEventTimestamp);
			ofNullable(getEventIdentifier()).ifPresent(builder::setEventIdentifier);
			ofNullable(getIndexFactor()).ifPresent(builder::setIndexFactor);
			ofNullable(getEmbeddedOptionType()).ifPresent(builder::setEmbeddedOptionType);
			ofNullable(getUpi()).ifPresent(builder::setUpi);
			ofNullable(getDeliveryType()).ifPresent(builder::setDeliveryType);
			ofNullable(getAssetClass()).ifPresent(builder::setAssetClass);
			ofNullable(getUnderlyingIdentificationType()).ifPresent(builder::setUnderlyingIdentificationType);
			ofNullable(getUnderlyingIdentification()).ifPresent(builder::setUnderlyingIdentification);
			ofNullable(getUnderlyingIndexIndicator()).ifPresent(builder::setUnderlyingIndexIndicator);
			ofNullable(getNameOfTheUnderlyingIndex()).ifPresent(builder::setNameOfTheUnderlyingIndex);
			ofNullable(getReferenceEntity()).ifPresent(builder::setReferenceEntity);
			ofNullable(getIndicatorOfTheFloatingRateOfLeg1()).ifPresent(builder::setIndicatorOfTheFloatingRateOfLeg1);
			ofNullable(getIndicatorOfTheFloatingRateOfLeg2()).ifPresent(builder::setIndicatorOfTheFloatingRateOfLeg2);
			ofNullable(getNameOfTheFloatingRateOfLeg1()).ifPresent(builder::setNameOfTheFloatingRateOfLeg1);
			ofNullable(getNameOfTheFloatingRateOfLeg2()).ifPresent(builder::setNameOfTheFloatingRateOfLeg2);
			ofNullable(getFloatingRateReferencePeriodOfLeg1TimePeriod()).ifPresent(builder::setFloatingRateReferencePeriodOfLeg1TimePeriod);
			ofNullable(getFloatingRateReferencePeriodOfLeg2TimePeriod()).ifPresent(builder::setFloatingRateReferencePeriodOfLeg2TimePeriod);
			ofNullable(getFloatingRateReferencePeriodOfLeg1Multiplier()).ifPresent(builder::setFloatingRateReferencePeriodOfLeg1Multiplier);
			ofNullable(getFloatingRateReferencePeriodOfLeg2Multiplier()).ifPresent(builder::setFloatingRateReferencePeriodOfLeg2Multiplier);
			ofNullable(getIsCrypto()).ifPresent(builder::setIsCrypto);
			ofNullable(getMaturityDateOfTheUnderlying()).ifPresent(builder::setMaturityDateOfTheUnderlying);
			ofNullable(getSeniority()).ifPresent(builder::setSeniority);
			ofNullable(getSeries()).ifPresent(builder::setSeries);
			ofNullable(getVersion()).ifPresent(builder::setVersion);
			ofNullable(getCdSIndexAttachmentPoint()).ifPresent(builder::setCdSIndexAttachmentPoint);
			ofNullable(getCdSIndexDetachmentPoint()).ifPresent(builder::setCdSIndexDetachmentPoint);
			ofNullable(getCustomBasketCode()).ifPresent(builder::setCustomBasketCode);
			ofNullable(getIdentifierOfBasketConstituents()).ifPresent(builder::setIdentifierOfBasketConstituents);
			ofNullable(getBasketConstituentIdentifierSource()).ifPresent(builder::setBasketConstituentIdentifierSource);
			ofNullable(getContractType()).ifPresent(builder::setContractType);
			ofNullable(getOptionStyle()).ifPresent(builder::setOptionStyle);
			ofNullable(getOptionType()).ifPresent(builder::setOptionType);
			ofNullable(getPackageIdentifier()).ifPresent(builder::setPackageIdentifier);
			ofNullable(getPackageTransactionPrice()).ifPresent(builder::setPackageTransactionPrice);
			ofNullable(getPackageTransactionPriceCurrency()).ifPresent(builder::setPackageTransactionPriceCurrency);
			ofNullable(getPackageTransactionPriceNotation()).ifPresent(builder::setPackageTransactionPriceNotation);
			ofNullable(getPackageTransactionSpread()).ifPresent(builder::setPackageTransactionSpread);
			ofNullable(getPackageTransactionSpreadCurrency()).ifPresent(builder::setPackageTransactionSpreadCurrency);
			ofNullable(getPackageTransactionSpreadNotation()).ifPresent(builder::setPackageTransactionSpreadNotation);
			ofNullable(getTechnicalRecordId()).ifPresent(builder::setTechnicalRecordId);
			ofNullable(getNonReportable()).ifPresent(builder::setNonReportable);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			JFSATransactionReport _that = getType().cast(o);
		
			if (!Objects.equals(effectiveDate, _that.getEffectiveDate())) return false;
			if (!Objects.equals(expirationDate, _that.getExpirationDate())) return false;
			if (!Objects.equals(earlyTerminationDate, _that.getEarlyTerminationDate())) return false;
			if (!Objects.equals(reportingTimestamp, _that.getReportingTimestamp())) return false;
			if (!Objects.equals(executionTimestamp, _that.getExecutionTimestamp())) return false;
			if (!Objects.equals(entityResponsibleForReporting, _that.getEntityResponsibleForReporting())) return false;
			if (!Objects.equals(counterparty1, _that.getCounterparty1())) return false;
			if (!Objects.equals(counterparty2, _that.getCounterparty2())) return false;
			if (!Objects.equals(counterparty2IdentifierType, _that.getCounterparty2IdentifierType())) return false;
			if (!Objects.equals(direction1, _that.getDirection1())) return false;
			if (!Objects.equals(direction2Leg1, _that.getDirection2Leg1())) return false;
			if (!Objects.equals(direction2Leg2, _that.getDirection2Leg2())) return false;
			if (!Objects.equals(submitterIdentifier, _that.getSubmitterIdentifier())) return false;
			if (!Objects.equals(newSDRIdentifier, _that.getNewSDRIdentifier())) return false;
			if (!Objects.equals(originalSwapSDRIdentifier, _that.getOriginalSwapSDRIdentifier())) return false;
			if (!Objects.equals(cleared, _that.getCleared())) return false;
			if (!Objects.equals(centralCounterparty, _that.getCentralCounterparty())) return false;
			if (!Objects.equals(clearingMember, _that.getClearingMember())) return false;
			if (!Objects.equals(platformIdentifier, _that.getPlatformIdentifier())) return false;
			if (!Objects.equals(confirmed, _that.getConfirmed())) return false;
			if (!Objects.equals(finalContractualSettlementDate, _that.getFinalContractualSettlementDate())) return false;
			if (!Objects.equals(confirmationTimestamp, _that.getConfirmationTimestamp())) return false;
			if (!Objects.equals(settlementCurrencyLeg1, _that.getSettlementCurrencyLeg1())) return false;
			if (!Objects.equals(settlementCurrencyLeg2, _that.getSettlementCurrencyLeg2())) return false;
			if (!Objects.equals(clearingAccountOrigin, _that.getClearingAccountOrigin())) return false;
			if (!Objects.equals(originalSwapUTI, _that.getOriginalSwapUTI())) return false;
			if (!Objects.equals(clearingReceiptTimestamp, _that.getClearingReceiptTimestamp())) return false;
			if (!Objects.equals(uti, _that.getUti())) return false;
			if (!Objects.equals(utiProprietary, _that.getUtiProprietary())) return false;
			if (!Objects.equals(priorUti, _that.getPriorUti())) return false;
			if (!Objects.equals(priorUtiProprietary, _that.getPriorUtiProprietary())) return false;
			if (!Objects.equals(fixedRateDayCountConventionLeg1, _that.getFixedRateDayCountConventionLeg1())) return false;
			if (!Objects.equals(floatingRateDayCountConventionOfLeg1, _that.getFloatingRateDayCountConventionOfLeg1())) return false;
			if (!Objects.equals(fixedRateDayCountConventionLeg2, _that.getFixedRateDayCountConventionLeg2())) return false;
			if (!Objects.equals(floatingRateDayCountConventionOfLeg2, _that.getFloatingRateDayCountConventionOfLeg2())) return false;
			if (!Objects.equals(fixedRatePaymentFrequencyPeriodLeg1, _that.getFixedRatePaymentFrequencyPeriodLeg1())) return false;
			if (!Objects.equals(fixedRatePaymentFrequencyPeriodLeg2, _that.getFixedRatePaymentFrequencyPeriodLeg2())) return false;
			if (!Objects.equals(floatingRatePaymentFrequencyPeriodLeg1, _that.getFloatingRatePaymentFrequencyPeriodLeg1())) return false;
			if (!Objects.equals(floatingRatePaymentFrequencyPeriodLeg2, _that.getFloatingRatePaymentFrequencyPeriodLeg2())) return false;
			if (!Objects.equals(fixedRatePaymentFrequencyPeriodMultiplierLeg1, _that.getFixedRatePaymentFrequencyPeriodMultiplierLeg1())) return false;
			if (!Objects.equals(fixedRatePaymentFrequencyPeriodMultiplierLeg2, _that.getFixedRatePaymentFrequencyPeriodMultiplierLeg2())) return false;
			if (!Objects.equals(floatingRatePaymentFrequencyPeriodMultiplierLeg1, _that.getFloatingRatePaymentFrequencyPeriodMultiplierLeg1())) return false;
			if (!Objects.equals(floatingRatePaymentFrequencyPeriodMultiplierLeg2, _that.getFloatingRatePaymentFrequencyPeriodMultiplierLeg2())) return false;
			if (!Objects.equals(fixingDateLeg1, _that.getFixingDateLeg1())) return false;
			if (!Objects.equals(fixingDateLeg2, _that.getFixingDateLeg2())) return false;
			if (!Objects.equals(floatingRateResetFrequencyPeriodLeg1, _that.getFloatingRateResetFrequencyPeriodLeg1())) return false;
			if (!Objects.equals(floatingRateResetFrequencyPeriodOfLeg2, _that.getFloatingRateResetFrequencyPeriodOfLeg2())) return false;
			if (!Objects.equals(floatingRateResetFrequencyPeriodMultiplierLeg1, _that.getFloatingRateResetFrequencyPeriodMultiplierLeg1())) return false;
			if (!Objects.equals(floatingRateResetFrequencyPeriodMultiplierLeg2, _that.getFloatingRateResetFrequencyPeriodMultiplierLeg2())) return false;
			if (!ListEquals.listEquals(otherPayment, _that.getOtherPayment())) return false;
			if (!Objects.equals(valuationAmount, _that.getValuationAmount())) return false;
			if (!Objects.equals(valuationCurrency, _that.getValuationCurrency())) return false;
			if (!Objects.equals(valuationMethod, _that.getValuationMethod())) return false;
			if (!Objects.equals(valuationTimestamp, _that.getValuationTimestamp())) return false;
			if (!Objects.equals(collateralPortfolioIndicator, _that.getCollateralPortfolioIndicator())) return false;
			if (!Objects.equals(initialMarginCollateralPortfolioCode, _that.getInitialMarginCollateralPortfolioCode())) return false;
			if (!Objects.equals(variationMarginCollateralPortfolioCode, _that.getVariationMarginCollateralPortfolioCode())) return false;
			if (!Objects.equals(price, _that.getPrice())) return false;
			if (!Objects.equals(priceNotation, _that.getPriceNotation())) return false;
			if (!Objects.equals(priceUnitOfMeasure, _that.getPriceUnitOfMeasure())) return false;
			if (!ListEquals.listEquals(priceSchedule, _that.getPriceSchedule())) return false;
			if (!Objects.equals(fixedRateLeg1, _that.getFixedRateLeg1())) return false;
			if (!Objects.equals(fixedRateLeg2, _that.getFixedRateLeg2())) return false;
			if (!Objects.equals(spreadOfLeg1, _that.getSpreadOfLeg1())) return false;
			if (!Objects.equals(spreadNotationOfLeg1, _that.getSpreadNotationOfLeg1())) return false;
			if (!Objects.equals(spreadOfLeg2, _that.getSpreadOfLeg2())) return false;
			if (!Objects.equals(spreadNotationOfLeg2, _that.getSpreadNotationOfLeg2())) return false;
			if (!Objects.equals(strikePrice, _that.getStrikePrice())) return false;
			if (!Objects.equals(strikePriceNotation, _that.getStrikePriceNotation())) return false;
			if (!ListEquals.listEquals(strikePriceSchedule, _that.getStrikePriceSchedule())) return false;
			if (!Objects.equals(optionPremiumAmount, _that.getOptionPremiumAmount())) return false;
			if (!Objects.equals(optionPremiumCurrency, _that.getOptionPremiumCurrency())) return false;
			if (!Objects.equals(optionPremiumPaymentDate, _that.getOptionPremiumPaymentDate())) return false;
			if (!Objects.equals(firstExerciseDate, _that.getFirstExerciseDate())) return false;
			if (!Objects.equals(exchangeRate, _that.getExchangeRate())) return false;
			if (!Objects.equals(exchangeRateBasis, _that.getExchangeRateBasis())) return false;
			if (!Objects.equals(notionalLeg1, _that.getNotionalLeg1())) return false;
			if (!Objects.equals(notionalLeg2, _that.getNotionalLeg2())) return false;
			if (!Objects.equals(callAmount, _that.getCallAmount())) return false;
			if (!Objects.equals(putAmount, _that.getPutAmount())) return false;
			if (!Objects.equals(notionalCurrencyOfLeg1, _that.getNotionalCurrencyOfLeg1())) return false;
			if (!Objects.equals(notionalCurrencyOfLeg2, _that.getNotionalCurrencyOfLeg2())) return false;
			if (!Objects.equals(callCurrency, _that.getCallCurrency())) return false;
			if (!Objects.equals(putCurrency, _that.getPutCurrency())) return false;
			if (!Objects.equals(quantityUnitOfMeasureLeg1, _that.getQuantityUnitOfMeasureLeg1())) return false;
			if (!Objects.equals(quantityUnitOfMeasureLeg2, _that.getQuantityUnitOfMeasureLeg2())) return false;
			if (!ListEquals.listEquals(notionalAmountScheduleLeg1, _that.getNotionalAmountScheduleLeg1())) return false;
			if (!ListEquals.listEquals(notionalAmountScheduleLeg2, _that.getNotionalAmountScheduleLeg2())) return false;
			if (!Objects.equals(totalNotionalQuantityLeg1, _that.getTotalNotionalQuantityLeg1())) return false;
			if (!Objects.equals(totalNotionalQuantityLeg2, _that.getTotalNotionalQuantityLeg2())) return false;
			if (!ListEquals.listEquals(notionalQuantityScheduleLeg1, _that.getNotionalQuantityScheduleLeg1())) return false;
			if (!ListEquals.listEquals(notionalQuantityScheduleLeg2, _that.getNotionalQuantityScheduleLeg2())) return false;
			if (!Objects.equals(actionType, _that.getActionType())) return false;
			if (!Objects.equals(eventType, _that.getEventType())) return false;
			if (!Objects.equals(eventTimestamp, _that.getEventTimestamp())) return false;
			if (!Objects.equals(eventIdentifier, _that.getEventIdentifier())) return false;
			if (!Objects.equals(indexFactor, _that.getIndexFactor())) return false;
			if (!Objects.equals(embeddedOptionType, _that.getEmbeddedOptionType())) return false;
			if (!Objects.equals(upi, _that.getUpi())) return false;
			if (!Objects.equals(deliveryType, _that.getDeliveryType())) return false;
			if (!Objects.equals(assetClass, _that.getAssetClass())) return false;
			if (!Objects.equals(underlyingIdentificationType, _that.getUnderlyingIdentificationType())) return false;
			if (!Objects.equals(underlyingIdentification, _that.getUnderlyingIdentification())) return false;
			if (!Objects.equals(underlyingIndexIndicator, _that.getUnderlyingIndexIndicator())) return false;
			if (!Objects.equals(nameOfTheUnderlyingIndex, _that.getNameOfTheUnderlyingIndex())) return false;
			if (!Objects.equals(referenceEntity, _that.getReferenceEntity())) return false;
			if (!Objects.equals(indicatorOfTheFloatingRateOfLeg1, _that.getIndicatorOfTheFloatingRateOfLeg1())) return false;
			if (!Objects.equals(indicatorOfTheFloatingRateOfLeg2, _that.getIndicatorOfTheFloatingRateOfLeg2())) return false;
			if (!Objects.equals(nameOfTheFloatingRateOfLeg1, _that.getNameOfTheFloatingRateOfLeg1())) return false;
			if (!Objects.equals(nameOfTheFloatingRateOfLeg2, _that.getNameOfTheFloatingRateOfLeg2())) return false;
			if (!Objects.equals(floatingRateReferencePeriodOfLeg1TimePeriod, _that.getFloatingRateReferencePeriodOfLeg1TimePeriod())) return false;
			if (!Objects.equals(floatingRateReferencePeriodOfLeg2TimePeriod, _that.getFloatingRateReferencePeriodOfLeg2TimePeriod())) return false;
			if (!Objects.equals(floatingRateReferencePeriodOfLeg1Multiplier, _that.getFloatingRateReferencePeriodOfLeg1Multiplier())) return false;
			if (!Objects.equals(floatingRateReferencePeriodOfLeg2Multiplier, _that.getFloatingRateReferencePeriodOfLeg2Multiplier())) return false;
			if (!Objects.equals(isCrypto, _that.getIsCrypto())) return false;
			if (!Objects.equals(maturityDateOfTheUnderlying, _that.getMaturityDateOfTheUnderlying())) return false;
			if (!Objects.equals(seniority, _that.getSeniority())) return false;
			if (!Objects.equals(series, _that.getSeries())) return false;
			if (!Objects.equals(version, _that.getVersion())) return false;
			if (!Objects.equals(cdSIndexAttachmentPoint, _that.getCdSIndexAttachmentPoint())) return false;
			if (!Objects.equals(cdSIndexDetachmentPoint, _that.getCdSIndexDetachmentPoint())) return false;
			if (!Objects.equals(customBasketCode, _that.getCustomBasketCode())) return false;
			if (!ListEquals.listEquals(identifierOfBasketConstituents, _that.getIdentifierOfBasketConstituents())) return false;
			if (!ListEquals.listEquals(basketConstituentIdentifierSource, _that.getBasketConstituentIdentifierSource())) return false;
			if (!Objects.equals(contractType, _that.getContractType())) return false;
			if (!Objects.equals(optionStyle, _that.getOptionStyle())) return false;
			if (!Objects.equals(optionType, _that.getOptionType())) return false;
			if (!Objects.equals(packageIdentifier, _that.getPackageIdentifier())) return false;
			if (!Objects.equals(packageTransactionPrice, _that.getPackageTransactionPrice())) return false;
			if (!Objects.equals(packageTransactionPriceCurrency, _that.getPackageTransactionPriceCurrency())) return false;
			if (!Objects.equals(packageTransactionPriceNotation, _that.getPackageTransactionPriceNotation())) return false;
			if (!Objects.equals(packageTransactionSpread, _that.getPackageTransactionSpread())) return false;
			if (!Objects.equals(packageTransactionSpreadCurrency, _that.getPackageTransactionSpreadCurrency())) return false;
			if (!Objects.equals(packageTransactionSpreadNotation, _that.getPackageTransactionSpreadNotation())) return false;
			if (!Objects.equals(technicalRecordId, _that.getTechnicalRecordId())) return false;
			if (!Objects.equals(nonReportable, _that.getNonReportable())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (effectiveDate != null ? effectiveDate.hashCode() : 0);
			_result = 31 * _result + (expirationDate != null ? expirationDate.hashCode() : 0);
			_result = 31 * _result + (earlyTerminationDate != null ? earlyTerminationDate.hashCode() : 0);
			_result = 31 * _result + (reportingTimestamp != null ? reportingTimestamp.hashCode() : 0);
			_result = 31 * _result + (executionTimestamp != null ? executionTimestamp.hashCode() : 0);
			_result = 31 * _result + (entityResponsibleForReporting != null ? entityResponsibleForReporting.hashCode() : 0);
			_result = 31 * _result + (counterparty1 != null ? counterparty1.hashCode() : 0);
			_result = 31 * _result + (counterparty2 != null ? counterparty2.hashCode() : 0);
			_result = 31 * _result + (counterparty2IdentifierType != null ? counterparty2IdentifierType.hashCode() : 0);
			_result = 31 * _result + (direction1 != null ? direction1.hashCode() : 0);
			_result = 31 * _result + (direction2Leg1 != null ? direction2Leg1.hashCode() : 0);
			_result = 31 * _result + (direction2Leg2 != null ? direction2Leg2.hashCode() : 0);
			_result = 31 * _result + (submitterIdentifier != null ? submitterIdentifier.hashCode() : 0);
			_result = 31 * _result + (newSDRIdentifier != null ? newSDRIdentifier.hashCode() : 0);
			_result = 31 * _result + (originalSwapSDRIdentifier != null ? originalSwapSDRIdentifier.hashCode() : 0);
			_result = 31 * _result + (cleared != null ? cleared.hashCode() : 0);
			_result = 31 * _result + (centralCounterparty != null ? centralCounterparty.hashCode() : 0);
			_result = 31 * _result + (clearingMember != null ? clearingMember.hashCode() : 0);
			_result = 31 * _result + (platformIdentifier != null ? platformIdentifier.hashCode() : 0);
			_result = 31 * _result + (confirmed != null ? confirmed.hashCode() : 0);
			_result = 31 * _result + (finalContractualSettlementDate != null ? finalContractualSettlementDate.hashCode() : 0);
			_result = 31 * _result + (confirmationTimestamp != null ? confirmationTimestamp.hashCode() : 0);
			_result = 31 * _result + (settlementCurrencyLeg1 != null ? settlementCurrencyLeg1.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (settlementCurrencyLeg2 != null ? settlementCurrencyLeg2.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (clearingAccountOrigin != null ? clearingAccountOrigin.hashCode() : 0);
			_result = 31 * _result + (originalSwapUTI != null ? originalSwapUTI.hashCode() : 0);
			_result = 31 * _result + (clearingReceiptTimestamp != null ? clearingReceiptTimestamp.hashCode() : 0);
			_result = 31 * _result + (uti != null ? uti.hashCode() : 0);
			_result = 31 * _result + (utiProprietary != null ? utiProprietary.hashCode() : 0);
			_result = 31 * _result + (priorUti != null ? priorUti.hashCode() : 0);
			_result = 31 * _result + (priorUtiProprietary != null ? priorUtiProprietary.hashCode() : 0);
			_result = 31 * _result + (fixedRateDayCountConventionLeg1 != null ? fixedRateDayCountConventionLeg1.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (floatingRateDayCountConventionOfLeg1 != null ? floatingRateDayCountConventionOfLeg1.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (fixedRateDayCountConventionLeg2 != null ? fixedRateDayCountConventionLeg2.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (floatingRateDayCountConventionOfLeg2 != null ? floatingRateDayCountConventionOfLeg2.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (fixedRatePaymentFrequencyPeriodLeg1 != null ? fixedRatePaymentFrequencyPeriodLeg1.hashCode() : 0);
			_result = 31 * _result + (fixedRatePaymentFrequencyPeriodLeg2 != null ? fixedRatePaymentFrequencyPeriodLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRatePaymentFrequencyPeriodLeg1 != null ? floatingRatePaymentFrequencyPeriodLeg1.hashCode() : 0);
			_result = 31 * _result + (floatingRatePaymentFrequencyPeriodLeg2 != null ? floatingRatePaymentFrequencyPeriodLeg2.hashCode() : 0);
			_result = 31 * _result + (fixedRatePaymentFrequencyPeriodMultiplierLeg1 != null ? fixedRatePaymentFrequencyPeriodMultiplierLeg1.hashCode() : 0);
			_result = 31 * _result + (fixedRatePaymentFrequencyPeriodMultiplierLeg2 != null ? fixedRatePaymentFrequencyPeriodMultiplierLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRatePaymentFrequencyPeriodMultiplierLeg1 != null ? floatingRatePaymentFrequencyPeriodMultiplierLeg1.hashCode() : 0);
			_result = 31 * _result + (floatingRatePaymentFrequencyPeriodMultiplierLeg2 != null ? floatingRatePaymentFrequencyPeriodMultiplierLeg2.hashCode() : 0);
			_result = 31 * _result + (fixingDateLeg1 != null ? fixingDateLeg1.hashCode() : 0);
			_result = 31 * _result + (fixingDateLeg2 != null ? fixingDateLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRateResetFrequencyPeriodLeg1 != null ? floatingRateResetFrequencyPeriodLeg1.hashCode() : 0);
			_result = 31 * _result + (floatingRateResetFrequencyPeriodOfLeg2 != null ? floatingRateResetFrequencyPeriodOfLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRateResetFrequencyPeriodMultiplierLeg1 != null ? floatingRateResetFrequencyPeriodMultiplierLeg1.hashCode() : 0);
			_result = 31 * _result + (floatingRateResetFrequencyPeriodMultiplierLeg2 != null ? floatingRateResetFrequencyPeriodMultiplierLeg2.hashCode() : 0);
			_result = 31 * _result + (otherPayment != null ? otherPayment.hashCode() : 0);
			_result = 31 * _result + (valuationAmount != null ? valuationAmount.hashCode() : 0);
			_result = 31 * _result + (valuationCurrency != null ? valuationCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (valuationMethod != null ? valuationMethod.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (valuationTimestamp != null ? valuationTimestamp.hashCode() : 0);
			_result = 31 * _result + (collateralPortfolioIndicator != null ? collateralPortfolioIndicator.hashCode() : 0);
			_result = 31 * _result + (initialMarginCollateralPortfolioCode != null ? initialMarginCollateralPortfolioCode.hashCode() : 0);
			_result = 31 * _result + (variationMarginCollateralPortfolioCode != null ? variationMarginCollateralPortfolioCode.hashCode() : 0);
			_result = 31 * _result + (price != null ? price.hashCode() : 0);
			_result = 31 * _result + (priceNotation != null ? priceNotation.hashCode() : 0);
			_result = 31 * _result + (priceUnitOfMeasure != null ? priceUnitOfMeasure.hashCode() : 0);
			_result = 31 * _result + (priceSchedule != null ? priceSchedule.hashCode() : 0);
			_result = 31 * _result + (fixedRateLeg1 != null ? fixedRateLeg1.hashCode() : 0);
			_result = 31 * _result + (fixedRateLeg2 != null ? fixedRateLeg2.hashCode() : 0);
			_result = 31 * _result + (spreadOfLeg1 != null ? spreadOfLeg1.hashCode() : 0);
			_result = 31 * _result + (spreadNotationOfLeg1 != null ? spreadNotationOfLeg1.hashCode() : 0);
			_result = 31 * _result + (spreadOfLeg2 != null ? spreadOfLeg2.hashCode() : 0);
			_result = 31 * _result + (spreadNotationOfLeg2 != null ? spreadNotationOfLeg2.hashCode() : 0);
			_result = 31 * _result + (strikePrice != null ? strikePrice.hashCode() : 0);
			_result = 31 * _result + (strikePriceNotation != null ? strikePriceNotation.hashCode() : 0);
			_result = 31 * _result + (strikePriceSchedule != null ? strikePriceSchedule.hashCode() : 0);
			_result = 31 * _result + (optionPremiumAmount != null ? optionPremiumAmount.hashCode() : 0);
			_result = 31 * _result + (optionPremiumCurrency != null ? optionPremiumCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (optionPremiumPaymentDate != null ? optionPremiumPaymentDate.hashCode() : 0);
			_result = 31 * _result + (firstExerciseDate != null ? firstExerciseDate.hashCode() : 0);
			_result = 31 * _result + (exchangeRate != null ? exchangeRate.hashCode() : 0);
			_result = 31 * _result + (exchangeRateBasis != null ? exchangeRateBasis.hashCode() : 0);
			_result = 31 * _result + (notionalLeg1 != null ? notionalLeg1.hashCode() : 0);
			_result = 31 * _result + (notionalLeg2 != null ? notionalLeg2.hashCode() : 0);
			_result = 31 * _result + (callAmount != null ? callAmount.hashCode() : 0);
			_result = 31 * _result + (putAmount != null ? putAmount.hashCode() : 0);
			_result = 31 * _result + (notionalCurrencyOfLeg1 != null ? notionalCurrencyOfLeg1.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (notionalCurrencyOfLeg2 != null ? notionalCurrencyOfLeg2.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (callCurrency != null ? callCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (putCurrency != null ? putCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (quantityUnitOfMeasureLeg1 != null ? quantityUnitOfMeasureLeg1.hashCode() : 0);
			_result = 31 * _result + (quantityUnitOfMeasureLeg2 != null ? quantityUnitOfMeasureLeg2.hashCode() : 0);
			_result = 31 * _result + (notionalAmountScheduleLeg1 != null ? notionalAmountScheduleLeg1.hashCode() : 0);
			_result = 31 * _result + (notionalAmountScheduleLeg2 != null ? notionalAmountScheduleLeg2.hashCode() : 0);
			_result = 31 * _result + (totalNotionalQuantityLeg1 != null ? totalNotionalQuantityLeg1.hashCode() : 0);
			_result = 31 * _result + (totalNotionalQuantityLeg2 != null ? totalNotionalQuantityLeg2.hashCode() : 0);
			_result = 31 * _result + (notionalQuantityScheduleLeg1 != null ? notionalQuantityScheduleLeg1.hashCode() : 0);
			_result = 31 * _result + (notionalQuantityScheduleLeg2 != null ? notionalQuantityScheduleLeg2.hashCode() : 0);
			_result = 31 * _result + (actionType != null ? actionType.hashCode() : 0);
			_result = 31 * _result + (eventType != null ? eventType.hashCode() : 0);
			_result = 31 * _result + (eventTimestamp != null ? eventTimestamp.hashCode() : 0);
			_result = 31 * _result + (eventIdentifier != null ? eventIdentifier.hashCode() : 0);
			_result = 31 * _result + (indexFactor != null ? indexFactor.hashCode() : 0);
			_result = 31 * _result + (embeddedOptionType != null ? embeddedOptionType.hashCode() : 0);
			_result = 31 * _result + (upi != null ? upi.hashCode() : 0);
			_result = 31 * _result + (deliveryType != null ? deliveryType.hashCode() : 0);
			_result = 31 * _result + (assetClass != null ? assetClass.hashCode() : 0);
			_result = 31 * _result + (underlyingIdentificationType != null ? underlyingIdentificationType.hashCode() : 0);
			_result = 31 * _result + (underlyingIdentification != null ? underlyingIdentification.hashCode() : 0);
			_result = 31 * _result + (underlyingIndexIndicator != null ? underlyingIndexIndicator.hashCode() : 0);
			_result = 31 * _result + (nameOfTheUnderlyingIndex != null ? nameOfTheUnderlyingIndex.hashCode() : 0);
			_result = 31 * _result + (referenceEntity != null ? referenceEntity.hashCode() : 0);
			_result = 31 * _result + (indicatorOfTheFloatingRateOfLeg1 != null ? indicatorOfTheFloatingRateOfLeg1.hashCode() : 0);
			_result = 31 * _result + (indicatorOfTheFloatingRateOfLeg2 != null ? indicatorOfTheFloatingRateOfLeg2.hashCode() : 0);
			_result = 31 * _result + (nameOfTheFloatingRateOfLeg1 != null ? nameOfTheFloatingRateOfLeg1.hashCode() : 0);
			_result = 31 * _result + (nameOfTheFloatingRateOfLeg2 != null ? nameOfTheFloatingRateOfLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRateReferencePeriodOfLeg1TimePeriod != null ? floatingRateReferencePeriodOfLeg1TimePeriod.hashCode() : 0);
			_result = 31 * _result + (floatingRateReferencePeriodOfLeg2TimePeriod != null ? floatingRateReferencePeriodOfLeg2TimePeriod.hashCode() : 0);
			_result = 31 * _result + (floatingRateReferencePeriodOfLeg1Multiplier != null ? floatingRateReferencePeriodOfLeg1Multiplier.hashCode() : 0);
			_result = 31 * _result + (floatingRateReferencePeriodOfLeg2Multiplier != null ? floatingRateReferencePeriodOfLeg2Multiplier.hashCode() : 0);
			_result = 31 * _result + (isCrypto != null ? isCrypto.hashCode() : 0);
			_result = 31 * _result + (maturityDateOfTheUnderlying != null ? maturityDateOfTheUnderlying.hashCode() : 0);
			_result = 31 * _result + (seniority != null ? seniority.hashCode() : 0);
			_result = 31 * _result + (series != null ? series.hashCode() : 0);
			_result = 31 * _result + (version != null ? version.hashCode() : 0);
			_result = 31 * _result + (cdSIndexAttachmentPoint != null ? cdSIndexAttachmentPoint.hashCode() : 0);
			_result = 31 * _result + (cdSIndexDetachmentPoint != null ? cdSIndexDetachmentPoint.hashCode() : 0);
			_result = 31 * _result + (customBasketCode != null ? customBasketCode.hashCode() : 0);
			_result = 31 * _result + (identifierOfBasketConstituents != null ? identifierOfBasketConstituents.hashCode() : 0);
			_result = 31 * _result + (basketConstituentIdentifierSource != null ? basketConstituentIdentifierSource.stream().map(Object::getClass).map(Class::getName).mapToInt(String::hashCode).sum() : 0);
			_result = 31 * _result + (contractType != null ? contractType.hashCode() : 0);
			_result = 31 * _result + (optionStyle != null ? optionStyle.hashCode() : 0);
			_result = 31 * _result + (optionType != null ? optionType.hashCode() : 0);
			_result = 31 * _result + (packageIdentifier != null ? packageIdentifier.hashCode() : 0);
			_result = 31 * _result + (packageTransactionPrice != null ? packageTransactionPrice.hashCode() : 0);
			_result = 31 * _result + (packageTransactionPriceCurrency != null ? packageTransactionPriceCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (packageTransactionPriceNotation != null ? packageTransactionPriceNotation.hashCode() : 0);
			_result = 31 * _result + (packageTransactionSpread != null ? packageTransactionSpread.hashCode() : 0);
			_result = 31 * _result + (packageTransactionSpreadCurrency != null ? packageTransactionSpreadCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (packageTransactionSpreadNotation != null ? packageTransactionSpreadNotation.hashCode() : 0);
			_result = 31 * _result + (technicalRecordId != null ? technicalRecordId.hashCode() : 0);
			_result = 31 * _result + (nonReportable != null ? nonReportable.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "JFSATransactionReport {" +
				"effectiveDate=" + this.effectiveDate + ", " +
				"expirationDate=" + this.expirationDate + ", " +
				"earlyTerminationDate=" + this.earlyTerminationDate + ", " +
				"reportingTimestamp=" + this.reportingTimestamp + ", " +
				"executionTimestamp=" + this.executionTimestamp + ", " +
				"entityResponsibleForReporting=" + this.entityResponsibleForReporting + ", " +
				"counterparty1=" + this.counterparty1 + ", " +
				"counterparty2=" + this.counterparty2 + ", " +
				"counterparty2IdentifierType=" + this.counterparty2IdentifierType + ", " +
				"direction1=" + this.direction1 + ", " +
				"direction2Leg1=" + this.direction2Leg1 + ", " +
				"direction2Leg2=" + this.direction2Leg2 + ", " +
				"submitterIdentifier=" + this.submitterIdentifier + ", " +
				"newSDRIdentifier=" + this.newSDRIdentifier + ", " +
				"originalSwapSDRIdentifier=" + this.originalSwapSDRIdentifier + ", " +
				"cleared=" + this.cleared + ", " +
				"centralCounterparty=" + this.centralCounterparty + ", " +
				"clearingMember=" + this.clearingMember + ", " +
				"platformIdentifier=" + this.platformIdentifier + ", " +
				"confirmed=" + this.confirmed + ", " +
				"finalContractualSettlementDate=" + this.finalContractualSettlementDate + ", " +
				"confirmationTimestamp=" + this.confirmationTimestamp + ", " +
				"settlementCurrencyLeg1=" + this.settlementCurrencyLeg1 + ", " +
				"settlementCurrencyLeg2=" + this.settlementCurrencyLeg2 + ", " +
				"clearingAccountOrigin=" + this.clearingAccountOrigin + ", " +
				"originalSwapUTI=" + this.originalSwapUTI + ", " +
				"clearingReceiptTimestamp=" + this.clearingReceiptTimestamp + ", " +
				"uti=" + this.uti + ", " +
				"utiProprietary=" + this.utiProprietary + ", " +
				"priorUti=" + this.priorUti + ", " +
				"priorUtiProprietary=" + this.priorUtiProprietary + ", " +
				"fixedRateDayCountConventionLeg1=" + this.fixedRateDayCountConventionLeg1 + ", " +
				"floatingRateDayCountConventionOfLeg1=" + this.floatingRateDayCountConventionOfLeg1 + ", " +
				"fixedRateDayCountConventionLeg2=" + this.fixedRateDayCountConventionLeg2 + ", " +
				"floatingRateDayCountConventionOfLeg2=" + this.floatingRateDayCountConventionOfLeg2 + ", " +
				"fixedRatePaymentFrequencyPeriodLeg1=" + this.fixedRatePaymentFrequencyPeriodLeg1 + ", " +
				"fixedRatePaymentFrequencyPeriodLeg2=" + this.fixedRatePaymentFrequencyPeriodLeg2 + ", " +
				"floatingRatePaymentFrequencyPeriodLeg1=" + this.floatingRatePaymentFrequencyPeriodLeg1 + ", " +
				"floatingRatePaymentFrequencyPeriodLeg2=" + this.floatingRatePaymentFrequencyPeriodLeg2 + ", " +
				"fixedRatePaymentFrequencyPeriodMultiplierLeg1=" + this.fixedRatePaymentFrequencyPeriodMultiplierLeg1 + ", " +
				"fixedRatePaymentFrequencyPeriodMultiplierLeg2=" + this.fixedRatePaymentFrequencyPeriodMultiplierLeg2 + ", " +
				"floatingRatePaymentFrequencyPeriodMultiplierLeg1=" + this.floatingRatePaymentFrequencyPeriodMultiplierLeg1 + ", " +
				"floatingRatePaymentFrequencyPeriodMultiplierLeg2=" + this.floatingRatePaymentFrequencyPeriodMultiplierLeg2 + ", " +
				"fixingDateLeg1=" + this.fixingDateLeg1 + ", " +
				"fixingDateLeg2=" + this.fixingDateLeg2 + ", " +
				"floatingRateResetFrequencyPeriodLeg1=" + this.floatingRateResetFrequencyPeriodLeg1 + ", " +
				"floatingRateResetFrequencyPeriodOfLeg2=" + this.floatingRateResetFrequencyPeriodOfLeg2 + ", " +
				"floatingRateResetFrequencyPeriodMultiplierLeg1=" + this.floatingRateResetFrequencyPeriodMultiplierLeg1 + ", " +
				"floatingRateResetFrequencyPeriodMultiplierLeg2=" + this.floatingRateResetFrequencyPeriodMultiplierLeg2 + ", " +
				"otherPayment=" + this.otherPayment + ", " +
				"valuationAmount=" + this.valuationAmount + ", " +
				"valuationCurrency=" + this.valuationCurrency + ", " +
				"valuationMethod=" + this.valuationMethod + ", " +
				"valuationTimestamp=" + this.valuationTimestamp + ", " +
				"collateralPortfolioIndicator=" + this.collateralPortfolioIndicator + ", " +
				"initialMarginCollateralPortfolioCode=" + this.initialMarginCollateralPortfolioCode + ", " +
				"variationMarginCollateralPortfolioCode=" + this.variationMarginCollateralPortfolioCode + ", " +
				"price=" + this.price + ", " +
				"priceNotation=" + this.priceNotation + ", " +
				"priceUnitOfMeasure=" + this.priceUnitOfMeasure + ", " +
				"priceSchedule=" + this.priceSchedule + ", " +
				"fixedRateLeg1=" + this.fixedRateLeg1 + ", " +
				"fixedRateLeg2=" + this.fixedRateLeg2 + ", " +
				"spreadOfLeg1=" + this.spreadOfLeg1 + ", " +
				"spreadNotationOfLeg1=" + this.spreadNotationOfLeg1 + ", " +
				"spreadOfLeg2=" + this.spreadOfLeg2 + ", " +
				"spreadNotationOfLeg2=" + this.spreadNotationOfLeg2 + ", " +
				"strikePrice=" + this.strikePrice + ", " +
				"strikePriceNotation=" + this.strikePriceNotation + ", " +
				"strikePriceSchedule=" + this.strikePriceSchedule + ", " +
				"optionPremiumAmount=" + this.optionPremiumAmount + ", " +
				"optionPremiumCurrency=" + this.optionPremiumCurrency + ", " +
				"optionPremiumPaymentDate=" + this.optionPremiumPaymentDate + ", " +
				"firstExerciseDate=" + this.firstExerciseDate + ", " +
				"exchangeRate=" + this.exchangeRate + ", " +
				"exchangeRateBasis=" + this.exchangeRateBasis + ", " +
				"notionalLeg1=" + this.notionalLeg1 + ", " +
				"notionalLeg2=" + this.notionalLeg2 + ", " +
				"callAmount=" + this.callAmount + ", " +
				"putAmount=" + this.putAmount + ", " +
				"notionalCurrencyOfLeg1=" + this.notionalCurrencyOfLeg1 + ", " +
				"notionalCurrencyOfLeg2=" + this.notionalCurrencyOfLeg2 + ", " +
				"callCurrency=" + this.callCurrency + ", " +
				"putCurrency=" + this.putCurrency + ", " +
				"quantityUnitOfMeasureLeg1=" + this.quantityUnitOfMeasureLeg1 + ", " +
				"quantityUnitOfMeasureLeg2=" + this.quantityUnitOfMeasureLeg2 + ", " +
				"notionalAmountScheduleLeg1=" + this.notionalAmountScheduleLeg1 + ", " +
				"notionalAmountScheduleLeg2=" + this.notionalAmountScheduleLeg2 + ", " +
				"totalNotionalQuantityLeg1=" + this.totalNotionalQuantityLeg1 + ", " +
				"totalNotionalQuantityLeg2=" + this.totalNotionalQuantityLeg2 + ", " +
				"notionalQuantityScheduleLeg1=" + this.notionalQuantityScheduleLeg1 + ", " +
				"notionalQuantityScheduleLeg2=" + this.notionalQuantityScheduleLeg2 + ", " +
				"actionType=" + this.actionType + ", " +
				"eventType=" + this.eventType + ", " +
				"eventTimestamp=" + this.eventTimestamp + ", " +
				"eventIdentifier=" + this.eventIdentifier + ", " +
				"indexFactor=" + this.indexFactor + ", " +
				"embeddedOptionType=" + this.embeddedOptionType + ", " +
				"upi=" + this.upi + ", " +
				"deliveryType=" + this.deliveryType + ", " +
				"assetClass=" + this.assetClass + ", " +
				"underlyingIdentificationType=" + this.underlyingIdentificationType + ", " +
				"underlyingIdentification=" + this.underlyingIdentification + ", " +
				"underlyingIndexIndicator=" + this.underlyingIndexIndicator + ", " +
				"nameOfTheUnderlyingIndex=" + this.nameOfTheUnderlyingIndex + ", " +
				"referenceEntity=" + this.referenceEntity + ", " +
				"indicatorOfTheFloatingRateOfLeg1=" + this.indicatorOfTheFloatingRateOfLeg1 + ", " +
				"indicatorOfTheFloatingRateOfLeg2=" + this.indicatorOfTheFloatingRateOfLeg2 + ", " +
				"nameOfTheFloatingRateOfLeg1=" + this.nameOfTheFloatingRateOfLeg1 + ", " +
				"nameOfTheFloatingRateOfLeg2=" + this.nameOfTheFloatingRateOfLeg2 + ", " +
				"floatingRateReferencePeriodOfLeg1TimePeriod=" + this.floatingRateReferencePeriodOfLeg1TimePeriod + ", " +
				"floatingRateReferencePeriodOfLeg2TimePeriod=" + this.floatingRateReferencePeriodOfLeg2TimePeriod + ", " +
				"floatingRateReferencePeriodOfLeg1Multiplier=" + this.floatingRateReferencePeriodOfLeg1Multiplier + ", " +
				"floatingRateReferencePeriodOfLeg2Multiplier=" + this.floatingRateReferencePeriodOfLeg2Multiplier + ", " +
				"isCrypto=" + this.isCrypto + ", " +
				"maturityDateOfTheUnderlying=" + this.maturityDateOfTheUnderlying + ", " +
				"seniority=" + this.seniority + ", " +
				"series=" + this.series + ", " +
				"version=" + this.version + ", " +
				"cdSIndexAttachmentPoint=" + this.cdSIndexAttachmentPoint + ", " +
				"cdSIndexDetachmentPoint=" + this.cdSIndexDetachmentPoint + ", " +
				"customBasketCode=" + this.customBasketCode + ", " +
				"identifierOfBasketConstituents=" + this.identifierOfBasketConstituents + ", " +
				"basketConstituentIdentifierSource=" + this.basketConstituentIdentifierSource + ", " +
				"contractType=" + this.contractType + ", " +
				"optionStyle=" + this.optionStyle + ", " +
				"optionType=" + this.optionType + ", " +
				"packageIdentifier=" + this.packageIdentifier + ", " +
				"packageTransactionPrice=" + this.packageTransactionPrice + ", " +
				"packageTransactionPriceCurrency=" + this.packageTransactionPriceCurrency + ", " +
				"packageTransactionPriceNotation=" + this.packageTransactionPriceNotation + ", " +
				"packageTransactionSpread=" + this.packageTransactionSpread + ", " +
				"packageTransactionSpreadCurrency=" + this.packageTransactionSpreadCurrency + ", " +
				"packageTransactionSpreadNotation=" + this.packageTransactionSpreadNotation + ", " +
				"technicalRecordId=" + this.technicalRecordId + ", " +
				"nonReportable=" + this.nonReportable +
			'}';
		}
	}

	/*********************** Builder Implementation of JFSATransactionReport  ***********************/
	class JFSATransactionReportBuilderImpl implements JFSATransactionReport.JFSATransactionReportBuilder {
	
		protected Date effectiveDate;
		protected Date expirationDate;
		protected Date earlyTerminationDate;
		protected ZonedDateTime reportingTimestamp;
		protected ZonedDateTime executionTimestamp;
		protected String entityResponsibleForReporting;
		protected String counterparty1;
		protected String counterparty2;
		protected Boolean counterparty2IdentifierType;
		protected String direction1;
		protected String direction2Leg1;
		protected String direction2Leg2;
		protected String submitterIdentifier;
		protected String newSDRIdentifier;
		protected String originalSwapSDRIdentifier;
		protected String cleared;
		protected String centralCounterparty;
		protected String clearingMember;
		protected String platformIdentifier;
		protected String confirmed;
		protected Date finalContractualSettlementDate;
		protected ZonedDateTime confirmationTimestamp;
		protected ISOCurrencyCodeEnum settlementCurrencyLeg1;
		protected ISOCurrencyCodeEnum settlementCurrencyLeg2;
		protected String clearingAccountOrigin;
		protected String originalSwapUTI;
		protected ZonedDateTime clearingReceiptTimestamp;
		protected String uti;
		protected String utiProprietary;
		protected String priorUti;
		protected String priorUtiProprietary;
		protected InterestComputationMethod4Code fixedRateDayCountConventionLeg1;
		protected InterestComputationMethod4Code floatingRateDayCountConventionOfLeg1;
		protected InterestComputationMethod4Code fixedRateDayCountConventionLeg2;
		protected InterestComputationMethod4Code floatingRateDayCountConventionOfLeg2;
		protected String fixedRatePaymentFrequencyPeriodLeg1;
		protected String fixedRatePaymentFrequencyPeriodLeg2;
		protected String floatingRatePaymentFrequencyPeriodLeg1;
		protected String floatingRatePaymentFrequencyPeriodLeg2;
		protected Integer fixedRatePaymentFrequencyPeriodMultiplierLeg1;
		protected Integer fixedRatePaymentFrequencyPeriodMultiplierLeg2;
		protected Integer floatingRatePaymentFrequencyPeriodMultiplierLeg1;
		protected Integer floatingRatePaymentFrequencyPeriodMultiplierLeg2;
		protected ZonedDateTime fixingDateLeg1;
		protected ZonedDateTime fixingDateLeg2;
		protected String floatingRateResetFrequencyPeriodLeg1;
		protected String floatingRateResetFrequencyPeriodOfLeg2;
		protected Integer floatingRateResetFrequencyPeriodMultiplierLeg1;
		protected Integer floatingRateResetFrequencyPeriodMultiplierLeg2;
		protected List<OtherPaymentReport.OtherPaymentReportBuilder> otherPayment = new ArrayList<>();
		protected BigDecimal valuationAmount;
		protected ISOCurrencyCodeEnum valuationCurrency;
		protected ValuationType1Code valuationMethod;
		protected ZonedDateTime valuationTimestamp;
		protected Boolean collateralPortfolioIndicator;
		protected String initialMarginCollateralPortfolioCode;
		protected String variationMarginCollateralPortfolioCode;
		protected PriceAndNotation.PriceAndNotationBuilder price;
		protected String priceNotation;
		protected String priceUnitOfMeasure;
		protected List<PriceScheduleReport.PriceScheduleReportBuilder> priceSchedule = new ArrayList<>();
		protected BigDecimal fixedRateLeg1;
		protected BigDecimal fixedRateLeg2;
		protected SpreadOfLeg1.SpreadOfLeg1Builder spreadOfLeg1;
		protected String spreadNotationOfLeg1;
		protected SpreadOfLeg2.SpreadOfLeg2Builder spreadOfLeg2;
		protected String spreadNotationOfLeg2;
		protected StrikePriceAndNotation.StrikePriceAndNotationBuilder strikePrice;
		protected String strikePriceNotation;
		protected List<StrikePriceScheduleReport.StrikePriceScheduleReportBuilder> strikePriceSchedule = new ArrayList<>();
		protected BigDecimal optionPremiumAmount;
		protected ISOCurrencyCodeEnum optionPremiumCurrency;
		protected Date optionPremiumPaymentDate;
		protected Date firstExerciseDate;
		protected BigDecimal exchangeRate;
		protected String exchangeRateBasis;
		protected BigDecimal notionalLeg1;
		protected BigDecimal notionalLeg2;
		protected BigDecimal callAmount;
		protected BigDecimal putAmount;
		protected ISOCurrencyCodeEnum notionalCurrencyOfLeg1;
		protected ISOCurrencyCodeEnum notionalCurrencyOfLeg2;
		protected ISOCurrencyCodeEnum callCurrency;
		protected ISOCurrencyCodeEnum putCurrency;
		protected String quantityUnitOfMeasureLeg1;
		protected String quantityUnitOfMeasureLeg2;
		protected List<NotionalAmountScheduleLeg1Report.NotionalAmountScheduleLeg1ReportBuilder> notionalAmountScheduleLeg1 = new ArrayList<>();
		protected List<NotionalAmountScheduleLeg2Report.NotionalAmountScheduleLeg2ReportBuilder> notionalAmountScheduleLeg2 = new ArrayList<>();
		protected BigDecimal totalNotionalQuantityLeg1;
		protected BigDecimal totalNotionalQuantityLeg2;
		protected List<NotionalQuantityScheduleLeg1Report.NotionalQuantityScheduleLeg1ReportBuilder> notionalQuantityScheduleLeg1 = new ArrayList<>();
		protected List<NotionalQuantityScheduleLeg2Report.NotionalQuantityScheduleLeg2ReportBuilder> notionalQuantityScheduleLeg2 = new ArrayList<>();
		protected String actionType;
		protected String eventType;
		protected ZonedDateTime eventTimestamp;
		protected String eventIdentifier;
		protected BigDecimal indexFactor;
		protected String embeddedOptionType;
		protected String upi;
		protected String deliveryType;
		protected String assetClass;
		protected String underlyingIdentificationType;
		protected String underlyingIdentification;
		protected String underlyingIndexIndicator;
		protected String nameOfTheUnderlyingIndex;
		protected String referenceEntity;
		protected String indicatorOfTheFloatingRateOfLeg1;
		protected String indicatorOfTheFloatingRateOfLeg2;
		protected String nameOfTheFloatingRateOfLeg1;
		protected String nameOfTheFloatingRateOfLeg2;
		protected String floatingRateReferencePeriodOfLeg1TimePeriod;
		protected String floatingRateReferencePeriodOfLeg2TimePeriod;
		protected Integer floatingRateReferencePeriodOfLeg1Multiplier;
		protected Integer floatingRateReferencePeriodOfLeg2Multiplier;
		protected Boolean isCrypto;
		protected Date maturityDateOfTheUnderlying;
		protected String seniority;
		protected Integer series;
		protected Integer version;
		protected BigDecimal cdSIndexAttachmentPoint;
		protected BigDecimal cdSIndexDetachmentPoint;
		protected String customBasketCode;
		protected List<String> identifierOfBasketConstituents = new ArrayList<>();
		protected List<ProductIdTypeEnum> basketConstituentIdentifierSource = new ArrayList<>();
		protected String contractType;
		protected String optionStyle;
		protected String optionType;
		protected String packageIdentifier;
		protected PackageTransactionPrice.PackageTransactionPriceBuilder packageTransactionPrice;
		protected ISOCurrencyCodeEnum packageTransactionPriceCurrency;
		protected String packageTransactionPriceNotation;
		protected PackageTransactionSpread.PackageTransactionSpreadBuilder packageTransactionSpread;
		protected ISOCurrencyCodeEnum packageTransactionSpreadCurrency;
		protected String packageTransactionSpreadNotation;
		protected String technicalRecordId;
		protected JFSANonReportable.JFSANonReportableBuilder nonReportable;
	
		public JFSATransactionReportBuilderImpl() {
		}
	
		@Override
		@RosettaAttribute("effectiveDate")
		public Date getEffectiveDate() {
			return effectiveDate;
		}
		
		@Override
		@RosettaAttribute("expirationDate")
		public Date getExpirationDate() {
			return expirationDate;
		}
		
		@Override
		@RosettaAttribute("earlyTerminationDate")
		public Date getEarlyTerminationDate() {
			return earlyTerminationDate;
		}
		
		@Override
		@RosettaAttribute("reportingTimestamp")
		public ZonedDateTime getReportingTimestamp() {
			return reportingTimestamp;
		}
		
		@Override
		@RosettaAttribute("executionTimestamp")
		public ZonedDateTime getExecutionTimestamp() {
			return executionTimestamp;
		}
		
		@Override
		@RosettaAttribute("entityResponsibleForReporting")
		public String getEntityResponsibleForReporting() {
			return entityResponsibleForReporting;
		}
		
		@Override
		@RosettaAttribute("counterparty1")
		public String getCounterparty1() {
			return counterparty1;
		}
		
		@Override
		@RosettaAttribute("counterparty2")
		public String getCounterparty2() {
			return counterparty2;
		}
		
		@Override
		@RosettaAttribute("counterparty2IdentifierType")
		public Boolean getCounterparty2IdentifierType() {
			return counterparty2IdentifierType;
		}
		
		@Override
		@RosettaAttribute("direction1")
		public String getDirection1() {
			return direction1;
		}
		
		@Override
		@RosettaAttribute("direction2Leg1")
		public String getDirection2Leg1() {
			return direction2Leg1;
		}
		
		@Override
		@RosettaAttribute("direction2Leg2")
		public String getDirection2Leg2() {
			return direction2Leg2;
		}
		
		@Override
		@RosettaAttribute("submitterIdentifier")
		public String getSubmitterIdentifier() {
			return submitterIdentifier;
		}
		
		@Override
		@RosettaAttribute("newSDRIdentifier")
		public String getNewSDRIdentifier() {
			return newSDRIdentifier;
		}
		
		@Override
		@RosettaAttribute("originalSwapSDRIdentifier")
		public String getOriginalSwapSDRIdentifier() {
			return originalSwapSDRIdentifier;
		}
		
		@Override
		@RosettaAttribute("cleared")
		public String getCleared() {
			return cleared;
		}
		
		@Override
		@RosettaAttribute("centralCounterparty")
		public String getCentralCounterparty() {
			return centralCounterparty;
		}
		
		@Override
		@RosettaAttribute("clearingMember")
		public String getClearingMember() {
			return clearingMember;
		}
		
		@Override
		@RosettaAttribute("platformIdentifier")
		public String getPlatformIdentifier() {
			return platformIdentifier;
		}
		
		@Override
		@RosettaAttribute("confirmed")
		public String getConfirmed() {
			return confirmed;
		}
		
		@Override
		@RosettaAttribute("finalContractualSettlementDate")
		public Date getFinalContractualSettlementDate() {
			return finalContractualSettlementDate;
		}
		
		@Override
		@RosettaAttribute("confirmationTimestamp")
		public ZonedDateTime getConfirmationTimestamp() {
			return confirmationTimestamp;
		}
		
		@Override
		@RosettaAttribute("settlementCurrencyLeg1")
		public ISOCurrencyCodeEnum getSettlementCurrencyLeg1() {
			return settlementCurrencyLeg1;
		}
		
		@Override
		@RosettaAttribute("settlementCurrencyLeg2")
		public ISOCurrencyCodeEnum getSettlementCurrencyLeg2() {
			return settlementCurrencyLeg2;
		}
		
		@Override
		@RosettaAttribute("clearingAccountOrigin")
		public String getClearingAccountOrigin() {
			return clearingAccountOrigin;
		}
		
		@Override
		@RosettaAttribute("originalSwapUTI")
		public String getOriginalSwapUTI() {
			return originalSwapUTI;
		}
		
		@Override
		@RosettaAttribute("clearingReceiptTimestamp")
		public ZonedDateTime getClearingReceiptTimestamp() {
			return clearingReceiptTimestamp;
		}
		
		@Override
		@RosettaAttribute("uti")
		public String getUti() {
			return uti;
		}
		
		@Override
		@RosettaAttribute("utiProprietary")
		public String getUtiProprietary() {
			return utiProprietary;
		}
		
		@Override
		@RosettaAttribute("priorUti")
		public String getPriorUti() {
			return priorUti;
		}
		
		@Override
		@RosettaAttribute("priorUtiProprietary")
		public String getPriorUtiProprietary() {
			return priorUtiProprietary;
		}
		
		@Override
		@RosettaAttribute("fixedRateDayCountConventionLeg1")
		public InterestComputationMethod4Code getFixedRateDayCountConventionLeg1() {
			return fixedRateDayCountConventionLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRateDayCountConventionOfLeg1")
		public InterestComputationMethod4Code getFloatingRateDayCountConventionOfLeg1() {
			return floatingRateDayCountConventionOfLeg1;
		}
		
		@Override
		@RosettaAttribute("fixedRateDayCountConventionLeg2")
		public InterestComputationMethod4Code getFixedRateDayCountConventionLeg2() {
			return fixedRateDayCountConventionLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRateDayCountConventionOfLeg2")
		public InterestComputationMethod4Code getFloatingRateDayCountConventionOfLeg2() {
			return floatingRateDayCountConventionOfLeg2;
		}
		
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodLeg1")
		public String getFixedRatePaymentFrequencyPeriodLeg1() {
			return fixedRatePaymentFrequencyPeriodLeg1;
		}
		
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodLeg2")
		public String getFixedRatePaymentFrequencyPeriodLeg2() {
			return fixedRatePaymentFrequencyPeriodLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodLeg1")
		public String getFloatingRatePaymentFrequencyPeriodLeg1() {
			return floatingRatePaymentFrequencyPeriodLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodLeg2")
		public String getFloatingRatePaymentFrequencyPeriodLeg2() {
			return floatingRatePaymentFrequencyPeriodLeg2;
		}
		
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodMultiplierLeg1")
		public Integer getFixedRatePaymentFrequencyPeriodMultiplierLeg1() {
			return fixedRatePaymentFrequencyPeriodMultiplierLeg1;
		}
		
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodMultiplierLeg2")
		public Integer getFixedRatePaymentFrequencyPeriodMultiplierLeg2() {
			return fixedRatePaymentFrequencyPeriodMultiplierLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodMultiplierLeg1")
		public Integer getFloatingRatePaymentFrequencyPeriodMultiplierLeg1() {
			return floatingRatePaymentFrequencyPeriodMultiplierLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodMultiplierLeg2")
		public Integer getFloatingRatePaymentFrequencyPeriodMultiplierLeg2() {
			return floatingRatePaymentFrequencyPeriodMultiplierLeg2;
		}
		
		@Override
		@RosettaAttribute("fixingDateLeg1")
		public ZonedDateTime getFixingDateLeg1() {
			return fixingDateLeg1;
		}
		
		@Override
		@RosettaAttribute("fixingDateLeg2")
		public ZonedDateTime getFixingDateLeg2() {
			return fixingDateLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodLeg1")
		public String getFloatingRateResetFrequencyPeriodLeg1() {
			return floatingRateResetFrequencyPeriodLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodOfLeg2")
		public String getFloatingRateResetFrequencyPeriodOfLeg2() {
			return floatingRateResetFrequencyPeriodOfLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodMultiplierLeg1")
		public Integer getFloatingRateResetFrequencyPeriodMultiplierLeg1() {
			return floatingRateResetFrequencyPeriodMultiplierLeg1;
		}
		
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodMultiplierLeg2")
		public Integer getFloatingRateResetFrequencyPeriodMultiplierLeg2() {
			return floatingRateResetFrequencyPeriodMultiplierLeg2;
		}
		
		@Override
		@RosettaAttribute("otherPayment")
		public List<? extends OtherPaymentReport.OtherPaymentReportBuilder> getOtherPayment() {
			return otherPayment;
		}
		
		public OtherPaymentReport.OtherPaymentReportBuilder getOrCreateOtherPayment(int _index) {
		
			if (otherPayment==null) {
				this.otherPayment = new ArrayList<>();
			}
			OtherPaymentReport.OtherPaymentReportBuilder result;
			return getIndex(otherPayment, _index, () -> {
						OtherPaymentReport.OtherPaymentReportBuilder newOtherPayment = OtherPaymentReport.builder();
						return newOtherPayment;
					});
		}
		
		@Override
		@RosettaAttribute("valuationAmount")
		public BigDecimal getValuationAmount() {
			return valuationAmount;
		}
		
		@Override
		@RosettaAttribute("valuationCurrency")
		public ISOCurrencyCodeEnum getValuationCurrency() {
			return valuationCurrency;
		}
		
		@Override
		@RosettaAttribute("valuationMethod")
		public ValuationType1Code getValuationMethod() {
			return valuationMethod;
		}
		
		@Override
		@RosettaAttribute("valuationTimestamp")
		public ZonedDateTime getValuationTimestamp() {
			return valuationTimestamp;
		}
		
		@Override
		@RosettaAttribute("collateralPortfolioIndicator")
		public Boolean getCollateralPortfolioIndicator() {
			return collateralPortfolioIndicator;
		}
		
		@Override
		@RosettaAttribute("initialMarginCollateralPortfolioCode")
		public String getInitialMarginCollateralPortfolioCode() {
			return initialMarginCollateralPortfolioCode;
		}
		
		@Override
		@RosettaAttribute("variationMarginCollateralPortfolioCode")
		public String getVariationMarginCollateralPortfolioCode() {
			return variationMarginCollateralPortfolioCode;
		}
		
		@Override
		@RosettaAttribute("price")
		public PriceAndNotation.PriceAndNotationBuilder getPrice() {
			return price;
		}
		
		@Override
		public PriceAndNotation.PriceAndNotationBuilder getOrCreatePrice() {
			PriceAndNotation.PriceAndNotationBuilder result;
			if (price!=null) {
				result = price;
			}
			else {
				result = price = PriceAndNotation.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("priceNotation")
		public String getPriceNotation() {
			return priceNotation;
		}
		
		@Override
		@RosettaAttribute("priceUnitOfMeasure")
		public String getPriceUnitOfMeasure() {
			return priceUnitOfMeasure;
		}
		
		@Override
		@RosettaAttribute("priceSchedule")
		public List<? extends PriceScheduleReport.PriceScheduleReportBuilder> getPriceSchedule() {
			return priceSchedule;
		}
		
		public PriceScheduleReport.PriceScheduleReportBuilder getOrCreatePriceSchedule(int _index) {
		
			if (priceSchedule==null) {
				this.priceSchedule = new ArrayList<>();
			}
			PriceScheduleReport.PriceScheduleReportBuilder result;
			return getIndex(priceSchedule, _index, () -> {
						PriceScheduleReport.PriceScheduleReportBuilder newPriceSchedule = PriceScheduleReport.builder();
						return newPriceSchedule;
					});
		}
		
		@Override
		@RosettaAttribute("fixedRateLeg1")
		public BigDecimal getFixedRateLeg1() {
			return fixedRateLeg1;
		}
		
		@Override
		@RosettaAttribute("fixedRateLeg2")
		public BigDecimal getFixedRateLeg2() {
			return fixedRateLeg2;
		}
		
		@Override
		@RosettaAttribute("spreadOfLeg1")
		public SpreadOfLeg1.SpreadOfLeg1Builder getSpreadOfLeg1() {
			return spreadOfLeg1;
		}
		
		@Override
		public SpreadOfLeg1.SpreadOfLeg1Builder getOrCreateSpreadOfLeg1() {
			SpreadOfLeg1.SpreadOfLeg1Builder result;
			if (spreadOfLeg1!=null) {
				result = spreadOfLeg1;
			}
			else {
				result = spreadOfLeg1 = SpreadOfLeg1.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("spreadNotationOfLeg1")
		public String getSpreadNotationOfLeg1() {
			return spreadNotationOfLeg1;
		}
		
		@Override
		@RosettaAttribute("spreadOfLeg2")
		public SpreadOfLeg2.SpreadOfLeg2Builder getSpreadOfLeg2() {
			return spreadOfLeg2;
		}
		
		@Override
		public SpreadOfLeg2.SpreadOfLeg2Builder getOrCreateSpreadOfLeg2() {
			SpreadOfLeg2.SpreadOfLeg2Builder result;
			if (spreadOfLeg2!=null) {
				result = spreadOfLeg2;
			}
			else {
				result = spreadOfLeg2 = SpreadOfLeg2.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("spreadNotationOfLeg2")
		public String getSpreadNotationOfLeg2() {
			return spreadNotationOfLeg2;
		}
		
		@Override
		@RosettaAttribute("strikePrice")
		public StrikePriceAndNotation.StrikePriceAndNotationBuilder getStrikePrice() {
			return strikePrice;
		}
		
		@Override
		public StrikePriceAndNotation.StrikePriceAndNotationBuilder getOrCreateStrikePrice() {
			StrikePriceAndNotation.StrikePriceAndNotationBuilder result;
			if (strikePrice!=null) {
				result = strikePrice;
			}
			else {
				result = strikePrice = StrikePriceAndNotation.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("strikePriceNotation")
		public String getStrikePriceNotation() {
			return strikePriceNotation;
		}
		
		@Override
		@RosettaAttribute("strikePriceSchedule")
		public List<? extends StrikePriceScheduleReport.StrikePriceScheduleReportBuilder> getStrikePriceSchedule() {
			return strikePriceSchedule;
		}
		
		public StrikePriceScheduleReport.StrikePriceScheduleReportBuilder getOrCreateStrikePriceSchedule(int _index) {
		
			if (strikePriceSchedule==null) {
				this.strikePriceSchedule = new ArrayList<>();
			}
			StrikePriceScheduleReport.StrikePriceScheduleReportBuilder result;
			return getIndex(strikePriceSchedule, _index, () -> {
						StrikePriceScheduleReport.StrikePriceScheduleReportBuilder newStrikePriceSchedule = StrikePriceScheduleReport.builder();
						return newStrikePriceSchedule;
					});
		}
		
		@Override
		@RosettaAttribute("optionPremiumAmount")
		public BigDecimal getOptionPremiumAmount() {
			return optionPremiumAmount;
		}
		
		@Override
		@RosettaAttribute("optionPremiumCurrency")
		public ISOCurrencyCodeEnum getOptionPremiumCurrency() {
			return optionPremiumCurrency;
		}
		
		@Override
		@RosettaAttribute("optionPremiumPaymentDate")
		public Date getOptionPremiumPaymentDate() {
			return optionPremiumPaymentDate;
		}
		
		@Override
		@RosettaAttribute("firstExerciseDate")
		public Date getFirstExerciseDate() {
			return firstExerciseDate;
		}
		
		@Override
		@RosettaAttribute("exchangeRate")
		public BigDecimal getExchangeRate() {
			return exchangeRate;
		}
		
		@Override
		@RosettaAttribute("exchangeRateBasis")
		public String getExchangeRateBasis() {
			return exchangeRateBasis;
		}
		
		@Override
		@RosettaAttribute("notionalLeg1")
		public BigDecimal getNotionalLeg1() {
			return notionalLeg1;
		}
		
		@Override
		@RosettaAttribute("notionalLeg2")
		public BigDecimal getNotionalLeg2() {
			return notionalLeg2;
		}
		
		@Override
		@RosettaAttribute("callAmount")
		public BigDecimal getCallAmount() {
			return callAmount;
		}
		
		@Override
		@RosettaAttribute("putAmount")
		public BigDecimal getPutAmount() {
			return putAmount;
		}
		
		@Override
		@RosettaAttribute("notionalCurrencyOfLeg1")
		public ISOCurrencyCodeEnum getNotionalCurrencyOfLeg1() {
			return notionalCurrencyOfLeg1;
		}
		
		@Override
		@RosettaAttribute("notionalCurrencyOfLeg2")
		public ISOCurrencyCodeEnum getNotionalCurrencyOfLeg2() {
			return notionalCurrencyOfLeg2;
		}
		
		@Override
		@RosettaAttribute("callCurrency")
		public ISOCurrencyCodeEnum getCallCurrency() {
			return callCurrency;
		}
		
		@Override
		@RosettaAttribute("putCurrency")
		public ISOCurrencyCodeEnum getPutCurrency() {
			return putCurrency;
		}
		
		@Override
		@RosettaAttribute("quantityUnitOfMeasureLeg1")
		public String getQuantityUnitOfMeasureLeg1() {
			return quantityUnitOfMeasureLeg1;
		}
		
		@Override
		@RosettaAttribute("quantityUnitOfMeasureLeg2")
		public String getQuantityUnitOfMeasureLeg2() {
			return quantityUnitOfMeasureLeg2;
		}
		
		@Override
		@RosettaAttribute("notionalAmountScheduleLeg1")
		public List<? extends NotionalAmountScheduleLeg1Report.NotionalAmountScheduleLeg1ReportBuilder> getNotionalAmountScheduleLeg1() {
			return notionalAmountScheduleLeg1;
		}
		
		public NotionalAmountScheduleLeg1Report.NotionalAmountScheduleLeg1ReportBuilder getOrCreateNotionalAmountScheduleLeg1(int _index) {
		
			if (notionalAmountScheduleLeg1==null) {
				this.notionalAmountScheduleLeg1 = new ArrayList<>();
			}
			NotionalAmountScheduleLeg1Report.NotionalAmountScheduleLeg1ReportBuilder result;
			return getIndex(notionalAmountScheduleLeg1, _index, () -> {
						NotionalAmountScheduleLeg1Report.NotionalAmountScheduleLeg1ReportBuilder newNotionalAmountScheduleLeg1 = NotionalAmountScheduleLeg1Report.builder();
						return newNotionalAmountScheduleLeg1;
					});
		}
		
		@Override
		@RosettaAttribute("notionalAmountScheduleLeg2")
		public List<? extends NotionalAmountScheduleLeg2Report.NotionalAmountScheduleLeg2ReportBuilder> getNotionalAmountScheduleLeg2() {
			return notionalAmountScheduleLeg2;
		}
		
		public NotionalAmountScheduleLeg2Report.NotionalAmountScheduleLeg2ReportBuilder getOrCreateNotionalAmountScheduleLeg2(int _index) {
		
			if (notionalAmountScheduleLeg2==null) {
				this.notionalAmountScheduleLeg2 = new ArrayList<>();
			}
			NotionalAmountScheduleLeg2Report.NotionalAmountScheduleLeg2ReportBuilder result;
			return getIndex(notionalAmountScheduleLeg2, _index, () -> {
						NotionalAmountScheduleLeg2Report.NotionalAmountScheduleLeg2ReportBuilder newNotionalAmountScheduleLeg2 = NotionalAmountScheduleLeg2Report.builder();
						return newNotionalAmountScheduleLeg2;
					});
		}
		
		@Override
		@RosettaAttribute("totalNotionalQuantityLeg1")
		public BigDecimal getTotalNotionalQuantityLeg1() {
			return totalNotionalQuantityLeg1;
		}
		
		@Override
		@RosettaAttribute("totalNotionalQuantityLeg2")
		public BigDecimal getTotalNotionalQuantityLeg2() {
			return totalNotionalQuantityLeg2;
		}
		
		@Override
		@RosettaAttribute("notionalQuantityScheduleLeg1")
		public List<? extends NotionalQuantityScheduleLeg1Report.NotionalQuantityScheduleLeg1ReportBuilder> getNotionalQuantityScheduleLeg1() {
			return notionalQuantityScheduleLeg1;
		}
		
		public NotionalQuantityScheduleLeg1Report.NotionalQuantityScheduleLeg1ReportBuilder getOrCreateNotionalQuantityScheduleLeg1(int _index) {
		
			if (notionalQuantityScheduleLeg1==null) {
				this.notionalQuantityScheduleLeg1 = new ArrayList<>();
			}
			NotionalQuantityScheduleLeg1Report.NotionalQuantityScheduleLeg1ReportBuilder result;
			return getIndex(notionalQuantityScheduleLeg1, _index, () -> {
						NotionalQuantityScheduleLeg1Report.NotionalQuantityScheduleLeg1ReportBuilder newNotionalQuantityScheduleLeg1 = NotionalQuantityScheduleLeg1Report.builder();
						return newNotionalQuantityScheduleLeg1;
					});
		}
		
		@Override
		@RosettaAttribute("notionalQuantityScheduleLeg2")
		public List<? extends NotionalQuantityScheduleLeg2Report.NotionalQuantityScheduleLeg2ReportBuilder> getNotionalQuantityScheduleLeg2() {
			return notionalQuantityScheduleLeg2;
		}
		
		public NotionalQuantityScheduleLeg2Report.NotionalQuantityScheduleLeg2ReportBuilder getOrCreateNotionalQuantityScheduleLeg2(int _index) {
		
			if (notionalQuantityScheduleLeg2==null) {
				this.notionalQuantityScheduleLeg2 = new ArrayList<>();
			}
			NotionalQuantityScheduleLeg2Report.NotionalQuantityScheduleLeg2ReportBuilder result;
			return getIndex(notionalQuantityScheduleLeg2, _index, () -> {
						NotionalQuantityScheduleLeg2Report.NotionalQuantityScheduleLeg2ReportBuilder newNotionalQuantityScheduleLeg2 = NotionalQuantityScheduleLeg2Report.builder();
						return newNotionalQuantityScheduleLeg2;
					});
		}
		
		@Override
		@RosettaAttribute("actionType")
		public String getActionType() {
			return actionType;
		}
		
		@Override
		@RosettaAttribute("eventType")
		public String getEventType() {
			return eventType;
		}
		
		@Override
		@RosettaAttribute("eventTimestamp")
		public ZonedDateTime getEventTimestamp() {
			return eventTimestamp;
		}
		
		@Override
		@RosettaAttribute("eventIdentifier")
		public String getEventIdentifier() {
			return eventIdentifier;
		}
		
		@Override
		@RosettaAttribute("indexFactor")
		public BigDecimal getIndexFactor() {
			return indexFactor;
		}
		
		@Override
		@RosettaAttribute("embeddedOptionType")
		public String getEmbeddedOptionType() {
			return embeddedOptionType;
		}
		
		@Override
		@RosettaAttribute("upi")
		public String getUpi() {
			return upi;
		}
		
		@Override
		@RosettaAttribute("deliveryType")
		public String getDeliveryType() {
			return deliveryType;
		}
		
		@Override
		@RosettaAttribute("assetClass")
		public String getAssetClass() {
			return assetClass;
		}
		
		@Override
		@RosettaAttribute("underlyingIdentificationType")
		public String getUnderlyingIdentificationType() {
			return underlyingIdentificationType;
		}
		
		@Override
		@RosettaAttribute("underlyingIdentification")
		public String getUnderlyingIdentification() {
			return underlyingIdentification;
		}
		
		@Override
		@RosettaAttribute("underlyingIndexIndicator")
		public String getUnderlyingIndexIndicator() {
			return underlyingIndexIndicator;
		}
		
		@Override
		@RosettaAttribute("nameOfTheUnderlyingIndex")
		public String getNameOfTheUnderlyingIndex() {
			return nameOfTheUnderlyingIndex;
		}
		
		@Override
		@RosettaAttribute("referenceEntity")
		public String getReferenceEntity() {
			return referenceEntity;
		}
		
		@Override
		@RosettaAttribute("indicatorOfTheFloatingRateOfLeg1")
		public String getIndicatorOfTheFloatingRateOfLeg1() {
			return indicatorOfTheFloatingRateOfLeg1;
		}
		
		@Override
		@RosettaAttribute("indicatorOfTheFloatingRateOfLeg2")
		public String getIndicatorOfTheFloatingRateOfLeg2() {
			return indicatorOfTheFloatingRateOfLeg2;
		}
		
		@Override
		@RosettaAttribute("nameOfTheFloatingRateOfLeg1")
		public String getNameOfTheFloatingRateOfLeg1() {
			return nameOfTheFloatingRateOfLeg1;
		}
		
		@Override
		@RosettaAttribute("nameOfTheFloatingRateOfLeg2")
		public String getNameOfTheFloatingRateOfLeg2() {
			return nameOfTheFloatingRateOfLeg2;
		}
		
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg1TimePeriod")
		public String getFloatingRateReferencePeriodOfLeg1TimePeriod() {
			return floatingRateReferencePeriodOfLeg1TimePeriod;
		}
		
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg2TimePeriod")
		public String getFloatingRateReferencePeriodOfLeg2TimePeriod() {
			return floatingRateReferencePeriodOfLeg2TimePeriod;
		}
		
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg1Multiplier")
		public Integer getFloatingRateReferencePeriodOfLeg1Multiplier() {
			return floatingRateReferencePeriodOfLeg1Multiplier;
		}
		
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg2Multiplier")
		public Integer getFloatingRateReferencePeriodOfLeg2Multiplier() {
			return floatingRateReferencePeriodOfLeg2Multiplier;
		}
		
		@Override
		@RosettaAttribute("isCrypto")
		public Boolean getIsCrypto() {
			return isCrypto;
		}
		
		@Override
		@RosettaAttribute("maturityDateOfTheUnderlying")
		public Date getMaturityDateOfTheUnderlying() {
			return maturityDateOfTheUnderlying;
		}
		
		@Override
		@RosettaAttribute("seniority")
		public String getSeniority() {
			return seniority;
		}
		
		@Override
		@RosettaAttribute("series")
		public Integer getSeries() {
			return series;
		}
		
		@Override
		@RosettaAttribute("version")
		public Integer getVersion() {
			return version;
		}
		
		@Override
		@RosettaAttribute("cdSIndexAttachmentPoint")
		public BigDecimal getCdSIndexAttachmentPoint() {
			return cdSIndexAttachmentPoint;
		}
		
		@Override
		@RosettaAttribute("cdSIndexDetachmentPoint")
		public BigDecimal getCdSIndexDetachmentPoint() {
			return cdSIndexDetachmentPoint;
		}
		
		@Override
		@RosettaAttribute("customBasketCode")
		public String getCustomBasketCode() {
			return customBasketCode;
		}
		
		@Override
		@RosettaAttribute("identifierOfBasketConstituents")
		public List<String> getIdentifierOfBasketConstituents() {
			return identifierOfBasketConstituents;
		}
		
		@Override
		@RosettaAttribute("basketConstituentIdentifierSource")
		public List<ProductIdTypeEnum> getBasketConstituentIdentifierSource() {
			return basketConstituentIdentifierSource;
		}
		
		@Override
		@RosettaAttribute("contractType")
		public String getContractType() {
			return contractType;
		}
		
		@Override
		@RosettaAttribute("optionStyle")
		public String getOptionStyle() {
			return optionStyle;
		}
		
		@Override
		@RosettaAttribute("optionType")
		public String getOptionType() {
			return optionType;
		}
		
		@Override
		@RosettaAttribute("packageIdentifier")
		public String getPackageIdentifier() {
			return packageIdentifier;
		}
		
		@Override
		@RosettaAttribute("packageTransactionPrice")
		public PackageTransactionPrice.PackageTransactionPriceBuilder getPackageTransactionPrice() {
			return packageTransactionPrice;
		}
		
		@Override
		public PackageTransactionPrice.PackageTransactionPriceBuilder getOrCreatePackageTransactionPrice() {
			PackageTransactionPrice.PackageTransactionPriceBuilder result;
			if (packageTransactionPrice!=null) {
				result = packageTransactionPrice;
			}
			else {
				result = packageTransactionPrice = PackageTransactionPrice.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("packageTransactionPriceCurrency")
		public ISOCurrencyCodeEnum getPackageTransactionPriceCurrency() {
			return packageTransactionPriceCurrency;
		}
		
		@Override
		@RosettaAttribute("packageTransactionPriceNotation")
		public String getPackageTransactionPriceNotation() {
			return packageTransactionPriceNotation;
		}
		
		@Override
		@RosettaAttribute("packageTransactionSpread")
		public PackageTransactionSpread.PackageTransactionSpreadBuilder getPackageTransactionSpread() {
			return packageTransactionSpread;
		}
		
		@Override
		public PackageTransactionSpread.PackageTransactionSpreadBuilder getOrCreatePackageTransactionSpread() {
			PackageTransactionSpread.PackageTransactionSpreadBuilder result;
			if (packageTransactionSpread!=null) {
				result = packageTransactionSpread;
			}
			else {
				result = packageTransactionSpread = PackageTransactionSpread.builder();
			}
			
			return result;
		}
		@Override
		@RosettaAttribute("packageTransactionSpreadCurrency")
		public ISOCurrencyCodeEnum getPackageTransactionSpreadCurrency() {
			return packageTransactionSpreadCurrency;
		}
		
		@Override
		@RosettaAttribute("packageTransactionSpreadNotation")
		public String getPackageTransactionSpreadNotation() {
			return packageTransactionSpreadNotation;
		}
		
		@Override
		@RosettaAttribute("technicalRecordId")
		public String getTechnicalRecordId() {
			return technicalRecordId;
		}
		
		@Override
		@RosettaAttribute("nonReportable")
		public JFSANonReportable.JFSANonReportableBuilder getNonReportable() {
			return nonReportable;
		}
		
		@Override
		public JFSANonReportable.JFSANonReportableBuilder getOrCreateNonReportable() {
			JFSANonReportable.JFSANonReportableBuilder result;
			if (nonReportable!=null) {
				result = nonReportable;
			}
			else {
				result = nonReportable = JFSANonReportable.builder();
			}
			
			return result;
		}
	
		@Override
		@RosettaAttribute("effectiveDate")
		public JFSATransactionReport.JFSATransactionReportBuilder setEffectiveDate(Date effectiveDate) {
			this.effectiveDate = effectiveDate==null?null:effectiveDate;
			return this;
		}
		@Override
		@RosettaAttribute("expirationDate")
		public JFSATransactionReport.JFSATransactionReportBuilder setExpirationDate(Date expirationDate) {
			this.expirationDate = expirationDate==null?null:expirationDate;
			return this;
		}
		@Override
		@RosettaAttribute("earlyTerminationDate")
		public JFSATransactionReport.JFSATransactionReportBuilder setEarlyTerminationDate(Date earlyTerminationDate) {
			this.earlyTerminationDate = earlyTerminationDate==null?null:earlyTerminationDate;
			return this;
		}
		@Override
		@RosettaAttribute("reportingTimestamp")
		public JFSATransactionReport.JFSATransactionReportBuilder setReportingTimestamp(ZonedDateTime reportingTimestamp) {
			this.reportingTimestamp = reportingTimestamp==null?null:reportingTimestamp;
			return this;
		}
		@Override
		@RosettaAttribute("executionTimestamp")
		public JFSATransactionReport.JFSATransactionReportBuilder setExecutionTimestamp(ZonedDateTime executionTimestamp) {
			this.executionTimestamp = executionTimestamp==null?null:executionTimestamp;
			return this;
		}
		@Override
		@RosettaAttribute("entityResponsibleForReporting")
		public JFSATransactionReport.JFSATransactionReportBuilder setEntityResponsibleForReporting(String entityResponsibleForReporting) {
			this.entityResponsibleForReporting = entityResponsibleForReporting==null?null:entityResponsibleForReporting;
			return this;
		}
		@Override
		@RosettaAttribute("counterparty1")
		public JFSATransactionReport.JFSATransactionReportBuilder setCounterparty1(String counterparty1) {
			this.counterparty1 = counterparty1==null?null:counterparty1;
			return this;
		}
		@Override
		@RosettaAttribute("counterparty2")
		public JFSATransactionReport.JFSATransactionReportBuilder setCounterparty2(String counterparty2) {
			this.counterparty2 = counterparty2==null?null:counterparty2;
			return this;
		}
		@Override
		@RosettaAttribute("counterparty2IdentifierType")
		public JFSATransactionReport.JFSATransactionReportBuilder setCounterparty2IdentifierType(Boolean counterparty2IdentifierType) {
			this.counterparty2IdentifierType = counterparty2IdentifierType==null?null:counterparty2IdentifierType;
			return this;
		}
		@Override
		@RosettaAttribute("direction1")
		public JFSATransactionReport.JFSATransactionReportBuilder setDirection1(String direction1) {
			this.direction1 = direction1==null?null:direction1;
			return this;
		}
		@Override
		@RosettaAttribute("direction2Leg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setDirection2Leg1(String direction2Leg1) {
			this.direction2Leg1 = direction2Leg1==null?null:direction2Leg1;
			return this;
		}
		@Override
		@RosettaAttribute("direction2Leg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setDirection2Leg2(String direction2Leg2) {
			this.direction2Leg2 = direction2Leg2==null?null:direction2Leg2;
			return this;
		}
		@Override
		@RosettaAttribute("submitterIdentifier")
		public JFSATransactionReport.JFSATransactionReportBuilder setSubmitterIdentifier(String submitterIdentifier) {
			this.submitterIdentifier = submitterIdentifier==null?null:submitterIdentifier;
			return this;
		}
		@Override
		@RosettaAttribute("newSDRIdentifier")
		public JFSATransactionReport.JFSATransactionReportBuilder setNewSDRIdentifier(String newSDRIdentifier) {
			this.newSDRIdentifier = newSDRIdentifier==null?null:newSDRIdentifier;
			return this;
		}
		@Override
		@RosettaAttribute("originalSwapSDRIdentifier")
		public JFSATransactionReport.JFSATransactionReportBuilder setOriginalSwapSDRIdentifier(String originalSwapSDRIdentifier) {
			this.originalSwapSDRIdentifier = originalSwapSDRIdentifier==null?null:originalSwapSDRIdentifier;
			return this;
		}
		@Override
		@RosettaAttribute("cleared")
		public JFSATransactionReport.JFSATransactionReportBuilder setCleared(String cleared) {
			this.cleared = cleared==null?null:cleared;
			return this;
		}
		@Override
		@RosettaAttribute("centralCounterparty")
		public JFSATransactionReport.JFSATransactionReportBuilder setCentralCounterparty(String centralCounterparty) {
			this.centralCounterparty = centralCounterparty==null?null:centralCounterparty;
			return this;
		}
		@Override
		@RosettaAttribute("clearingMember")
		public JFSATransactionReport.JFSATransactionReportBuilder setClearingMember(String clearingMember) {
			this.clearingMember = clearingMember==null?null:clearingMember;
			return this;
		}
		@Override
		@RosettaAttribute("platformIdentifier")
		public JFSATransactionReport.JFSATransactionReportBuilder setPlatformIdentifier(String platformIdentifier) {
			this.platformIdentifier = platformIdentifier==null?null:platformIdentifier;
			return this;
		}
		@Override
		@RosettaAttribute("confirmed")
		public JFSATransactionReport.JFSATransactionReportBuilder setConfirmed(String confirmed) {
			this.confirmed = confirmed==null?null:confirmed;
			return this;
		}
		@Override
		@RosettaAttribute("finalContractualSettlementDate")
		public JFSATransactionReport.JFSATransactionReportBuilder setFinalContractualSettlementDate(Date finalContractualSettlementDate) {
			this.finalContractualSettlementDate = finalContractualSettlementDate==null?null:finalContractualSettlementDate;
			return this;
		}
		@Override
		@RosettaAttribute("confirmationTimestamp")
		public JFSATransactionReport.JFSATransactionReportBuilder setConfirmationTimestamp(ZonedDateTime confirmationTimestamp) {
			this.confirmationTimestamp = confirmationTimestamp==null?null:confirmationTimestamp;
			return this;
		}
		@Override
		@RosettaAttribute("settlementCurrencyLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setSettlementCurrencyLeg1(ISOCurrencyCodeEnum settlementCurrencyLeg1) {
			this.settlementCurrencyLeg1 = settlementCurrencyLeg1==null?null:settlementCurrencyLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("settlementCurrencyLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setSettlementCurrencyLeg2(ISOCurrencyCodeEnum settlementCurrencyLeg2) {
			this.settlementCurrencyLeg2 = settlementCurrencyLeg2==null?null:settlementCurrencyLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("clearingAccountOrigin")
		public JFSATransactionReport.JFSATransactionReportBuilder setClearingAccountOrigin(String clearingAccountOrigin) {
			this.clearingAccountOrigin = clearingAccountOrigin==null?null:clearingAccountOrigin;
			return this;
		}
		@Override
		@RosettaAttribute("originalSwapUTI")
		public JFSATransactionReport.JFSATransactionReportBuilder setOriginalSwapUTI(String originalSwapUTI) {
			this.originalSwapUTI = originalSwapUTI==null?null:originalSwapUTI;
			return this;
		}
		@Override
		@RosettaAttribute("clearingReceiptTimestamp")
		public JFSATransactionReport.JFSATransactionReportBuilder setClearingReceiptTimestamp(ZonedDateTime clearingReceiptTimestamp) {
			this.clearingReceiptTimestamp = clearingReceiptTimestamp==null?null:clearingReceiptTimestamp;
			return this;
		}
		@Override
		@RosettaAttribute("uti")
		public JFSATransactionReport.JFSATransactionReportBuilder setUti(String uti) {
			this.uti = uti==null?null:uti;
			return this;
		}
		@Override
		@RosettaAttribute("utiProprietary")
		public JFSATransactionReport.JFSATransactionReportBuilder setUtiProprietary(String utiProprietary) {
			this.utiProprietary = utiProprietary==null?null:utiProprietary;
			return this;
		}
		@Override
		@RosettaAttribute("priorUti")
		public JFSATransactionReport.JFSATransactionReportBuilder setPriorUti(String priorUti) {
			this.priorUti = priorUti==null?null:priorUti;
			return this;
		}
		@Override
		@RosettaAttribute("priorUtiProprietary")
		public JFSATransactionReport.JFSATransactionReportBuilder setPriorUtiProprietary(String priorUtiProprietary) {
			this.priorUtiProprietary = priorUtiProprietary==null?null:priorUtiProprietary;
			return this;
		}
		@Override
		@RosettaAttribute("fixedRateDayCountConventionLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixedRateDayCountConventionLeg1(InterestComputationMethod4Code fixedRateDayCountConventionLeg1) {
			this.fixedRateDayCountConventionLeg1 = fixedRateDayCountConventionLeg1==null?null:fixedRateDayCountConventionLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateDayCountConventionOfLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateDayCountConventionOfLeg1(InterestComputationMethod4Code floatingRateDayCountConventionOfLeg1) {
			this.floatingRateDayCountConventionOfLeg1 = floatingRateDayCountConventionOfLeg1==null?null:floatingRateDayCountConventionOfLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("fixedRateDayCountConventionLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixedRateDayCountConventionLeg2(InterestComputationMethod4Code fixedRateDayCountConventionLeg2) {
			this.fixedRateDayCountConventionLeg2 = fixedRateDayCountConventionLeg2==null?null:fixedRateDayCountConventionLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateDayCountConventionOfLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateDayCountConventionOfLeg2(InterestComputationMethod4Code floatingRateDayCountConventionOfLeg2) {
			this.floatingRateDayCountConventionOfLeg2 = floatingRateDayCountConventionOfLeg2==null?null:floatingRateDayCountConventionOfLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixedRatePaymentFrequencyPeriodLeg1(String fixedRatePaymentFrequencyPeriodLeg1) {
			this.fixedRatePaymentFrequencyPeriodLeg1 = fixedRatePaymentFrequencyPeriodLeg1==null?null:fixedRatePaymentFrequencyPeriodLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixedRatePaymentFrequencyPeriodLeg2(String fixedRatePaymentFrequencyPeriodLeg2) {
			this.fixedRatePaymentFrequencyPeriodLeg2 = fixedRatePaymentFrequencyPeriodLeg2==null?null:fixedRatePaymentFrequencyPeriodLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRatePaymentFrequencyPeriodLeg1(String floatingRatePaymentFrequencyPeriodLeg1) {
			this.floatingRatePaymentFrequencyPeriodLeg1 = floatingRatePaymentFrequencyPeriodLeg1==null?null:floatingRatePaymentFrequencyPeriodLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRatePaymentFrequencyPeriodLeg2(String floatingRatePaymentFrequencyPeriodLeg2) {
			this.floatingRatePaymentFrequencyPeriodLeg2 = floatingRatePaymentFrequencyPeriodLeg2==null?null:floatingRatePaymentFrequencyPeriodLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodMultiplierLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixedRatePaymentFrequencyPeriodMultiplierLeg1(Integer fixedRatePaymentFrequencyPeriodMultiplierLeg1) {
			this.fixedRatePaymentFrequencyPeriodMultiplierLeg1 = fixedRatePaymentFrequencyPeriodMultiplierLeg1==null?null:fixedRatePaymentFrequencyPeriodMultiplierLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("fixedRatePaymentFrequencyPeriodMultiplierLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixedRatePaymentFrequencyPeriodMultiplierLeg2(Integer fixedRatePaymentFrequencyPeriodMultiplierLeg2) {
			this.fixedRatePaymentFrequencyPeriodMultiplierLeg2 = fixedRatePaymentFrequencyPeriodMultiplierLeg2==null?null:fixedRatePaymentFrequencyPeriodMultiplierLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodMultiplierLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRatePaymentFrequencyPeriodMultiplierLeg1(Integer floatingRatePaymentFrequencyPeriodMultiplierLeg1) {
			this.floatingRatePaymentFrequencyPeriodMultiplierLeg1 = floatingRatePaymentFrequencyPeriodMultiplierLeg1==null?null:floatingRatePaymentFrequencyPeriodMultiplierLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRatePaymentFrequencyPeriodMultiplierLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRatePaymentFrequencyPeriodMultiplierLeg2(Integer floatingRatePaymentFrequencyPeriodMultiplierLeg2) {
			this.floatingRatePaymentFrequencyPeriodMultiplierLeg2 = floatingRatePaymentFrequencyPeriodMultiplierLeg2==null?null:floatingRatePaymentFrequencyPeriodMultiplierLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("fixingDateLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixingDateLeg1(ZonedDateTime fixingDateLeg1) {
			this.fixingDateLeg1 = fixingDateLeg1==null?null:fixingDateLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("fixingDateLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixingDateLeg2(ZonedDateTime fixingDateLeg2) {
			this.fixingDateLeg2 = fixingDateLeg2==null?null:fixingDateLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateResetFrequencyPeriodLeg1(String floatingRateResetFrequencyPeriodLeg1) {
			this.floatingRateResetFrequencyPeriodLeg1 = floatingRateResetFrequencyPeriodLeg1==null?null:floatingRateResetFrequencyPeriodLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodOfLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateResetFrequencyPeriodOfLeg2(String floatingRateResetFrequencyPeriodOfLeg2) {
			this.floatingRateResetFrequencyPeriodOfLeg2 = floatingRateResetFrequencyPeriodOfLeg2==null?null:floatingRateResetFrequencyPeriodOfLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodMultiplierLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateResetFrequencyPeriodMultiplierLeg1(Integer floatingRateResetFrequencyPeriodMultiplierLeg1) {
			this.floatingRateResetFrequencyPeriodMultiplierLeg1 = floatingRateResetFrequencyPeriodMultiplierLeg1==null?null:floatingRateResetFrequencyPeriodMultiplierLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateResetFrequencyPeriodMultiplierLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateResetFrequencyPeriodMultiplierLeg2(Integer floatingRateResetFrequencyPeriodMultiplierLeg2) {
			this.floatingRateResetFrequencyPeriodMultiplierLeg2 = floatingRateResetFrequencyPeriodMultiplierLeg2==null?null:floatingRateResetFrequencyPeriodMultiplierLeg2;
			return this;
		}
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addOtherPayment(OtherPaymentReport otherPayment) {
			if (otherPayment!=null) this.otherPayment.add(otherPayment.toBuilder());
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addOtherPayment(OtherPaymentReport otherPayment, int _idx) {
			getIndex(this.otherPayment, _idx, () -> otherPayment.toBuilder());
			return this;
		}
		@Override 
		public JFSATransactionReport.JFSATransactionReportBuilder addOtherPayment(List<? extends OtherPaymentReport> otherPayments) {
			if (otherPayments != null) {
				for (OtherPaymentReport toAdd : otherPayments) {
					this.otherPayment.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("otherPayment")
		public JFSATransactionReport.JFSATransactionReportBuilder setOtherPayment(List<? extends OtherPaymentReport> otherPayments) {
			if (otherPayments == null)  {
				this.otherPayment = new ArrayList<>();
			}
			else {
				this.otherPayment = otherPayments.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		@RosettaAttribute("valuationAmount")
		public JFSATransactionReport.JFSATransactionReportBuilder setValuationAmount(BigDecimal valuationAmount) {
			this.valuationAmount = valuationAmount==null?null:valuationAmount;
			return this;
		}
		@Override
		@RosettaAttribute("valuationCurrency")
		public JFSATransactionReport.JFSATransactionReportBuilder setValuationCurrency(ISOCurrencyCodeEnum valuationCurrency) {
			this.valuationCurrency = valuationCurrency==null?null:valuationCurrency;
			return this;
		}
		@Override
		@RosettaAttribute("valuationMethod")
		public JFSATransactionReport.JFSATransactionReportBuilder setValuationMethod(ValuationType1Code valuationMethod) {
			this.valuationMethod = valuationMethod==null?null:valuationMethod;
			return this;
		}
		@Override
		@RosettaAttribute("valuationTimestamp")
		public JFSATransactionReport.JFSATransactionReportBuilder setValuationTimestamp(ZonedDateTime valuationTimestamp) {
			this.valuationTimestamp = valuationTimestamp==null?null:valuationTimestamp;
			return this;
		}
		@Override
		@RosettaAttribute("collateralPortfolioIndicator")
		public JFSATransactionReport.JFSATransactionReportBuilder setCollateralPortfolioIndicator(Boolean collateralPortfolioIndicator) {
			this.collateralPortfolioIndicator = collateralPortfolioIndicator==null?null:collateralPortfolioIndicator;
			return this;
		}
		@Override
		@RosettaAttribute("initialMarginCollateralPortfolioCode")
		public JFSATransactionReport.JFSATransactionReportBuilder setInitialMarginCollateralPortfolioCode(String initialMarginCollateralPortfolioCode) {
			this.initialMarginCollateralPortfolioCode = initialMarginCollateralPortfolioCode==null?null:initialMarginCollateralPortfolioCode;
			return this;
		}
		@Override
		@RosettaAttribute("variationMarginCollateralPortfolioCode")
		public JFSATransactionReport.JFSATransactionReportBuilder setVariationMarginCollateralPortfolioCode(String variationMarginCollateralPortfolioCode) {
			this.variationMarginCollateralPortfolioCode = variationMarginCollateralPortfolioCode==null?null:variationMarginCollateralPortfolioCode;
			return this;
		}
		@Override
		@RosettaAttribute("price")
		public JFSATransactionReport.JFSATransactionReportBuilder setPrice(PriceAndNotation price) {
			this.price = price==null?null:price.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("priceNotation")
		public JFSATransactionReport.JFSATransactionReportBuilder setPriceNotation(String priceNotation) {
			this.priceNotation = priceNotation==null?null:priceNotation;
			return this;
		}
		@Override
		@RosettaAttribute("priceUnitOfMeasure")
		public JFSATransactionReport.JFSATransactionReportBuilder setPriceUnitOfMeasure(String priceUnitOfMeasure) {
			this.priceUnitOfMeasure = priceUnitOfMeasure==null?null:priceUnitOfMeasure;
			return this;
		}
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addPriceSchedule(PriceScheduleReport priceSchedule) {
			if (priceSchedule!=null) this.priceSchedule.add(priceSchedule.toBuilder());
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addPriceSchedule(PriceScheduleReport priceSchedule, int _idx) {
			getIndex(this.priceSchedule, _idx, () -> priceSchedule.toBuilder());
			return this;
		}
		@Override 
		public JFSATransactionReport.JFSATransactionReportBuilder addPriceSchedule(List<? extends PriceScheduleReport> priceSchedules) {
			if (priceSchedules != null) {
				for (PriceScheduleReport toAdd : priceSchedules) {
					this.priceSchedule.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("priceSchedule")
		public JFSATransactionReport.JFSATransactionReportBuilder setPriceSchedule(List<? extends PriceScheduleReport> priceSchedules) {
			if (priceSchedules == null)  {
				this.priceSchedule = new ArrayList<>();
			}
			else {
				this.priceSchedule = priceSchedules.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		@RosettaAttribute("fixedRateLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixedRateLeg1(BigDecimal fixedRateLeg1) {
			this.fixedRateLeg1 = fixedRateLeg1==null?null:fixedRateLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("fixedRateLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setFixedRateLeg2(BigDecimal fixedRateLeg2) {
			this.fixedRateLeg2 = fixedRateLeg2==null?null:fixedRateLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("spreadOfLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setSpreadOfLeg1(SpreadOfLeg1 spreadOfLeg1) {
			this.spreadOfLeg1 = spreadOfLeg1==null?null:spreadOfLeg1.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("spreadNotationOfLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setSpreadNotationOfLeg1(String spreadNotationOfLeg1) {
			this.spreadNotationOfLeg1 = spreadNotationOfLeg1==null?null:spreadNotationOfLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("spreadOfLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setSpreadOfLeg2(SpreadOfLeg2 spreadOfLeg2) {
			this.spreadOfLeg2 = spreadOfLeg2==null?null:spreadOfLeg2.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("spreadNotationOfLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setSpreadNotationOfLeg2(String spreadNotationOfLeg2) {
			this.spreadNotationOfLeg2 = spreadNotationOfLeg2==null?null:spreadNotationOfLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("strikePrice")
		public JFSATransactionReport.JFSATransactionReportBuilder setStrikePrice(StrikePriceAndNotation strikePrice) {
			this.strikePrice = strikePrice==null?null:strikePrice.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("strikePriceNotation")
		public JFSATransactionReport.JFSATransactionReportBuilder setStrikePriceNotation(String strikePriceNotation) {
			this.strikePriceNotation = strikePriceNotation==null?null:strikePriceNotation;
			return this;
		}
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addStrikePriceSchedule(StrikePriceScheduleReport strikePriceSchedule) {
			if (strikePriceSchedule!=null) this.strikePriceSchedule.add(strikePriceSchedule.toBuilder());
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addStrikePriceSchedule(StrikePriceScheduleReport strikePriceSchedule, int _idx) {
			getIndex(this.strikePriceSchedule, _idx, () -> strikePriceSchedule.toBuilder());
			return this;
		}
		@Override 
		public JFSATransactionReport.JFSATransactionReportBuilder addStrikePriceSchedule(List<? extends StrikePriceScheduleReport> strikePriceSchedules) {
			if (strikePriceSchedules != null) {
				for (StrikePriceScheduleReport toAdd : strikePriceSchedules) {
					this.strikePriceSchedule.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("strikePriceSchedule")
		public JFSATransactionReport.JFSATransactionReportBuilder setStrikePriceSchedule(List<? extends StrikePriceScheduleReport> strikePriceSchedules) {
			if (strikePriceSchedules == null)  {
				this.strikePriceSchedule = new ArrayList<>();
			}
			else {
				this.strikePriceSchedule = strikePriceSchedules.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		@RosettaAttribute("optionPremiumAmount")
		public JFSATransactionReport.JFSATransactionReportBuilder setOptionPremiumAmount(BigDecimal optionPremiumAmount) {
			this.optionPremiumAmount = optionPremiumAmount==null?null:optionPremiumAmount;
			return this;
		}
		@Override
		@RosettaAttribute("optionPremiumCurrency")
		public JFSATransactionReport.JFSATransactionReportBuilder setOptionPremiumCurrency(ISOCurrencyCodeEnum optionPremiumCurrency) {
			this.optionPremiumCurrency = optionPremiumCurrency==null?null:optionPremiumCurrency;
			return this;
		}
		@Override
		@RosettaAttribute("optionPremiumPaymentDate")
		public JFSATransactionReport.JFSATransactionReportBuilder setOptionPremiumPaymentDate(Date optionPremiumPaymentDate) {
			this.optionPremiumPaymentDate = optionPremiumPaymentDate==null?null:optionPremiumPaymentDate;
			return this;
		}
		@Override
		@RosettaAttribute("firstExerciseDate")
		public JFSATransactionReport.JFSATransactionReportBuilder setFirstExerciseDate(Date firstExerciseDate) {
			this.firstExerciseDate = firstExerciseDate==null?null:firstExerciseDate;
			return this;
		}
		@Override
		@RosettaAttribute("exchangeRate")
		public JFSATransactionReport.JFSATransactionReportBuilder setExchangeRate(BigDecimal exchangeRate) {
			this.exchangeRate = exchangeRate==null?null:exchangeRate;
			return this;
		}
		@Override
		@RosettaAttribute("exchangeRateBasis")
		public JFSATransactionReport.JFSATransactionReportBuilder setExchangeRateBasis(String exchangeRateBasis) {
			this.exchangeRateBasis = exchangeRateBasis==null?null:exchangeRateBasis;
			return this;
		}
		@Override
		@RosettaAttribute("notionalLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setNotionalLeg1(BigDecimal notionalLeg1) {
			this.notionalLeg1 = notionalLeg1==null?null:notionalLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("notionalLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setNotionalLeg2(BigDecimal notionalLeg2) {
			this.notionalLeg2 = notionalLeg2==null?null:notionalLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("callAmount")
		public JFSATransactionReport.JFSATransactionReportBuilder setCallAmount(BigDecimal callAmount) {
			this.callAmount = callAmount==null?null:callAmount;
			return this;
		}
		@Override
		@RosettaAttribute("putAmount")
		public JFSATransactionReport.JFSATransactionReportBuilder setPutAmount(BigDecimal putAmount) {
			this.putAmount = putAmount==null?null:putAmount;
			return this;
		}
		@Override
		@RosettaAttribute("notionalCurrencyOfLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setNotionalCurrencyOfLeg1(ISOCurrencyCodeEnum notionalCurrencyOfLeg1) {
			this.notionalCurrencyOfLeg1 = notionalCurrencyOfLeg1==null?null:notionalCurrencyOfLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("notionalCurrencyOfLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setNotionalCurrencyOfLeg2(ISOCurrencyCodeEnum notionalCurrencyOfLeg2) {
			this.notionalCurrencyOfLeg2 = notionalCurrencyOfLeg2==null?null:notionalCurrencyOfLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("callCurrency")
		public JFSATransactionReport.JFSATransactionReportBuilder setCallCurrency(ISOCurrencyCodeEnum callCurrency) {
			this.callCurrency = callCurrency==null?null:callCurrency;
			return this;
		}
		@Override
		@RosettaAttribute("putCurrency")
		public JFSATransactionReport.JFSATransactionReportBuilder setPutCurrency(ISOCurrencyCodeEnum putCurrency) {
			this.putCurrency = putCurrency==null?null:putCurrency;
			return this;
		}
		@Override
		@RosettaAttribute("quantityUnitOfMeasureLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setQuantityUnitOfMeasureLeg1(String quantityUnitOfMeasureLeg1) {
			this.quantityUnitOfMeasureLeg1 = quantityUnitOfMeasureLeg1==null?null:quantityUnitOfMeasureLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("quantityUnitOfMeasureLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setQuantityUnitOfMeasureLeg2(String quantityUnitOfMeasureLeg2) {
			this.quantityUnitOfMeasureLeg2 = quantityUnitOfMeasureLeg2==null?null:quantityUnitOfMeasureLeg2;
			return this;
		}
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg1(NotionalAmountScheduleLeg1Report notionalAmountScheduleLeg1) {
			if (notionalAmountScheduleLeg1!=null) this.notionalAmountScheduleLeg1.add(notionalAmountScheduleLeg1.toBuilder());
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg1(NotionalAmountScheduleLeg1Report notionalAmountScheduleLeg1, int _idx) {
			getIndex(this.notionalAmountScheduleLeg1, _idx, () -> notionalAmountScheduleLeg1.toBuilder());
			return this;
		}
		@Override 
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg1(List<? extends NotionalAmountScheduleLeg1Report> notionalAmountScheduleLeg1s) {
			if (notionalAmountScheduleLeg1s != null) {
				for (NotionalAmountScheduleLeg1Report toAdd : notionalAmountScheduleLeg1s) {
					this.notionalAmountScheduleLeg1.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("notionalAmountScheduleLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setNotionalAmountScheduleLeg1(List<? extends NotionalAmountScheduleLeg1Report> notionalAmountScheduleLeg1s) {
			if (notionalAmountScheduleLeg1s == null)  {
				this.notionalAmountScheduleLeg1 = new ArrayList<>();
			}
			else {
				this.notionalAmountScheduleLeg1 = notionalAmountScheduleLeg1s.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg2(NotionalAmountScheduleLeg2Report notionalAmountScheduleLeg2) {
			if (notionalAmountScheduleLeg2!=null) this.notionalAmountScheduleLeg2.add(notionalAmountScheduleLeg2.toBuilder());
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg2(NotionalAmountScheduleLeg2Report notionalAmountScheduleLeg2, int _idx) {
			getIndex(this.notionalAmountScheduleLeg2, _idx, () -> notionalAmountScheduleLeg2.toBuilder());
			return this;
		}
		@Override 
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalAmountScheduleLeg2(List<? extends NotionalAmountScheduleLeg2Report> notionalAmountScheduleLeg2s) {
			if (notionalAmountScheduleLeg2s != null) {
				for (NotionalAmountScheduleLeg2Report toAdd : notionalAmountScheduleLeg2s) {
					this.notionalAmountScheduleLeg2.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("notionalAmountScheduleLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setNotionalAmountScheduleLeg2(List<? extends NotionalAmountScheduleLeg2Report> notionalAmountScheduleLeg2s) {
			if (notionalAmountScheduleLeg2s == null)  {
				this.notionalAmountScheduleLeg2 = new ArrayList<>();
			}
			else {
				this.notionalAmountScheduleLeg2 = notionalAmountScheduleLeg2s.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		@RosettaAttribute("totalNotionalQuantityLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setTotalNotionalQuantityLeg1(BigDecimal totalNotionalQuantityLeg1) {
			this.totalNotionalQuantityLeg1 = totalNotionalQuantityLeg1==null?null:totalNotionalQuantityLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("totalNotionalQuantityLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setTotalNotionalQuantityLeg2(BigDecimal totalNotionalQuantityLeg2) {
			this.totalNotionalQuantityLeg2 = totalNotionalQuantityLeg2==null?null:totalNotionalQuantityLeg2;
			return this;
		}
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg1(NotionalQuantityScheduleLeg1Report notionalQuantityScheduleLeg1) {
			if (notionalQuantityScheduleLeg1!=null) this.notionalQuantityScheduleLeg1.add(notionalQuantityScheduleLeg1.toBuilder());
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg1(NotionalQuantityScheduleLeg1Report notionalQuantityScheduleLeg1, int _idx) {
			getIndex(this.notionalQuantityScheduleLeg1, _idx, () -> notionalQuantityScheduleLeg1.toBuilder());
			return this;
		}
		@Override 
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg1(List<? extends NotionalQuantityScheduleLeg1Report> notionalQuantityScheduleLeg1s) {
			if (notionalQuantityScheduleLeg1s != null) {
				for (NotionalQuantityScheduleLeg1Report toAdd : notionalQuantityScheduleLeg1s) {
					this.notionalQuantityScheduleLeg1.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("notionalQuantityScheduleLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setNotionalQuantityScheduleLeg1(List<? extends NotionalQuantityScheduleLeg1Report> notionalQuantityScheduleLeg1s) {
			if (notionalQuantityScheduleLeg1s == null)  {
				this.notionalQuantityScheduleLeg1 = new ArrayList<>();
			}
			else {
				this.notionalQuantityScheduleLeg1 = notionalQuantityScheduleLeg1s.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg2(NotionalQuantityScheduleLeg2Report notionalQuantityScheduleLeg2) {
			if (notionalQuantityScheduleLeg2!=null) this.notionalQuantityScheduleLeg2.add(notionalQuantityScheduleLeg2.toBuilder());
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg2(NotionalQuantityScheduleLeg2Report notionalQuantityScheduleLeg2, int _idx) {
			getIndex(this.notionalQuantityScheduleLeg2, _idx, () -> notionalQuantityScheduleLeg2.toBuilder());
			return this;
		}
		@Override 
		public JFSATransactionReport.JFSATransactionReportBuilder addNotionalQuantityScheduleLeg2(List<? extends NotionalQuantityScheduleLeg2Report> notionalQuantityScheduleLeg2s) {
			if (notionalQuantityScheduleLeg2s != null) {
				for (NotionalQuantityScheduleLeg2Report toAdd : notionalQuantityScheduleLeg2s) {
					this.notionalQuantityScheduleLeg2.add(toAdd.toBuilder());
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("notionalQuantityScheduleLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setNotionalQuantityScheduleLeg2(List<? extends NotionalQuantityScheduleLeg2Report> notionalQuantityScheduleLeg2s) {
			if (notionalQuantityScheduleLeg2s == null)  {
				this.notionalQuantityScheduleLeg2 = new ArrayList<>();
			}
			else {
				this.notionalQuantityScheduleLeg2 = notionalQuantityScheduleLeg2s.stream()
					.map(_a->_a.toBuilder())
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		@RosettaAttribute("actionType")
		public JFSATransactionReport.JFSATransactionReportBuilder setActionType(String actionType) {
			this.actionType = actionType==null?null:actionType;
			return this;
		}
		@Override
		@RosettaAttribute("eventType")
		public JFSATransactionReport.JFSATransactionReportBuilder setEventType(String eventType) {
			this.eventType = eventType==null?null:eventType;
			return this;
		}
		@Override
		@RosettaAttribute("eventTimestamp")
		public JFSATransactionReport.JFSATransactionReportBuilder setEventTimestamp(ZonedDateTime eventTimestamp) {
			this.eventTimestamp = eventTimestamp==null?null:eventTimestamp;
			return this;
		}
		@Override
		@RosettaAttribute("eventIdentifier")
		public JFSATransactionReport.JFSATransactionReportBuilder setEventIdentifier(String eventIdentifier) {
			this.eventIdentifier = eventIdentifier==null?null:eventIdentifier;
			return this;
		}
		@Override
		@RosettaAttribute("indexFactor")
		public JFSATransactionReport.JFSATransactionReportBuilder setIndexFactor(BigDecimal indexFactor) {
			this.indexFactor = indexFactor==null?null:indexFactor;
			return this;
		}
		@Override
		@RosettaAttribute("embeddedOptionType")
		public JFSATransactionReport.JFSATransactionReportBuilder setEmbeddedOptionType(String embeddedOptionType) {
			this.embeddedOptionType = embeddedOptionType==null?null:embeddedOptionType;
			return this;
		}
		@Override
		@RosettaAttribute("upi")
		public JFSATransactionReport.JFSATransactionReportBuilder setUpi(String upi) {
			this.upi = upi==null?null:upi;
			return this;
		}
		@Override
		@RosettaAttribute("deliveryType")
		public JFSATransactionReport.JFSATransactionReportBuilder setDeliveryType(String deliveryType) {
			this.deliveryType = deliveryType==null?null:deliveryType;
			return this;
		}
		@Override
		@RosettaAttribute("assetClass")
		public JFSATransactionReport.JFSATransactionReportBuilder setAssetClass(String assetClass) {
			this.assetClass = assetClass==null?null:assetClass;
			return this;
		}
		@Override
		@RosettaAttribute("underlyingIdentificationType")
		public JFSATransactionReport.JFSATransactionReportBuilder setUnderlyingIdentificationType(String underlyingIdentificationType) {
			this.underlyingIdentificationType = underlyingIdentificationType==null?null:underlyingIdentificationType;
			return this;
		}
		@Override
		@RosettaAttribute("underlyingIdentification")
		public JFSATransactionReport.JFSATransactionReportBuilder setUnderlyingIdentification(String underlyingIdentification) {
			this.underlyingIdentification = underlyingIdentification==null?null:underlyingIdentification;
			return this;
		}
		@Override
		@RosettaAttribute("underlyingIndexIndicator")
		public JFSATransactionReport.JFSATransactionReportBuilder setUnderlyingIndexIndicator(String underlyingIndexIndicator) {
			this.underlyingIndexIndicator = underlyingIndexIndicator==null?null:underlyingIndexIndicator;
			return this;
		}
		@Override
		@RosettaAttribute("nameOfTheUnderlyingIndex")
		public JFSATransactionReport.JFSATransactionReportBuilder setNameOfTheUnderlyingIndex(String nameOfTheUnderlyingIndex) {
			this.nameOfTheUnderlyingIndex = nameOfTheUnderlyingIndex==null?null:nameOfTheUnderlyingIndex;
			return this;
		}
		@Override
		@RosettaAttribute("referenceEntity")
		public JFSATransactionReport.JFSATransactionReportBuilder setReferenceEntity(String referenceEntity) {
			this.referenceEntity = referenceEntity==null?null:referenceEntity;
			return this;
		}
		@Override
		@RosettaAttribute("indicatorOfTheFloatingRateOfLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setIndicatorOfTheFloatingRateOfLeg1(String indicatorOfTheFloatingRateOfLeg1) {
			this.indicatorOfTheFloatingRateOfLeg1 = indicatorOfTheFloatingRateOfLeg1==null?null:indicatorOfTheFloatingRateOfLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("indicatorOfTheFloatingRateOfLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setIndicatorOfTheFloatingRateOfLeg2(String indicatorOfTheFloatingRateOfLeg2) {
			this.indicatorOfTheFloatingRateOfLeg2 = indicatorOfTheFloatingRateOfLeg2==null?null:indicatorOfTheFloatingRateOfLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("nameOfTheFloatingRateOfLeg1")
		public JFSATransactionReport.JFSATransactionReportBuilder setNameOfTheFloatingRateOfLeg1(String nameOfTheFloatingRateOfLeg1) {
			this.nameOfTheFloatingRateOfLeg1 = nameOfTheFloatingRateOfLeg1==null?null:nameOfTheFloatingRateOfLeg1;
			return this;
		}
		@Override
		@RosettaAttribute("nameOfTheFloatingRateOfLeg2")
		public JFSATransactionReport.JFSATransactionReportBuilder setNameOfTheFloatingRateOfLeg2(String nameOfTheFloatingRateOfLeg2) {
			this.nameOfTheFloatingRateOfLeg2 = nameOfTheFloatingRateOfLeg2==null?null:nameOfTheFloatingRateOfLeg2;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg1TimePeriod")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateReferencePeriodOfLeg1TimePeriod(String floatingRateReferencePeriodOfLeg1TimePeriod) {
			this.floatingRateReferencePeriodOfLeg1TimePeriod = floatingRateReferencePeriodOfLeg1TimePeriod==null?null:floatingRateReferencePeriodOfLeg1TimePeriod;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg2TimePeriod")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateReferencePeriodOfLeg2TimePeriod(String floatingRateReferencePeriodOfLeg2TimePeriod) {
			this.floatingRateReferencePeriodOfLeg2TimePeriod = floatingRateReferencePeriodOfLeg2TimePeriod==null?null:floatingRateReferencePeriodOfLeg2TimePeriod;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg1Multiplier")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateReferencePeriodOfLeg1Multiplier(Integer floatingRateReferencePeriodOfLeg1Multiplier) {
			this.floatingRateReferencePeriodOfLeg1Multiplier = floatingRateReferencePeriodOfLeg1Multiplier==null?null:floatingRateReferencePeriodOfLeg1Multiplier;
			return this;
		}
		@Override
		@RosettaAttribute("floatingRateReferencePeriodOfLeg2Multiplier")
		public JFSATransactionReport.JFSATransactionReportBuilder setFloatingRateReferencePeriodOfLeg2Multiplier(Integer floatingRateReferencePeriodOfLeg2Multiplier) {
			this.floatingRateReferencePeriodOfLeg2Multiplier = floatingRateReferencePeriodOfLeg2Multiplier==null?null:floatingRateReferencePeriodOfLeg2Multiplier;
			return this;
		}
		@Override
		@RosettaAttribute("isCrypto")
		public JFSATransactionReport.JFSATransactionReportBuilder setIsCrypto(Boolean isCrypto) {
			this.isCrypto = isCrypto==null?null:isCrypto;
			return this;
		}
		@Override
		@RosettaAttribute("maturityDateOfTheUnderlying")
		public JFSATransactionReport.JFSATransactionReportBuilder setMaturityDateOfTheUnderlying(Date maturityDateOfTheUnderlying) {
			this.maturityDateOfTheUnderlying = maturityDateOfTheUnderlying==null?null:maturityDateOfTheUnderlying;
			return this;
		}
		@Override
		@RosettaAttribute("seniority")
		public JFSATransactionReport.JFSATransactionReportBuilder setSeniority(String seniority) {
			this.seniority = seniority==null?null:seniority;
			return this;
		}
		@Override
		@RosettaAttribute("series")
		public JFSATransactionReport.JFSATransactionReportBuilder setSeries(Integer series) {
			this.series = series==null?null:series;
			return this;
		}
		@Override
		@RosettaAttribute("version")
		public JFSATransactionReport.JFSATransactionReportBuilder setVersion(Integer version) {
			this.version = version==null?null:version;
			return this;
		}
		@Override
		@RosettaAttribute("cdSIndexAttachmentPoint")
		public JFSATransactionReport.JFSATransactionReportBuilder setCdSIndexAttachmentPoint(BigDecimal cdSIndexAttachmentPoint) {
			this.cdSIndexAttachmentPoint = cdSIndexAttachmentPoint==null?null:cdSIndexAttachmentPoint;
			return this;
		}
		@Override
		@RosettaAttribute("cdSIndexDetachmentPoint")
		public JFSATransactionReport.JFSATransactionReportBuilder setCdSIndexDetachmentPoint(BigDecimal cdSIndexDetachmentPoint) {
			this.cdSIndexDetachmentPoint = cdSIndexDetachmentPoint==null?null:cdSIndexDetachmentPoint;
			return this;
		}
		@Override
		@RosettaAttribute("customBasketCode")
		public JFSATransactionReport.JFSATransactionReportBuilder setCustomBasketCode(String customBasketCode) {
			this.customBasketCode = customBasketCode==null?null:customBasketCode;
			return this;
		}
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addIdentifierOfBasketConstituents(String identifierOfBasketConstituents) {
			if (identifierOfBasketConstituents!=null) this.identifierOfBasketConstituents.add(identifierOfBasketConstituents);
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addIdentifierOfBasketConstituents(String identifierOfBasketConstituents, int _idx) {
			getIndex(this.identifierOfBasketConstituents, _idx, () -> identifierOfBasketConstituents);
			return this;
		}
		@Override 
		public JFSATransactionReport.JFSATransactionReportBuilder addIdentifierOfBasketConstituents(List<? extends String> identifierOfBasketConstituentss) {
			if (identifierOfBasketConstituentss != null) {
				for (String toAdd : identifierOfBasketConstituentss) {
					this.identifierOfBasketConstituents.add(toAdd);
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("identifierOfBasketConstituents")
		public JFSATransactionReport.JFSATransactionReportBuilder setIdentifierOfBasketConstituents(List<? extends String> identifierOfBasketConstituentss) {
			if (identifierOfBasketConstituentss == null)  {
				this.identifierOfBasketConstituents = new ArrayList<>();
			}
			else {
				this.identifierOfBasketConstituents = identifierOfBasketConstituentss.stream()
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addBasketConstituentIdentifierSource(ProductIdTypeEnum basketConstituentIdentifierSource) {
			if (basketConstituentIdentifierSource!=null) this.basketConstituentIdentifierSource.add(basketConstituentIdentifierSource);
			return this;
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder addBasketConstituentIdentifierSource(ProductIdTypeEnum basketConstituentIdentifierSource, int _idx) {
			getIndex(this.basketConstituentIdentifierSource, _idx, () -> basketConstituentIdentifierSource);
			return this;
		}
		@Override 
		public JFSATransactionReport.JFSATransactionReportBuilder addBasketConstituentIdentifierSource(List<? extends ProductIdTypeEnum> basketConstituentIdentifierSources) {
			if (basketConstituentIdentifierSources != null) {
				for (ProductIdTypeEnum toAdd : basketConstituentIdentifierSources) {
					this.basketConstituentIdentifierSource.add(toAdd);
				}
			}
			return this;
		}
		
		@Override 
		@RosettaAttribute("basketConstituentIdentifierSource")
		public JFSATransactionReport.JFSATransactionReportBuilder setBasketConstituentIdentifierSource(List<? extends ProductIdTypeEnum> basketConstituentIdentifierSources) {
			if (basketConstituentIdentifierSources == null)  {
				this.basketConstituentIdentifierSource = new ArrayList<>();
			}
			else {
				this.basketConstituentIdentifierSource = basketConstituentIdentifierSources.stream()
					.collect(Collectors.toCollection(()->new ArrayList<>()));
			}
			return this;
		}
		
		@Override
		@RosettaAttribute("contractType")
		public JFSATransactionReport.JFSATransactionReportBuilder setContractType(String contractType) {
			this.contractType = contractType==null?null:contractType;
			return this;
		}
		@Override
		@RosettaAttribute("optionStyle")
		public JFSATransactionReport.JFSATransactionReportBuilder setOptionStyle(String optionStyle) {
			this.optionStyle = optionStyle==null?null:optionStyle;
			return this;
		}
		@Override
		@RosettaAttribute("optionType")
		public JFSATransactionReport.JFSATransactionReportBuilder setOptionType(String optionType) {
			this.optionType = optionType==null?null:optionType;
			return this;
		}
		@Override
		@RosettaAttribute("packageIdentifier")
		public JFSATransactionReport.JFSATransactionReportBuilder setPackageIdentifier(String packageIdentifier) {
			this.packageIdentifier = packageIdentifier==null?null:packageIdentifier;
			return this;
		}
		@Override
		@RosettaAttribute("packageTransactionPrice")
		public JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionPrice(PackageTransactionPrice packageTransactionPrice) {
			this.packageTransactionPrice = packageTransactionPrice==null?null:packageTransactionPrice.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("packageTransactionPriceCurrency")
		public JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionPriceCurrency(ISOCurrencyCodeEnum packageTransactionPriceCurrency) {
			this.packageTransactionPriceCurrency = packageTransactionPriceCurrency==null?null:packageTransactionPriceCurrency;
			return this;
		}
		@Override
		@RosettaAttribute("packageTransactionPriceNotation")
		public JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionPriceNotation(String packageTransactionPriceNotation) {
			this.packageTransactionPriceNotation = packageTransactionPriceNotation==null?null:packageTransactionPriceNotation;
			return this;
		}
		@Override
		@RosettaAttribute("packageTransactionSpread")
		public JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionSpread(PackageTransactionSpread packageTransactionSpread) {
			this.packageTransactionSpread = packageTransactionSpread==null?null:packageTransactionSpread.toBuilder();
			return this;
		}
		@Override
		@RosettaAttribute("packageTransactionSpreadCurrency")
		public JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionSpreadCurrency(ISOCurrencyCodeEnum packageTransactionSpreadCurrency) {
			this.packageTransactionSpreadCurrency = packageTransactionSpreadCurrency==null?null:packageTransactionSpreadCurrency;
			return this;
		}
		@Override
		@RosettaAttribute("packageTransactionSpreadNotation")
		public JFSATransactionReport.JFSATransactionReportBuilder setPackageTransactionSpreadNotation(String packageTransactionSpreadNotation) {
			this.packageTransactionSpreadNotation = packageTransactionSpreadNotation==null?null:packageTransactionSpreadNotation;
			return this;
		}
		@Override
		@RosettaAttribute("technicalRecordId")
		public JFSATransactionReport.JFSATransactionReportBuilder setTechnicalRecordId(String technicalRecordId) {
			this.technicalRecordId = technicalRecordId==null?null:technicalRecordId;
			return this;
		}
		@Override
		@RosettaAttribute("nonReportable")
		public JFSATransactionReport.JFSATransactionReportBuilder setNonReportable(JFSANonReportable nonReportable) {
			this.nonReportable = nonReportable==null?null:nonReportable.toBuilder();
			return this;
		}
		
		@Override
		public JFSATransactionReport build() {
			return new JFSATransactionReport.JFSATransactionReportImpl(this);
		}
		
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder toBuilder() {
			return this;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder prune() {
			otherPayment = otherPayment.stream().filter(b->b!=null).<OtherPaymentReport.OtherPaymentReportBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			if (price!=null && !price.prune().hasData()) price = null;
			priceSchedule = priceSchedule.stream().filter(b->b!=null).<PriceScheduleReport.PriceScheduleReportBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			if (spreadOfLeg1!=null && !spreadOfLeg1.prune().hasData()) spreadOfLeg1 = null;
			if (spreadOfLeg2!=null && !spreadOfLeg2.prune().hasData()) spreadOfLeg2 = null;
			if (strikePrice!=null && !strikePrice.prune().hasData()) strikePrice = null;
			strikePriceSchedule = strikePriceSchedule.stream().filter(b->b!=null).<StrikePriceScheduleReport.StrikePriceScheduleReportBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			notionalAmountScheduleLeg1 = notionalAmountScheduleLeg1.stream().filter(b->b!=null).<NotionalAmountScheduleLeg1Report.NotionalAmountScheduleLeg1ReportBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			notionalAmountScheduleLeg2 = notionalAmountScheduleLeg2.stream().filter(b->b!=null).<NotionalAmountScheduleLeg2Report.NotionalAmountScheduleLeg2ReportBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			notionalQuantityScheduleLeg1 = notionalQuantityScheduleLeg1.stream().filter(b->b!=null).<NotionalQuantityScheduleLeg1Report.NotionalQuantityScheduleLeg1ReportBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			notionalQuantityScheduleLeg2 = notionalQuantityScheduleLeg2.stream().filter(b->b!=null).<NotionalQuantityScheduleLeg2Report.NotionalQuantityScheduleLeg2ReportBuilder>map(b->b.prune()).filter(b->b.hasData()).collect(Collectors.toList());
			if (packageTransactionPrice!=null && !packageTransactionPrice.prune().hasData()) packageTransactionPrice = null;
			if (packageTransactionSpread!=null && !packageTransactionSpread.prune().hasData()) packageTransactionSpread = null;
			if (nonReportable!=null && !nonReportable.prune().hasData()) nonReportable = null;
			return this;
		}
		
		@Override
		public boolean hasData() {
			if (getEffectiveDate()!=null) return true;
			if (getExpirationDate()!=null) return true;
			if (getEarlyTerminationDate()!=null) return true;
			if (getReportingTimestamp()!=null) return true;
			if (getExecutionTimestamp()!=null) return true;
			if (getEntityResponsibleForReporting()!=null) return true;
			if (getCounterparty1()!=null) return true;
			if (getCounterparty2()!=null) return true;
			if (getCounterparty2IdentifierType()!=null) return true;
			if (getDirection1()!=null) return true;
			if (getDirection2Leg1()!=null) return true;
			if (getDirection2Leg2()!=null) return true;
			if (getSubmitterIdentifier()!=null) return true;
			if (getNewSDRIdentifier()!=null) return true;
			if (getOriginalSwapSDRIdentifier()!=null) return true;
			if (getCleared()!=null) return true;
			if (getCentralCounterparty()!=null) return true;
			if (getClearingMember()!=null) return true;
			if (getPlatformIdentifier()!=null) return true;
			if (getConfirmed()!=null) return true;
			if (getFinalContractualSettlementDate()!=null) return true;
			if (getConfirmationTimestamp()!=null) return true;
			if (getSettlementCurrencyLeg1()!=null) return true;
			if (getSettlementCurrencyLeg2()!=null) return true;
			if (getClearingAccountOrigin()!=null) return true;
			if (getOriginalSwapUTI()!=null) return true;
			if (getClearingReceiptTimestamp()!=null) return true;
			if (getUti()!=null) return true;
			if (getUtiProprietary()!=null) return true;
			if (getPriorUti()!=null) return true;
			if (getPriorUtiProprietary()!=null) return true;
			if (getFixedRateDayCountConventionLeg1()!=null) return true;
			if (getFloatingRateDayCountConventionOfLeg1()!=null) return true;
			if (getFixedRateDayCountConventionLeg2()!=null) return true;
			if (getFloatingRateDayCountConventionOfLeg2()!=null) return true;
			if (getFixedRatePaymentFrequencyPeriodLeg1()!=null) return true;
			if (getFixedRatePaymentFrequencyPeriodLeg2()!=null) return true;
			if (getFloatingRatePaymentFrequencyPeriodLeg1()!=null) return true;
			if (getFloatingRatePaymentFrequencyPeriodLeg2()!=null) return true;
			if (getFixedRatePaymentFrequencyPeriodMultiplierLeg1()!=null) return true;
			if (getFixedRatePaymentFrequencyPeriodMultiplierLeg2()!=null) return true;
			if (getFloatingRatePaymentFrequencyPeriodMultiplierLeg1()!=null) return true;
			if (getFloatingRatePaymentFrequencyPeriodMultiplierLeg2()!=null) return true;
			if (getFixingDateLeg1()!=null) return true;
			if (getFixingDateLeg2()!=null) return true;
			if (getFloatingRateResetFrequencyPeriodLeg1()!=null) return true;
			if (getFloatingRateResetFrequencyPeriodOfLeg2()!=null) return true;
			if (getFloatingRateResetFrequencyPeriodMultiplierLeg1()!=null) return true;
			if (getFloatingRateResetFrequencyPeriodMultiplierLeg2()!=null) return true;
			if (getOtherPayment()!=null && getOtherPayment().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getValuationAmount()!=null) return true;
			if (getValuationCurrency()!=null) return true;
			if (getValuationMethod()!=null) return true;
			if (getValuationTimestamp()!=null) return true;
			if (getCollateralPortfolioIndicator()!=null) return true;
			if (getInitialMarginCollateralPortfolioCode()!=null) return true;
			if (getVariationMarginCollateralPortfolioCode()!=null) return true;
			if (getPrice()!=null && getPrice().hasData()) return true;
			if (getPriceNotation()!=null) return true;
			if (getPriceUnitOfMeasure()!=null) return true;
			if (getPriceSchedule()!=null && getPriceSchedule().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getFixedRateLeg1()!=null) return true;
			if (getFixedRateLeg2()!=null) return true;
			if (getSpreadOfLeg1()!=null && getSpreadOfLeg1().hasData()) return true;
			if (getSpreadNotationOfLeg1()!=null) return true;
			if (getSpreadOfLeg2()!=null && getSpreadOfLeg2().hasData()) return true;
			if (getSpreadNotationOfLeg2()!=null) return true;
			if (getStrikePrice()!=null && getStrikePrice().hasData()) return true;
			if (getStrikePriceNotation()!=null) return true;
			if (getStrikePriceSchedule()!=null && getStrikePriceSchedule().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getOptionPremiumAmount()!=null) return true;
			if (getOptionPremiumCurrency()!=null) return true;
			if (getOptionPremiumPaymentDate()!=null) return true;
			if (getFirstExerciseDate()!=null) return true;
			if (getExchangeRate()!=null) return true;
			if (getExchangeRateBasis()!=null) return true;
			if (getNotionalLeg1()!=null) return true;
			if (getNotionalLeg2()!=null) return true;
			if (getCallAmount()!=null) return true;
			if (getPutAmount()!=null) return true;
			if (getNotionalCurrencyOfLeg1()!=null) return true;
			if (getNotionalCurrencyOfLeg2()!=null) return true;
			if (getCallCurrency()!=null) return true;
			if (getPutCurrency()!=null) return true;
			if (getQuantityUnitOfMeasureLeg1()!=null) return true;
			if (getQuantityUnitOfMeasureLeg2()!=null) return true;
			if (getNotionalAmountScheduleLeg1()!=null && getNotionalAmountScheduleLeg1().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getNotionalAmountScheduleLeg2()!=null && getNotionalAmountScheduleLeg2().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getTotalNotionalQuantityLeg1()!=null) return true;
			if (getTotalNotionalQuantityLeg2()!=null) return true;
			if (getNotionalQuantityScheduleLeg1()!=null && getNotionalQuantityScheduleLeg1().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getNotionalQuantityScheduleLeg2()!=null && getNotionalQuantityScheduleLeg2().stream().filter(Objects::nonNull).anyMatch(a->a.hasData())) return true;
			if (getActionType()!=null) return true;
			if (getEventType()!=null) return true;
			if (getEventTimestamp()!=null) return true;
			if (getEventIdentifier()!=null) return true;
			if (getIndexFactor()!=null) return true;
			if (getEmbeddedOptionType()!=null) return true;
			if (getUpi()!=null) return true;
			if (getDeliveryType()!=null) return true;
			if (getAssetClass()!=null) return true;
			if (getUnderlyingIdentificationType()!=null) return true;
			if (getUnderlyingIdentification()!=null) return true;
			if (getUnderlyingIndexIndicator()!=null) return true;
			if (getNameOfTheUnderlyingIndex()!=null) return true;
			if (getReferenceEntity()!=null) return true;
			if (getIndicatorOfTheFloatingRateOfLeg1()!=null) return true;
			if (getIndicatorOfTheFloatingRateOfLeg2()!=null) return true;
			if (getNameOfTheFloatingRateOfLeg1()!=null) return true;
			if (getNameOfTheFloatingRateOfLeg2()!=null) return true;
			if (getFloatingRateReferencePeriodOfLeg1TimePeriod()!=null) return true;
			if (getFloatingRateReferencePeriodOfLeg2TimePeriod()!=null) return true;
			if (getFloatingRateReferencePeriodOfLeg1Multiplier()!=null) return true;
			if (getFloatingRateReferencePeriodOfLeg2Multiplier()!=null) return true;
			if (getIsCrypto()!=null) return true;
			if (getMaturityDateOfTheUnderlying()!=null) return true;
			if (getSeniority()!=null) return true;
			if (getSeries()!=null) return true;
			if (getVersion()!=null) return true;
			if (getCdSIndexAttachmentPoint()!=null) return true;
			if (getCdSIndexDetachmentPoint()!=null) return true;
			if (getCustomBasketCode()!=null) return true;
			if (getIdentifierOfBasketConstituents()!=null && !getIdentifierOfBasketConstituents().isEmpty()) return true;
			if (getBasketConstituentIdentifierSource()!=null && !getBasketConstituentIdentifierSource().isEmpty()) return true;
			if (getContractType()!=null) return true;
			if (getOptionStyle()!=null) return true;
			if (getOptionType()!=null) return true;
			if (getPackageIdentifier()!=null) return true;
			if (getPackageTransactionPrice()!=null && getPackageTransactionPrice().hasData()) return true;
			if (getPackageTransactionPriceCurrency()!=null) return true;
			if (getPackageTransactionPriceNotation()!=null) return true;
			if (getPackageTransactionSpread()!=null && getPackageTransactionSpread().hasData()) return true;
			if (getPackageTransactionSpreadCurrency()!=null) return true;
			if (getPackageTransactionSpreadNotation()!=null) return true;
			if (getTechnicalRecordId()!=null) return true;
			if (getNonReportable()!=null && getNonReportable().hasData()) return true;
			return false;
		}
	
		@SuppressWarnings("unchecked")
		@Override
		public JFSATransactionReport.JFSATransactionReportBuilder merge(RosettaModelObjectBuilder other, BuilderMerger merger) {
			JFSATransactionReport.JFSATransactionReportBuilder o = (JFSATransactionReport.JFSATransactionReportBuilder) other;
			
			merger.mergeRosetta(getOtherPayment(), o.getOtherPayment(), this::getOrCreateOtherPayment);
			merger.mergeRosetta(getPrice(), o.getPrice(), this::setPrice);
			merger.mergeRosetta(getPriceSchedule(), o.getPriceSchedule(), this::getOrCreatePriceSchedule);
			merger.mergeRosetta(getSpreadOfLeg1(), o.getSpreadOfLeg1(), this::setSpreadOfLeg1);
			merger.mergeRosetta(getSpreadOfLeg2(), o.getSpreadOfLeg2(), this::setSpreadOfLeg2);
			merger.mergeRosetta(getStrikePrice(), o.getStrikePrice(), this::setStrikePrice);
			merger.mergeRosetta(getStrikePriceSchedule(), o.getStrikePriceSchedule(), this::getOrCreateStrikePriceSchedule);
			merger.mergeRosetta(getNotionalAmountScheduleLeg1(), o.getNotionalAmountScheduleLeg1(), this::getOrCreateNotionalAmountScheduleLeg1);
			merger.mergeRosetta(getNotionalAmountScheduleLeg2(), o.getNotionalAmountScheduleLeg2(), this::getOrCreateNotionalAmountScheduleLeg2);
			merger.mergeRosetta(getNotionalQuantityScheduleLeg1(), o.getNotionalQuantityScheduleLeg1(), this::getOrCreateNotionalQuantityScheduleLeg1);
			merger.mergeRosetta(getNotionalQuantityScheduleLeg2(), o.getNotionalQuantityScheduleLeg2(), this::getOrCreateNotionalQuantityScheduleLeg2);
			merger.mergeRosetta(getPackageTransactionPrice(), o.getPackageTransactionPrice(), this::setPackageTransactionPrice);
			merger.mergeRosetta(getPackageTransactionSpread(), o.getPackageTransactionSpread(), this::setPackageTransactionSpread);
			merger.mergeRosetta(getNonReportable(), o.getNonReportable(), this::setNonReportable);
			
			merger.mergeBasic(getEffectiveDate(), o.getEffectiveDate(), this::setEffectiveDate);
			merger.mergeBasic(getExpirationDate(), o.getExpirationDate(), this::setExpirationDate);
			merger.mergeBasic(getEarlyTerminationDate(), o.getEarlyTerminationDate(), this::setEarlyTerminationDate);
			merger.mergeBasic(getReportingTimestamp(), o.getReportingTimestamp(), this::setReportingTimestamp);
			merger.mergeBasic(getExecutionTimestamp(), o.getExecutionTimestamp(), this::setExecutionTimestamp);
			merger.mergeBasic(getEntityResponsibleForReporting(), o.getEntityResponsibleForReporting(), this::setEntityResponsibleForReporting);
			merger.mergeBasic(getCounterparty1(), o.getCounterparty1(), this::setCounterparty1);
			merger.mergeBasic(getCounterparty2(), o.getCounterparty2(), this::setCounterparty2);
			merger.mergeBasic(getCounterparty2IdentifierType(), o.getCounterparty2IdentifierType(), this::setCounterparty2IdentifierType);
			merger.mergeBasic(getDirection1(), o.getDirection1(), this::setDirection1);
			merger.mergeBasic(getDirection2Leg1(), o.getDirection2Leg1(), this::setDirection2Leg1);
			merger.mergeBasic(getDirection2Leg2(), o.getDirection2Leg2(), this::setDirection2Leg2);
			merger.mergeBasic(getSubmitterIdentifier(), o.getSubmitterIdentifier(), this::setSubmitterIdentifier);
			merger.mergeBasic(getNewSDRIdentifier(), o.getNewSDRIdentifier(), this::setNewSDRIdentifier);
			merger.mergeBasic(getOriginalSwapSDRIdentifier(), o.getOriginalSwapSDRIdentifier(), this::setOriginalSwapSDRIdentifier);
			merger.mergeBasic(getCleared(), o.getCleared(), this::setCleared);
			merger.mergeBasic(getCentralCounterparty(), o.getCentralCounterparty(), this::setCentralCounterparty);
			merger.mergeBasic(getClearingMember(), o.getClearingMember(), this::setClearingMember);
			merger.mergeBasic(getPlatformIdentifier(), o.getPlatformIdentifier(), this::setPlatformIdentifier);
			merger.mergeBasic(getConfirmed(), o.getConfirmed(), this::setConfirmed);
			merger.mergeBasic(getFinalContractualSettlementDate(), o.getFinalContractualSettlementDate(), this::setFinalContractualSettlementDate);
			merger.mergeBasic(getConfirmationTimestamp(), o.getConfirmationTimestamp(), this::setConfirmationTimestamp);
			merger.mergeBasic(getSettlementCurrencyLeg1(), o.getSettlementCurrencyLeg1(), this::setSettlementCurrencyLeg1);
			merger.mergeBasic(getSettlementCurrencyLeg2(), o.getSettlementCurrencyLeg2(), this::setSettlementCurrencyLeg2);
			merger.mergeBasic(getClearingAccountOrigin(), o.getClearingAccountOrigin(), this::setClearingAccountOrigin);
			merger.mergeBasic(getOriginalSwapUTI(), o.getOriginalSwapUTI(), this::setOriginalSwapUTI);
			merger.mergeBasic(getClearingReceiptTimestamp(), o.getClearingReceiptTimestamp(), this::setClearingReceiptTimestamp);
			merger.mergeBasic(getUti(), o.getUti(), this::setUti);
			merger.mergeBasic(getUtiProprietary(), o.getUtiProprietary(), this::setUtiProprietary);
			merger.mergeBasic(getPriorUti(), o.getPriorUti(), this::setPriorUti);
			merger.mergeBasic(getPriorUtiProprietary(), o.getPriorUtiProprietary(), this::setPriorUtiProprietary);
			merger.mergeBasic(getFixedRateDayCountConventionLeg1(), o.getFixedRateDayCountConventionLeg1(), this::setFixedRateDayCountConventionLeg1);
			merger.mergeBasic(getFloatingRateDayCountConventionOfLeg1(), o.getFloatingRateDayCountConventionOfLeg1(), this::setFloatingRateDayCountConventionOfLeg1);
			merger.mergeBasic(getFixedRateDayCountConventionLeg2(), o.getFixedRateDayCountConventionLeg2(), this::setFixedRateDayCountConventionLeg2);
			merger.mergeBasic(getFloatingRateDayCountConventionOfLeg2(), o.getFloatingRateDayCountConventionOfLeg2(), this::setFloatingRateDayCountConventionOfLeg2);
			merger.mergeBasic(getFixedRatePaymentFrequencyPeriodLeg1(), o.getFixedRatePaymentFrequencyPeriodLeg1(), this::setFixedRatePaymentFrequencyPeriodLeg1);
			merger.mergeBasic(getFixedRatePaymentFrequencyPeriodLeg2(), o.getFixedRatePaymentFrequencyPeriodLeg2(), this::setFixedRatePaymentFrequencyPeriodLeg2);
			merger.mergeBasic(getFloatingRatePaymentFrequencyPeriodLeg1(), o.getFloatingRatePaymentFrequencyPeriodLeg1(), this::setFloatingRatePaymentFrequencyPeriodLeg1);
			merger.mergeBasic(getFloatingRatePaymentFrequencyPeriodLeg2(), o.getFloatingRatePaymentFrequencyPeriodLeg2(), this::setFloatingRatePaymentFrequencyPeriodLeg2);
			merger.mergeBasic(getFixedRatePaymentFrequencyPeriodMultiplierLeg1(), o.getFixedRatePaymentFrequencyPeriodMultiplierLeg1(), this::setFixedRatePaymentFrequencyPeriodMultiplierLeg1);
			merger.mergeBasic(getFixedRatePaymentFrequencyPeriodMultiplierLeg2(), o.getFixedRatePaymentFrequencyPeriodMultiplierLeg2(), this::setFixedRatePaymentFrequencyPeriodMultiplierLeg2);
			merger.mergeBasic(getFloatingRatePaymentFrequencyPeriodMultiplierLeg1(), o.getFloatingRatePaymentFrequencyPeriodMultiplierLeg1(), this::setFloatingRatePaymentFrequencyPeriodMultiplierLeg1);
			merger.mergeBasic(getFloatingRatePaymentFrequencyPeriodMultiplierLeg2(), o.getFloatingRatePaymentFrequencyPeriodMultiplierLeg2(), this::setFloatingRatePaymentFrequencyPeriodMultiplierLeg2);
			merger.mergeBasic(getFixingDateLeg1(), o.getFixingDateLeg1(), this::setFixingDateLeg1);
			merger.mergeBasic(getFixingDateLeg2(), o.getFixingDateLeg2(), this::setFixingDateLeg2);
			merger.mergeBasic(getFloatingRateResetFrequencyPeriodLeg1(), o.getFloatingRateResetFrequencyPeriodLeg1(), this::setFloatingRateResetFrequencyPeriodLeg1);
			merger.mergeBasic(getFloatingRateResetFrequencyPeriodOfLeg2(), o.getFloatingRateResetFrequencyPeriodOfLeg2(), this::setFloatingRateResetFrequencyPeriodOfLeg2);
			merger.mergeBasic(getFloatingRateResetFrequencyPeriodMultiplierLeg1(), o.getFloatingRateResetFrequencyPeriodMultiplierLeg1(), this::setFloatingRateResetFrequencyPeriodMultiplierLeg1);
			merger.mergeBasic(getFloatingRateResetFrequencyPeriodMultiplierLeg2(), o.getFloatingRateResetFrequencyPeriodMultiplierLeg2(), this::setFloatingRateResetFrequencyPeriodMultiplierLeg2);
			merger.mergeBasic(getValuationAmount(), o.getValuationAmount(), this::setValuationAmount);
			merger.mergeBasic(getValuationCurrency(), o.getValuationCurrency(), this::setValuationCurrency);
			merger.mergeBasic(getValuationMethod(), o.getValuationMethod(), this::setValuationMethod);
			merger.mergeBasic(getValuationTimestamp(), o.getValuationTimestamp(), this::setValuationTimestamp);
			merger.mergeBasic(getCollateralPortfolioIndicator(), o.getCollateralPortfolioIndicator(), this::setCollateralPortfolioIndicator);
			merger.mergeBasic(getInitialMarginCollateralPortfolioCode(), o.getInitialMarginCollateralPortfolioCode(), this::setInitialMarginCollateralPortfolioCode);
			merger.mergeBasic(getVariationMarginCollateralPortfolioCode(), o.getVariationMarginCollateralPortfolioCode(), this::setVariationMarginCollateralPortfolioCode);
			merger.mergeBasic(getPriceNotation(), o.getPriceNotation(), this::setPriceNotation);
			merger.mergeBasic(getPriceUnitOfMeasure(), o.getPriceUnitOfMeasure(), this::setPriceUnitOfMeasure);
			merger.mergeBasic(getFixedRateLeg1(), o.getFixedRateLeg1(), this::setFixedRateLeg1);
			merger.mergeBasic(getFixedRateLeg2(), o.getFixedRateLeg2(), this::setFixedRateLeg2);
			merger.mergeBasic(getSpreadNotationOfLeg1(), o.getSpreadNotationOfLeg1(), this::setSpreadNotationOfLeg1);
			merger.mergeBasic(getSpreadNotationOfLeg2(), o.getSpreadNotationOfLeg2(), this::setSpreadNotationOfLeg2);
			merger.mergeBasic(getStrikePriceNotation(), o.getStrikePriceNotation(), this::setStrikePriceNotation);
			merger.mergeBasic(getOptionPremiumAmount(), o.getOptionPremiumAmount(), this::setOptionPremiumAmount);
			merger.mergeBasic(getOptionPremiumCurrency(), o.getOptionPremiumCurrency(), this::setOptionPremiumCurrency);
			merger.mergeBasic(getOptionPremiumPaymentDate(), o.getOptionPremiumPaymentDate(), this::setOptionPremiumPaymentDate);
			merger.mergeBasic(getFirstExerciseDate(), o.getFirstExerciseDate(), this::setFirstExerciseDate);
			merger.mergeBasic(getExchangeRate(), o.getExchangeRate(), this::setExchangeRate);
			merger.mergeBasic(getExchangeRateBasis(), o.getExchangeRateBasis(), this::setExchangeRateBasis);
			merger.mergeBasic(getNotionalLeg1(), o.getNotionalLeg1(), this::setNotionalLeg1);
			merger.mergeBasic(getNotionalLeg2(), o.getNotionalLeg2(), this::setNotionalLeg2);
			merger.mergeBasic(getCallAmount(), o.getCallAmount(), this::setCallAmount);
			merger.mergeBasic(getPutAmount(), o.getPutAmount(), this::setPutAmount);
			merger.mergeBasic(getNotionalCurrencyOfLeg1(), o.getNotionalCurrencyOfLeg1(), this::setNotionalCurrencyOfLeg1);
			merger.mergeBasic(getNotionalCurrencyOfLeg2(), o.getNotionalCurrencyOfLeg2(), this::setNotionalCurrencyOfLeg2);
			merger.mergeBasic(getCallCurrency(), o.getCallCurrency(), this::setCallCurrency);
			merger.mergeBasic(getPutCurrency(), o.getPutCurrency(), this::setPutCurrency);
			merger.mergeBasic(getQuantityUnitOfMeasureLeg1(), o.getQuantityUnitOfMeasureLeg1(), this::setQuantityUnitOfMeasureLeg1);
			merger.mergeBasic(getQuantityUnitOfMeasureLeg2(), o.getQuantityUnitOfMeasureLeg2(), this::setQuantityUnitOfMeasureLeg2);
			merger.mergeBasic(getTotalNotionalQuantityLeg1(), o.getTotalNotionalQuantityLeg1(), this::setTotalNotionalQuantityLeg1);
			merger.mergeBasic(getTotalNotionalQuantityLeg2(), o.getTotalNotionalQuantityLeg2(), this::setTotalNotionalQuantityLeg2);
			merger.mergeBasic(getActionType(), o.getActionType(), this::setActionType);
			merger.mergeBasic(getEventType(), o.getEventType(), this::setEventType);
			merger.mergeBasic(getEventTimestamp(), o.getEventTimestamp(), this::setEventTimestamp);
			merger.mergeBasic(getEventIdentifier(), o.getEventIdentifier(), this::setEventIdentifier);
			merger.mergeBasic(getIndexFactor(), o.getIndexFactor(), this::setIndexFactor);
			merger.mergeBasic(getEmbeddedOptionType(), o.getEmbeddedOptionType(), this::setEmbeddedOptionType);
			merger.mergeBasic(getUpi(), o.getUpi(), this::setUpi);
			merger.mergeBasic(getDeliveryType(), o.getDeliveryType(), this::setDeliveryType);
			merger.mergeBasic(getAssetClass(), o.getAssetClass(), this::setAssetClass);
			merger.mergeBasic(getUnderlyingIdentificationType(), o.getUnderlyingIdentificationType(), this::setUnderlyingIdentificationType);
			merger.mergeBasic(getUnderlyingIdentification(), o.getUnderlyingIdentification(), this::setUnderlyingIdentification);
			merger.mergeBasic(getUnderlyingIndexIndicator(), o.getUnderlyingIndexIndicator(), this::setUnderlyingIndexIndicator);
			merger.mergeBasic(getNameOfTheUnderlyingIndex(), o.getNameOfTheUnderlyingIndex(), this::setNameOfTheUnderlyingIndex);
			merger.mergeBasic(getReferenceEntity(), o.getReferenceEntity(), this::setReferenceEntity);
			merger.mergeBasic(getIndicatorOfTheFloatingRateOfLeg1(), o.getIndicatorOfTheFloatingRateOfLeg1(), this::setIndicatorOfTheFloatingRateOfLeg1);
			merger.mergeBasic(getIndicatorOfTheFloatingRateOfLeg2(), o.getIndicatorOfTheFloatingRateOfLeg2(), this::setIndicatorOfTheFloatingRateOfLeg2);
			merger.mergeBasic(getNameOfTheFloatingRateOfLeg1(), o.getNameOfTheFloatingRateOfLeg1(), this::setNameOfTheFloatingRateOfLeg1);
			merger.mergeBasic(getNameOfTheFloatingRateOfLeg2(), o.getNameOfTheFloatingRateOfLeg2(), this::setNameOfTheFloatingRateOfLeg2);
			merger.mergeBasic(getFloatingRateReferencePeriodOfLeg1TimePeriod(), o.getFloatingRateReferencePeriodOfLeg1TimePeriod(), this::setFloatingRateReferencePeriodOfLeg1TimePeriod);
			merger.mergeBasic(getFloatingRateReferencePeriodOfLeg2TimePeriod(), o.getFloatingRateReferencePeriodOfLeg2TimePeriod(), this::setFloatingRateReferencePeriodOfLeg2TimePeriod);
			merger.mergeBasic(getFloatingRateReferencePeriodOfLeg1Multiplier(), o.getFloatingRateReferencePeriodOfLeg1Multiplier(), this::setFloatingRateReferencePeriodOfLeg1Multiplier);
			merger.mergeBasic(getFloatingRateReferencePeriodOfLeg2Multiplier(), o.getFloatingRateReferencePeriodOfLeg2Multiplier(), this::setFloatingRateReferencePeriodOfLeg2Multiplier);
			merger.mergeBasic(getIsCrypto(), o.getIsCrypto(), this::setIsCrypto);
			merger.mergeBasic(getMaturityDateOfTheUnderlying(), o.getMaturityDateOfTheUnderlying(), this::setMaturityDateOfTheUnderlying);
			merger.mergeBasic(getSeniority(), o.getSeniority(), this::setSeniority);
			merger.mergeBasic(getSeries(), o.getSeries(), this::setSeries);
			merger.mergeBasic(getVersion(), o.getVersion(), this::setVersion);
			merger.mergeBasic(getCdSIndexAttachmentPoint(), o.getCdSIndexAttachmentPoint(), this::setCdSIndexAttachmentPoint);
			merger.mergeBasic(getCdSIndexDetachmentPoint(), o.getCdSIndexDetachmentPoint(), this::setCdSIndexDetachmentPoint);
			merger.mergeBasic(getCustomBasketCode(), o.getCustomBasketCode(), this::setCustomBasketCode);
			merger.mergeBasic(getIdentifierOfBasketConstituents(), o.getIdentifierOfBasketConstituents(), (Consumer<String>) this::addIdentifierOfBasketConstituents);
			merger.mergeBasic(getBasketConstituentIdentifierSource(), o.getBasketConstituentIdentifierSource(), (Consumer<ProductIdTypeEnum>) this::addBasketConstituentIdentifierSource);
			merger.mergeBasic(getContractType(), o.getContractType(), this::setContractType);
			merger.mergeBasic(getOptionStyle(), o.getOptionStyle(), this::setOptionStyle);
			merger.mergeBasic(getOptionType(), o.getOptionType(), this::setOptionType);
			merger.mergeBasic(getPackageIdentifier(), o.getPackageIdentifier(), this::setPackageIdentifier);
			merger.mergeBasic(getPackageTransactionPriceCurrency(), o.getPackageTransactionPriceCurrency(), this::setPackageTransactionPriceCurrency);
			merger.mergeBasic(getPackageTransactionPriceNotation(), o.getPackageTransactionPriceNotation(), this::setPackageTransactionPriceNotation);
			merger.mergeBasic(getPackageTransactionSpreadCurrency(), o.getPackageTransactionSpreadCurrency(), this::setPackageTransactionSpreadCurrency);
			merger.mergeBasic(getPackageTransactionSpreadNotation(), o.getPackageTransactionSpreadNotation(), this::setPackageTransactionSpreadNotation);
			merger.mergeBasic(getTechnicalRecordId(), o.getTechnicalRecordId(), this::setTechnicalRecordId);
			return this;
		}
	
		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || !(o instanceof RosettaModelObject) || !getType().equals(((RosettaModelObject)o).getType())) return false;
		
			JFSATransactionReport _that = getType().cast(o);
		
			if (!Objects.equals(effectiveDate, _that.getEffectiveDate())) return false;
			if (!Objects.equals(expirationDate, _that.getExpirationDate())) return false;
			if (!Objects.equals(earlyTerminationDate, _that.getEarlyTerminationDate())) return false;
			if (!Objects.equals(reportingTimestamp, _that.getReportingTimestamp())) return false;
			if (!Objects.equals(executionTimestamp, _that.getExecutionTimestamp())) return false;
			if (!Objects.equals(entityResponsibleForReporting, _that.getEntityResponsibleForReporting())) return false;
			if (!Objects.equals(counterparty1, _that.getCounterparty1())) return false;
			if (!Objects.equals(counterparty2, _that.getCounterparty2())) return false;
			if (!Objects.equals(counterparty2IdentifierType, _that.getCounterparty2IdentifierType())) return false;
			if (!Objects.equals(direction1, _that.getDirection1())) return false;
			if (!Objects.equals(direction2Leg1, _that.getDirection2Leg1())) return false;
			if (!Objects.equals(direction2Leg2, _that.getDirection2Leg2())) return false;
			if (!Objects.equals(submitterIdentifier, _that.getSubmitterIdentifier())) return false;
			if (!Objects.equals(newSDRIdentifier, _that.getNewSDRIdentifier())) return false;
			if (!Objects.equals(originalSwapSDRIdentifier, _that.getOriginalSwapSDRIdentifier())) return false;
			if (!Objects.equals(cleared, _that.getCleared())) return false;
			if (!Objects.equals(centralCounterparty, _that.getCentralCounterparty())) return false;
			if (!Objects.equals(clearingMember, _that.getClearingMember())) return false;
			if (!Objects.equals(platformIdentifier, _that.getPlatformIdentifier())) return false;
			if (!Objects.equals(confirmed, _that.getConfirmed())) return false;
			if (!Objects.equals(finalContractualSettlementDate, _that.getFinalContractualSettlementDate())) return false;
			if (!Objects.equals(confirmationTimestamp, _that.getConfirmationTimestamp())) return false;
			if (!Objects.equals(settlementCurrencyLeg1, _that.getSettlementCurrencyLeg1())) return false;
			if (!Objects.equals(settlementCurrencyLeg2, _that.getSettlementCurrencyLeg2())) return false;
			if (!Objects.equals(clearingAccountOrigin, _that.getClearingAccountOrigin())) return false;
			if (!Objects.equals(originalSwapUTI, _that.getOriginalSwapUTI())) return false;
			if (!Objects.equals(clearingReceiptTimestamp, _that.getClearingReceiptTimestamp())) return false;
			if (!Objects.equals(uti, _that.getUti())) return false;
			if (!Objects.equals(utiProprietary, _that.getUtiProprietary())) return false;
			if (!Objects.equals(priorUti, _that.getPriorUti())) return false;
			if (!Objects.equals(priorUtiProprietary, _that.getPriorUtiProprietary())) return false;
			if (!Objects.equals(fixedRateDayCountConventionLeg1, _that.getFixedRateDayCountConventionLeg1())) return false;
			if (!Objects.equals(floatingRateDayCountConventionOfLeg1, _that.getFloatingRateDayCountConventionOfLeg1())) return false;
			if (!Objects.equals(fixedRateDayCountConventionLeg2, _that.getFixedRateDayCountConventionLeg2())) return false;
			if (!Objects.equals(floatingRateDayCountConventionOfLeg2, _that.getFloatingRateDayCountConventionOfLeg2())) return false;
			if (!Objects.equals(fixedRatePaymentFrequencyPeriodLeg1, _that.getFixedRatePaymentFrequencyPeriodLeg1())) return false;
			if (!Objects.equals(fixedRatePaymentFrequencyPeriodLeg2, _that.getFixedRatePaymentFrequencyPeriodLeg2())) return false;
			if (!Objects.equals(floatingRatePaymentFrequencyPeriodLeg1, _that.getFloatingRatePaymentFrequencyPeriodLeg1())) return false;
			if (!Objects.equals(floatingRatePaymentFrequencyPeriodLeg2, _that.getFloatingRatePaymentFrequencyPeriodLeg2())) return false;
			if (!Objects.equals(fixedRatePaymentFrequencyPeriodMultiplierLeg1, _that.getFixedRatePaymentFrequencyPeriodMultiplierLeg1())) return false;
			if (!Objects.equals(fixedRatePaymentFrequencyPeriodMultiplierLeg2, _that.getFixedRatePaymentFrequencyPeriodMultiplierLeg2())) return false;
			if (!Objects.equals(floatingRatePaymentFrequencyPeriodMultiplierLeg1, _that.getFloatingRatePaymentFrequencyPeriodMultiplierLeg1())) return false;
			if (!Objects.equals(floatingRatePaymentFrequencyPeriodMultiplierLeg2, _that.getFloatingRatePaymentFrequencyPeriodMultiplierLeg2())) return false;
			if (!Objects.equals(fixingDateLeg1, _that.getFixingDateLeg1())) return false;
			if (!Objects.equals(fixingDateLeg2, _that.getFixingDateLeg2())) return false;
			if (!Objects.equals(floatingRateResetFrequencyPeriodLeg1, _that.getFloatingRateResetFrequencyPeriodLeg1())) return false;
			if (!Objects.equals(floatingRateResetFrequencyPeriodOfLeg2, _that.getFloatingRateResetFrequencyPeriodOfLeg2())) return false;
			if (!Objects.equals(floatingRateResetFrequencyPeriodMultiplierLeg1, _that.getFloatingRateResetFrequencyPeriodMultiplierLeg1())) return false;
			if (!Objects.equals(floatingRateResetFrequencyPeriodMultiplierLeg2, _that.getFloatingRateResetFrequencyPeriodMultiplierLeg2())) return false;
			if (!ListEquals.listEquals(otherPayment, _that.getOtherPayment())) return false;
			if (!Objects.equals(valuationAmount, _that.getValuationAmount())) return false;
			if (!Objects.equals(valuationCurrency, _that.getValuationCurrency())) return false;
			if (!Objects.equals(valuationMethod, _that.getValuationMethod())) return false;
			if (!Objects.equals(valuationTimestamp, _that.getValuationTimestamp())) return false;
			if (!Objects.equals(collateralPortfolioIndicator, _that.getCollateralPortfolioIndicator())) return false;
			if (!Objects.equals(initialMarginCollateralPortfolioCode, _that.getInitialMarginCollateralPortfolioCode())) return false;
			if (!Objects.equals(variationMarginCollateralPortfolioCode, _that.getVariationMarginCollateralPortfolioCode())) return false;
			if (!Objects.equals(price, _that.getPrice())) return false;
			if (!Objects.equals(priceNotation, _that.getPriceNotation())) return false;
			if (!Objects.equals(priceUnitOfMeasure, _that.getPriceUnitOfMeasure())) return false;
			if (!ListEquals.listEquals(priceSchedule, _that.getPriceSchedule())) return false;
			if (!Objects.equals(fixedRateLeg1, _that.getFixedRateLeg1())) return false;
			if (!Objects.equals(fixedRateLeg2, _that.getFixedRateLeg2())) return false;
			if (!Objects.equals(spreadOfLeg1, _that.getSpreadOfLeg1())) return false;
			if (!Objects.equals(spreadNotationOfLeg1, _that.getSpreadNotationOfLeg1())) return false;
			if (!Objects.equals(spreadOfLeg2, _that.getSpreadOfLeg2())) return false;
			if (!Objects.equals(spreadNotationOfLeg2, _that.getSpreadNotationOfLeg2())) return false;
			if (!Objects.equals(strikePrice, _that.getStrikePrice())) return false;
			if (!Objects.equals(strikePriceNotation, _that.getStrikePriceNotation())) return false;
			if (!ListEquals.listEquals(strikePriceSchedule, _that.getStrikePriceSchedule())) return false;
			if (!Objects.equals(optionPremiumAmount, _that.getOptionPremiumAmount())) return false;
			if (!Objects.equals(optionPremiumCurrency, _that.getOptionPremiumCurrency())) return false;
			if (!Objects.equals(optionPremiumPaymentDate, _that.getOptionPremiumPaymentDate())) return false;
			if (!Objects.equals(firstExerciseDate, _that.getFirstExerciseDate())) return false;
			if (!Objects.equals(exchangeRate, _that.getExchangeRate())) return false;
			if (!Objects.equals(exchangeRateBasis, _that.getExchangeRateBasis())) return false;
			if (!Objects.equals(notionalLeg1, _that.getNotionalLeg1())) return false;
			if (!Objects.equals(notionalLeg2, _that.getNotionalLeg2())) return false;
			if (!Objects.equals(callAmount, _that.getCallAmount())) return false;
			if (!Objects.equals(putAmount, _that.getPutAmount())) return false;
			if (!Objects.equals(notionalCurrencyOfLeg1, _that.getNotionalCurrencyOfLeg1())) return false;
			if (!Objects.equals(notionalCurrencyOfLeg2, _that.getNotionalCurrencyOfLeg2())) return false;
			if (!Objects.equals(callCurrency, _that.getCallCurrency())) return false;
			if (!Objects.equals(putCurrency, _that.getPutCurrency())) return false;
			if (!Objects.equals(quantityUnitOfMeasureLeg1, _that.getQuantityUnitOfMeasureLeg1())) return false;
			if (!Objects.equals(quantityUnitOfMeasureLeg2, _that.getQuantityUnitOfMeasureLeg2())) return false;
			if (!ListEquals.listEquals(notionalAmountScheduleLeg1, _that.getNotionalAmountScheduleLeg1())) return false;
			if (!ListEquals.listEquals(notionalAmountScheduleLeg2, _that.getNotionalAmountScheduleLeg2())) return false;
			if (!Objects.equals(totalNotionalQuantityLeg1, _that.getTotalNotionalQuantityLeg1())) return false;
			if (!Objects.equals(totalNotionalQuantityLeg2, _that.getTotalNotionalQuantityLeg2())) return false;
			if (!ListEquals.listEquals(notionalQuantityScheduleLeg1, _that.getNotionalQuantityScheduleLeg1())) return false;
			if (!ListEquals.listEquals(notionalQuantityScheduleLeg2, _that.getNotionalQuantityScheduleLeg2())) return false;
			if (!Objects.equals(actionType, _that.getActionType())) return false;
			if (!Objects.equals(eventType, _that.getEventType())) return false;
			if (!Objects.equals(eventTimestamp, _that.getEventTimestamp())) return false;
			if (!Objects.equals(eventIdentifier, _that.getEventIdentifier())) return false;
			if (!Objects.equals(indexFactor, _that.getIndexFactor())) return false;
			if (!Objects.equals(embeddedOptionType, _that.getEmbeddedOptionType())) return false;
			if (!Objects.equals(upi, _that.getUpi())) return false;
			if (!Objects.equals(deliveryType, _that.getDeliveryType())) return false;
			if (!Objects.equals(assetClass, _that.getAssetClass())) return false;
			if (!Objects.equals(underlyingIdentificationType, _that.getUnderlyingIdentificationType())) return false;
			if (!Objects.equals(underlyingIdentification, _that.getUnderlyingIdentification())) return false;
			if (!Objects.equals(underlyingIndexIndicator, _that.getUnderlyingIndexIndicator())) return false;
			if (!Objects.equals(nameOfTheUnderlyingIndex, _that.getNameOfTheUnderlyingIndex())) return false;
			if (!Objects.equals(referenceEntity, _that.getReferenceEntity())) return false;
			if (!Objects.equals(indicatorOfTheFloatingRateOfLeg1, _that.getIndicatorOfTheFloatingRateOfLeg1())) return false;
			if (!Objects.equals(indicatorOfTheFloatingRateOfLeg2, _that.getIndicatorOfTheFloatingRateOfLeg2())) return false;
			if (!Objects.equals(nameOfTheFloatingRateOfLeg1, _that.getNameOfTheFloatingRateOfLeg1())) return false;
			if (!Objects.equals(nameOfTheFloatingRateOfLeg2, _that.getNameOfTheFloatingRateOfLeg2())) return false;
			if (!Objects.equals(floatingRateReferencePeriodOfLeg1TimePeriod, _that.getFloatingRateReferencePeriodOfLeg1TimePeriod())) return false;
			if (!Objects.equals(floatingRateReferencePeriodOfLeg2TimePeriod, _that.getFloatingRateReferencePeriodOfLeg2TimePeriod())) return false;
			if (!Objects.equals(floatingRateReferencePeriodOfLeg1Multiplier, _that.getFloatingRateReferencePeriodOfLeg1Multiplier())) return false;
			if (!Objects.equals(floatingRateReferencePeriodOfLeg2Multiplier, _that.getFloatingRateReferencePeriodOfLeg2Multiplier())) return false;
			if (!Objects.equals(isCrypto, _that.getIsCrypto())) return false;
			if (!Objects.equals(maturityDateOfTheUnderlying, _that.getMaturityDateOfTheUnderlying())) return false;
			if (!Objects.equals(seniority, _that.getSeniority())) return false;
			if (!Objects.equals(series, _that.getSeries())) return false;
			if (!Objects.equals(version, _that.getVersion())) return false;
			if (!Objects.equals(cdSIndexAttachmentPoint, _that.getCdSIndexAttachmentPoint())) return false;
			if (!Objects.equals(cdSIndexDetachmentPoint, _that.getCdSIndexDetachmentPoint())) return false;
			if (!Objects.equals(customBasketCode, _that.getCustomBasketCode())) return false;
			if (!ListEquals.listEquals(identifierOfBasketConstituents, _that.getIdentifierOfBasketConstituents())) return false;
			if (!ListEquals.listEquals(basketConstituentIdentifierSource, _that.getBasketConstituentIdentifierSource())) return false;
			if (!Objects.equals(contractType, _that.getContractType())) return false;
			if (!Objects.equals(optionStyle, _that.getOptionStyle())) return false;
			if (!Objects.equals(optionType, _that.getOptionType())) return false;
			if (!Objects.equals(packageIdentifier, _that.getPackageIdentifier())) return false;
			if (!Objects.equals(packageTransactionPrice, _that.getPackageTransactionPrice())) return false;
			if (!Objects.equals(packageTransactionPriceCurrency, _that.getPackageTransactionPriceCurrency())) return false;
			if (!Objects.equals(packageTransactionPriceNotation, _that.getPackageTransactionPriceNotation())) return false;
			if (!Objects.equals(packageTransactionSpread, _that.getPackageTransactionSpread())) return false;
			if (!Objects.equals(packageTransactionSpreadCurrency, _that.getPackageTransactionSpreadCurrency())) return false;
			if (!Objects.equals(packageTransactionSpreadNotation, _that.getPackageTransactionSpreadNotation())) return false;
			if (!Objects.equals(technicalRecordId, _that.getTechnicalRecordId())) return false;
			if (!Objects.equals(nonReportable, _that.getNonReportable())) return false;
			return true;
		}
		
		@Override
		public int hashCode() {
			int _result = 0;
			_result = 31 * _result + (effectiveDate != null ? effectiveDate.hashCode() : 0);
			_result = 31 * _result + (expirationDate != null ? expirationDate.hashCode() : 0);
			_result = 31 * _result + (earlyTerminationDate != null ? earlyTerminationDate.hashCode() : 0);
			_result = 31 * _result + (reportingTimestamp != null ? reportingTimestamp.hashCode() : 0);
			_result = 31 * _result + (executionTimestamp != null ? executionTimestamp.hashCode() : 0);
			_result = 31 * _result + (entityResponsibleForReporting != null ? entityResponsibleForReporting.hashCode() : 0);
			_result = 31 * _result + (counterparty1 != null ? counterparty1.hashCode() : 0);
			_result = 31 * _result + (counterparty2 != null ? counterparty2.hashCode() : 0);
			_result = 31 * _result + (counterparty2IdentifierType != null ? counterparty2IdentifierType.hashCode() : 0);
			_result = 31 * _result + (direction1 != null ? direction1.hashCode() : 0);
			_result = 31 * _result + (direction2Leg1 != null ? direction2Leg1.hashCode() : 0);
			_result = 31 * _result + (direction2Leg2 != null ? direction2Leg2.hashCode() : 0);
			_result = 31 * _result + (submitterIdentifier != null ? submitterIdentifier.hashCode() : 0);
			_result = 31 * _result + (newSDRIdentifier != null ? newSDRIdentifier.hashCode() : 0);
			_result = 31 * _result + (originalSwapSDRIdentifier != null ? originalSwapSDRIdentifier.hashCode() : 0);
			_result = 31 * _result + (cleared != null ? cleared.hashCode() : 0);
			_result = 31 * _result + (centralCounterparty != null ? centralCounterparty.hashCode() : 0);
			_result = 31 * _result + (clearingMember != null ? clearingMember.hashCode() : 0);
			_result = 31 * _result + (platformIdentifier != null ? platformIdentifier.hashCode() : 0);
			_result = 31 * _result + (confirmed != null ? confirmed.hashCode() : 0);
			_result = 31 * _result + (finalContractualSettlementDate != null ? finalContractualSettlementDate.hashCode() : 0);
			_result = 31 * _result + (confirmationTimestamp != null ? confirmationTimestamp.hashCode() : 0);
			_result = 31 * _result + (settlementCurrencyLeg1 != null ? settlementCurrencyLeg1.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (settlementCurrencyLeg2 != null ? settlementCurrencyLeg2.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (clearingAccountOrigin != null ? clearingAccountOrigin.hashCode() : 0);
			_result = 31 * _result + (originalSwapUTI != null ? originalSwapUTI.hashCode() : 0);
			_result = 31 * _result + (clearingReceiptTimestamp != null ? clearingReceiptTimestamp.hashCode() : 0);
			_result = 31 * _result + (uti != null ? uti.hashCode() : 0);
			_result = 31 * _result + (utiProprietary != null ? utiProprietary.hashCode() : 0);
			_result = 31 * _result + (priorUti != null ? priorUti.hashCode() : 0);
			_result = 31 * _result + (priorUtiProprietary != null ? priorUtiProprietary.hashCode() : 0);
			_result = 31 * _result + (fixedRateDayCountConventionLeg1 != null ? fixedRateDayCountConventionLeg1.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (floatingRateDayCountConventionOfLeg1 != null ? floatingRateDayCountConventionOfLeg1.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (fixedRateDayCountConventionLeg2 != null ? fixedRateDayCountConventionLeg2.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (floatingRateDayCountConventionOfLeg2 != null ? floatingRateDayCountConventionOfLeg2.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (fixedRatePaymentFrequencyPeriodLeg1 != null ? fixedRatePaymentFrequencyPeriodLeg1.hashCode() : 0);
			_result = 31 * _result + (fixedRatePaymentFrequencyPeriodLeg2 != null ? fixedRatePaymentFrequencyPeriodLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRatePaymentFrequencyPeriodLeg1 != null ? floatingRatePaymentFrequencyPeriodLeg1.hashCode() : 0);
			_result = 31 * _result + (floatingRatePaymentFrequencyPeriodLeg2 != null ? floatingRatePaymentFrequencyPeriodLeg2.hashCode() : 0);
			_result = 31 * _result + (fixedRatePaymentFrequencyPeriodMultiplierLeg1 != null ? fixedRatePaymentFrequencyPeriodMultiplierLeg1.hashCode() : 0);
			_result = 31 * _result + (fixedRatePaymentFrequencyPeriodMultiplierLeg2 != null ? fixedRatePaymentFrequencyPeriodMultiplierLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRatePaymentFrequencyPeriodMultiplierLeg1 != null ? floatingRatePaymentFrequencyPeriodMultiplierLeg1.hashCode() : 0);
			_result = 31 * _result + (floatingRatePaymentFrequencyPeriodMultiplierLeg2 != null ? floatingRatePaymentFrequencyPeriodMultiplierLeg2.hashCode() : 0);
			_result = 31 * _result + (fixingDateLeg1 != null ? fixingDateLeg1.hashCode() : 0);
			_result = 31 * _result + (fixingDateLeg2 != null ? fixingDateLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRateResetFrequencyPeriodLeg1 != null ? floatingRateResetFrequencyPeriodLeg1.hashCode() : 0);
			_result = 31 * _result + (floatingRateResetFrequencyPeriodOfLeg2 != null ? floatingRateResetFrequencyPeriodOfLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRateResetFrequencyPeriodMultiplierLeg1 != null ? floatingRateResetFrequencyPeriodMultiplierLeg1.hashCode() : 0);
			_result = 31 * _result + (floatingRateResetFrequencyPeriodMultiplierLeg2 != null ? floatingRateResetFrequencyPeriodMultiplierLeg2.hashCode() : 0);
			_result = 31 * _result + (otherPayment != null ? otherPayment.hashCode() : 0);
			_result = 31 * _result + (valuationAmount != null ? valuationAmount.hashCode() : 0);
			_result = 31 * _result + (valuationCurrency != null ? valuationCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (valuationMethod != null ? valuationMethod.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (valuationTimestamp != null ? valuationTimestamp.hashCode() : 0);
			_result = 31 * _result + (collateralPortfolioIndicator != null ? collateralPortfolioIndicator.hashCode() : 0);
			_result = 31 * _result + (initialMarginCollateralPortfolioCode != null ? initialMarginCollateralPortfolioCode.hashCode() : 0);
			_result = 31 * _result + (variationMarginCollateralPortfolioCode != null ? variationMarginCollateralPortfolioCode.hashCode() : 0);
			_result = 31 * _result + (price != null ? price.hashCode() : 0);
			_result = 31 * _result + (priceNotation != null ? priceNotation.hashCode() : 0);
			_result = 31 * _result + (priceUnitOfMeasure != null ? priceUnitOfMeasure.hashCode() : 0);
			_result = 31 * _result + (priceSchedule != null ? priceSchedule.hashCode() : 0);
			_result = 31 * _result + (fixedRateLeg1 != null ? fixedRateLeg1.hashCode() : 0);
			_result = 31 * _result + (fixedRateLeg2 != null ? fixedRateLeg2.hashCode() : 0);
			_result = 31 * _result + (spreadOfLeg1 != null ? spreadOfLeg1.hashCode() : 0);
			_result = 31 * _result + (spreadNotationOfLeg1 != null ? spreadNotationOfLeg1.hashCode() : 0);
			_result = 31 * _result + (spreadOfLeg2 != null ? spreadOfLeg2.hashCode() : 0);
			_result = 31 * _result + (spreadNotationOfLeg2 != null ? spreadNotationOfLeg2.hashCode() : 0);
			_result = 31 * _result + (strikePrice != null ? strikePrice.hashCode() : 0);
			_result = 31 * _result + (strikePriceNotation != null ? strikePriceNotation.hashCode() : 0);
			_result = 31 * _result + (strikePriceSchedule != null ? strikePriceSchedule.hashCode() : 0);
			_result = 31 * _result + (optionPremiumAmount != null ? optionPremiumAmount.hashCode() : 0);
			_result = 31 * _result + (optionPremiumCurrency != null ? optionPremiumCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (optionPremiumPaymentDate != null ? optionPremiumPaymentDate.hashCode() : 0);
			_result = 31 * _result + (firstExerciseDate != null ? firstExerciseDate.hashCode() : 0);
			_result = 31 * _result + (exchangeRate != null ? exchangeRate.hashCode() : 0);
			_result = 31 * _result + (exchangeRateBasis != null ? exchangeRateBasis.hashCode() : 0);
			_result = 31 * _result + (notionalLeg1 != null ? notionalLeg1.hashCode() : 0);
			_result = 31 * _result + (notionalLeg2 != null ? notionalLeg2.hashCode() : 0);
			_result = 31 * _result + (callAmount != null ? callAmount.hashCode() : 0);
			_result = 31 * _result + (putAmount != null ? putAmount.hashCode() : 0);
			_result = 31 * _result + (notionalCurrencyOfLeg1 != null ? notionalCurrencyOfLeg1.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (notionalCurrencyOfLeg2 != null ? notionalCurrencyOfLeg2.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (callCurrency != null ? callCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (putCurrency != null ? putCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (quantityUnitOfMeasureLeg1 != null ? quantityUnitOfMeasureLeg1.hashCode() : 0);
			_result = 31 * _result + (quantityUnitOfMeasureLeg2 != null ? quantityUnitOfMeasureLeg2.hashCode() : 0);
			_result = 31 * _result + (notionalAmountScheduleLeg1 != null ? notionalAmountScheduleLeg1.hashCode() : 0);
			_result = 31 * _result + (notionalAmountScheduleLeg2 != null ? notionalAmountScheduleLeg2.hashCode() : 0);
			_result = 31 * _result + (totalNotionalQuantityLeg1 != null ? totalNotionalQuantityLeg1.hashCode() : 0);
			_result = 31 * _result + (totalNotionalQuantityLeg2 != null ? totalNotionalQuantityLeg2.hashCode() : 0);
			_result = 31 * _result + (notionalQuantityScheduleLeg1 != null ? notionalQuantityScheduleLeg1.hashCode() : 0);
			_result = 31 * _result + (notionalQuantityScheduleLeg2 != null ? notionalQuantityScheduleLeg2.hashCode() : 0);
			_result = 31 * _result + (actionType != null ? actionType.hashCode() : 0);
			_result = 31 * _result + (eventType != null ? eventType.hashCode() : 0);
			_result = 31 * _result + (eventTimestamp != null ? eventTimestamp.hashCode() : 0);
			_result = 31 * _result + (eventIdentifier != null ? eventIdentifier.hashCode() : 0);
			_result = 31 * _result + (indexFactor != null ? indexFactor.hashCode() : 0);
			_result = 31 * _result + (embeddedOptionType != null ? embeddedOptionType.hashCode() : 0);
			_result = 31 * _result + (upi != null ? upi.hashCode() : 0);
			_result = 31 * _result + (deliveryType != null ? deliveryType.hashCode() : 0);
			_result = 31 * _result + (assetClass != null ? assetClass.hashCode() : 0);
			_result = 31 * _result + (underlyingIdentificationType != null ? underlyingIdentificationType.hashCode() : 0);
			_result = 31 * _result + (underlyingIdentification != null ? underlyingIdentification.hashCode() : 0);
			_result = 31 * _result + (underlyingIndexIndicator != null ? underlyingIndexIndicator.hashCode() : 0);
			_result = 31 * _result + (nameOfTheUnderlyingIndex != null ? nameOfTheUnderlyingIndex.hashCode() : 0);
			_result = 31 * _result + (referenceEntity != null ? referenceEntity.hashCode() : 0);
			_result = 31 * _result + (indicatorOfTheFloatingRateOfLeg1 != null ? indicatorOfTheFloatingRateOfLeg1.hashCode() : 0);
			_result = 31 * _result + (indicatorOfTheFloatingRateOfLeg2 != null ? indicatorOfTheFloatingRateOfLeg2.hashCode() : 0);
			_result = 31 * _result + (nameOfTheFloatingRateOfLeg1 != null ? nameOfTheFloatingRateOfLeg1.hashCode() : 0);
			_result = 31 * _result + (nameOfTheFloatingRateOfLeg2 != null ? nameOfTheFloatingRateOfLeg2.hashCode() : 0);
			_result = 31 * _result + (floatingRateReferencePeriodOfLeg1TimePeriod != null ? floatingRateReferencePeriodOfLeg1TimePeriod.hashCode() : 0);
			_result = 31 * _result + (floatingRateReferencePeriodOfLeg2TimePeriod != null ? floatingRateReferencePeriodOfLeg2TimePeriod.hashCode() : 0);
			_result = 31 * _result + (floatingRateReferencePeriodOfLeg1Multiplier != null ? floatingRateReferencePeriodOfLeg1Multiplier.hashCode() : 0);
			_result = 31 * _result + (floatingRateReferencePeriodOfLeg2Multiplier != null ? floatingRateReferencePeriodOfLeg2Multiplier.hashCode() : 0);
			_result = 31 * _result + (isCrypto != null ? isCrypto.hashCode() : 0);
			_result = 31 * _result + (maturityDateOfTheUnderlying != null ? maturityDateOfTheUnderlying.hashCode() : 0);
			_result = 31 * _result + (seniority != null ? seniority.hashCode() : 0);
			_result = 31 * _result + (series != null ? series.hashCode() : 0);
			_result = 31 * _result + (version != null ? version.hashCode() : 0);
			_result = 31 * _result + (cdSIndexAttachmentPoint != null ? cdSIndexAttachmentPoint.hashCode() : 0);
			_result = 31 * _result + (cdSIndexDetachmentPoint != null ? cdSIndexDetachmentPoint.hashCode() : 0);
			_result = 31 * _result + (customBasketCode != null ? customBasketCode.hashCode() : 0);
			_result = 31 * _result + (identifierOfBasketConstituents != null ? identifierOfBasketConstituents.hashCode() : 0);
			_result = 31 * _result + (basketConstituentIdentifierSource != null ? basketConstituentIdentifierSource.stream().map(Object::getClass).map(Class::getName).mapToInt(String::hashCode).sum() : 0);
			_result = 31 * _result + (contractType != null ? contractType.hashCode() : 0);
			_result = 31 * _result + (optionStyle != null ? optionStyle.hashCode() : 0);
			_result = 31 * _result + (optionType != null ? optionType.hashCode() : 0);
			_result = 31 * _result + (packageIdentifier != null ? packageIdentifier.hashCode() : 0);
			_result = 31 * _result + (packageTransactionPrice != null ? packageTransactionPrice.hashCode() : 0);
			_result = 31 * _result + (packageTransactionPriceCurrency != null ? packageTransactionPriceCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (packageTransactionPriceNotation != null ? packageTransactionPriceNotation.hashCode() : 0);
			_result = 31 * _result + (packageTransactionSpread != null ? packageTransactionSpread.hashCode() : 0);
			_result = 31 * _result + (packageTransactionSpreadCurrency != null ? packageTransactionSpreadCurrency.getClass().getName().hashCode() : 0);
			_result = 31 * _result + (packageTransactionSpreadNotation != null ? packageTransactionSpreadNotation.hashCode() : 0);
			_result = 31 * _result + (technicalRecordId != null ? technicalRecordId.hashCode() : 0);
			_result = 31 * _result + (nonReportable != null ? nonReportable.hashCode() : 0);
			return _result;
		}
		
		@Override
		public String toString() {
			return "JFSATransactionReportBuilder {" +
				"effectiveDate=" + this.effectiveDate + ", " +
				"expirationDate=" + this.expirationDate + ", " +
				"earlyTerminationDate=" + this.earlyTerminationDate + ", " +
				"reportingTimestamp=" + this.reportingTimestamp + ", " +
				"executionTimestamp=" + this.executionTimestamp + ", " +
				"entityResponsibleForReporting=" + this.entityResponsibleForReporting + ", " +
				"counterparty1=" + this.counterparty1 + ", " +
				"counterparty2=" + this.counterparty2 + ", " +
				"counterparty2IdentifierType=" + this.counterparty2IdentifierType + ", " +
				"direction1=" + this.direction1 + ", " +
				"direction2Leg1=" + this.direction2Leg1 + ", " +
				"direction2Leg2=" + this.direction2Leg2 + ", " +
				"submitterIdentifier=" + this.submitterIdentifier + ", " +
				"newSDRIdentifier=" + this.newSDRIdentifier + ", " +
				"originalSwapSDRIdentifier=" + this.originalSwapSDRIdentifier + ", " +
				"cleared=" + this.cleared + ", " +
				"centralCounterparty=" + this.centralCounterparty + ", " +
				"clearingMember=" + this.clearingMember + ", " +
				"platformIdentifier=" + this.platformIdentifier + ", " +
				"confirmed=" + this.confirmed + ", " +
				"finalContractualSettlementDate=" + this.finalContractualSettlementDate + ", " +
				"confirmationTimestamp=" + this.confirmationTimestamp + ", " +
				"settlementCurrencyLeg1=" + this.settlementCurrencyLeg1 + ", " +
				"settlementCurrencyLeg2=" + this.settlementCurrencyLeg2 + ", " +
				"clearingAccountOrigin=" + this.clearingAccountOrigin + ", " +
				"originalSwapUTI=" + this.originalSwapUTI + ", " +
				"clearingReceiptTimestamp=" + this.clearingReceiptTimestamp + ", " +
				"uti=" + this.uti + ", " +
				"utiProprietary=" + this.utiProprietary + ", " +
				"priorUti=" + this.priorUti + ", " +
				"priorUtiProprietary=" + this.priorUtiProprietary + ", " +
				"fixedRateDayCountConventionLeg1=" + this.fixedRateDayCountConventionLeg1 + ", " +
				"floatingRateDayCountConventionOfLeg1=" + this.floatingRateDayCountConventionOfLeg1 + ", " +
				"fixedRateDayCountConventionLeg2=" + this.fixedRateDayCountConventionLeg2 + ", " +
				"floatingRateDayCountConventionOfLeg2=" + this.floatingRateDayCountConventionOfLeg2 + ", " +
				"fixedRatePaymentFrequencyPeriodLeg1=" + this.fixedRatePaymentFrequencyPeriodLeg1 + ", " +
				"fixedRatePaymentFrequencyPeriodLeg2=" + this.fixedRatePaymentFrequencyPeriodLeg2 + ", " +
				"floatingRatePaymentFrequencyPeriodLeg1=" + this.floatingRatePaymentFrequencyPeriodLeg1 + ", " +
				"floatingRatePaymentFrequencyPeriodLeg2=" + this.floatingRatePaymentFrequencyPeriodLeg2 + ", " +
				"fixedRatePaymentFrequencyPeriodMultiplierLeg1=" + this.fixedRatePaymentFrequencyPeriodMultiplierLeg1 + ", " +
				"fixedRatePaymentFrequencyPeriodMultiplierLeg2=" + this.fixedRatePaymentFrequencyPeriodMultiplierLeg2 + ", " +
				"floatingRatePaymentFrequencyPeriodMultiplierLeg1=" + this.floatingRatePaymentFrequencyPeriodMultiplierLeg1 + ", " +
				"floatingRatePaymentFrequencyPeriodMultiplierLeg2=" + this.floatingRatePaymentFrequencyPeriodMultiplierLeg2 + ", " +
				"fixingDateLeg1=" + this.fixingDateLeg1 + ", " +
				"fixingDateLeg2=" + this.fixingDateLeg2 + ", " +
				"floatingRateResetFrequencyPeriodLeg1=" + this.floatingRateResetFrequencyPeriodLeg1 + ", " +
				"floatingRateResetFrequencyPeriodOfLeg2=" + this.floatingRateResetFrequencyPeriodOfLeg2 + ", " +
				"floatingRateResetFrequencyPeriodMultiplierLeg1=" + this.floatingRateResetFrequencyPeriodMultiplierLeg1 + ", " +
				"floatingRateResetFrequencyPeriodMultiplierLeg2=" + this.floatingRateResetFrequencyPeriodMultiplierLeg2 + ", " +
				"otherPayment=" + this.otherPayment + ", " +
				"valuationAmount=" + this.valuationAmount + ", " +
				"valuationCurrency=" + this.valuationCurrency + ", " +
				"valuationMethod=" + this.valuationMethod + ", " +
				"valuationTimestamp=" + this.valuationTimestamp + ", " +
				"collateralPortfolioIndicator=" + this.collateralPortfolioIndicator + ", " +
				"initialMarginCollateralPortfolioCode=" + this.initialMarginCollateralPortfolioCode + ", " +
				"variationMarginCollateralPortfolioCode=" + this.variationMarginCollateralPortfolioCode + ", " +
				"price=" + this.price + ", " +
				"priceNotation=" + this.priceNotation + ", " +
				"priceUnitOfMeasure=" + this.priceUnitOfMeasure + ", " +
				"priceSchedule=" + this.priceSchedule + ", " +
				"fixedRateLeg1=" + this.fixedRateLeg1 + ", " +
				"fixedRateLeg2=" + this.fixedRateLeg2 + ", " +
				"spreadOfLeg1=" + this.spreadOfLeg1 + ", " +
				"spreadNotationOfLeg1=" + this.spreadNotationOfLeg1 + ", " +
				"spreadOfLeg2=" + this.spreadOfLeg2 + ", " +
				"spreadNotationOfLeg2=" + this.spreadNotationOfLeg2 + ", " +
				"strikePrice=" + this.strikePrice + ", " +
				"strikePriceNotation=" + this.strikePriceNotation + ", " +
				"strikePriceSchedule=" + this.strikePriceSchedule + ", " +
				"optionPremiumAmount=" + this.optionPremiumAmount + ", " +
				"optionPremiumCurrency=" + this.optionPremiumCurrency + ", " +
				"optionPremiumPaymentDate=" + this.optionPremiumPaymentDate + ", " +
				"firstExerciseDate=" + this.firstExerciseDate + ", " +
				"exchangeRate=" + this.exchangeRate + ", " +
				"exchangeRateBasis=" + this.exchangeRateBasis + ", " +
				"notionalLeg1=" + this.notionalLeg1 + ", " +
				"notionalLeg2=" + this.notionalLeg2 + ", " +
				"callAmount=" + this.callAmount + ", " +
				"putAmount=" + this.putAmount + ", " +
				"notionalCurrencyOfLeg1=" + this.notionalCurrencyOfLeg1 + ", " +
				"notionalCurrencyOfLeg2=" + this.notionalCurrencyOfLeg2 + ", " +
				"callCurrency=" + this.callCurrency + ", " +
				"putCurrency=" + this.putCurrency + ", " +
				"quantityUnitOfMeasureLeg1=" + this.quantityUnitOfMeasureLeg1 + ", " +
				"quantityUnitOfMeasureLeg2=" + this.quantityUnitOfMeasureLeg2 + ", " +
				"notionalAmountScheduleLeg1=" + this.notionalAmountScheduleLeg1 + ", " +
				"notionalAmountScheduleLeg2=" + this.notionalAmountScheduleLeg2 + ", " +
				"totalNotionalQuantityLeg1=" + this.totalNotionalQuantityLeg1 + ", " +
				"totalNotionalQuantityLeg2=" + this.totalNotionalQuantityLeg2 + ", " +
				"notionalQuantityScheduleLeg1=" + this.notionalQuantityScheduleLeg1 + ", " +
				"notionalQuantityScheduleLeg2=" + this.notionalQuantityScheduleLeg2 + ", " +
				"actionType=" + this.actionType + ", " +
				"eventType=" + this.eventType + ", " +
				"eventTimestamp=" + this.eventTimestamp + ", " +
				"eventIdentifier=" + this.eventIdentifier + ", " +
				"indexFactor=" + this.indexFactor + ", " +
				"embeddedOptionType=" + this.embeddedOptionType + ", " +
				"upi=" + this.upi + ", " +
				"deliveryType=" + this.deliveryType + ", " +
				"assetClass=" + this.assetClass + ", " +
				"underlyingIdentificationType=" + this.underlyingIdentificationType + ", " +
				"underlyingIdentification=" + this.underlyingIdentification + ", " +
				"underlyingIndexIndicator=" + this.underlyingIndexIndicator + ", " +
				"nameOfTheUnderlyingIndex=" + this.nameOfTheUnderlyingIndex + ", " +
				"referenceEntity=" + this.referenceEntity + ", " +
				"indicatorOfTheFloatingRateOfLeg1=" + this.indicatorOfTheFloatingRateOfLeg1 + ", " +
				"indicatorOfTheFloatingRateOfLeg2=" + this.indicatorOfTheFloatingRateOfLeg2 + ", " +
				"nameOfTheFloatingRateOfLeg1=" + this.nameOfTheFloatingRateOfLeg1 + ", " +
				"nameOfTheFloatingRateOfLeg2=" + this.nameOfTheFloatingRateOfLeg2 + ", " +
				"floatingRateReferencePeriodOfLeg1TimePeriod=" + this.floatingRateReferencePeriodOfLeg1TimePeriod + ", " +
				"floatingRateReferencePeriodOfLeg2TimePeriod=" + this.floatingRateReferencePeriodOfLeg2TimePeriod + ", " +
				"floatingRateReferencePeriodOfLeg1Multiplier=" + this.floatingRateReferencePeriodOfLeg1Multiplier + ", " +
				"floatingRateReferencePeriodOfLeg2Multiplier=" + this.floatingRateReferencePeriodOfLeg2Multiplier + ", " +
				"isCrypto=" + this.isCrypto + ", " +
				"maturityDateOfTheUnderlying=" + this.maturityDateOfTheUnderlying + ", " +
				"seniority=" + this.seniority + ", " +
				"series=" + this.series + ", " +
				"version=" + this.version + ", " +
				"cdSIndexAttachmentPoint=" + this.cdSIndexAttachmentPoint + ", " +
				"cdSIndexDetachmentPoint=" + this.cdSIndexDetachmentPoint + ", " +
				"customBasketCode=" + this.customBasketCode + ", " +
				"identifierOfBasketConstituents=" + this.identifierOfBasketConstituents + ", " +
				"basketConstituentIdentifierSource=" + this.basketConstituentIdentifierSource + ", " +
				"contractType=" + this.contractType + ", " +
				"optionStyle=" + this.optionStyle + ", " +
				"optionType=" + this.optionType + ", " +
				"packageIdentifier=" + this.packageIdentifier + ", " +
				"packageTransactionPrice=" + this.packageTransactionPrice + ", " +
				"packageTransactionPriceCurrency=" + this.packageTransactionPriceCurrency + ", " +
				"packageTransactionPriceNotation=" + this.packageTransactionPriceNotation + ", " +
				"packageTransactionSpread=" + this.packageTransactionSpread + ", " +
				"packageTransactionSpreadCurrency=" + this.packageTransactionSpreadCurrency + ", " +
				"packageTransactionSpreadNotation=" + this.packageTransactionSpreadNotation + ", " +
				"technicalRecordId=" + this.technicalRecordId + ", " +
				"nonReportable=" + this.nonReportable +
			'}';
		}
	}
}
